package edu.stanford.nlp.io;

/** This class adapts between a <code>java.io.StreamTokenizer</code>
 *  and a <code>edu.stanford.nlp.io.StreamTokenizer</code>.
 *  
 *  @author Christopher Manning
 *  @version 2003/01/09
 */
public class StreamTokenizerAdapter implements StreamTokenizer {

    private java.io.StreamTokenizer st;

    private String eolString = "<EOL>";

    /** Create a new <code>StreamTokenizer</code>.  In general, it is 
     *  recommended that the passed in <code>StreamTokenizer</code> should
     *  have had <code>resetSyntax()</code> done to it, so that numbers are
     *  returned as entered as of type <code>String</code>, though this
     *  code will cope as best it can.
     *  @param st The internal <code>java.io.StreamTokenizer</code>
     */
    public StreamTokenizerAdapter(java.io.StreamTokenizer st) {
	this.st = st;
    }

    /** Returns the next token from the input Reader.  At end of file, it
     *	returns <code>null</code>.
     *  @return The token
     *  @throws IOException If there is some I/O error
     */
    public String next() throws java.io.IOException {
	int ttype = st.nextToken();
	switch (ttype) {
	case java.io.StreamTokenizer.TT_EOL:
	    return eolString;
	case java.io.StreamTokenizer.TT_EOF:
	    return null;
	case java.io.StreamTokenizer.TT_WORD:
	    return st.sval;
	case java.io.StreamTokenizer.TT_NUMBER:
	    return Double.toString(st.nval);
	default:
	    char[] t = { (char) ttype };    // (array initialization)
	    return new String(t);
	}
    }

    /** Pushback the last read token from the tokenizer for re-reading.
     *  That is, the next call to <code>next()</code> will return the same
     *  token as the last call to <code>next()</code>.  This allows just
     *  a single level of lookahead in token reading.
     */
    public void pushBack() {
	st.pushBack();
    }

    /** Set the <code>String</code> returned when the inner tokenizer 
     *  returns an end-of-line token.  This will only happen if the 
     *  inner tokenizer has been set to <code>eolIsSignificant(true)</code>.
     *  @param eolString The String used to represent eol.  It is not allowed
     *  to be <code>null</code> (which would confuse line ends and file end)
     */
    public void setEolString(String eolString) {
	if (eolString == null) {
	    throw new IllegalArgumentException("eolString cannot be null");
	}
	this.eolString = eolString;
    }

    /** Say whether the <code>String</code> is the end-of-line token for 
     *  this tokenizer.
     *  @param str The String being tested
     *  @return Whether it is the end-of-line token
     */
    public boolean isEol(String str) {
	return eolString.equals(str);
    }

}
