package edu.stanford.nlp.math;

/** The class <code>SloppyMath</code> contains methods for performing basic
 *  numeric operations.  In some cases, such as max and min, they cut a few
 *  corners in  
 *  the implementation for the sake of efficiency.  In particular, they may
 *  not handle special notions like NaN and -0.0 correctly.  This was the
 *  origin of the class name, but some other operations are just useful
 *  math additions, such as logSum.
 * 
 *  @author Christopher Manning
 *  @version 2003/01/02
 */
public final class SloppyMath {

    /**
     * Returns the greater of two <code>float</code> values.  That is,
     * the result is the argument closer to positive infinity. If the
     * arguments have the same value, the result is that same
     * value.  Does none of the special checks for NaN or -0.0f that 
     * <code>Math.max</code> does.
     *
     * @param   a   an argument.
     * @param   b   another argument.
     * @return  the larger of <code>a</code> and <code>b</code>.
     */
    public static float max(float a, float b) {
	return (a >= b) ? a : b;
    }


    /**
     * Returns the greater of two <code>double</code> values.  That
     * is, the result is the argument closer to positive infinity. If
     * the arguments have the same value, the result is that same
     * value.  Does none of the special checks for NaN or -0.0f that
     * <code>Math.max</code> does.
     *
     * @param   a   an argument.
     * @param   b   another argument.
     * @return  the larger of <code>a</code> and <code>b</code>.
     */
    public static double max(double a, double b) {
	return (a >= b) ? a : b;
    }


    /**
     * Returns the smaller of two <code>float</code> values.  That is,
     * the result is the value closer to negative infinity. If the
     * arguments have the same value, the result is that same
     * value.  Does none of the special checks for NaN or -0.0f that
     * <code>Math.max</code> does.
     *
     * @param   a   an argument.
     * @param   b   another argument.
     * @return  the smaller of <code>a</code> and <code>b.</code>
     */
    public static float min(float a, float b) {
	return (a <= b) ? a : b;
    }


    /**
     * Returns the smaller of two <code>double</code> values.  That
     * is, the result is the value closer to negative infinity. If the
     * arguments have the same value, the result is that same
     * value.  Does none of the special checks for NaN or -0.0f that
     * <code>Math.max</code> does.
     *
     * @param   a   an argument.
     * @param   b   another argument.
     * @return  the smaller of <code>a</code> and <code>b</code>.
     */
    public static double min(double a, double b) {
	return (a <= b) ? a : b;
    }


    /** Returns true if the argument is a "dangerous" double to have
     *  around, namely one that is infinite, NaN or zero.
     */
    public static boolean isDangerous(double d) {
	return Double.isInfinite(d) || Double.isNaN(d) || d == 0.0;
    }


    /** Returns true if the argument is a "very dangerous" double to have
     *  around, namely one that is infinite or NaN.
     */
    public static boolean isVeryDangerous(double d) {
	return Double.isInfinite(d) || Double.isNaN(d);
    }


    /** If a difference is bigger than this in log terms, then the sum or
     *  difference of them will just be the larger (to 12 or so decimal
     *  places). 
     */
    private static final double LOGTOLERANCE = 30.0;


    /** Returns the log of the sum of an array of numbers, which are
     *  themselves input in log form.  This is all natural logarithms.
     *  Reasonable care is taken to do this as efficiently as possible
     *  (under the assumption that the numbers might differ greatly in
     *  magnitude), with high accuracy, and without numerical overflow.
     *  @param inputs An array of numbers [log(x1), ..., log(xn)]
     *  @return log(x1 + ... + xn)
     */
    public static double logSum(double[] logInputs) {
	int leng = logInputs.length;
	if (leng == 0) {
	    throw new IllegalArgumentException();
	}
	int maxIdx = 0;
	double max = Double.NEGATIVE_INFINITY;
	for (int i = 0; i < leng; i++) {
	    if (logInputs[i] > max) {
		maxIdx = i;
		max = logInputs[i];
	    }
	}
	boolean haveTerms = false;
	double intermediate = 0.0;
	double cutoff = max - LOGTOLERANCE;
	// we avoid rearranging the array and so test indices each time!
	for (int i = 0; i < leng; i++) {
	    if (i != maxIdx && logInputs[i] >= cutoff) {
		haveTerms = true;
		intermediate += Math.exp(logInputs[i] - max);
	    }
	}
	if (haveTerms) {
	    return max + Math.log(1.0 + intermediate);
	} else {
	    return max;
	}
    }		


    /** Returns the log of the sum of two numbers, which are
     *  themselves input in log form.  This uses natural logarithms.
     *  Reasonable care is taken to do this as efficiently as possible
     *  (under the assumption that the numbers might differ greatly in
     *  magnitude), with high accuracy, and without numerical overflow.
     *  Also, handle correctly the case of arguments being -Inf (e.g.,
     *  probability 0).
     *  @param lx First number, in log form
     *  @param ly Second number, in log form
     *  @return log(exp(lx) + exp(ly))
     */
    public static double logAdd(double lx, double ly) {
	double max, negDiff;
	if (lx > ly) {
	    max = lx;
	    negDiff = ly - lx;
	} else {
	    max = ly;
	    negDiff = lx - ly;
	}
	if (max == Double.NEGATIVE_INFINITY) {
	    return max;
	} else if (negDiff < - LOGTOLERANCE) {
	    return max;
	} else {
	    return max + Math.log(1.0 + Math.exp(negDiff));
	}
    }		


    /** Find a hypergeometric distribution.  This uses exact math, trying as
     *  hard as one can to avoid numeric overflow by interleaving
     *  multiplications and divisions.
     *  Warning: the way this is written is only efficient if r,m < n/2.
     *  For other situations, some more code should be added.
     *  @param k The number of black balls drawn
     *  @param n The total number of balls
     *  @param r The number of black balls
     *  @param m The number of balls drawn
     *  @return The hypergeometric value
     */
    public static double hypergeometric(int k, int n, int r, int m) {
	if (k < 0 || r > n || m > n || n <= 0 || m < 0 | r < 0) {
	    throw new IllegalArgumentException("Invalid hypergeometric");
	}

	// exploit symmetry of problem
	if (m > n / 2) {
	    m = n - m;
	    k = r - k;
	}
	if (r > n / 2) {
	    r = n - r;
	    k = m - k;
	}
	if (m > r) {
	    int temp = m;
	    m = r;
	    r = temp;
	}
	// now we have that k <= m <= r <= n/2

	if (k < (m + r) - n || k > m) {
	    return 0.0;
	}

	// Do limit cases explicitly 
	// It's unclear whether this is a good idea.  I put it in fearing
	// numerical errors when the numbers seemed off, but actually there
	// was a bug in the Fisher's exact routine.
	if (r == n) {
	    if (k == m) {
		return 1.0;
	    } else {
		return 0.0;
	    }
	} else if (r == n - 1) {
	    if (k == m) {
		return (n - m)/ (double) n;
	    } else if (k == m - 1) {
		return m / (double) n;
	    } else {
		return 0.0;
	    }
	} else if (m == 1) {
	    if (k == 0) {
		return (n - r)/ (double) n;
	    } else if (k == 1) {
		return r / (double) n;
	    } else {
		return 0.0;
	    }
	} else if (m == 0) {
	    if (k == 0) {
		return 1.0;
	    } else {
		return 0.0;
	    }
	} else if (k == 0) {
	    double ans = 1.0;
	    for (int m0 = 0; m0 < m; m0++) {
		ans *= ((n-r)-m0);
		ans /= (n-m0);
	    }
	    return ans;
	}
	    
	double ans = 1.0;
	// do (n-r)x...x((n-r)-((m-k)-1))/n x...x (n-((m-k-1)))
	// leaving rest of denominator to get to multimply by (n-(m-1))
	// that's k things which goes into next loop
	for (int nr = n - r, n0 = n; nr > (n-r) - (m-k); nr--, n0--) {
	    // System.out.println("Multiplying by " + nr);
	    ans *= nr;
	    // System.out.println("Dividing by " + n0);
	    ans /= n0;
	}
	// System.out.println("Done phase 1");
	for (int k0 = 0; k0 < k; k0++) {
	    ans *= (m-k0);
	    // System.out.println("Multiplying by " + (m-k0));
	    ans /= ((n-(m-k0))+1);
	    // System.out.println("Dividing by " + ((n-(m+k0)+1)));
	    ans *= (r-k0);
	    // System.out.println("Multiplying by " + (r-k0));
	    ans /= (k0+1);
	    // System.out.println("Dividing by " + (k0+1));
	}
	return ans;
    }


    /** Find a one-tailed Fisher's exact probability.  Chance of having seen
     *  this or a more extreme departure from what you would have expected
     *  given independence.  I.e., k >= the value passed in.
     *  Warning: this was done just for collocations, where you are 
     *  concerned with the case of k being larger than predicted.  It doesn't
     *  correctly handle other cases, such as k being smaller than expected.
     *  @param k The number of black balls drawn
     *  @param n The total number of balls
     *  @param r The number of black balls
     *  @param m The number of balls drawn
     *  @return The Fisher's exact p-value
     */
    public static double oneTailedFishersExact(int k, int n, int r, int m) {
	if (k < 0 || k < (m + r) - n || k > r || k > m || r > n || m > n) {
	    throw new IllegalArgumentException("Invalid Fisher's exact: " +
		    "k="+k+" n="+n+" r="+r+" m="+m+" k<0="+(k<0)+" k<(m+r)-n="+
                    (k<(m+r)-n)+" k>r="+(k>r)+" k>m="+(k>m)+" r>n="+(r>n)+
		    "m>n="+(m>n));
	}
	// exploit symmetry of problem
	if (m > n / 2) {
	    m = n - m;
	    k = r - k;
	}
	if (r > n / 2) {
	    r = n - r;
	    k = m - k;
	}
	if (m > r) {
	    int temp = m;
	    m = r;
	    r = temp;
	}
	// now we have that k <= m <= r <= n/2

	double total = 0.0;
	if (k > m / 2) {
	    // sum from k to m
	    for (int k0 = k; k0 <= m; k0++) {
		// System.out.println("Calling hypg(" + k0 + "; " + n +
		// 		   ", " + r + ", " + m + ")");
		total += SloppyMath.hypergeometric(k0, n, r, m);
	    }
	} else {
	    // sum from max(0, (m+r)-n) to k-1, and then subtract from 1
	    int min = Math.max(0, (m+r)-n);
	    for (int k0 = min; k0 < k; k0++) {
		// System.out.println("Calling hypg(" + k0 + "; " + n +
		// 		   ", " + r + ", " + m + ")");
		total += SloppyMath.hypergeometric(k0, n, r, m);
	    }
	    total = 1.0 - total;
	}
	return total;
    }


    /** Find a 2x2 chi-square value. 
     *  Note: could do this more neatly using simplified formula for 2x2 case.
     *  @param k The number of black balls drawn
     *  @param n The total number of balls
     *  @param r The number of black balls
     *  @param m The number of balls drawn
     *  @return The Fisher's exact p-value
     */
    public static double chiSquare2by2(int k, int n, int r, int m) {
	int[][] cg = { { k, r - k}, {m - k, n - (k + (r-k) + (m-k)) } };
	int[] cgr = { r, n - r};
	int[] cgc = { m, n - m};
	double total = 0.0;
	for (int i = 0; i < 2; i++) {
	    for (int j = 0; j < 2; j++) {
		double exp = (double) cgr[i] * cgc[j] / n;
		total += (cg[i][j] - exp) * (cg[i][j] - exp) / exp;
	    }
	}
	return total;
    }


    /** Tests the hypergeometric distribution code.
     *  @param args The 4 arguments: k (cell), n (total), r (row), m (col)
     */
    public static void main(String[] args) {
	if (args.length == 0) {
	    System.out.println("Log adds of neg infinity numbers, etc.");
	    System.out.println("(logs) -Inf + -Inf = " + 
			       logAdd(Double.NEGATIVE_INFINITY, 
				      Double.NEGATIVE_INFINITY));
	    System.out.println("(logs) -Inf + -7 = " +
			       logAdd(Double.NEGATIVE_INFINITY, -7.0));
	    System.out.println("(logs) -7 + -Inf = " +
			       logAdd(-7.0, Double.NEGATIVE_INFINITY));
	    System.out.println("(logs) -50 + -7 = " +
			       logAdd(-50.0, -7.0));
	    System.out.println("(logs) -11 + -7 = " +
			       logAdd(-11.0, -7.0));
	    System.out.println("(logs) -7 + -11 = " +
			       logAdd(-7.0, -11.0));
	    System.out.println("real 1/2 + 1/2 = " +
			       logAdd(Math.log(0.5), Math.log(0.5)));
	} else if (args.length != 4) {
	    System.err.println(
		     "usage: edu.stanford.nlp.math.SloppyMath k n r m");
	} else {
	    int k = Integer.parseInt(args[0]);
	    int n = Integer.parseInt(args[1]);
	    int r = Integer.parseInt(args[2]);
	    int m = Integer.parseInt(args[3]);
	    double ans = SloppyMath.hypergeometric(k, n, r, m);
	    System.out.println("hypg(" + k + "; " + n + ", " + r + ", "  +
			       m + ") = " + ans);
	    ans = SloppyMath.oneTailedFishersExact(k, n, r, m);
	    System.out.println("1-tailed Fisher's exact(" + k + "; " + n +
			       ", " + r + ", "  + m + ") = " + ans);
	    double ansChi = SloppyMath.chiSquare2by2(k, n, r, m);
	    System.out.println("chiSquare(" + k + "; " + n +
			       ", " + r + ", "  + m + ") = " + ansChi);

	    System.out.println("Swapping arguments should give same hypg:");
	    ans = SloppyMath.hypergeometric(k, n, r, m);
	    System.out.println("hypg(" + k + "; " + n + ", " + m + ", "  +
			       r + ") = " + ans);
	    int othrow = n - m;
	    int othcol = n - r;
	    int cell12 = m - k;
	    int cell21 = r - k;
	    int cell22 = othrow - (r - k);
	    ans = SloppyMath.hypergeometric(cell12, n, othcol, m);
	    System.out.println("hypg(" + cell12 + "; " + n + ", " + othcol +
			       ", "  + m + ") = " + ans);
	    ans = SloppyMath.hypergeometric(cell21, n, r, othrow);
	    System.out.println("hypg(" + cell21 + "; " + n + ", " + r +
			       ", "  + othrow + ") = " + ans);
	    ans = SloppyMath.hypergeometric(cell22, n, othcol, othrow);
	    System.out.println("hypg(" + cell22 + "; " + n + ", " + othcol +
			       ", "  + othrow + ") = " + ans);
	}
    }

}
