package edu.stanford.nlp.parser.lexparser;

import java.io.*;
import java.util.*;
import java.net.*;
import java.util.zip.*;
import java.util.regex.*;

import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.util.*;
import edu.stanford.nlp.math.SloppyMath;
import edu.stanford.nlp.io.NumberRangeFileFilter;


/** @author Dan Klein */

class GLCBinarizerFactory {

  // Binarizes such that head-argument structure is respected
  // -- Looks only at the value of input tree nodes
  // -- Produces LSTrees with CWT labels.
  static class GLCHeadRespectingBinarizer implements TreeTransformer {
    private boolean grandParentAnnotate = Train.gPA;
    private boolean parentAnnotate = Train.PA;
    private boolean splitIN = true; // XXXX.  Fix! Train.splitIN;

    private HeadFinder hf;
    private TreeFactory tf;
    private int declare = 4; // number left when you switch out->in

    protected static Numberer ruleNumberer = new Numberer();
    protected static Counter ruleCounter = new Counter();
    protected static boolean markRuleNums = false;

    private String localTreeString(Tree tree) {
      StringBuffer sb = new StringBuffer();
      sb.append(tree.label().value());
      if (! tree.isLeaf()) {
	sb.append(" ->");
	for (int c=0; c<tree.children().length; c++) {
	  sb.append(" "+tree.children()[c].label().value());
	}
      }
      return sb.toString();
    }

    private String catCat(List trees) {
      StringBuffer sb = new StringBuffer();
      for (int c=0; c<trees.size(); c++) {
	Tree tree = (Tree)trees.get(c);
	Label label = tree.label();
	if (c > 0)
	  sb.append(" ");
	sb.append(label.value());
      }
      return sb.toString();
    }

    private void insideBinarizeLocalTree(Tree tree, int headLoc, String headWord, String headTag, String topCat) {
      if (tree.children().length == 1)
	return;
      if (headLoc > 0) {
	Tree child = tree.children()[0];
	List rest = tree.getChildrenAsList().subList(1,tree.children().length);
	String subString = "@?"+topCat+" : "+catCat(rest);
	Label subLabel = new CategoryWordTag(subString,
					     headWord,
					     headTag);
	Tree subTree = tf.newTreeNode(subLabel, rest);
	insideBinarizeLocalTree(subTree, headLoc-1, headWord, headTag, topCat);
	List children = new ArrayList(2);
	children.add(child);
	children.add(subTree);
	tree.setChildren(children);
	return;
      }
      int last = tree.children().length-1;
      Tree child = tree.children()[last];
      List rest = tree.getChildrenAsList().subList(0,last);
      String subString = "@?"+topCat+" : "+catCat(rest);
      Label subLabel = new CategoryWordTag(subString,
					   headWord,
					   headTag);
      Tree subTree = tf.newTreeNode(subLabel, rest);
      insideBinarizeLocalTree(subTree, headLoc, headWord, headTag, topCat);
      List children = new ArrayList(2);
      children.add(subTree);
      children.add(child);
      tree.setChildren(children);
    }

    private void outsideBinarizeLocalTree(Tree tree, int headLoc, String headWord, String headTag, String topCat, StringBuffer leftSB, StringBuffer rightSB) {
      if (tree.children().length == 1)
	return;
      if (tree.children().length < declare) {
	insideBinarizeLocalTree(tree, headLoc, headWord, headTag, topCat);
	return;
      }
      // carve off a left if there is one
      if (headLoc > 0) {
	leftSB.append(" ");
	leftSB.append(tree.children()[0].label().value());
	String labelStr = "@"+topCat+" :"+leftSB.toString()+" ...";
	Label label = new CategoryWordTag(labelStr, headWord, headTag);
	Tree splitTree = tf.newTreeNode(label, tree.getChildrenAsList().subList(1, tree.children().length));
	outsideBinarizeLocalTree(splitTree, headLoc-1, headWord, headTag, topCat, leftSB, rightSB);
	List children = new ArrayList(2);
	children.add(tree.children()[0]);
	children.add(splitTree);
	tree.setChildren(children);
	return;
      }
      int last = tree.children().length-1;
      rightSB.insert(0," "+tree.children()[last].label().value());
      String labelString = "@"+topCat+" :"+leftSB.toString()+" ..."+rightSB.toString();
      Label label = new CategoryWordTag(labelString, headWord, headTag);
      Tree splitTree = tf.newTreeNode(label, tree.getChildrenAsList().subList(0, tree.children().length-1));
      outsideBinarizeLocalTree(splitTree, headLoc, headWord, headTag, topCat, leftSB, rightSB);
      List children = new ArrayList(2);
      children.add(splitTree);
      children.add(tree.children()[last]);
      tree.setChildren(children);
    }

    private Tree binarize(Tree tree) {
      if (tree.isLeaf()) {
	return tree;
      }
      if (tree.isPreTerminal()) {
	return tf.newTreeNode(new CategoryWordTag(tree.label().value(),
						  tree.children()[0].label().value(),
						  tree.label().value()),
			      Collections.singletonList(binarize(tree.children()[0])));
	//Collections.singletonList(tf.newLeaf(tree.label().value())));
      }
      // otherwise, recurse
      List children = new ArrayList(tree.children().length);
      for (int c=0; c<tree.children().length; c++) {
	children.add(binarize(tree.children()[c]));
      }
      CategoryWordTag label = new CategoryWordTag(tree.label().value(), "", "");
      Tree newTree = tf.newTreeNode(label, children);
      Tree headChild = hf.determineHead(newTree);
      int headLoc = newTree.getChildrenAsList().indexOf(headChild);
      String headWord = ((HasWord)headChild.label()).word();
      String headTag = ((HasTag)headChild.label()).tag();
      label.setWord(headWord);
      label.setTag(headTag);
      // now we have a flat tree with binarized children,
      // so binarize this local level
      outsideBinarizeLocalTree(newTree, headLoc, headWord, headTag, tree.label().value(), new StringBuffer(), new StringBuffer());
      if (markRuleNums) {
	String ruleStr = localTreeString(tree);
	label.setCategory(tree.label().value()+"^"+ruleStr);
	ruleCounter.add(ruleStr, 1);
      }
      return newTree;
    }

    private void addRoot(Tree t) {
      t.setLabel(new CategoryWordTag("ROOT",Lexicon.BOUNDARY,Lexicon.BOUNDARY_TAG));
      List preTermChildList = new ArrayList();
      Tree boundaryTerm = tf.newLeaf(new StringLabel(Lexicon.BOUNDARY));//CategoryWordTag(Lexicon.BOUNDARY,Lexicon.BOUNDARY,""));
      preTermChildList.add(boundaryTerm);
      Tree boundaryPreTerm = tf.newTreeNode(new CategoryWordTag(Lexicon.BOUNDARY_TAG,Lexicon.BOUNDARY,Lexicon.BOUNDARY_TAG),preTermChildList);
      List childList = t.getChildrenAsList();
      Tree synthTree = tf.newTreeNode(new CategoryWordTag("@ROOT : ... *",
							  ((HasWord)t.children()[0].label()).word(),
							  ((HasTag)t.children()[0].label()).tag()),
				      childList);
      List topChildren = new ArrayList(2);
      topChildren.add(synthTree);
      topChildren.add(boundaryPreTerm);
      t.setChildren(topChildren);
    }

    public Tree transformTree(Tree t, TreeFactory dummy) {
      Tree binarizedTree = binarize(t);
      addRoot(binarizedTree);
      //binarizedTree.pennPrint();
      return binarizedTree;
    }

    public GLCHeadRespectingBinarizer (HeadFinder hf, int declare){
      this.hf = hf;
      this.tf = new LabeledScoredTreeFactory(new CategoryWordTagFactory());
      this.declare = declare;
    }
  }
  
  public static TreeTransformer getGLCHeadRespectingBinarizer(HeadFinder hf, int declare) {
    return new GLCHeadRespectingBinarizer(hf,declare);
  }
}

class LeftHeadFinder implements HeadFinder {
  public Tree determineHead(Tree t) {
    if (t.isLeaf())
      return null;
    return t.children()[0];
  }
}

class SimpleLexicon extends Lexicon {

  protected Map rulesByWord;
  protected Map rulesToScores;

  Numberer tagNumberer;
  Numberer wordNumberer;
  
  public Iterator ruleIteratorByWord(int word, int pos) {
    Collection rules = knownRulesByWord(word);
    if (rules == null || rules.isEmpty()) {
      rules = knownRulesByWord(wordNumberer.number(Lexicon.UNKNOWN_WORD));
    }
    return rules.iterator();
  }

  public double score(Object o, int pos) {
    IntTaggedWord tagging = (IntTaggedWord)o;
    Double score = (Double)rulesToScores.get(tagging);
    if (score == null)
      return Double.NEGATIVE_INFINITY;
    return score.doubleValue();
  }

  public void addTagging(IntTaggedWord tagging, double score) {
    rulesToScores.put(tagging, new Double(score));
    Collection rules = knownRulesByWord(tagging.word);
    rules.add(tagging);
  }

  protected Collection knownRulesByWord(int word) {
    Collection rules = (Collection)rulesByWord.get(new Integer(word));
    if (rules == null) {
      rules = new ArrayList();
      rulesByWord.put(new Integer(word), rules);
    }
    return rules;
  }

  public void writeTo(Writer out) throws IOException {
    for (Iterator ruleI = rulesToScores.keySet().iterator(); ruleI.hasNext();) {
      IntTaggedWord tagging = (IntTaggedWord)ruleI.next();
      out.write("<"+tagNumberer.object(tagging.tag)+"> -> <"+
		wordNumberer.object(tagging.word)+"> %% "+
		score(tagging, 0)+"\n");
    }
    out.flush();
  }

  public SimpleLexicon() {
    wordNumberer = Numberer.getGlobalNumberer("words");
    tagNumberer = Numberer.getGlobalNumberer("tags");
    rulesByWord = new HashMap();
    rulesToScores = new HashMap();
  }

  public SimpleLexicon(BufferedReader in) throws IOException {
    this();
    Pattern unaryRulePattern = Pattern.compile("<(.+)> -> <(.+)> %% (.+)");
    while (in.ready()) {
      String str = in.readLine();
      Matcher unaryMatcher = unaryRulePattern.matcher(str);
      if (unaryMatcher.matches()) {
	double score = Double.parseDouble(unaryMatcher.group(3));
	String parentStr = unaryMatcher.group(1);
	int parent = tagNumberer.number(parentStr);
	String childStr = unaryMatcher.group(2);
	int child = wordNumberer.number(childStr);
	IntTaggedWord tagging = new IntTaggedWord(child, parent);
	addTagging(tagging, score);
      }
    } 
  }
}

class SmoothedGrammarAndSimpleLexiconExtractor extends AbstractTreeExtractor {

  protected int HISTORY_DEPTH() { return 1; }
  protected boolean BAYESIAN_SMOOTH() { return true; }
  protected boolean SMOOTH() { return false; }
  protected boolean SPLIT_F_TAGS() { return false; }

  protected double k() { return 0.01; }

  protected Numberer stateNumberer = Numberer.getGlobalNumberer("states");
  protected Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  protected Numberer wordNumberer = Numberer.getGlobalNumberer("words");

  protected Numberer localTreeNumberer = Numberer.getGlobalNumberer("localTrees");

  protected Map ruleToLabel = new HashMap();

  protected Counter rulePairs = new Counter();
  protected Counter labelPairs = new Counter();

  protected Counter wordCounter = new Counter();
  protected Counter tagCounter = new Counter();
  protected Counter taggingCounter = new Counter();
  protected Counter unseenTagCounter = new Counter();
  protected Counter unseenTaggingCounter = new Counter();

  protected boolean markRules = GLCBinarizerFactory.GLCHeadRespectingBinarizer.markRuleNums;
  protected Numberer ruleNumberer = GLCBinarizerFactory.GLCHeadRespectingBinarizer.ruleNumberer;
  protected Counter ruleCounter = GLCBinarizerFactory.GLCHeadRespectingBinarizer.ruleCounter;
  

  // LEXICON

  protected void tallyPreTerminal(Tree t, LinkedList parents) {
    String tagStr = t.label().value();
    String wordStr = t.children()[0].label().value();

    if (SPLIT_F_TAGS() &&
	parents.size() > 1 &&
	isFunctionTag(stateNumberer.number(tagStr))) {
      String parent = historyToString(new ArrayList(parents.subList(1,2)));
      tagStr = tagStr+parent;
    }

    int tag = tagNumberer.number(tagStr);
    int word = wordNumberer.number(wordStr);
    Integer tagI = new Integer(tag);
    Integer wordI = new Integer(word);
    IntTaggedWord tagging = new IntTaggedWord(word, tag);

    tagCounter.add(tagI, 1);
    wordCounter.add(wordI, 1);
    taggingCounter.add(tagging, 1);

    if (wordCounter.countOf(wordI) < 2) {
      if (!isClosedTag(stateNumberer.number(tagStr))) {
	int unk = wordNumberer.number(Lexicon.UNKNOWN_WORD);
	IntTaggedWord utagging = new IntTaggedWord(unk, tag);
	unseenTaggingCounter.add(utagging,1);
	unseenTagCounter.add(tagI,1);
      }
    }

  }

  // GRAMMAR

  public void tallyTree(Tree t) {
    LinkedList ll = new LinkedList();
    LinkedList rr = new LinkedList();
    tallyTree(t, ll, rr);
  }

  protected String stripRule(Tree t) {
    String str = t.label().value();
    int split = str.indexOf('^');
    String node = str.substring(0,split);
    String rule = ""+ruleNumberer.number(str.substring(split+1, str.length()));
    t.label().setValue(node);
    return rule;
  }

  protected void tallyTree(Tree t, LinkedList parents, LinkedList rules) {
    // traverse tree, building parent list
    String str = t.label().value();
    boolean strIsPassive = (str.indexOf('@') == -1);
    if (strIsPassive && markRules) {
      String ruleStr = stripRule(t);
      if (ruleCounter.countOf(ruleStr) > 5) {
	parents.addFirst(ruleStr);
      } else {
	parents.addFirst(null);
      }
      str = t.label().value();
    }
    if (strIsPassive) {
      parents.addFirst(str);
    }
    if (! t.isLeaf()) {
      if (! t.children()[0].isLeaf()) {
	tallyInternalNode(t, parents);
	for (int c=0; c<t.children().length; c++) {
	  Tree child = t.children()[c];
	  tallyTree(child, parents, rules);
	}
      } else {
	tallyPreTerminal(t, parents);
	//tagNumberer.number(t.label().value());
      }
    }
    if (strIsPassive) {
      parents.removeFirst();
    }
    if (strIsPassive && markRules) {
      parents.removeFirst();
    }
  }

  protected Rule ltToRule(Tree lt) {
    Rule r = null;
    if (lt.children().length == 1) {
      UnaryRule ur = new UnaryRule();
      ur.parent = stateNumberer.number(lt.label().value());
      ur.child = stateNumberer.number(lt.children()[0].label().value());
      r = ur;
    } else {
      BinaryRule br = new BinaryRule();
      br.parent = stateNumberer.number(lt.label().value());
      br.leftChild = stateNumberer.number(lt.children()[0].label().value());
      br.rightChild = stateNumberer.number(lt.children()[1].label().value());
      r = br;
    }
    return r;
  }

  protected static Boolean[] isSynthetic = new Boolean[1000000];
  protected boolean isSynthetic(int state) {
    if (isSynthetic[state] == null) {
      if (((String)stateNumberer.object(state)).indexOf('@') > -1)
	isSynthetic[state] = Boolean.TRUE;
      else
	isSynthetic[state] = Boolean.FALSE;
    }
    return isSynthetic[state].booleanValue();
	
  }

  protected static Boolean[] isTag = new Boolean[1000000];
  protected boolean isTag(int state) {
    if (isTag[state] == null) {
      if (tagNumberer.hasSeen(stateNumberer.object(state)))
	isTag[state] = Boolean.TRUE;
      else
	isTag[state] = Boolean.FALSE;
    }
    return isTag[state].booleanValue();
  }


  protected Set functionTagStrings = null;
  protected static Boolean[] isFunctionTag = new Boolean[1000000];
  protected boolean isFunctionTag(int state) {
    if(functionTagStrings == null) {
      functionTagStrings = new HashSet();
      functionTagStrings.add("IN");
      functionTagStrings.add("CC");
      functionTagStrings.add("DT");
      functionTagStrings.add("RB");
    }
    if (isFunctionTag[state] == null) {
      if (functionTagStrings.contains(stateNumberer.object(state)))
	isFunctionTag[state] = Boolean.TRUE;
      else
	isFunctionTag[state] = Boolean.FALSE;
    }
    return isFunctionTag[state].booleanValue();
  }

  protected Set functionTagStrings2 = null;
  protected static Boolean[] isFunctionTag2 = new Boolean[1000000];
  protected boolean isFunctionTag2(int state) {
    if(functionTagStrings2 == null) {
      functionTagStrings2 = new HashSet();
      functionTagStrings2.add("DT");
    }
    if (isFunctionTag2[state] == null) {
      if (functionTagStrings2.contains(stateNumberer.object(state)))
	isFunctionTag2[state] = Boolean.TRUE;
      else
	isFunctionTag2[state] = Boolean.FALSE;
    }
    return isFunctionTag2[state].booleanValue();
  }

  protected Set closedTagStrings = null;
  protected static Boolean[] isClosedTag = new Boolean[1000000];
  protected boolean isClosedTag(int state) {
    if(closedTagStrings == null) {
      closedTagStrings = new HashSet();
      closedTagStrings.add("#");
      closedTagStrings.add("*");
      closedTagStrings.add("RP");
      closedTagStrings.add("DT");
      closedTagStrings.add(".");
      closedTagStrings.add(",");
      closedTagStrings.add(":");
      closedTagStrings.add("``");
      closedTagStrings.add("''");
      closedTagStrings.add("-RRB-");
      closedTagStrings.add("-LRB-");
      closedTagStrings.add("PRP");
      closedTagStrings.add("PRP$");
      closedTagStrings.add("IN");
      closedTagStrings.add("POS");
      closedTagStrings.add("MD");
      closedTagStrings.add("TO");
      closedTagStrings.add("UH");
      closedTagStrings.add("$");
      closedTagStrings.add("SYM");
      closedTagStrings.add("LS");
      closedTagStrings.add("EX");
      closedTagStrings.add("WRB");
      closedTagStrings.add("WP$");
      closedTagStrings.add("WP");
      closedTagStrings.add("WDT");
      closedTagStrings.add("CC");
    }
    if (isClosedTag[state] == null) {
      String base = (String)stateNumberer.object(state);
      if (base.indexOf('^') > -1) {
	base = base.substring(0, base.indexOf('^'));
      }
      if (closedTagStrings.contains(base))
	isClosedTag[state] = Boolean.TRUE;
      else
	isClosedTag[state] = Boolean.FALSE;
    }
    return isClosedTag[state].booleanValue();
  }

  protected Rule specifyRule(Rule rule, List history) {
    Rule r = null;
    String topHistoryStr = "";
    String bottomHistoryStr = "";
    if (! markRules) {
      topHistoryStr = historyToString(history.subList(1,history.size()));
      bottomHistoryStr = historyToString(history.subList(0,Math.min(HISTORY_DEPTH()-1, history.size())));
    }
    if (rule instanceof UnaryRule) {
      UnaryRule ur = new UnaryRule();
      UnaryRule urule = (UnaryRule)rule;
      ur.parent = stateNumberer.number(stateNumberer.object(urule.parent)+topHistoryStr);
      if (isSynthetic(urule.child)) {
	ur.child = stateNumberer.number(stateNumberer.object(urule.child)+topHistoryStr);
      } else if (isFunctionTag(urule.child)) {
	if (SPLIT_F_TAGS() &&
	    history.size() > 0 &&
	    isFunctionTag(urule.child)) {
	  String parent = historyToString(history.subList(0,1));
	  ur.child = stateNumberer.number(stateNumberer.object(urule.child)+parent);
	} else {
	  ur.child = urule.child;
	}
      } else {
	ur.child = stateNumberer.number(stateNumberer.object(urule.child)+bottomHistoryStr);
      }
      r = ur;
    } else {
      BinaryRule br = new BinaryRule();
      BinaryRule brule = (BinaryRule)rule;
      br.parent = stateNumberer.number(stateNumberer.object(brule.parent)+topHistoryStr);
      if (isSynthetic(brule.leftChild)) {
	br.leftChild = stateNumberer.number(stateNumberer.object(brule.leftChild)+topHistoryStr);
      } else if (isFunctionTag(brule.leftChild)) {
	if (SPLIT_F_TAGS() &&
	    history.size() > 0 &&
	    isFunctionTag(brule.leftChild)) {
	  String parent = historyToString(history.subList(0,1));
	  br.leftChild = stateNumberer.number(stateNumberer.object(brule.leftChild)+parent);
	} else {
	  br.leftChild = brule.leftChild;
	}
      } else {
	br.leftChild = stateNumberer.number(stateNumberer.object(brule.leftChild)+bottomHistoryStr);
      }
      if (isSynthetic(brule.rightChild)) {
	br.rightChild = stateNumberer.number(stateNumberer.object(brule.rightChild)+topHistoryStr);
      } else if (isFunctionTag(brule.rightChild)) {
	if (SPLIT_F_TAGS() &&
	    history.size() > 0 &&
	    isFunctionTag(brule.rightChild)) {
	  String parent = historyToString(history.subList(0,1));
	  br.rightChild = stateNumberer.number(stateNumberer.object(brule.rightChild)+parent);
	} else {
	  br.rightChild = brule.rightChild;
	}
      } else {
	br.rightChild = stateNumberer.number(stateNumberer.object(brule.rightChild)+bottomHistoryStr);
      }
      r = br;
    }
    return r;
  }


  protected void tallyInternalNode(Tree lt, List parents) {
    StringBuffer historySB = new StringBuffer();
    // form base rule
    String label = lt.label().value();
    Rule baseR = ltToRule(lt);
    ruleToLabel.put(baseR,label);
    // act on each history depth
    for (int depth = 1; depth <= Math.min(HISTORY_DEPTH(),parents.size()); depth++) {
      List history = new ArrayList(parents.subList(0,depth));
      // tally each history level / rewrite pair
      rulePairs.add(new Pair(baseR,history),1);
      labelPairs.add(new Pair(label,history),1);
    }
  }

  protected Map historyToString = new HashMap();
  protected String historyToString(List history) {
    String str = (String)historyToString.get(history);
    if (str == null) { 
      StringBuffer sb = new StringBuffer();
      for (int i=0; i<history.size(); i++) {
	sb.append('^');
	sb.append(history.get(i));
      }
      str = sb.toString();
      historyToString.put(history, str);
    }
    return str;
  }

  protected Object formResult() {
    PCFGGrammar grammar = formGrammar();
    SimpleLexicon lexicon = formLexicon();
    return new Pair(grammar, lexicon);
  }

  protected SimpleLexicon formLexicon() {
    SimpleLexicon lexicon = new SimpleLexicon();
    for (Iterator taggingI = taggingCounter.seenSet().iterator(); taggingI.hasNext();) {
      IntTaggedWord tagging = (IntTaggedWord)taggingI.next();
      double c_tw = taggingCounter.countOf(tagging);
      double c_t = tagCounter.countOf(new Integer(tagging.tag));
      double score = Math.log(c_tw/c_t);
      lexicon.addTagging(tagging, score);
    }
    for (Iterator taggingI = unseenTaggingCounter.seenSet().iterator(); taggingI.hasNext();) {
      IntTaggedWord tagging = (IntTaggedWord)taggingI.next();
      double c_tu = unseenTaggingCounter.countOf(tagging);
      //double c_u = unseenTaggingCounter.total();
      //double c_s = tagCounter.total();
      double c_ts = tagCounter.countOf(new Integer(tagging.tag));
      double score = Math.log(c_tu/c_ts); // P(t|u)*P(u)/P(t)
      lexicon.addTagging(tagging, score);
    }
    return lexicon;
  }

  protected PCFGGrammar formGrammar() {
    Set brs = new HashSet();
    Set urs = new HashSet();
    // scan each rule / history pair
    int ruleCount = 0;
    for (Iterator pairI = rulePairs.seenSet().iterator(); pairI.hasNext();) {
      if (ruleCount % 100 == 0)
	System.err.println("Rules multiplied: "+ruleCount);
      ruleCount++;
      Pair rulePair = (Pair)pairI.next();
      Rule baseRule = (Rule)rulePair.first;
      String baseLabel = (String)ruleToLabel.get(baseRule);
      List history = (List)rulePair.second;
      if (history.size() != HISTORY_DEPTH() &&
	  ! history.get(history.size()-1).equals("ROOT"))
	continue;
      double totalProb = 0;
      for (int depth = (k() > 0.0 ? 1 : history.size()); depth <= HISTORY_DEPTH() && depth <= history.size(); depth++) {
	List subHistory = history.subList(0,depth);
	double c_label = labelPairs.countOf(new Pair(baseLabel,subHistory));
	double c_rule = rulePairs.countOf(new Pair(baseRule,subHistory));
	//System.out.println("Multiplying out "+baseRule+" with history "+subHistory);
	//System.out.println("Count of "+baseLabel+" with "+subHistory+" is "+c_label);
	//System.out.println("Count of "+baseRule+" with "+subHistory+" is "+c_rule );

	if (! BAYESIAN_SMOOTH()) {
	  // linear interpolation
	  double weight = (depth == 1 ? 0.01 :
			   (depth == 2 ? 0.09 : 0.9));
	  if (HISTORY_DEPTH() == 1)
	    weight /= 0.01;
	  if (HISTORY_DEPTH() == 2)
	    weight /= 0.1;
	  double prob = weight*(c_rule)/(c_label);
	  totalProb += prob;
	} else {
	  // smooth against a prior
	  if (depth == 1)
	    totalProb = c_rule/c_label;
	  else
	    totalProb = (c_rule+k()*totalProb)/(c_label+k());
	}
	if (totalProb == 0.0)
	  continue;
	Rule rule = specifyRule(baseRule, subHistory);
	rule.score = Math.log(totalProb);
	//System.out.println("Created  "+rule+" with score "+rule.score);
	if (rule instanceof UnaryRule)
	  urs.add(rule);
	else
	  brs.add(rule);
      }
    }
    // label back-off unaries
    for (Iterator pairI = labelPairs.seenSet().iterator(); pairI.hasNext();) {
      Pair labelPair = (Pair)pairI.next();
      String baseLabel = (String)labelPair.first;
      List history = (List)labelPair.second;
      if (history.size() > 1 && baseLabel.charAt(0) != '@') {
	UnaryRule ur = new UnaryRule();
	String parentStr = historyToString(history);
	String childStr = historyToString(history.subList(0,history.size()-1));
	ur.parent = stateNumberer.number(parentStr.substring(1));
	ur.child = stateNumberer.number(childStr.substring(1));
	if (! BAYESIAN_SMOOTH()) {
	  ur.score = Math.log(0.99);
	} else {
	  double c_label = labelPairs.countOf(labelPair);
	  double pForget = k()/(c_label+k());
	  if (pForget == 0.0)
	    continue;
	  ur.score = Math.log(pForget);
	}
	urs.add(ur);
	//System.out.println("Created: "+ur);
      }
    }

    System.out.println("Total states: "+stateNumberer.total());
    BinaryGrammar bg = new BinaryGrammar(stateNumberer.total());
    UnaryGrammar ug = new UnaryGrammar(stateNumberer.total());
    for (Iterator brI = brs.iterator(); brI.hasNext();) {
      BinaryRule br = (BinaryRule)brI.next();
      bg.addRule(br);
    }
    for (Iterator urI = urs.iterator(); urI.hasNext();) {
      UnaryRule ur = (UnaryRule)urI.next();
      ug.addRule(ur);
    }
    ug.purgeRules();
    return new PCFGGrammar(bg,ug);
  }
}


class NullScorer implements Scorer {
  public double oScore(Edge edge) { return 0; }
  public double iScore(Edge edge) { return 0; }
  public boolean oPossible(Hook hook) { return true; }
  public boolean iPossible(Hook hook) { return true; }
  public boolean parse(List words) { return true; }
}

class LexiconSerializer {
  static class LexiconData implements Serializable {
    public Lexicon lexicon;
    public Map numberers;
    public LexiconData(Lexicon lexicon, Map numberers) {
      this.lexicon = lexicon;
      this.numberers = numberers;
    }
  }

  public static void serializeLexicon(Lexicon lexicon, String filename) {
    LexiconData ld = new LexiconData(lexicon, Numberer.getNumberers());
    try {
      ObjectOutputStream out;
      if (filename.endsWith(".gz")) {
	out = new ObjectOutputStream(new GZIPOutputStream(
		new BufferedOutputStream(new FileOutputStream(filename))));
      } else {
	out = new ObjectOutputStream(
		new BufferedOutputStream(new FileOutputStream(filename)));
      }
      out.writeObject(ld);
      out.close();
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }


  public static Lexicon deserializeLexicon(String filenameOrUrl) {
    LexiconData ld = null;
    try {
      ObjectInputStream in;
      InputStream is;
      if (filenameOrUrl.startsWith("http://")) {
	URL u = new URL(filenameOrUrl);
	URLConnection uc = u.openConnection();
	is = uc.getInputStream();
      } else {
	is = new FileInputStream(filenameOrUrl);
      }
      if (filenameOrUrl.endsWith(".gz")) {
	in = new ObjectInputStream(new GZIPInputStream(new BufferedInputStream(
					      is)));
      } else {
	in = new ObjectInputStream(new BufferedInputStream(is));
      }
      ld = (LexiconData) in.readObject();
      in.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    Numberer.setNumberers(ld.numberers);
    return ld.lexicon;
  }

}


class PCFGGrammar {
  BinaryGrammar bg;
  UnaryGrammar ug;

  public int numStates() { return stateNumberer().total(); }

  public int goalState() { return stateNumberer().number("ROOT"); }

  public boolean isTerminal(int state) {
    return Numberer.getGlobalNumberer("tags").hasSeen(stateNumberer().object(state));
  }

  public Iterator binaryRuleIteratorByLeftChild(int left) {
    return bg.ruleIteratorByLeftChild(left);
  }
  public Iterator binaryRuleIteratorByRightChild(int right) {
    return bg.ruleIteratorByRightChild(right);
  }
  public Iterator binaryRuleIteratorByParent(int parent) {
    return bg.ruleIteratorByParent(parent);
  }

  public Iterator closedUnaryRuleIteratorByParent(int parent) {
    return ug.closedRuleIteratorByParent(parent);
  }
  public Iterator closedUnaryRuleIteratorByChild(int child) {
    return ug.closedRuleIteratorByChild(child);
  }
  public Iterator closedUnaryRuleIterator() {
    return ug.closedRuleIterator();
  }

  public BinaryGrammar binaryGrammar() {
    return bg;
  }

  public UnaryGrammar unaryGrammar() {
    return ug;
  }

  public String stateSpace() {
    return bg.stateSpace();
  }

  public Numberer stateNumberer() {
    return Numberer.getGlobalNumberer(stateSpace());
  }

  public void writeTo(Writer out) throws IOException{
    out.write(stateNumberer().total()+"\n");
    // write binary rules
    for (Iterator brI = bg.ruleIterator(); brI.hasNext();) {
      BinaryRule br = (BinaryRule)brI.next();
      out.write("<"+stateNumberer().object(br.parent)+">"+
		" -> "+
		"<"+stateNumberer().object(br.leftChild)+">"+
		" "+
		"<"+stateNumberer().object(br.rightChild)+">"+
		" %% "+
		br.score()+
		"\n");
    }
    // write unary rules
    for (Iterator urI = ug.ruleIterator(); urI.hasNext();) {
      UnaryRule ur = (UnaryRule)urI.next();
      out.write("<"+stateNumberer().object(ur.parent)+">"+
		" -> "+
		"<"+stateNumberer().object(ur.child)+">"+
		" %% "+
		ur.score()+
		"\n");
    }
    out.flush();
  }

  PCFGGrammar(BinaryGrammar bg, UnaryGrammar ug) {
    this.bg = bg;
    this.ug = ug;
  }

  PCFGGrammar(Pair p) {
    this((BinaryGrammar)p.second, (UnaryGrammar)p.first);
  }

  PCFGGrammar(BufferedReader in) throws IOException {
    String str = in.readLine();
    int numStates = Integer.parseInt(str);
    bg = new BinaryGrammar(numStates);
    ug = new UnaryGrammar(numStates);
    // read and dispatch
    Pattern binaryRulePattern = Pattern.compile("<(.+)> -> <(.+)> <(.+)> %% (.+)");
    Pattern unaryRulePattern = Pattern.compile("<(.+)> -> <(.+)> %% (.+)");
    while (in.ready()) {
      str = in.readLine();
      Matcher binaryMatcher = binaryRulePattern.matcher(str);
      Matcher unaryMatcher = unaryRulePattern.matcher(str);
      if (binaryMatcher.matches()) {
	double score = Double.parseDouble(binaryMatcher.group(4));
	String parentStr = binaryMatcher.group(1);
	int parent = stateNumberer().number(parentStr);
	String leftStr = binaryMatcher.group(2);
	int left = stateNumberer().number(leftStr);
	String rightStr = binaryMatcher.group(3);
	int right = stateNumberer().number(rightStr);
	BinaryRule br = new BinaryRule();
	br.parent = parent;
	br.leftChild = left;
	br.rightChild = right;
	br.score = score;
	bg.addRule(br);
      } else if (unaryMatcher.matches()) {
	double score = Double.parseDouble(unaryMatcher.group(3));
	String parentStr = unaryMatcher.group(1);
	int parent = stateNumberer().number(parentStr);
	String childStr = unaryMatcher.group(2);
	int child = stateNumberer().number(childStr);
	UnaryRule ur = new UnaryRule();
	ur.parent = parent;
	ur.child = child;
	ur.score = score;
	ug.addRule(ur);
      }
    }
    //bg.splitRules();
    ug.purgeRules();
  }
}

class HeadlessEdge extends Edge {
  public Hook backHook;
  public boolean isEdge() { return true; }
  public double score() { return iScore+oScore; }
  public String toString() { return "Edge("+Numberer.getGlobalNumberer("states").object(state)+":"+start+"-"+end+")"; }
  public int hashCode() { return (state<<1)^(start<<17)^(end<<26); }
  public boolean equals(Object o) {
    if (this == o)
      return true;
    try {
      Item i = (Item)o;
      if (!i.isEdge())
	return false;
      Edge e = (Edge)o;
      if (state == e.state &&
	  start == e.start &&
	  end == e.end)
	return true;
      return false;
    } catch (ClassCastException e) {
      return false;
    }
  }
}

class HeadlessHook extends Hook {
  boolean isPreHook = false;
  public boolean isPreHook() { return isPreHook; }
  public boolean isPostHook() { return !isPreHook; }
  public double score() { return iScore+oScore; }
  public String toString() { return (isPreHook()?"Pre":"Post")+"Hook("+Numberer.getGlobalNumberer("states").object(state)+"|"+Numberer.getGlobalNumberer("states").object(subState)+":"+start+"-"+end+")"; }
  /*
    public int hashCode() { return 1+(state<<14)^(subState<<16)^(head<<22)^(start<<1)^(end<<7); }

    public boolean equals(Object o) {
    if (this == o)
    return true;
    return false;
    }
  */
}


class EdgeInterner {

  int length;
  int numStates;

  Edge[][][] edges;

  public Edge intern(Edge edge) {
    Edge current = edges[edge.start][edge.end][edge.state];
    if (current == null) {
      edges[edge.start][edge.end][edge.state] = edge;
      return edge;
    }
    return current;
  }

  public void clear() {
    for (int start=0; start<length; start++) {
      for (int end=1; end<=length; end++) {
	Arrays.fill(edges[start][end], null);
      }
    }
  }

  public EdgeInterner(int length, int numStates) {
    this.length = length;
    this.numStates = numStates;
    edges = new Edge[length][][];
    for (int start=0; start<length; start++) {
      edges[start] = new Edge[length+1][];
      for (int end=1; end<=length; end++) {
	edges[start][end] = new Edge[numStates];
      }
    }
  }
}


class ArrayChart {
  private List[][] preHooksByR;
  private List[][] postHooksByL;

  private List[][] edgesByL;
  private List[][] edgesByR;

  private PCFGGrammar grammar;
  private int length;

  private List empty = new ArrayList(0);

  public Collection getPreHooks(Edge edge) {
    List result = preHooksByR[edge.state][edge.end];
    return (result == null ? empty : result);
  }

  public Collection getPostHooks(Edge edge) {
    List result = postHooksByL[edge.state][edge.start];
    return (result == null ? empty : result);
  }

  public Collection getEdges(Hook hook) {
    List result = null;
    if (hook.isPreHook()) {
      result = edgesByR[hook.subState][hook.start];
    } else {
      result = edgesByL[hook.subState][hook.end];
    }
    return (result == null ? empty : result);
  }

  private void insert(List[][] map, int state, int loc, Object item) {
    List list = (List)map[state][loc];
    if (list == null) {
      list = new ArrayList();
      map[state][loc] = list;
    }
    list.add(item);
  }

  public void addEdge(Edge edge) {
    insert(edgesByL,edge.state, edge.start, edge);
    insert(edgesByR,edge.state, edge.end, edge);
  }

  public void addHook(Hook hook) {
    if (hook.isPreHook()) {
      insert(preHooksByR, hook.subState, hook.start, hook);
    } else {
      insert(postHooksByL, hook.subState, hook.end, hook);
    }
  }

  public ArrayChart(PCFGGrammar grammar, int length) {
    this.grammar = grammar;
    this.length = length;

    int numStates = grammar.stateNumberer().total();

    preHooksByR = new List[numStates][length+1];
    postHooksByL = new List[numStates][length+1];
    edgesByL = new List[numStates][length+1];
    edgesByR = new List[numStates][length+1];
  }
}

class PCFGParser {
  private static final boolean verbose = false;

  private ArrayChart chart;
  private Heap agenda;
  private int length;
  private int[] words;
  private Edge goal;
  private EdgeInterner interner;
  private Scorer scorer;

  private PCFGGrammar grammar;
  private BinaryGrammar bg;
  private UnaryGrammar ug;
  private Lexicon lex;
  private List[] taggedWordList;

  private Numberer wordNumberer = Numberer.getGlobalNumberer("words");
  private Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  private Numberer stateNumberer = Numberer.getGlobalNumberer("states");

  private TreeFactory tf = new LabeledScoredTreeFactory();

  // temp
  private long relaxHook1 = 0;
  private long relaxHook2 = 0;
  private long relaxHook3 = 0;
  private long relaxHook4 = 0;

  private long builtHooks = 0;
  private long builtEdges = 0;
  private long extractedHooks = 0;
  private long extractedEdges = 0;


  private double TOL = 1e-10;

  private boolean better(double x, double y) {
    if ((x-y)/(Math.abs(x)+Math.abs(y)+1e-100) > TOL)
      return true;
    return false;
  }

  private Tree extractParse(Edge edge) {
    //String head = (String)wordNumberer.object(words[edge.head]);
    //String tag = (String)tagNumberer.object(edge.tag);
    String state = (String)stateNumberer.object(edge.state);
    Label label = new StringLabel(state);//new CategoryWordTag(state,head,tag);
    if (edge.backEdge == null && edge.backHook == null) {
      // leaf, but needs word terminal
      List childList = new ArrayList(1);
      childList.add(tf.newLeaf(new StringLabel((String)wordNumberer.object(words[edge.start]))));//(head)));
      return tf.newTreeNode(label,childList);
    }
    List children = new ArrayList();
    if (edge.backHook == null) {
      // unary
      children.add(extractParse(edge.backEdge));
      return tf.newTreeNode(label,children);
    }
    // binary
    if (edge.backHook.isPreHook()) {
      children.add(extractParse(edge.backEdge));
      children.add(extractParse(edge.backHook.backEdge));
    } else {
      children.add(extractParse(edge.backHook.backEdge));
      children.add(extractParse(edge.backEdge));
    }
    return tf.newTreeNode(label,children);
  }

  public Tree getBestParse() {
    return extractParse(goal);
  }

  private Edge tempEdge = new HeadlessEdge();
  private void combine(Edge edge, Hook hook) {
    if (verbose) System.out.println("Combining: "+edge+" and "+hook);
    // make result edge
    if (hook.isPreHook()) {
      tempEdge.start = edge.start;
      tempEdge.end = hook.end;
    } else {
      tempEdge.start = hook.start;
      tempEdge.end = edge.end;
    }
    tempEdge.state = hook.state;
    //tempEdge.head = tempEdge.start; // DIFFERENT FROM BiLex PARSER
    //tempEdge.tag = 0; //hook.tag;
    tempEdge.iScore = hook.iScore+edge.iScore;
    tempEdge.backEdge = edge;
    tempEdge.backHook = hook;
    relaxTempEdge();
  }

  private void relaxTempEdge() {
    if (false) {
      if (tempEdge.iScore > scorer.iScore(tempEdge)+1e-4) {
	System.out.println(tempEdge+" has i "+tempEdge.iScore+" iE "+scorer.iScore(tempEdge));
      }
    }
    Edge resultEdge = (Edge)interner.intern(tempEdge);
    if (verbose) System.out.println("Formed "+resultEdge+" "+(resultEdge == tempEdge ? "new" : "old")+" "+tempEdge.iScore+" was "+resultEdge.iScore+" better? "+better(tempEdge.iScore, resultEdge.iScore));
    if (resultEdge == tempEdge) {
      //System.out.println(resultEdge+" was new.");
      tempEdge = new HeadlessEdge();
      discoverEdge(resultEdge);
    } else {
      if (better(tempEdge.iScore, resultEdge.iScore) && resultEdge.oScore > Double.NEGATIVE_INFINITY) {
	double back = resultEdge.iScore;
	Edge backE = resultEdge.backEdge;
	Hook backH = resultEdge.backHook;
	resultEdge.iScore = tempEdge.iScore;
	resultEdge.backEdge = tempEdge.backEdge;
	resultEdge.backHook = tempEdge.backHook;
	try {
	  agenda.decreaseKey(resultEdge);
	} catch(NullPointerException e) {
	  if (false) {
	    System.out.println("");
	    System.out.println("Old backEdge: "+backE+" i "+backE.iScore+" o "+backE.oScore+" s "+backE.score());
	    System.out.println("Old backEdge: "+backE+" iE "+scorer.iScore(backE));
	    System.out.println("Old backHook: "+backH+" i "+backH.iScore+" o "+backH.oScore+" s "+backH.score());
	    System.out.println("New backEdge: "+tempEdge.backEdge+" i "+tempEdge.backEdge.iScore+" o "+tempEdge.backEdge.oScore+" s "+tempEdge.backEdge.score());
	    System.out.println("New backEdge: "+tempEdge.backEdge+" iE "+scorer.iScore(tempEdge.backEdge));
	    System.out.println("New backHook: "+tempEdge.backHook+" i "+tempEdge.backHook.iScore+" o "+tempEdge.backHook.oScore+" s "+tempEdge.backHook.score());
	    System.out.println("ERROR: Formed "+resultEdge+" i "+tempEdge.iScore+" o "+resultEdge.oScore+" s "+resultEdge.score());
	    System.out.println("ERROR: Formed "+resultEdge+" "+(resultEdge == tempEdge ? "new" : "old")+" "+tempEdge.iScore+" was "+back+" better? "+better(tempEdge.iScore, back));
	  }
	}
      }
    }
  }
  private void discoverEdge(Edge edge) {
    // create new edge
    edge.oScore = scorer.oScore(edge);
    if (edge.oScore > Double.NEGATIVE_INFINITY) {
      agenda.add(edge);
      builtEdges++;
    }
  }
  /*
  private void discoverHook(Hook hook) {
    hook.oScore = buildOScore(hook);
    if (hook.oScore == Double.NEGATIVE_INFINITY)
      relaxHook4++;
    builtHooks++;
    agenda.add(hook);
  }
  private Edge iTemp = new Edge();
  private Edge oTemp = new Edge();
  private double buildOScore(Hook hook) {
    double bestOScore = Double.NEGATIVE_INFINITY;
    iTemp.head = 0;//hook.head;
    iTemp.tag = 0;//hook.tag;
    iTemp.state = hook.subState;
    oTemp.head = 0;//hook.head;
    oTemp.tag = 0;//hook.tag;
    oTemp.state = hook.state;
    if (hook.isPreHook()) {
      iTemp.end = hook.start;
      oTemp.end = hook.end;
      for (int start = 0; start < hook.start; start++) {
	iTemp.start = start;
	oTemp.start = start;
	double oScore = scorer.oScore(oTemp)+scorer.iScore(iTemp);
	//System.out.println("Score for "+hook+" is i "+iTemp+" ("+scorer.iScore(iTemp)+") o "+oTemp+" ("+scorer.oScore(oTemp)+")");
	bestOScore = SloppyMath.max(bestOScore,oScore);
      }
    } else {
      iTemp.start = hook.end;
      oTemp.start = hook.start;
      for (int end = hook.end; end <= length; end++) {
	iTemp.end = end;
	oTemp.end = end;
	double oScore = scorer.oScore(oTemp)+scorer.iScore(iTemp);
	bestOScore = SloppyMath.max(bestOScore,oScore);
      }
    }
    return bestOScore;
  }
  */
  private HeadlessHook tempHook = new HeadlessHook();
  private void projectHooks(Edge edge) {
    // form hooks
    // POST HOOKS
    for (Iterator rI = bg.ruleIteratorByLeftChild(edge.state); rI.hasNext(); ) {
      BinaryRule br = (BinaryRule)rI.next();
      //	  ! scorer.iPossibleL(project(br.rightChild),edge.end))
      //	if (! Test.exhaustiveTest)
      //	  continue;
      //for (int head=edge.end; head<length; head++) {
      //for (int head = 0; head < 1; head++) {
	//for (Iterator iTWI=taggedWordList[head].iterator(); iTWI.hasNext();) {
	//IntTaggedWord iTW = (IntTaggedWord)iTWI.next();
	//int tag = iTW.tag;
	//int word = iTW.word;
	  tempHook.start = edge.start;
	  tempHook.end = edge.end;
	  //tempHook.head = edge.end;//head;
	  //tempHook.tag = 0;//tag;
	  tempHook.isPreHook = false;
	  tempHook.state = br.parent;
	  tempHook.subState = br.rightChild;
	  if (! scorer.oPossible(tempHook))
	    continue;
	  //if (! chart.isBuiltL(tempHook.subState,tempHook.end,tempHook.head,tempHook.tag))
	  //  continue;
	  tempHook.iScore = edge.iScore+br.score;//+depScore[binDistance[head][edge.end]][head][tag][edge.head][edge.tag]+
	  //headStopL[edge.head][edge.tag][edge.start]+
	  //headStopR[edge.head][edge.tag][edge.end];
	  tempHook.backEdge = edge;
	  chart.addHook(tempHook);
	  builtHooks++;
	  combineHook(tempHook);
	  tempHook = new HeadlessHook();
	  //relaxTempHook();
	  //}
	  //}
    }
    // PRE HOOKS
    for (Iterator rI = bg.ruleIteratorByRightChild(edge.state); rI.hasNext(); ) {
      BinaryRule br = (BinaryRule)rI.next();
      //	  ! fscorer.iPossibleR(project(br.leftChild),edge.start))
      //	if (! Test.exhaustiveTest)
      //	  continue;
      //for (int head=0; head<edge.start; head++) {
      //for (int head = 0; head < 1; head++) {
	//for (Iterator iTWI=taggedWordList[head].iterator(); iTWI.hasNext();) {
	//IntTaggedWord iTW = (IntTaggedWord)iTWI.next();
	//int tag = iTW.tag;
	//int word = iTW.word;
	  tempHook.start = edge.start;
	  tempHook.end = edge.end;
	  //tempHook.head = 0;//head;
	  //tempHook.tag = 0;//tag;
	  tempHook.isPreHook = true;
	  tempHook.state = br.parent;
	  tempHook.subState = br.leftChild;
	  if (! scorer.oPossible(tempHook))
	    continue;
	  //if (! chart.isBuiltR(tempHook.subState,tempHook.start,tempHook.head,tempHook.tag))
	  // continue;
	  tempHook.iScore = edge.iScore+br.score;//+depScore[binDistance[head][edge.start]][head][tag][edge.head][edge.tag]+
	  //headStopL[edge.head][edge.tag][edge.start]+
	  //headStopR[edge.head][edge.tag][edge.end];
	  tempHook.backEdge = edge;
	  chart.addHook(tempHook);
	  builtHooks++;
	  combineHook(tempHook);
	  tempHook = new HeadlessHook();
	  //relaxTempHook();
	  //}
    }
    //}
  }
  /*
  private void registerReal(Edge real) {
    chart.registerRealEdge(real);
  }
  private void triggerHooks(Edge edge) {
    // we might have built a synth edge, enabling some old real edges to project hooks
    boolean newL = ! chart.isBuiltL(edge.state,edge.start,edge.head,edge.tag);
    boolean newR = ! chart.isBuiltR(edge.state,edge.end,edge.head,edge.tag);
    //if (newL)
    //  System.out.println("Triggering on L: "+edge);
    //if (newR)
    //  System.out.println("Triggering on R: "+edge);
    chart.registerEdgeIndexes(edge);
    if (newR) {
      // PRE HOOKS
      BinaryRule[] rules = bg.splitRulesWithLC(edge.state);
      for (int i=0; i<rules.length; i++) {
	BinaryRule br = rules[i];
	Collection realEdges = chart.getRealEdgesWithL(br.rightChild,edge.end);
	for (Iterator realI = realEdges.iterator(); realI.hasNext();) {
	  Edge real = (Edge)realI.next();
	  tempHook.start = real.start;
	  tempHook.end = real.end;
	  tempHook.state = br.parent;
	  tempHook.subState = br.leftChild;
	  tempHook.head = 0;//edge.head;
	  tempHook.tag = 0;//edge.tag;
	  tempHook.backEdge = real;
	  tempHook.iScore = real.iScore+br.score;//+depScore[binDistance[edge.head][edge.end]][edge.head][edge.tag][real.head][real.tag]+
	  //headStopL[real.head][real.tag][real.start]+
	  //headStopR[real.head][real.tag][real.end];
	  relaxTempHook();
	}
      }
    }
    if (newL) {
      // POST HOOKS
      BinaryRule[] rules = bg.splitRulesWithRC(edge.state);
      for (int i=0; i<rules.length; i++) {
	BinaryRule br = rules[i];
	Collection realEdges = chart.getRealEdgesWithR(br.leftChild,edge.start);
	for (Iterator realI = realEdges.iterator(); realI.hasNext();) {
	  Edge real = (Edge)realI.next();
	  tempHook.start = real.start;
	  tempHook.end = real.end;
	  tempHook.state = br.parent;
	  tempHook.subState = br.rightChild;
	  tempHook.head = 0;//edge.head;
	  tempHook.tag = 0;//edge.tag;
	  tempHook.backEdge = real;
	  tempHook.iScore = real.iScore+br.score;//+depScore[binDistance[edge.head][edge.start]][edge.head][edge.tag][real.head][real.tag]+
	  //headStopL[real.head][real.tag][real.start]+
	  //headStopR[real.head][real.tag][real.end];
	  relaxTempHook();
	}
      }
    }
  }

  private void relaxTempHook() {
    relaxHook1++;
    if (! Test.exhaustiveTest)
      if (! scorer.oPossible(tempHook) || ! scorer.iPossible(tempHook))
	return;
    relaxHook2++;
    Hook resultHook = (Hook)interner.intern(tempHook);
    if (verbose) System.out.println("Formed "+resultHook+" "+(resultHook == tempHook ? "new" : "old")+" "+tempHook.iScore+" was "+resultHook.iScore);
    if (resultHook == tempHook) {
      relaxHook3++;
      tempHook = new Hook();
      discoverHook(resultHook);
    }
    if (better(tempHook.iScore, resultHook.iScore)) {
      resultHook.iScore = tempHook.iScore;
      resultHook.backEdge = tempHook.backEdge;
      try {
	agenda.decreaseKey(resultHook);
      } catch (NullPointerException e) {
      }
    }
  }
  */

  private void projectUnaries(Edge edge) {
    for (Iterator rI = ug.ruleIteratorByChild(edge.state); rI.hasNext(); ) {
      UnaryRule ur = (UnaryRule)rI.next();
      if (ur.child == ur.parent)
	continue;
      tempEdge.start = edge.start;
      tempEdge.end = edge.end;
      //tempEdge.head = edge.start;//edge.head;
      //tempEdge.tag = 0;//edge.tag;
      tempEdge.state = ur.parent;
      tempEdge.backEdge = edge;
      tempEdge.backHook = null;
      tempEdge.iScore = edge.iScore+ur.score;
      relaxTempEdge();
    }
  }

  private void combineEdge(Edge edge) {
    // fetch existing hooks that can combine with it
    Collection preHooks = chart.getPreHooks(edge);
    Collection postHooks = chart.getPostHooks(edge);
    // combine them
    for (Iterator preI=preHooks.iterator(); preI.hasNext();) {
      Hook hook = (Hook)preI.next();
      combine(edge,hook);
    }
    for (Iterator postI=postHooks.iterator(); postI.hasNext();) {
      Hook hook = (Hook)postI.next();
      combine(edge,hook);
    }
  }

  private void combineHook(Hook hook) {
    Collection edges = chart.getEdges(hook);
    // combine them
    for (Iterator edgeI=edges.iterator(); edgeI.hasNext();) {
      Edge edge = (Edge)edgeI.next();
      combine(edge,hook);
    }
  }

  private void processEdge(Edge edge) {
    // add to chart
    //if (verbose) System.out.println("Adding to chart: "+edge);
    if (! bg.isSynthetic(edge.state)) {
      chart.addEdge(edge);
      combineEdge(edge);
    }      
    // do projections
    //if (verbose) System.out.println("Projecting: "+edge);
    if (bg.isSynthetic(edge.state)) {
      projectHooks(edge);
    } else {
      projectUnaries(edge);
      //registerReal(edge);
    }
    //triggerHooks(edge);
  }

  /*
  private void processHook(Hook hook) {
    // add to chart
    //if (verbose) System.out.println("Adding to chart: "+hook);
    chart.addHook(hook);
    Collection edges = chart.getEdges(hook);
    for (Iterator edgeI=edges.iterator(); edgeI.hasNext();) {
      Edge edge = (Edge)edgeI.next();
      combine(edge,hook);
    }
  }
  private void processItem(Item item) {
    if (item.isEdge())
      processEdge((Edge)item);
    else
      processHook((Hook)item);
  }
  private void discoverItem(Item item) {
    if (item.isEdge())
      discoverEdge((Edge)item);
    else
      discoverHook((Hook)item);
  }
  */
  private Item makeInitialItem(int pos, int state, double iScore) {
    Edge edge = new HeadlessEdge();
    edge.start = pos;
    edge.end = pos+1;
    edge.state = state;
    //edge.head = pos;//pos;
    //edge.tag = 0;//tag;
    edge.iScore = iScore;
    return edge;
  }
  private List makeInitialItems(List wordList) {
    List itemList = new ArrayList();
    int length = wordList.size();
    int numTags = tagNumberer.total();
    words = new int[length];
    taggedWordList = new List[length];
    int terminalCount = 0;
    for (int i=0; i<length; i++) {
      taggedWordList[i] = new ArrayList(numTags);
      String wordStr = wordList.get(i).toString();
      //if (! wordNumberer.hasSeen(wordStr))
      //	wordStr = Lexicon.UNKNOWN_WORD;
      int word = wordNumberer.number(wordStr);
      words[i] = word;
      for (Iterator tagI = lex.ruleIteratorByWord(word,i); tagI.hasNext(); ) {
	IntTaggedWord tagging = (IntTaggedWord)tagI.next();
	int tag = tagging.tag;
	int state = Numberer.translate("tags","states",tag);
	//itemList.add(makeInitialItem(i,tag,state,1.0*tagging.score));
	// THIS WILL CAUSE BUGS!!!  Don't use with another A* scorers
	tempEdge.state = state;
	//tempEdge.head = i;//i;
	tempEdge.start = i;
	tempEdge.end = i+1;
	//tempEdge.tag = 0;//tag;
	itemList.add(makeInitialItem(i,state,lex.score(tagging, i)));//scorer.iScore(tempEdge)));
	terminalCount++;
	taggedWordList[i].add(new IntTaggedWord(word,tag));
      }
    }
    if (Test.verbose) {
      System.out.println("Terminals:"+terminalCount);
    }
    return itemList;
  }

  private void setGoal(int length) {
    goal = new HeadlessEdge();
    goal.start = 0;
    goal.end = length;
    goal.state = stateNumberer.number("ROOT");
    //goal.tag = 0;//tagNumberer.number(Lexicon.BOUNDARY_TAG);
    //goal.head = 0;//length-1;
    //goal = (Edge)interner.intern(goal);
  }
  private void initialize(List words) {
    length = words.size();
    //interner = new Interner();
    agenda = new ArrayHeap(new ScoredComparator(ScoredComparator.DESCENDING));
    chart = new ArrayChart(grammar, words.size());
    setGoal(length);
    List initialItems = makeInitialItems(words);
    //scoreDependencies();
    for (int i=0; i<initialItems.size(); i++) {
      Edge edge = (Edge)initialItems.get(i);
      edge = (Edge)interner.intern(edge);
      //if (verbose) System.out.println("Initial: "+item);
      discoverEdge(edge);
    }
  }

  public boolean parse(List words) {
    scorer.parse(words);
    int spanFound = 0;
    long last = 0;
    int exHook = 0;
    relaxHook1 = 0;
    relaxHook2 = 0;
    relaxHook3 = 0;
    relaxHook4 = 0;
    builtHooks = 0;
    builtEdges = 0;
    extractedHooks = 0;
    extractedEdges = 0;
    if (Test.verbose) {
      Timing.tick("Starting combined parse.");
    }
    initialize(words);
    while (!agenda.isEmpty()) {
      Item item = (Item)agenda.extractMin();
      if (!item.isEdge()) {
	exHook++;
	extractedHooks++; 
      } else {
	extractedEdges++;
      }
      if (relaxHook1 > last + 1000000) {
	last = relaxHook1;
	if (Test.verbose) {
	  System.out.println("Proposed hooks:   "+relaxHook1);
	  System.out.println("Unfiltered hooks: "+relaxHook2);
	  System.out.println("Built hooks:      "+relaxHook3);
	  System.out.println("Waste hooks:      "+relaxHook4);
	  System.out.println("Extracted hooks:  "+exHook);
	  System.out.println("Built edges:      "+builtEdges);
	  System.out.println("Extracted edges:  "+extractedEdges);
	}
      }
      if (item.end-item.start > spanFound) {
	spanFound = item.end-item.start;
	if (Test.verbose) {
	  System.out.print(spanFound+" ");
	  //System.out.println(item);
	}
      }
      //if (item.end < 5) System.out.println("Extracted: "+item+" iScore "+item.iScore+" oScore "+item.oScore+" score "+item.score());
      if (item.equals(goal)) {
	if (Test.verbose) {
	  System.out.println("Found goal!");
	  //System.out.println("Comb iScore "+goal.iScore);
	  Timing.tick("Done, parse found.");
	  System.out.println("Built items:      "+(builtEdges+builtHooks));
	  System.out.println("Built hooks:      "+builtHooks);
	  System.out.println("Built edges:      "+builtEdges);
	  System.out.println("Extracted items:  "+(extractedEdges+extractedHooks));
	  System.out.println("Extracted hooks:  "+extractedHooks);
	  System.out.println("Extracted edges:  "+extractedEdges);
	  //postMortem();
	}
	goal = (Edge)item;
	interner.clear();
	agenda = null;
	return true;
      }
      if (item.score() == Double.NEGATIVE_INFINITY) {
	Timing.tick("Hit A*-blocked edges, failing.");
	return false;
      }
      if (verbose && item.score() != Double.NEGATIVE_INFINITY)
	System.out.println("Removing from agenda: "+item+" score i "+item.iScore+" + o "+item.oScore+" = "+item.score());
      if (verbose && item.score() != Double.NEGATIVE_INFINITY) System.out.println("Backtrace: "+(item.backEdge != null ? item.backEdge.toString() : "null")+" "+(item.isEdge() ? (((Edge)item).backHook != null ? ((Edge)item).backHook.toString() : "") : ""));
      processEdge((Edge)item);
    }
    Timing.tick("Done, parse NOT found.");
    return false;
  }

  private void postMortem() {
    int numHooks = 0;
    int numEdges = 0;
    int numUnmatchedHooks = 0;
    int total = agenda.size();
    int done = 0;
    while (! agenda.isEmpty()) {
      Item item = (Item)agenda.extractMin();
      done++;
      //if(done % (total/10) == 0)
      //	System.out.println("Scanning: "+100*done/total);
      if (item.isEdge()) {
	numEdges++;
      } else {
	numHooks++;
	Collection edges = chart.getEdges((Hook)item);
	if (edges.size() == 0)
	  numUnmatchedHooks++;
      }
    }
    System.out.println("--- Agenda Post-Mortem ---");
    System.out.println("Edges:           "+numEdges);
    System.out.println("Hooks:           "+numHooks);
    System.out.println("Unmatched Hooks: "+numUnmatchedHooks);
  }

  //protected int project(int state) {
  //  return projection.project(state);
  //}

  PCFGParser(Scorer scorer, PCFGGrammar grammar, Lexicon lex) {
    this.scorer = scorer;
    this.grammar = grammar;
    this.bg = grammar.binaryGrammar();
    this.ug = grammar.unaryGrammar();
    this.lex = lex;
    int length = 41;
    int numTags = Numberer.getGlobalNumberer("tags").total();
    interner = new EdgeInterner(length, grammar.stateNumberer().total());
  }
}


class AStarPCFGParser {

  public static void main(String[] args) throws IOException, 
					        FileNotFoundException {
    boolean trainMode = false;
    boolean testMode = false;
    String path = "/dfs/ah/1/tmp/klein/corpora/Treebank3/parsed/mrg/wsj";
    String grammarFile = "";
    String lexFile = "";
    int high = 0;
    int low = 0;
    int length = 40;
    int num = 1000000;
    try {
      int pos = 0;
      while (pos < args.length) {
	if (args[pos].equals("-train")) {
	  trainMode = true;
	  grammarFile = args[pos+1];
	  lexFile = args[pos+2];
	  pos+=3;
	  low = 200;
	  high = 2172;
	  continue;
	}
	if (args[pos].equals("-test")) {
	  testMode = true;
	  grammarFile = args[pos+1];
	  lexFile = args[pos+2];
	  pos+=3;
	  low = 2300;
	  high = 2399;
	  continue;
	}
	if (args[pos].equals("-path")) {
	  path = args[pos+1];
	  pos+=2;
	  continue;
	}
	if (args[pos].equals("-length")) {
	  length = Integer.parseInt(args[pos+1]);
	  pos+=2;
	  continue;
	}
	if (args[pos].equals("-num")) {
	  num = Integer.parseInt(args[pos+1]);
	  pos+=2;
	  continue;
	}
	if (args[pos].equals("-range")) {
	  low = Integer.parseInt(args[pos+1]);
	  high = Integer.parseInt(args[pos+2]);
	  pos+=3;
	  continue;
	}
	throw new IllegalArgumentException();
      }	
      if (trainMode == testMode) {
	throw new IllegalArgumentException();
      }
    } catch (Exception e) {
      System.err.println("Usage: AStarPCFGParser [(-train | -test) grammarFile lexiconFile] [-range low high] [-path path] [-length maxLength] [-num maxNum]");
      System.exit(0);
    }

    Treebank treebank = new MemoryTreebank(new TreeReaderFactory() {
	public TreeReader newTreeReader(Reader in) {
	  return new PennTreeReader(in, 
				new LabeledScoredTreeFactory(
			        new StringLabelFactory()),
				new BobChrisTreeNormalizer());
	    }
      });

    System.out.println("Currently "+new Date());
    System.out.print("Reading trees...");
    treebank.loadPath(path, new NumberRangeFileFilter(low, high, true));
    Timing.tick("done.");
    if (trainMode) {
      train(treebank, grammarFile, lexFile);
    }
    if (testMode) {
      test(treebank, grammarFile, lexFile, length, num);
    }
  }

  protected static Collection binarize(Collection trees) {
    TreeTransformer binarizer = GLCBinarizerFactory.getGLCHeadRespectingBinarizer(new LeftHeadFinder(), 4);
    List binaryTrees = new ArrayList();
    for (Iterator tI=trees.iterator(); tI.hasNext();) {
      Tree tree = (Tree)tI.next();
      Tree binaryTree = binarizer.transformTree(tree, null);
      binaryTrees.add(binaryTree);
    }
    return binaryTrees;
  }

  protected static void train(Collection trees, String grammarFile, String lexFile) throws IOException {
    System.out.print("Binarizing trees...");
    Collection binaryTrees = binarize(trees);
    Timing.tick("done.");

    // extract grammars
    Extractor pcfgExtractor = new SmoothedGrammarAndSimpleLexiconExtractor();
    System.out.print("Extracting Grammar and Lexicon...");
    Pair gramLex = (Pair)pcfgExtractor.extract(binaryTrees);
    PCFGGrammar grammar = (PCFGGrammar)gramLex.first;
    SimpleLexicon lex = (SimpleLexicon)gramLex.second;
    Timing.tick("done.");

    System.out.print("Writing lexicon...");
    lex.writeTo(new BufferedWriter(new FileWriter(lexFile)));
    Timing.tick("done.");
    System.out.print("Writing grammar...");
    grammar.writeTo(new BufferedWriter(new FileWriter(grammarFile)));
    Timing.tick("done.");
  }

  protected static void test(Collection trees, String grammarFile, String lexFile, int length, int maxNum) throws IOException {
    System.out.print("Binarizing trees...");
    Collection binaryTrees = binarize(trees);
    Timing.tick("done.");

    System.out.print("Reading lexicon...");
    Lexicon lexicon = new SimpleLexicon(new BufferedReader(new FileReader(lexFile)));
    Timing.tick("done.");
    System.out.print("Reading pcfg...");
    PCFGGrammar grammar = new PCFGGrammar(new BufferedReader(new FileReader(grammarFile)));
    Timing.tick("done.");

    System.out.print("Building astar scorer...");
    Scorer aStarScorer = buildScorer(grammar, lexicon, length);
    Timing.tick("done.");

    System.out.print("Testing...");
    PCFGParser parser = new PCFGParser(aStarScorer, grammar, lexicon);
    LabeledConstituentEval pcfgPE = new LabeledConstituentEval("pcfg  PE");
    int count = 0;

    if (Test.evalb)
      EvalB.initEVALBfiles();

    int num = 0;
    for (Iterator tI = binaryTrees.iterator(); tI.hasNext();) {
      Tree tree = (Tree)tI.next();
      count++;
      int treeLength = tree.yield().size()-1;
      if (treeLength > length)
	continue;
      num++;
      if (num > maxNum)
	continue;
      System.out.println("-------------------------------------");
      System.out.println("Number: "+count);
      System.out.println("Length: "+treeLength);
      long timeMil1 = System.currentTimeMillis();
      parser.parse(tree.yield());
      long timeMil2 = System.currentTimeMillis();
      long elapsed = timeMil2-timeMil1;
      System.out.println("Time: "+(int)elapsed/1000+" sec.");
      Tree tree1 = parser.getBestParse();
      Tree tree1db = BinarizerFactory.getDebinarizer().transformTree(tree1,null);
      Tree treedb = BinarizerFactory.getDebinarizer().transformTree(tree, null);
      tree1db.pennPrint();
      treedb.pennPrint();
      TreeTransformer tc = new TreeCollinizer();
      TreeTransformer tcEvalb = new TreeCollinizer(false);
      pcfgPE.evaluate(tc.transformTree(tree1db, null),
		      tc.transformTree(treedb, null));

      if (Test.evalb) {
	EvalB.writeEVALBline(tcEvalb.transformTree(tree1db, null),
		       tcEvalb.transformTree(treedb, null));
      }
    }

    if (Test.evalb)
      EvalB.closeEVALBfiles();

    System.out.println("ParseEval for PCFG:");
    pcfgPE.display(false);

    Timing.tick("done.");
  }

  protected static Scorer buildScorer(PCFGGrammar grammar, Lexicon lex, int length) {
    //return new NullScorer();
    //return new FilterScorer(lex, grammar);
    //return new TwinScorer(new SXOScorer(grammar, length), new TagScorer(lex));
    return new MaxScorer(new TwinScorer(new SXOScorer(grammar, length), new TagScorer(lex)), new FilterScorer(lex, grammar));
  }

}
