package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.trees.international.*;
import edu.stanford.nlp.util.StringUtils;

import java.io.*;
import java.util.*;


/**
 * Parameter file for parsing the Penn Chinese Treebank.  Includes
 * category enrichments specific to the Penn Chinese Treebank.
 *
 * @author Roger Levy
 */

public class ChineseTreebankParserParams implements TreebankLangParserParams,
						    Serializable {

  private ChineseTreebankLanguagePack tlp = new ChineseTreebankLanguagePack();
  private String inputEncoding = tlp.getEncoding();
  private String outputEncoding = tlp.getEncoding();


  public void setInputEncoding(String encoding) {
    inputEncoding = encoding;
  }

  public void setOutputEncoding(String encoding) {
    outputEncoding = encoding;
  }

  public HeadFinder headFinder() {
    return new ChineseHeadFinder();
  }

  public Lexicon lex() {
    return new ChineseLexicon();
  }

  public MemoryTreebank memoryTreebank() {
    final TreeNormalizer tn = new BobChrisTreeNormalizer();

    return new MemoryTreebank(new TreeReaderFactory() {
	public TreeReader newTreeReader(Reader in) {
	  return new PennTreeReader(in, 
				new LabeledScoredTreeFactory(
			        new StringLabelFactory()),
				tn,
				new CHTBTokenizer(in));
	    }
      },
			      inputEncoding);
  }

  /** returns a MemoryTreebank appropriate to the testing treebank source */
  public MemoryTreebank testMemoryTreebank() {
    return memoryTreebank();
  }


  /** the tree transformer used to produce trees for evaluation.  Will be
   *  applied both to the
   */
  public TreeTransformer collinizer() {
    return new ChineseCollinizer();
  }
  public TreeTransformer collinizerEvalb() {
    return new ChineseCollinizer(false);
  }

  /** Returns a <code>ChineseTreebankLanguagePack</code> */
  public TreebankLanguagePack treebankLanguagePack() {
    return new ChineseTreebankLanguagePack();
  }

  /** the PrintWriter used to print output. It's the responsibility of
   * pw to deal properly with character encodings for the relevant
   * treebank */
  public PrintWriter pw() {
    return pw(System.out);
  }

  /** the PrintWriter used to print output. It's the responsibility of
   * pw to deal properly with character encodings for the relevant
   * treebank */
  public PrintWriter pw(OutputStream o) {

    try{
      return new PrintWriter(new OutputStreamWriter(o,
						    outputEncoding), true);
    } catch(UnsupportedEncodingException e) {
      System.out.println("Warning: desired encoding not accepted: "  +  e);
      return new PrintWriter(o,true);
    }
  }

  // Automatically generated by ParentAnnotationStats -- preferably don't edit
  private static String[] splitters1 = new String[] {"VA^VCD", "NP^NP", "NP^VP", "NP^IP", "NP^DNP", "NP^PP", "NP^LCP", "NP^PRN", "NP^QP", "PP^IP", "PP^NP", "NN^FRAG", "NN^NP", "NT^FRAG", "NT^NP", "NR^FRAG", "NR^NP", "VV^FRAG", "VV^VRD", "VV^VCD", "VV^VP", "VV^VSB", "VP^VP", "VP^IP", "VP^DVP", "IP^ROOT", "IP^IP", "IP^CP", "IP^VP", "IP^PP", "IP^NP", "IP^LCP", "CP^IP", "QP^NP", "QP^PP", "QP^VP", "ADVP^CP", "CC^VP", "CC^NP", "CC^IP", "CC^QP", "PU^NP", "PU^FRAG", "PU^IP", "PU^VP", "PU^PRN", "PU^QP", "PU^LST", "NP^DNP~QP", "NT^NP~NP", "NT^NP~VP", "NT^NP~IP", "NT^NP~LCP", "NT^NP~PP", "NT^NP~PRN", "NT^NP~QP", "NT^NP~DNP", "NP^NP~VP", "NP^NP~NP", "NP^NP~IP", "NP^NP~PP", "NP^NP~DNP", "NP^NP~LCP", "NN^NP~VP", "NN^NP~IP", "NN^NP~NP", "NN^NP~PP", "NN^NP~DNP", "NN^NP~LCP", "NN^NP~UCP", "NN^NP~QP", "NN^NP~PRN", "M^CLP~DP", "M^CLP~QP", "M^CLP~NP", "M^CLP~CLP", "CD^QP~VP", "CD^QP~NP", "CD^QP~QP", "CD^QP~LCP", "CD^QP~PP", "CD^QP~DNP", "CD^QP~DP", "CD^QP~IP", "IP^IP~IP", "IP^IP~ROOT", "IP^IP~VP", "LC^LCP~PP", "LC^LCP~IP", "NP^VP~IP", "NP^VP~VP", "AD^ADVP~IP", "AD^ADVP~QP", "AD^ADVP~VP", "AD^ADVP~NP", "AD^ADVP~PP", "AD^ADVP~ADVP", "NP^IP~ROOT", "NP^IP~IP", "NP^IP~CP", "NP^IP~VP", "DT^DP~PP", "P^PP~IP", "P^PP~NP", "P^PP~VP", "P^PP~DNP", "VV^VP~IP", "VV^VP~VP", "PU^IP~IP", "PU^IP~VP", "PU^IP~ROOT", "PU^IP~CP", "JJ^ADJP~DNP", "JJ^ADJP~ADJP", "NR^NP~IP", "NR^NP~NP", "NR^NP~PP", "NR^NP~VP", "NR^NP~DNP", "NR^NP~LCP", "NR^NP~PRN", "NP^PP~NP", "NP^PP~IP", "NP^PP~DNP", "VA^VP~VP", "VA^VP~IP", "VA^VP~DVP", "VP^VP~VP", "VP^VP~IP", "VP^VP~DVP", "VP^IP~ROOT", "VP^IP~CP", "VP^IP~IP", "VP^IP~VP", "VP^IP~PP", "VP^IP~LCP", "VP^IP~NP", "PN^NP~NP", "PN^NP~IP", "PN^NP~PP"};
  private static String[] splitters2 = new String[] {"VA^VCD", "NP^NP", "NP^VP", "NP^IP", "NP^DNP", "NP^PP", "NP^LCP", "NN^FRAG", "NN^NP", "NT^FRAG", "NT^NP", "NR^FRAG", "NR^NP", "VV^FRAG", "VV^VRD", "VV^VCD", "VV^VP", "VV^VSB", "VP^VP", "VP^IP", "VP^DVP", "IP^ROOT", "IP^IP", "IP^CP", "IP^VP", "IP^PP", "CP^IP", "ADVP^CP", "CC^VP", "CC^NP", "PU^NP", "PU^FRAG", "PU^IP", "PU^VP", "PU^PRN", "NT^NP~NP", "NT^NP~VP", "NT^NP~IP", "NT^NP~LCP", "NT^NP~PP", "NP^NP~VP", "NP^NP~NP", "NP^NP~IP", "NP^NP~PP", "NP^NP~DNP", "NN^NP~VP", "NN^NP~IP", "NN^NP~NP", "NN^NP~PP", "NN^NP~DNP", "NN^NP~LCP", "NN^NP~UCP", "NN^NP~QP", "NN^NP~PRN", "M^CLP~DP", "CD^QP~VP", "CD^QP~NP", "CD^QP~QP", "CD^QP~LCP", "CD^QP~PP", "CD^QP~DNP", "CD^QP~DP", "LC^LCP~PP", "NP^VP~IP", "NP^VP~VP", "AD^ADVP~IP", "AD^ADVP~QP", "AD^ADVP~VP", "AD^ADVP~NP", "NP^IP~ROOT", "NP^IP~IP", "NP^IP~CP", "NP^IP~VP", "P^PP~IP", "P^PP~NP", "P^PP~VP", "P^PP~DNP", "VV^VP~IP", "VV^VP~VP", "PU^IP~IP", "PU^IP~VP", "PU^IP~ROOT", "PU^IP~CP", "JJ^ADJP~DNP", "NR^NP~IP", "NR^NP~NP", "NR^NP~PP", "NR^NP~VP", "NR^NP~DNP", "NR^NP~LCP", "NP^PP~NP", "VA^VP~VP", "VA^VP~IP", "VP^VP~VP", "VP^IP~ROOT", "VP^IP~CP", "VP^IP~IP", "VP^IP~VP", "VP^IP~PP", "VP^IP~LCP", "VP^IP~NP", "PN^NP~NP"};
  private static String[] splitters3 = new String[] {"NP^NP", "NP^VP", "NP^IP", "NP^DNP", "NP^PP", "NP^LCP", "NN^FRAG", "NN^NP", "NT^FRAG", "NR^FRAG", "NR^NP", "VV^FRAG", "VV^VRD", "VV^VCD", "VV^VP", "VV^VSB", "VP^VP", "VP^IP", "IP^ROOT", "IP^IP", "IP^CP", "IP^VP", "PU^NP", "PU^FRAG", "PU^IP", "PU^VP", "PU^PRN", "NP^NP~VP", "NN^NP~VP", "NN^NP~IP", "NN^NP~NP", "NN^NP~PP", "NN^NP~DNP", "NN^NP~LCP", "M^CLP~DP", "CD^QP~VP", "CD^QP~NP", "CD^QP~QP", "AD^ADVP~IP", "AD^ADVP~QP", "AD^ADVP~VP", "P^PP~IP", "VV^VP~IP", "VV^VP~VP", "PU^IP~IP", "PU^IP~VP", "NR^NP~IP", "NR^NP~NP", "NR^NP~PP", "NR^NP~VP", "VP^VP~VP", "VP^IP~ROOT", "VP^IP~CP", "VP^IP~IP", "VP^IP~VP"};
  private static String[] splitters4 = new String[] {"NP^NP", "NP^VP", "NP^IP", "NN^FRAG", "NT^FRAG", "NR^FRAG", "VV^FRAG", "VV^VRD", "VV^VCD", "VP^VP", "VP^IP", "IP^ROOT", "IP^IP", "IP^CP", "IP^VP", "PU^NP", "PU^FRAG", "PU^IP", "PU^VP", "NN^NP~VP", "NN^NP~IP", "NN^NP~NP", "NN^NP~PP", "NN^NP~DNP", "NN^NP~LCP", "CD^QP~VP", "CD^QP~NP", "AD^ADVP~IP", "VV^VP~IP", "VV^VP~VP", "NR^NP~IP", "VP^IP~ROOT", "VP^IP~CP"};
  // these ones were built by hand.
  // one can't tag split under FRAG or everything breaks, because of those
  // big flat FRAGs....
  private static String[] splitters5 =
    new String[] {"NN^FRAG", "NT^FRAG", "NR^FRAG", "VV^FRAG", 
		  "VV^VCD", "VV^VRD",
		  "NP^NP", "VP^VP",
		  "IP^ROOT", "IP^IP",
		  "PU^NP", "PU^FRAG",
		  "P^PP~VP", "P^PP~IP"};
  private static String[] splitters6 =
    new String[] {"VV^VCD", "VV^VRD",
		  "NP^NP", "VP^VP",
		  "IP^ROOT", "IP^IP",
		  "PU^NP",
		  "P^PP~VP", "P^PP~IP"};
  private static String[] splitters7 =
    new String[] {"NP^NP", "VP^VP",
		  "IP^ROOT", "IP^IP",
		  "PU^NP",
		  "P^PP~VP", "P^PP~IP"};
  private static String[] splitters8 =
    new String[] {"IP^ROOT", "IP^IP",
		  "PU^NP",
		  "P^PP~VP", "P^PP~IP"};
  private static String[] splitters9 =
    new String[] {"VV^VCD", "VV^VRD",
		  "NP^NP", "VP^VP",
		  "IP^ROOT", "IP^IP",
		  "P^PP~VP", "P^PP~IP"};
  private static String[] splitters10 =
    new String[] {"NP^NP", "VP^VP",
		  "IP^ROOT", "IP^IP",
		  "P^PP~VP", "P^PP~IP"};


  public String[] splitters() {
    switch (selectiveSplitLevel) {
    case 1:
      return splitters1;
    case 2:
      return splitters2;
    case 3:
      return splitters3;
    case 4:
      return splitters4;
    case 5:
      return splitters5;
    case 6:
      return splitters6;
    case 7:
      return splitters7;
    case 8:
      return splitters8;
    case 9:
      return splitters9;
    case 10:
      return splitters10;
    default:
      return new String[0];
    }
  }


  public String[] sisterSplitters() {
    return new String[0];
  }

  /** transformTree does all language-specific tree
   *  transformations. Any parameterizations should be inside the
   *  specific TreebankLangParserarams class.
   */
  public TreeHeadPair transformTree(Tree t, Tree root, TreeHeadPair thp) {
    String parentStr;
    String grandParentStr;
    Tree parent; 
    Tree grandParent;
    if(root == null || t.equals(root)) {
      parent = null;
      parentStr = "";
      grandParent = null;
      grandParentStr = "";
    } else {
      parent = t.parent(root);
      parentStr = parent.label().value();
    }
    if(parent == null || parent.equals(root)) {
      grandParent = null;
      grandParentStr = "";
    } else {
      grandParent = parent.parent(root);
      grandParentStr = grandParent.label().value();
    }

    Tree tree = thp.tree;
    TaggedWord head = thp.head;

    String baseParentStr = tlp.basicCategory(parentStr);

    if (t == null || t.isLeaf()) 
      return thp;
    else if (t.isPreTerminal()) { // it's a POS tag
      String word = head.word();
      String tag = tree.label().value();
      String baseTag = tlp.basicCategory(tag);

      List leftAunts = listBasicCategories(SisterAnnotationStats.leftSisterLabels(parent,grandParent));
      List rightAunts = listBasicCategories(SisterAnnotationStats.rightSisterLabels(parent,grandParent));

      // Chinese-specific punctuation splits
      if (chineseSplitDouHao) {
	if(tlp.chineseDouHaoAcceptFilter().accept(word) && 
	   baseTag.equals("PU")) {
	  tag = tag+"-DOU";
	  }
	}

      if (chineseSplitPunct && baseTag.equals("PU")) {
	if(tlp.chineseDouHaoAcceptFilter().accept(word)) {
	  tag = tag+"-DOU";
	  // System.out.println("Punct: Split dou hao"); // debugging
	}
	else if(tlp.chineseCommaAcceptFilter().accept(word)) {
	  tag = tag+"-COMMA";
	  // System.out.println("Punct: Split comma"); // debugging
	}
	else if(tlp.chineseColonAcceptFilter().accept(word)) {
	  tag = tag+"-COLON";
	  // System.out.println("Punct: Split colon"); // debugging
	}
	else if(tlp.chineseQuoteMarkAcceptFilter().accept(word)) {
	  if (chineseSplitPunctLR) {
	    if (tlp.chineseLeftQuoteMarkAcceptFilter().accept(word)) {
	      tag += "-LQUOTE";
	    } else {
	      tag += "-RQUOTE";
	    }
	  } else {
	    tag = tag+"-QUOTE";
	  }
	  // System.out.println("Punct: Split quote"); // debugging
	}
	else if(tlp.chineseEndSentenceAcceptFilter().accept(word)) {
	  tag = tag+"-ENDSENT";
	  // System.out.println("Punct: Split end sent"); // debugging
	}
	else if (tlp.chineseParenthesisAcceptFilter().accept(word)) {
	  if (chineseSplitPunctLR) {
	    if (tlp.chineseLeftParenthesisAcceptFilter().accept(word)) {
	      tag += "-LPAREN";
	    } else {
	      tag += "-RPAREN";
	    }
	  } else {
	    tag += "-PAREN";
	  }
	  // System.out.println("Punct: Split paren"); // debugging
	}
	else if (tlp.chineseDashAcceptFilter().accept(word)) {
	  tag = tag+"-DASH";
	  // System.out.println("Punct: Split dash"); // debugging
	} else if(tlp.chineseOtherAcceptFilter().accept(word)) {
	  tag = tag+"-OTHER";
	} else {
	  pw(System.err).println("Unknown punct: " + tag + " " + word);
	}
      }


      // Chinese-specific POS tag splits (non-punctuation)

      if (tagWordSize) {
	int l = word.length();
	tag += "-" + l + "CHARS";
      }

      if (mergeNNVV & baseTag.equals("NN")) {
	tag = "VV";
      }
	
      if ((chineseSelectiveTagPA || chineseVerySelectiveTagPA) &&
	  (baseTag.equals("CC") || baseTag.equals("P"))) {
	tag += "-" + baseParentStr;
      }
      if (chineseSelectiveTagPA && (baseTag.equals("VV"))) {
	tag += "-" + baseParentStr;
      }

      if (markMultiNtag && tag.startsWith("N")) {
	for(int i = 0; i < parent.children().length;i++) {
	  if (parent.children()[i].label().value().startsWith("N") && parent.children()[i] != t) {
	    tag += "=N";
	    //System.out.println("Found multi=N rewrite");
	  }
	}
      }
	  
      if (markVVsisterIP && baseTag.equals("VV")) {
	boolean seenIP = false;
	for(int i = 0; i < parent.children().length; i++) {
	  if (parent.children()[i].label().value().startsWith("IP"))
	    seenIP=true;
	}
	if(seenIP) {
	  tag += "-IP";
	  //System.out.println("Found VV with IP sister"); // testing
	}
      }

      if (markPsisterIP && baseTag.equals("P")) {
	boolean seenIP = false;
	for(int i = 0; i < parent.children().length; i++) {
	  if (parent.children()[i].label().value().startsWith("IP"))
	    seenIP=true;
	}
	if(seenIP) {
	  tag += "-IP";
	}
      }
	
      if (markADgrandchildOfIP && baseTag.equals("AD") && tlp.basicCategory(grandParentStr).equals("IP")) {
	tag += "~IP";
	//System.out.println("Found AD with IP grandparent"); // testing
      }
	
      if (gpaAD && baseTag.equals("AD")) {
	tag += "~" + grandParentStr;
	//System.out.println("Found AD with grandparent " + grandParentStr); // testing
      }

      if (markPostverbalP && leftAunts.contains("VV") && baseTag.equals("P")) {
	//System.out.println("Found post-verbal P");
	tag += "^=lVV";
      }


      // end Chinese-specific tag splits

      Label label = new CategoryWordTag(tag, word, tag);
      thp.tree.setLabel(label);
      
      thp.head.setTag(tag);
    }
    else { 
      // it's a phrasal category
      String word = head.word();
      String tag = head.tag();
      String category = tree.label().value();
      String baseCategory = tlp.basicCategory(category);

      Tree[] kids = tree.children();

      // Chinese-specific category splits
      List leftSis = listBasicCategories(SisterAnnotationStats.leftSisterLabels(t,parent));
      List rightSis = listBasicCategories(SisterAnnotationStats.rightSisterLabels(t,parent));

      if (paRootDtr && baseParentStr.equals("ROOT")) {
	category+="^ROOT";
      }

      if (markIPsisterBA && baseCategory.equals("IP")) {
	if(leftSis.contains("BA")) {
	  category+= "=BA";
	  //System.out.println("Found IP sister of BA");
	}
      }
	
      if (markIPsisterVVorP && baseCategory.equals("IP")) {
	if(leftSis.contains("VV") || leftSis.contains("P")) {
	  category += "=VVP";
	}
      }

      if (markIPsisDEC && baseCategory.equals("IP")) {
	if(rightSis.contains("DEC")) {
	  category+="=DEC";
	  //System.out.println("Found prenominal IP");
	}
      }

      if (chineseSplitVP3 && category.startsWith("VP")) {
	boolean hasCC = false;
	boolean hasPU = false;
	boolean hasLexV = false;
	for (int i = 0; i < kids.length; i++) {
	  if (kids[i].label().value().startsWith("CC")) {
	    hasCC = true;
	  }
	  if (kids[i].label().value().startsWith("PU")) {
	    hasPU = true;
	  }
	  if (StringUtils.lookingAt(kids[i].label().value(),
				    "(V[ACEV]|VCD|VCP|VNV|VPT|VRD|VSB)")) {
	    hasLexV = true;
	  }
	}
	if(hasCC || (hasPU && !hasLexV)) {
	  category += "-CRD";
	  //System.out.println("Found coordinate VP"); // testing
	}
	else if(hasLexV) {
	  category += "-COMP";
	  //System.out.println("Found complementing VP"); // testing
	}
	else {
	  category += "-ADJT";
	  //System.out.println("Found adjoining VP"); // testing
	}
      }

      if (markVPadjunct && parentStr.startsWith("VP")) {
	  Tree[] sisters = parent.children();
	  boolean hasVPsister = false;
	  boolean hasCC = false;
	  boolean hasPU = false;
	  boolean hasLexV = false;
	  for(int i = 0; i < sisters.length; i++) {
	    if (sisters[i].label().value().startsWith("VP")) {
	      hasVPsister = true;
	    }
	    if (sisters[i].label().value().startsWith("CC")) {
	      hasCC = true;
	    }
	    if (sisters[i].label().value().startsWith("PU")) {
	      hasPU = true;
	    }
	    if (StringUtils.lookingAt(sisters[i].label().value(), "(V[ACEV]|VCD|VCP|VNV|VPT|VRD|VSB)")) {
	      hasLexV = true;
	  }
	  }
	  if(hasVPsister && !(hasCC || hasPU || hasLexV)) {
	    category += "-VPADJ";
	    //System.out.println("Found adjunct of VP"); // testing
	  }
      }

      if (markNPmodNP && baseCategory.equals("NP") && baseParentStr.equals("NP")) {
	  if (rightSis.contains("NP")) {
	    category += "=MODIFIERNP";
	    //System.out.println("Found NP modifier of NP"); // testing
	  }
      }

      if (markModifiedNP && baseCategory.equals("NP") && baseParentStr.equals("NP")) {
	  if(rightSis.size() == 0 && (leftSis.contains("ADJP") || leftSis.contains("NP") || leftSis.contains("DNP") || leftSis.contains("QP") || leftSis.contains("CP") || leftSis.contains("PP"))) {
	    category += "=MODIFIEDNP";
	    //System.out.println("Found modified NP"); // testing
	  }
      }

      if (markNPconj && baseCategory.equals("NP") && baseParentStr.equals("NP")) {
	  if(rightSis.contains("CC") || rightSis.contains("PU") || leftSis.contains("CC") || leftSis.contains("PU")) {
	    category += "=CONJ";
	    //System.out.println("Found NP conjunct"); // testing
	  }
      }

      if (markIPconj && baseCategory.equals("IP") && baseParentStr.equals("IP")) {
	  Tree[] sisters = parent.children();
	  boolean hasCommaSis = false;
	  boolean hasIPSis = false;
	  for(int i = 0; i < sisters.length; i++) {
	    if(tlp.basicCategory(sisters[i].label().value()).equals("PU") && tlp.chineseCommaAcceptFilter().accept(sisters[i].children()[0].label().toString())) {
	      hasCommaSis = true;
	      //System.out.println("Found CommaSis"); // testing
	    }
	    if(tlp.basicCategory(sisters[i].label().value()).equals("IP") && sisters[i] != t) {
	      hasIPSis = true;
	    }
	  }
	  if(hasCommaSis && hasIPSis) {
	    category += "-CONJ";
	    //System.out.println("Found IP conjunct"); // testing
	  }
      }
	
      if (unaryIP && baseCategory.equals("IP") && t.children().length==1) {
	  category+= "-U";
	  //System.out.println("Found unary IP"); //testing
      }
      if(unaryCP && baseCategory.equals("CP") && t.children().length==1) {
	  category+= "-U";
	  //System.out.println("Found unary CP"); //testing
      }

      if (splitBaseNP && baseCategory.equals("NP")) {
	  if (t.isPrePreTerminal()) {
	    category = category + "-B";
	  }
      }

      //if (Test.verbose) System.err.println(baseCategory + " " + leftSis.toString()); //debugging

      if(markPostverbalPP && leftSis.contains("VV") && baseCategory.equals("PP")) {
	  //System.out.println("Found post-verbal PP");
	  category += "=lVV";
      }

      if((markADgrandchildOfIP || gpaAD) && listBasicCategories(SisterAnnotationStats.kidLabels(t)).contains("AD")) {
	  category += "^ADVP";
      }

      Label label = new CategoryWordTag(category, word, tag);
      thp.tree.setLabel(label);

      thp.head = head;
    }
    return thp;
  }


  /** Chinese: Split the dou hao (a punctuation mark separating
   * members of a list) from other punctuation.  Good but included below.
   */
  public static boolean chineseSplitDouHao = false;
  /** Chinese: split Chinese punctuation several ways, along the lines
   * of English punctuation plus another category for the dou hao.  Good.
   */
  public static boolean chineseSplitPunct = true;
  /** Chinese: split left right/paren quote (if chineseSplitPunct is also
   *  true.  Only very marginal gains, but seems positive.
   */
  public static boolean chineseSplitPunctLR = false;

  /** Chinese: mark VVs that are sister of IP (communication &
   *  small-clause-taking verbs).  Good: give 0.5%
   */
  public static boolean markVVsisterIP = true;

  /** Chinese: mark P's that are sister of IP.  Negative effect */
  public static boolean markPsisterIP = true;

  /** Chinese: mark IP's that are sister of VV or P.  These rarely
   *  have punctuation. Small positive effect. */
  public static boolean markIPsisterVVorP = true;


  /** Chinese: mark ADs that are grandchild of IP. */
  public static boolean markADgrandchildOfIP = false;
  /** Grandparent annotate all AD.  Seems slightly negative. */
  public static boolean gpaAD = true;

  // using tagPA on Chinese 100k is negative.

  public static boolean chineseVerySelectiveTagPA = false;
  public static boolean chineseSelectiveTagPA = false;

  /** Chinese: mark IPs that are sister of BA.  These always have
   * overt NP.  Very slightly positive. */
  public static boolean markIPsisterBA = true;

  /** Chinese: mark phrases that are adjuncts of VP (these tend to be
   * locatives/temporals, and have a specific distribution).
   * Necessary even with chineseSplitVP3 and parent annotation because
   * parent annotation happens with unsplit parent categories.
   * Slightly positive.
   */
  public static boolean markVPadjunct = true;

  /** Chinese: mark NP modifiers of NPs. Quite positive (0.5%) */
  public static boolean markNPmodNP = true;

  /** Chinese: mark left-modified NPs (rightmost NPs with a left-side
   * mod).  Slightly positive. */
  public static boolean markModifiedNP = true;

  /** Chinese: mark NPs that are conjuncts.  Negative on small set. */
  public static boolean markNPconj = true;

  /** Chinese: mark nominal tags that are part of multi-nominal
   * rewrites.  Doesn't seem any good. */
  public static boolean markMultiNtag = false;

  /** Chinese: mark IPs that are part of prenominal modifiers. Negative. */
  public static boolean markIPsisDEC = true;

  /** Chinese: mark IPs that are conjuncts.  Or those that have
   * (adjuncts or subjects) */
  public static boolean markIPconj = false;
  public static boolean markIPadjsubj = false;

  /** Chinese: split VPs into VP-COMP, VP-CRD, VP-ADJ.  Negative value.  */
  public static boolean chineseSplitVP3 = true;

  /** Chinese: postverbal PPs; there aren't that many of
   * them. Implementation can't be done easily yet.*/
  //public static final boolean chineseMarkPostverbalP = true;


  /** Chinese: if an IP has no subject (including no empty-category
   * subject), then it should only have an NP (adjunct) daughter if
   * it's a coordinate IP and the NP scopes over the conjunct
   * IPs. (sometimes this NP daughter is adjoined in an IP -> NP
   * IP_coord structure, sometimes the IP conjuncts are at the same
   * level as the NP).  In other cases NP adjuncts should be inside
   * VP.  So: an IP dominating neither a non-subject NP nor another IP
   * should have no NP daughters.  BUT this generalization breaks down
   * when you try to extend it to IPs ignoring their empty subjects.
   * So the simplest thing to do would be to mark non-subject dtrs of
   * IP....  but I think we need to leave the SBJ functional tagging
   * on categories to be consistent about this.
   *
   * Update: I tried retaining SBJ markers with
   * SbjRetainingTreeNormalizer but it works worse than using
   * markVPadjunct.
   */

  /** Chinese: merge NN and VV.  A lark. */
  public static boolean mergeNNVV = false;

  // XXXX upto in testing

  /** Chinese: unary category marking */
  public static boolean unaryIP = false;
  public static boolean unaryCP = false;

  /** Chinese: parent annotate daughter of root.  Meant only for
   * selectivesplit=false. */
  public static boolean paRootDtr = false; // true

  /** Chinese: mark P with a left aunt VV, and PP with a left sister
   * VV.  Note that it's necessary to mark both to thread the
   * context-marking.  Used to identify post-verbal P's, which are
   * rare. */
  public static boolean markPostverbalP = false;
  public static boolean markPostverbalPP = false;
  

  /** How selectively to split. */
  public static int selectiveSplitLevel = 1;

  /** Mark base NPs.  Good. */
  public static boolean splitBaseNP = false;

  /** Annotate tags for number of characters contained. */
  public static boolean tagWordSize = false;

  public void display() {
    String chineseParams = "Using ChineseTreebankParserParams"
      +" chineseSplitDouHao="+chineseSplitDouHao
      +" chineseSplitPunct="+chineseSplitPunct
      +" chineseSplitPunctLR="+chineseSplitPunctLR
      +" markVVsisterIP="+markVVsisterIP
      +" markVPadjunct="+markVPadjunct
      +" chineseSplitVP3="+chineseSplitVP3
      +" mergeNNVV="+mergeNNVV
      +" unaryIP="+unaryIP
      +" unaryCP="+unaryCP
      +" paRootDtr="+paRootDtr
      +" markPsisterIP="+markPsisterIP
      +" markIPsisterVVorP="+markIPsisterVVorP
      +" markADgrandchildOfIP="+markADgrandchildOfIP
      +" gpaAD="+gpaAD
      +" markIPsisterBA="+markIPsisterBA
      +" markNPmodNP="+markNPmodNP
      +" markNPconj="+markNPconj
      +" markMultiNtag="+markMultiNtag
      +" markIPsisDEC="+markIPsisDEC
      +" markIPconj="+markIPconj
      +" markIPadjsubj="+markIPadjsubj
      +" markPostverbalP="+markPostverbalP
      +" markPostverbalPP="+markPostverbalPP
      + " selSplitLevel=" + selectiveSplitLevel
      + " baseNP=" + splitBaseNP
      ;
    System.out.println(chineseParams);
  }
    

  private List listBasicCategories(List l) {
    List l1 = new ArrayList();
    for(Iterator i = l.iterator();i.hasNext();) {
      l1.add(tlp.basicCategory((String)i.next()));
    }
    return l1;
  }

  /** Set language-specific options according to flags.
   *  This routine should process the option starting in args[i] (which
   *  might potentially be several arguments long if it takes arguments).
   *  It should return the index after the last index it consumed in
   *  processing.  In particular, if it cannot process the current option,
   *  the return value should be i.
   */
  public int setOptionFlag(String[] args, int i) {
    if (args[i].equalsIgnoreCase("-vSelSplitLevel") &&
	       (i+1 < args.length)) {
      selectiveSplitLevel = Integer.parseInt(args[i+1]);
      i+=2;
    } else if (args[i].equalsIgnoreCase("-paRootDtr")) {
      paRootDtr = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-unaryIP")) {
      unaryIP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-unaryCP")) {
      unaryCP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-markPostverbalP")) {
      markPostverbalP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-markPostverbalPP")) {
      markPostverbalPP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-baseNP")) {
      splitBaseNP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-markVVsisterIP")) {
      markVVsisterIP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-markPsisterIP")) {
      markPsisterIP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-markIPsisterVVorP")) {
      markIPsisterVVorP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-markIPsisterBA")) {
      markIPsisterBA = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-gpaAD")) {
      gpaAD = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-markVPadjunct")) {
      markVPadjunct = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-markNPmodNP")) {
      markNPmodNP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-markModifiedNP")) {
      markModifiedNP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-markNPconj")) {
      markNPconj = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-chineseSplitPunct")) {
      chineseSplitPunct = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-chineseSplitPunctLR")) {
      chineseSplitPunct = true;
      chineseSplitPunctLR = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-chineseSelectiveTagPA")) {
      chineseSelectiveTagPA = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-chineseVerySelectiveTagPA")) {
      chineseVerySelectiveTagPA = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-markIPsisDEC")) {
      markIPsisDEC = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-chineseSplitVP3")) {
      chineseSplitVP3 = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-tagWordSize")) {
      tagWordSize = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-vanilla")) {
      chineseSplitDouHao = false;
      chineseSplitPunct = false;
      chineseSplitPunctLR = false;
      markVVsisterIP = false;
      markPsisterIP = false;
      markIPsisterVVorP = false;
      markADgrandchildOfIP = false;
      gpaAD = false;
      markIPsisterBA = false;
      markVPadjunct = false;
      markNPmodNP = false;
      markModifiedNP = false;
      markNPconj = false;
      markMultiNtag = false;
      markIPsisDEC = false;
      markIPconj = false;
      markIPadjsubj = false;
      chineseSplitVP3 = false;
      mergeNNVV = false;
      unaryIP = false;
      unaryCP = false;
      paRootDtr = false;
      markPostverbalP = false;
      markPostverbalPP = false;
      splitBaseNP = false;
      selectiveSplitLevel = 0;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-acl03chinese")) {
      Train.markovOrder = 1;
      Train.markovFactor = true;
      chineseSplitDouHao = false;
      chineseSplitPunct = true;
      chineseSplitPunctLR = true;
      markVVsisterIP = true;
      markPsisterIP = true;
      markIPsisterVVorP = true;
      markADgrandchildOfIP = false;
      gpaAD = true;
      markIPsisterBA = false;
      markVPadjunct = true;
      markNPmodNP = true;
      markModifiedNP = true;
      markNPconj = true;
      markMultiNtag = false;
      markIPsisDEC = true;
      markIPconj = false;
      markIPadjsubj = false;
      chineseSplitVP3 = true;
      mergeNNVV = false;
      unaryIP = true;
      unaryCP = true;
      paRootDtr = true;
      markPostverbalP = false;
      markPostverbalPP = false;
      splitBaseNP = false;
      selectiveSplitLevel = 0;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-chinesePCFG")) {
      Train.markovOrder = 2;
      Train.markovFactor = true;
      Train.HSEL_CUT = 5;
      Train.PA = true;
      Train.gPA = true;
      Train.selectiveSplit = false;
      chineseSplitDouHao = false;
      chineseSplitPunct = true;
      chineseSplitPunctLR = true;
      markVVsisterIP = true;
      markPsisterIP = false;
      markIPsisterVVorP = true;
      markADgrandchildOfIP = false;
      gpaAD = false;
      markIPsisterBA = true;
      markVPadjunct = true;
      markNPmodNP = true;
      markModifiedNP = true;
      markNPconj = false;
      markMultiNtag = false;
      markIPsisDEC = false;
      markIPconj = false;
      markIPadjsubj = false;
      chineseSplitVP3 = false;
      mergeNNVV = false;
      unaryIP = false;
      unaryCP = false;
      paRootDtr = false;
      markPostverbalP = false;
      markPostverbalPP = false;
      splitBaseNP = false;
      selectiveSplitLevel = 0;
      chineseVerySelectiveTagPA = true;
      i+=1;
    }

    return i;
  }

  private static long serialVersionUID = 2;

}
