package edu.stanford.nlp.parser.lexparser;

import java.io.*;
import java.util.regex.*;
import java.util.*;
import edu.stanford.nlp.util.*;
import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.io.EncodingPrintWriter; // for debugging only
import java.lang.*;

/** Stores, trains, and scores with an unknown word model.  A couple
 * of filters deterministically force rewrites for certain proper
 * nouns, dates, and cardinal and ordinal numbers; when none of these
 * filters are met, either the distribution of terminals with the same
 * first character is used, or Good-Turing smoothing is used. Although
 * this is developed for Chinese, the training and storage methods
 * could be used cross-linguistically.
 *
 *
 * @author Roger Levy
 */
public class ChineseUnknownWordModel implements Serializable {

  private static final String encoding = "GB18030"; // used only for debugging

  private static final boolean useFirst = true;
  private static final boolean useGT = true; 

  private static final String unknown = "UNK";


  /* These strings are stored in ascii-stype Unicode encoding.  To
   * edit them, either use the Unicode codes or use native2ascii or a
   * similar program to convert the file into a Chinese encoding, then
   * convert back. */
  private static final String dateMatch = ".*[\u5e74\u6708\u65e5\u53f7]$";
  private static final String numberMatch = ".*[\uff10\uff11\uff12\uff13\uff14\uff15\uff16\uff17\uff18\uff19\uff11\u4e00\u4e8c\u4e09\u56db\u4e94\u516d\u4e03\u516b\u4e5d\u5341\u767e\u5343\u4e07\u4ebf].*";
  private static final String ordinalMatch = "^\u7b2c.*";
  private static final String properNameMatch = ".*\u00b7.*";

  private Map tagHash = new HashMap();
  private Set seenFirst = new HashSet();

  private Map unknownGT = new HashMap();


  public double score(IntTaggedWord itw) {
    // treat an IntTaggedWord by changing it into a TaggedWord
    return score(itw.toTaggedWord());
  }


  /* returns the log-prob score of a particular TaggedWord in the
   * unknown word model.
   * @param tw the tag->word reproduction in TaggedWord form
   */
  public double score(TaggedWord tw) {
    double logProb; 

    String word = tw.word().toString();
    String tag = tw.tag().toString();

    // testing
    //EncodingPrintWriter.out.println("Scoring unknown word " + word + " with tag " + tag,encoding);
    // end testing    

    
    if(word.matches(dateMatch)) {
      //EncodingPrintWriter.out.println("Date match for " + word,encoding);
      if(tag.equals("NT")) {
	logProb = 0.0;
      }
      else {
	logProb = Double.NEGATIVE_INFINITY;
      }
    }
    else if(word.matches(numberMatch)) {
      //EncodingPrintWriter.out.println("Number match for " + word,encoding);
      if(tag.equals("CD") && (! word.matches(ordinalMatch))) {
	logProb = 0.0;
      } else
      if(tag.equals("OD") && word.matches(ordinalMatch)) {
	logProb = 0.0;
      } 
      else {
	logProb = Double.NEGATIVE_INFINITY;
      }
    } else if(word.matches(properNameMatch)) {
      //EncodingPrintWriter.out.println("Proper name match for " + word,encoding);
      if(tag.equals("NR")) {
	logProb = 0.0;
      }
      else {
	logProb = Double.NEGATIVE_INFINITY;
      }
    }
    else {
      first:
      if(useFirst) {
	String first = word.substring(0,1);
	if(! seenFirst.contains(first)) {
	  if(useGT) {
	    logProb = scoreGT(tag);
	    break first;
	  }
	  else
	    first = unknown;
	}
	//System.out.println("using first-character model for for unknown word "+  word + " for tag " + tag);

	/* get the Counter of terminal rewrites for the relevant tag */
	Counter wordProbs = (Counter) tagHash.get(tag);

	/* if the proposed tag has never been seen before, issue a
	 * warning and return probability 0 */
	if(wordProbs == null) {
	  //System.err.println("Warning: proposed tag is unseen in training data!");
	  logProb = Double.NEGATIVE_INFINITY;
	}
	else if(wordProbs.hasSeen(first)) {
	  logProb = (double) wordProbs.countOf(first);
	}
	else
	  logProb = (double) wordProbs.countOf(unknown);
      }
      else if (useGT) 
	logProb = scoreGT(tag);
      else {
	System.err.println("Warning: no unknown word model in place!\nGiving the combination " + word + " " + tag + " zero probability.");
	logProb = Double.NEGATIVE_INFINITY; // should never get this!
      }
    }

    //EncodingPrintWriter.out.println("Unknown word estimate for " + word + " as " + tag + ": " + logProb,encoding); //debugging
    return logProb;
  }

  private double scoreGT(String tag) {
    //System.out.println("using GT for unknown word and tag " + tag);
    double logProb;
    if(unknownGT.containsKey(tag))
      logProb = ((Double) unknownGT.get(tag)).doubleValue();
    else
      logProb = Double.NEGATIVE_INFINITY;
    return logProb;
  }


  /** trains the first-character based unknown word model.
   *  @param trees the collection of trees to be trained over
   */
  public void train(Collection trees) {
    if(useFirst)
      System.out.println("treating unknown word as the average of their equivalents by first-character identity.");
    if(useGT) 
      System.out.println("using Good-Turing smoothing for unknown words.");

    trainUnknownGT(trees);

    HashMap c = new HashMap(); // counts

    Counter tc = new Counter();

    for (Iterator i = trees.iterator(); i.hasNext(); ){
      Tree t = (Tree) i.next();
      List words = t.taggedYield();
      for (Iterator j = words.iterator(); j.hasNext(); ){
        TaggedWord tw = (TaggedWord) (j.next());
	String word = tw.word();
	String first = tw.word().substring(0,1);
	String tag = tw.tag();
	if(! c.containsKey(tag)) {
	  c.put(tag,new Counter());
	}
	((Counter)c.get(tag)).increment(first);

	tc.increment(tag);

	seenFirst.add(first);

      }
    }


    for (Iterator i = c.keySet().iterator(); i.hasNext(); ){
      String tag = (String) i.next();
      Counter wc = (Counter)c.get(tag); // counts for words given a tag

      /* outer iteration is over tags */
      if (! tagHash.containsKey(tag)) {
	tagHash.put(tag,new Counter());
      }

      /* the UNKNOWN first character is assumed to be seen once in each tag */
      // this is really sort of broken!
      tc.increment(tag);
      wc.setCount(unknown,1.0);

      /* inner iteration is over words */
      for (Iterator j = wc.seenSet().iterator(); j.hasNext(); ){
	String first = (String) j.next();
	double prob = Math.log(((double) wc.countOf(first))/((double)tc.countOf(tag)));
	((Counter)tagHash.get(tag)).setCount(first,prob);
	//if (Test.verbose) 
	  //EncodingPrintWriter.out.println(tag + " rewrites as " + first + " firstchar with probability " + prob,encoding);
      }
    }
  }

  /* trains Good-Turing estimation of unknown words */
  private void trainUnknownGT(Collection trees) {

    Counter twCount = new Counter();
    Counter wtCount = new Counter();
    Counter tagCount = new Counter();
    Counter r1 = new Counter(); // for each tag, # of words seen once
    Counter r0 = new Counter(); // for each tag, # of words not seen
    Set seenWords = new HashSet();

    int tokens = 0;

    /* get TaggedWord and total tag counts, and get set of all
     * words attested in training */
    for (Iterator i = trees.iterator(); i.hasNext(); ){
      Tree t = (Tree) i.next();
      List words = t.taggedYield();
      for (Iterator j = words.iterator(); j.hasNext(); ){
	tokens++;
        TaggedWord tw = (TaggedWord) (j.next());
	WordTag wt = toWordTag(tw);
	String word = wt.word();
	String tag = wt.tag();
	//if (Test.verbose) EncodingPrintWriter.out.println("recording instance of " + wt.toString(),encoding); // testing

	wtCount.increment(wt);// TaggedWord has crummy equality conditions
	twCount.increment(tw);//testing
	//if (Test.verbose) EncodingPrintWriter.out.println("This is the " + wtCount.countOf(wt) + "th occurrence of" + wt.toString(),encoding); // testing	
	tagCount.increment(tag);
	boolean alreadySeen = seenWords.add(word);

	// if (Test.verbose) if(! alreadySeen) EncodingPrintWriter.out.println("already seen " + wt.toString(),encoding); // testing

      }
    }
    
    // testing: get some stats here
    System.out.println("Total tokens: " + tokens);
    System.out.println("Total WordTag types: " + wtCount.seenSet().size());
    System.out.println("Total TaggedWord types: " + twCount.seenSet().size());
    System.out.println("Total tag types: " + tagCount.seenSet().size());
    System.out.println("Total word types: " + seenWords.size());


    /* find # of once-seen words for each tag */
    for (Iterator i = wtCount.seenSet().iterator(); i.hasNext();) {
      WordTag wt = (WordTag) i.next();
      if (wtCount.countOf(wt) == 1) {
	r1.increment(wt.tag());
      }
    }

    /* find # of unseen words for each tag */
    for (Iterator i = tagCount.seenSet().iterator(); i.hasNext();) {
      String tag = (String) i.next();
      for(Iterator j = seenWords.iterator(); j.hasNext();){
	String word = (String) j.next();
	WordTag wt = new WordTag(word,tag);
	//EncodingPrintWriter.out.println("seeking " + wt.toString(),encoding); // testing
	if(! (wtCount.hasSeen(wt))) {
	  r0.increment(tag);
	  //EncodingPrintWriter.out.println("unseen " + wt.toString(),encoding); // testing
	}
	else {
	  //EncodingPrintWriter.out.println("count for " + wt.toString() + " is " + wtCount.countOf(wt),encoding);
	}
      }
    }

    /* set unseen word probability for each tag */
    for (Iterator i = tagCount.seenSet().iterator(); i.hasNext();) {
      String tag = (String) i.next();
      //System.out.println("Tag " + tag + ".  Word types for which seen once: " + r1.countOf(tag) + ".  Word types for which unseen: " + r0.countOf(tag) + ".  Total count token for tag: " + tagCount.countOf(tag)); // testing
      
      double logprob = Math.log(r1.countOf(tag)/(tagCount.countOf(tag)*r0.countOf(tag)));

      unknownGT.put(tag,new Double(logprob));
    }

    /* testing only: print the GT-smoothed model */
    //System.out.println("The GT-smoothing model:");
    //System.out.println(unknownGT.toString());
    //EncodingPrintWriter.out.println(wtCount.toString(),encoding);


  }

  public static void main(String[] args) {
    System.out.println("Testing unknown matching");
    String s = "\u5218\u00b7\u9769\u547d";
    if (s.matches(properNameMatch)) {
	System.out.println("hooray names!");
    } else {
	System.out.println("Uh-oh names!");
    }
    String s1 = "\uff13\uff10\uff10\uff10";
    if (s1.matches(numberMatch)) {
	System.out.println("hooray numbers!");
    } else {
	System.out.println("Uh-oh numbers!");
    }	
    String s11 = "\u767e\u5206\u4e4b\u56db\u5341\u4e09\u70b9\u4e8c";
    if (s1.matches(numberMatch)) {
	System.out.println("hooray numbers!");
    } else {
	System.out.println("Uh-oh numbers!");
    }	
    String s12 = "\u767e\u5206\u4e4b\u4e09\u5341\u516b\u70b9\u516d";
    if (s1.matches(numberMatch)) {
	System.out.println("hooray numbers!");
    } else {
	System.out.println("Uh-oh numbers!");
    }	
    String s2 = "\u4e09\u6708";
    if(s2.matches(dateMatch)) {
	System.out.println("hooray dates!");
    } else {
	System.out.println("Uh-oh dates!");
    }

    System.out.println("Testing tagged word");
    Counter c = new Counter();
    TaggedWord tw1 = new TaggedWord("w","t");
    c.increment(tw1);
    TaggedWord tw2 = new TaggedWord("w","t2");
    System.out.println(c.hasSeen(tw2));
    System.out.println(tw1.equals(tw2));

    WordTag wt1 = toWordTag(tw1);
    WordTag wt2 = toWordTag(tw2);
    WordTag wt3 = new WordTag("w","t2");
    System.out.println(wt1.equals(wt2));
    System.out.println(wt2.equals(wt3));
  }

  private static WordTag toWordTag(TaggedWord tw) {
    return new WordTag(tw.word(),tw.tag());
  }

  private static long serialVersionUID = 221L;

}
	
