package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.trees.*;
import java.io.*;
import java.util.*;

/** Parser parameters for the Penn English Treebank (WSJ, Brown,
 *  Switchboard)
 *
 *  @author Roger Levy
 *  @version   03/05/2003
 */

public class EnglishTreebankParserParams implements TreebankLangParserParams {

  private TreebankLanguagePack tlp;
  private TreeFactory tf = 
    new LabeledScoredTreeFactory(new CategoryWordTagFactory());


  /* manipulating input and output encodings shouldn't be needed for
   * English, generally */
  private String inputEncoding;
  private String outputEncoding;

  public EnglishTreebankParserParams() {
    tlp = treebankLanguagePack();
    inputEncoding = tlp.getEncoding();
    outputEncoding = tlp.getEncoding();
  }

  
  public void setInputEncoding(String encoding) {
    inputEncoding = encoding;
  }

  public void setOutputEncoding(String encoding) {
    outputEncoding = encoding;
  }


  private HeadFinder etbHeadFinder = new CollinsHeadFinder();

  public HeadFinder headFinder() {
    return etbHeadFinder;
  }


  /** Reads in trees from the source you want.  It's the
   *  responsibility of tr to deal properly with character-set encoding
   *  of the input.  It also is the responsibility of tr to properly
   *  normalize trees. 
   */
  public MemoryTreebank memoryTreebank() { 
    return new MemoryTreebank(new TreeReaderFactory() {
 	public TreeReader newTreeReader(Reader in) {
 	  return new PennTreeReader(in, 
	      new LabeledScoredTreeFactory(
				 new StringLabelFactory()),
	      new NPTmpRetainingTreeNormalizer(EnglishTrain.splitNPTMP,
					       EnglishTrain.splitTMP,
 					       EnglishTrain.splitSGapped == 5,
 					       Train.leaveItAll));
 	}
      });
  }

  /** returns a MemoryTreebank appropriate to the testing treebank source */
  public MemoryTreebank testMemoryTreebank() {
    return new MemoryTreebank(new TreeReaderFactory() {
	public TreeReader newTreeReader(Reader in) {
	  return new PennTreeReader(in, 
				new LabeledScoredTreeFactory(
						     new StringLabelFactory()),
				new BobChrisTreeNormalizer());
	}
      });
  }


  public Lexicon lex() {
    return new Lexicon();
  }

  /** the tree transformer used to produce trees for evaluation.  Will
   * be applied both to the */
  public TreeTransformer collinizer() {
    return new TreeCollinizer();
  }

  public TreeTransformer collinizerEvalb() {
    return new TreeCollinizer(false);
  }

  /** contains Treebank-specific (but not parser-specific) info such
   * as what is punctuation, and also information about the structure
   * of labels */
  public TreebankLanguagePack treebankLanguagePack() {
    return new PennTreebankLanguagePack();
  }

  /** the PrintWriter used to print output. It's the responsibility of
   * pw to deal properly with character encodings for the relevant
   * treebank */
  public PrintWriter pw() {
    return pw(System.out);
  }

  /** the PrintWriter used to print output to OutputStream o. It's the
   * responsibility of pw to deal properly with character encodings
   * for the relevant treebank */
  public PrintWriter pw(OutputStream o) {
    return new PrintWriter(o,true);
  }
  
  
  // Automatically generated by ParentAnnotationStats -- preferably don't edit
  private static String[] splitters1 = new String[] {"RB^VP", "RB^ADVP", "RB^ADJP", "RB^QP", "RB^NP", "RB^CONJP", "RB^PP", "RB^SBAR", "RB^S", "RB^PRT", "RB^WHADVP", "RB^FRAG", "RB^INTJ", "RB^WHNP", "RB^SQ", "ADJP^NP", "ADJP^VP", "ADJP^ADJP", "ADJP^S", "ADJP^PP", "ADJP^UCP", "ADJP^FRAG", "ADJP^PRN", "ADJP^ROOT", "-RRB-^NP", "NX^NP", "NX^NX", "JJ^ADJP", "JJ^NP", "JJ^PP", "JJ^ADVP", "JJ^QP", "JJ^VP", "JJ^NX", "JJ^UCP", "JJ^WHNP", "JJ^WHADJP", "JJ^WHADVP", "JJ^NAC", "JJ^PRN", "-LRB-^NP", "NP^S", "NP^NP", "NP^PP", "NP^VP", "NP^ROOT", "NP^SINV", "NP^ADVP", "NP^ADJP", "NP^PRN", "NP^FRAG", "NP^SQ", "NP^WHNP", "NP^UCP", "NP^SBAR", "NP^X", "PP^NP", "PP^VP", "PP^S", "PP^ADVP", "PP^PP", "PP^ADJP", "PP^ROOT", "PP^PRN", "PP^FRAG", "PP^UCP", "NN^ADJP", "NN^QP", "NN^NX", "NN^VP", "NN^UCP", "NN^ADVP", "NN^WHNP", "NN^NP", "NN^SBAR", "NN^NAC", "NN^PRN", "NN^PP", "NN^INTJ", "VBP^SQ", "VBP^NP", "S^ROOT", "S^SBAR", "S^VP", "S^S", "S^PP", "S^SINV", "S^NP", "S^ADJP", "S^PRN", "S^FRAG", "VBN^NP", "VBN^ADJP", "VBN^PP", "VBN^VP", ":^PRN", ":^S", ":^VP", ":^SINV", "RBR^ADJP", "RBR^ADVP", "RBR^NP", "RBR^QP", "VBZ^NP", "VBZ^SQ", "DT^SBAR", "DT^ADVP", "DT^QP", "DT^WHNP", "DT^VP", "DT^ADJP", "DT^S", "DT^PP", "DT^NP", "NNS^NX", "NNS^QP", "NNS^VP", "NNS^WHNP", "NNS^NP", "NNS^UCP", "NNS^ADJP", "JJS^QP", "JJS^NP", "JJS^ADVP", "NNP^NAC", "NNP^NX", "NNP^ADJP", "NNP^NP", "NNP^VP", "NNP^UCP", "NNP^ADVP", "NNP^PRN", "NNP^SBAR", "NNP^PP", "NNP^WHNP", "NNP^X", "JJR^QP", "JJR^ADJP", "JJR^NP", "JJR^ADVP", "WDT^NP", "VP^VP", "VP^S", "VP^NP", "VP^SINV", "VP^SQ", "VP^PRN", "VP^UCP", "VP^FRAG", "VP^ROOT", "VP^PP", "CD^QP", "CD^NP", "CD^ADJP", "CD^ADVP", "CD^NAC", "CD^NX", "VBG^PP", "VBG^NP", "VBG^VP", "VBG^ADJP", "MD^SQ", "WRB^WHADJP", "WRB^WHNP", "IN^SBAR", "IN^PP", "IN^QP", "IN^ADVP", "IN^WHNP", "IN^NP", "IN^CONJP", "IN^S", "IN^PRT", "IN^ADJP", "IN^WHPP", "IN^VP", "IN^WHADVP", "VBD^NP", "VBD^SINV", "VBD^SQ", "VBD^ADJP", "QP^ADJP", "WHNP^WHNP", "WHNP^SBARQ", "WHNP^WHPP", "PRN^S", "PRN^NP", "PRN^VP", "PRN^ADJP", "PRN^PP", "VB^NP", "VB^INTJ", "VB^ADJP", "VB^VP", "SBAR^NP", "SBAR^VP", "SBAR^S", "SBAR^PP", "SBAR^ADJP", "SBAR^ROOT", "SBAR^SBAR", "SBAR^ADVP", "SBAR^PRN", "SBAR^FRAG", "ADVP^S", "ADVP^ADVP", "ADVP^NP", "ADVP^PP", "ADVP^VP", "ADVP^ADJP", "ADVP^PRN", "ADVP^ROOT", "CC^S", "CC^NP", "CC^VP", "CC^QP", "CC^PP", "CC^ADJP", "CC^SINV", "CC^NX", "CC^SBAR", "CC^UCP", "CC^FRAG", "UCP^VP", "UCP^NP", "FRAG^ROOT", "FRAG^FRAG", "FRAG^SBAR", "FRAG^VP", "FRAG^S", "SINV^PRN", "SINV^ROOT", "SINV^SBAR", "SINV^S", "PDT^QP", ".^SBARQ", ".^SQ", ".^S", ".^FRAG", "SQ^ROOT", "SQ^SBARQ", "S^SBAR~NP", "S^SBAR~VP", "S^SBAR~S", "S^SBAR~SBAR", "IN^PP~VP", "IN^PP~S", "IN^PP~NP", "IN^PP~ADJP", "IN^PP~ADVP", "IN^PP~PP", "IN^PP~SINV", "IN^PP~PRN", "IN^PP~NAC", "IN^PP~UCP", "IN^PP~FRAG", "VBN^NP~PP", "VBN^NP~NP", "VBN^NP~S", "VBN^NP~VP", "IN^ADVP~S", "IN^ADVP~VP", "IN^ADVP~PP", "IN^ADVP~NP", "VP^NP~VP", "VP^NP~NP", "VP^NP~PP", "VP^NP~S", "VP^NP~ROOT", "NP^PRN~S", "NP^PRN~VP", "JJS^NP~PP", "JJS^NP~S", "JJS^NP~NP", "NP^VP~VP", "NP^VP~S", "NP^VP~NP", "NP^VP~SQ", "VBG^PP~NP", "VBG^PP~VP", "VBG^PP~S", "JJ^NP~PP", "JJ^NP~S", "JJ^NP~VP", "JJ^NP~NP", "JJ^NP~ROOT", "JJ^NP~ADVP", "JJ^NP~SINV", "JJ^NP~FRAG", "JJ^NP~PRN", "JJ^NP~UCP", "JJ^NP~SQ", "JJ^NP~ADJP", "JJR^NP~PP", "JJR^NP~S", "JJR^NP~VP", "TO^PP~S", "WDT^WHNP~WHPP", "JJ^ADJP~NP", "JJ^ADJP~VP", "JJ^ADJP~S", "JJ^ADJP~ADJP", "JJ^ADJP~PP", "JJ^ADJP~UCP", "JJ^ADJP~FRAG", "JJ^ADJP~SINV", "JJ^ADJP~PRN", "JJ^ADJP~SQ", "RB^ADJP~VP", "RB^ADJP~NP", "RB^ADJP~ADJP", "RB^ADJP~S", "RB^ADJP~PP", "RB^ADJP~UCP", "RBR^ADVP~PP", "RBR^ADVP~VP", "RBR^ADVP~ADJP", "RBR^ADVP~S", "S^S~VP", "S^S~SBAR", "S^S~ROOT", "S^S~S", "S^S~SINV", "VBG^NP~PP", "VBG^NP~NP", "VBG^NP~VP", "VBG^NP~S", "NP^ADJP~VP", "NP^ADJP~NP", "NP^PP~NP", "NP^PP~VP", "NP^PP~S", "NP^PP~PP", "NP^PP~ADJP", "NP^PP~ADVP", "NP^PP~NAC", "NP^PP~PRN", "NP^PP~SINV", "NP^PP~UCP", "NP^PP~FRAG", "NP^PP~NX", "NP^PP~WHNP", "RB^PP~S", "RB^PP~NP", "RB^PP~VP", "NNS^NP~S", "NNS^NP~PP", "NNS^NP~VP", "NNS^NP~NP", "NNS^NP~ADVP", "NNS^NP~ROOT", "NNS^NP~ADJP", "NNS^NP~PRN", "NNS^NP~SINV", "NNS^NP~FRAG", "NNS^NP~SBAR", "NNS^NP~SQ", "NNS^NP~UCP", "RB^NP~PP", "RB^NP~S", "RB^NP~VP", "RB^NP~NP", "RB^NP~ROOT", "RB^NP~ADJP", "RB^NP~ADVP", "CC^NP~NP", "CC^NP~VP", "CC^NP~S", "CC^NP~PP", "VP^SINV~SBAR", "VP^SINV~S", "ADJP^ADJP~NP", "ADJP^ADJP~VP", "CC^VP~VP", "S^NP~ROOT", "PRP^NP~VP", "PRP^NP~PP", "PRP^NP~S", "PRP^NP~NP", "PRP^NP~SQ", "VB^VP~S", "VB^VP~VP", "VB^VP~SQ", "VB^VP~SINV", "VB^VP~FRAG", "IN^WHPP~WHNP", "VBN^VP~NP", "VBN^VP~S", "VBN^VP~VP", "VBN^VP~SINV", "VBN^VP~PRN", "VBN^VP~UCP", "VBN^VP~SQ", "NN^ADJP~VP", "NN^ADJP~ADJP", "NN^ADJP~NP", "NN^ADJP~S", "DT^NP~S", "DT^NP~PP", "DT^NP~VP", "DT^NP~NP", "DT^NP~ADVP", "DT^NP~WHNP", "DT^NP~ADJP", "WHNP^SBAR~PP", "WHNP^SBAR~VP", "WHNP^SBAR~NP", "WHNP^SBAR~S", "S^VP~NP", "S^VP~S", "S^VP~VP", "ADVP^PP~ADVP", "PRN^NP~S", "PRN^NP~VP", "PRN^NP~NP", "PRN^NP~PP", "NNPS^NP~PP", "NNPS^NP~S", "NNPS^NP~NP", "NNPS^NP~VP", "NNPS^NP~ROOT", "JJ^PP~VP", "SBAR^SBAR~NP", "RB^ADVP~S", "RB^ADVP~VP", "RB^ADVP~NP", "RB^ADVP~PP", "RB^ADVP~ADVP", "RB^ADVP~SINV", "RB^ADVP~SBAR", "RB^ADVP~ADJP", "RB^ADVP~PRN", "RB^ADVP~UCP", "RB^ADVP~FRAG", "RB^ADVP~SQ", "RB^ADVP~ROOT", "RB^ADVP~RRC", "ADVP^NP~NP", "ADVP^NP~VP", "ADVP^NP~PP", "VBP^VP~VP", "VBP^VP~SINV", "VBP^VP~S", "JJR^ADJP~NP", "JJR^ADJP~S", "IN^SBAR~S", "IN^SBAR~VP", "IN^SBAR~NP", "IN^SBAR~ADJP", "IN^SBAR~PP", "IN^SBAR~ADVP", "IN^SBAR~SBAR", "IN^SBAR~PRN", "NP^ADVP~PP", "NP^ADVP~VP", "NP^ADVP~S", "NP^ADVP~NP", "NP^NP~VP", "NP^NP~SINV", "NP^NP~NP", "NP^NP~S", "NP^NP~PP", "NP^NP~ROOT", "NP^NP~PRN", "NP^NP~FRAG", "NP^NP~ADVP", "NP^NP~ADJP", "NP^NP~UCP", "NP^NP~SQ", "NP^NP~X", "SBAR^VP~VP", "SBAR^VP~S", "CD^QP~ADJP", "CD^QP~NP", "ADVP^VP~VP", "ADVP^VP~S", "ADVP^VP~NP", "NN^VP~VP", "NN^VP~S", "NN^VP~NP", "RB^S~ROOT", "PP^NP~NP", "PP^NP~VP", "PP^NP~S", "PP^NP~PP", "VBZ^VP~SINV", "VBZ^VP~VP", "VBZ^VP~S", "VBZ^VP~SQ", "QP^NP~PP", "QP^NP~S", "QP^NP~VP", "QP^NP~NP", "QP^NP~ADVP", "QP^NP~PRN", "QP^NP~ADJP", "WRB^WHADVP~SBARQ", "VP^S~VP", "VP^S~ROOT", "VP^S~PP", "VP^S~SBAR", "VP^S~NP", "VP^S~S", "VP^S~ADJP", "VP^S~PRN", "VP^S~SINV", "VP^S~FRAG", "VP^VP~VP", "VP^VP~S", "VP^VP~NP", ":^NP~NP", ":^NP~S", ":^NP~ROOT", "VBD^VP~VP", "VBD^VP~SINV", "VBD^VP~NP", "VBD^VP~S", "CC^S~SBAR", "NNP^NP~PP", "NNP^NP~S", "NNP^NP~NP", "NNP^NP~VP", "NNP^NP~SINV", "NNP^NP~ROOT", "NNP^NP~PRN", "NNP^NP~ADJP", "NNP^NP~FRAG", "NNP^NP~SQ", "NNP^NP~NX", "NNP^NP~UCP", "NP^FRAG~FRAG", "NP^FRAG~ROOT", "SBAR^NP~VP", "SBAR^NP~S", "SBAR^NP~PP", "NP^SINV~SBAR", "NP^SINV~S", "NP^SINV~ROOT", "NP^SINV~PRN", "NN^NP~PP", "NN^NP~S", "NN^NP~NP", "NN^NP~VP", "NN^NP~ADVP", "NN^NP~SINV", "NN^NP~ROOT", "NN^NP~PRN", "NN^NP~FRAG", "NN^NP~ADJP", "NN^NP~SQ", "NN^NP~UCP", "NN^NP~NX", "NN^NP~X", "NN^NP~WHNP", "NN^NP~SBAR", "ADJP^NP~NP", "ADJP^NP~S", "ADJP^NP~VP", "ADJP^NP~PP", "ADJP^NP~ROOT", "NNP^NX~NP", "VBG^VP~NP", "VBG^VP~VP", "VBG^VP~S", "VBG^VP~SINV", "VBG^VP~SQ", "NN^NX~NP", "JJ^ADVP~S", "JJ^ADVP~VP", "ADJP^VP~VP", "ADJP^VP~S", "PRP$^NP~S", "PRP$^NP~PP", "PRP$^NP~VP", "NP^S~SBAR", "NP^S~ROOT", "NP^S~S", "NP^S~VP", "NP^S~SINV", "NP^S~PRN", "NP^S~PP", "NP^S~NP", "VBN^ADJP~VP", "VBN^ADJP~NP", "VBN^ADJP~ADJP", "VBN^ADJP~S", "IN^NP~S", "CD^NP~PP", "CD^NP~NP", "CD^NP~VP", "CD^NP~ADVP", "CD^NP~S", "CD^NP~ADJP", "CD^NP~PRN", "CD^NP~ROOT", "CD^NP~SINV", "CD^NP~FRAG", "RB^SBAR~S", "PP^VP~S", "PP^VP~NP", "PP^VP~VP", "WP^WHNP~SBARQ", "WP^WHNP~WHPP"};
  private static String[] splitters2 = new String[] {"RB^VP", "RB^ADVP", "RB^ADJP", "RB^QP", "RB^NP", "RB^CONJP", "RB^PP", "RB^SBAR", "RB^S", "RB^PRT", "RB^WHADVP", "RB^FRAG", "ADJP^NP", "ADJP^VP", "ADJP^ADJP", "ADJP^S", "ADJP^PP", "NX^NP", "NX^NX", "JJ^ADJP", "JJ^NP", "JJ^PP", "JJ^ADVP", "JJ^QP", "JJ^VP", "JJ^NX", "JJ^UCP", "JJ^WHNP", "JJ^WHADJP", "NP^S", "NP^NP", "NP^PP", "NP^VP", "NP^ROOT", "NP^SINV", "NP^ADVP", "NP^ADJP", "NP^PRN", "NP^FRAG", "NP^SQ", "NP^WHNP", "NP^UCP", "NP^SBAR", "PP^NP", "PP^VP", "PP^S", "PP^ADVP", "PP^PP", "PP^ADJP", "PP^ROOT", "PP^PRN", "NN^ADJP", "NN^QP", "NN^NX", "NN^VP", "NN^UCP", "NN^ADVP", "NN^WHNP", "NN^NP", "NN^SBAR", "VBP^SQ", "S^ROOT", "S^SBAR", "S^VP", "S^S", "S^PP", "S^SINV", "S^NP", "S^ADJP", "S^PRN", "VBN^NP", "VBN^ADJP", "VBN^PP", "VBN^VP", ":^PRN", ":^S", "RBR^ADJP", "RBR^ADVP", "VBZ^NP", "VBZ^SQ", "DT^SBAR", "DT^ADVP", "DT^QP", "DT^WHNP", "DT^VP", "DT^ADJP", "DT^S", "DT^PP", "NNS^NX", "NNS^QP", "NNS^VP", "NNS^WHNP", "JJS^QP", "JJS^NP", "JJS^ADVP", "NNP^NAC", "NNP^NX", "NNP^ADJP", "NNP^NP", "NNP^VP", "NNP^UCP", "NNP^ADVP", "NNP^PRN", "NNP^SBAR", "JJR^QP", "JJR^ADJP", "VP^VP", "VP^S", "VP^NP", "VP^SINV", "VP^SQ", "VP^PRN", "VP^UCP", "VP^FRAG", "CD^QP", "CD^NP", "CD^ADJP", "CD^ADVP", "VBG^PP", "VBG^NP", "VBG^VP", "VBG^ADJP", "MD^SQ", "IN^SBAR", "IN^PP", "IN^QP", "IN^ADVP", "IN^WHNP", "IN^NP", "IN^CONJP", "IN^S", "IN^PRT", "IN^ADJP", "IN^WHPP", "IN^VP", "VBD^NP", "VBD^SINV", "QP^ADJP", "WHNP^WHNP", "WHNP^SBARQ", "WHNP^WHPP", "PRN^S", "PRN^NP", "PRN^VP", "VB^NP", "SBAR^NP", "SBAR^VP", "SBAR^S", "SBAR^PP", "SBAR^ADJP", "SBAR^ROOT", "SBAR^SBAR", "ADVP^S", "ADVP^ADVP", "ADVP^NP", "ADVP^PP", "ADVP^VP", "ADVP^ADJP", "CC^S", "CC^NP", "CC^VP", "CC^QP", "CC^PP", "CC^ADJP", "CC^SINV", "UCP^VP", "UCP^NP", "FRAG^ROOT", "FRAG^FRAG", "FRAG^SBAR", "SINV^PRN", "SINV^ROOT", "SINV^SBAR", "SINV^S", ".^SBARQ", ".^SQ", ".^S", "SQ^ROOT", "S^SBAR~NP", "S^SBAR~VP", "S^SBAR~S", "IN^PP~VP", "IN^PP~S", "IN^PP~NP", "IN^PP~ADJP", "IN^PP~ADVP", "IN^PP~PP", "IN^PP~SINV", "IN^PP~PRN", "IN^PP~NAC", "VBN^NP~PP", "VBN^NP~NP", "VBN^NP~S", "VBN^NP~VP", "VP^NP~VP", "NP^VP~VP", "NP^VP~S", "NP^VP~NP", "VBG^PP~NP", "VBG^PP~VP", "VBG^PP~S", "JJ^NP~PP", "JJ^NP~S", "JJ^NP~VP", "JJ^NP~NP", "JJ^NP~ROOT", "JJ^NP~ADVP", "JJ^NP~SINV", "JJ^NP~FRAG", "JJ^NP~PRN", "JJ^NP~UCP", "TO^PP~S", "WDT^WHNP~WHPP", "JJ^ADJP~NP", "JJ^ADJP~VP", "JJ^ADJP~S", "JJ^ADJP~ADJP", "JJ^ADJP~PP", "JJ^ADJP~UCP", "JJ^ADJP~FRAG", "JJ^ADJP~SINV", "JJ^ADJP~PRN", "RB^ADJP~VP", "RB^ADJP~NP", "RB^ADJP~ADJP", "RB^ADJP~S", "RBR^ADVP~PP", "S^S~VP", "S^S~SBAR", "VBG^NP~PP", "VBG^NP~NP", "VBG^NP~VP", "VBG^NP~S", "NP^PP~NP", "NP^PP~VP", "NP^PP~S", "NP^PP~PP", "NP^PP~ADJP", "NP^PP~ADVP", "NP^PP~NAC", "NP^PP~PRN", "NP^PP~SINV", "NNS^NP~S", "NNS^NP~PP", "NNS^NP~VP", "NNS^NP~NP", "NNS^NP~ADVP", "NNS^NP~ROOT", "NNS^NP~ADJP", "NNS^NP~PRN", "NNS^NP~SINV", "RB^NP~PP", "RB^NP~S", "RB^NP~VP", "RB^NP~NP", "CC^NP~NP", "CC^NP~VP", "CC^NP~S", "PRP^NP~VP", "PRP^NP~PP", "PRP^NP~S", "PRP^NP~NP", "VB^VP~S", "VB^VP~VP", "VB^VP~SQ", "VBN^VP~NP", "VBN^VP~S", "VBN^VP~VP", "VBN^VP~SINV", "VBN^VP~PRN", "DT^NP~S", "DT^NP~PP", "DT^NP~VP", "DT^NP~NP", "DT^NP~ADVP", "WHNP^SBAR~PP", "WHNP^SBAR~VP", "WHNP^SBAR~NP", "S^VP~NP", "S^VP~S", "NNPS^NP~PP", "NNPS^NP~S", "NNPS^NP~NP", "NNPS^NP~VP", "RB^ADVP~S", "RB^ADVP~VP", "RB^ADVP~NP", "RB^ADVP~PP", "RB^ADVP~ADVP", "RB^ADVP~SINV", "RB^ADVP~SBAR", "RB^ADVP~ADJP", "RB^ADVP~PRN", "RB^ADVP~UCP", "RB^ADVP~FRAG", "VBP^VP~VP", "IN^SBAR~S", "IN^SBAR~VP", "IN^SBAR~NP", "IN^SBAR~ADJP", "IN^SBAR~PP", "IN^SBAR~ADVP", "IN^SBAR~SBAR", "NP^ADVP~PP", "NP^ADVP~VP", "NP^NP~VP", "NP^NP~SINV", "NP^NP~NP", "NP^NP~S", "NP^NP~PP", "NP^NP~ROOT", "NP^NP~PRN", "NP^NP~FRAG", "NP^NP~ADVP", "NP^NP~ADJP", "SBAR^VP~VP", "SBAR^VP~S", "CD^QP~ADJP", "CD^QP~NP", "ADVP^VP~VP", "ADVP^VP~S", "PP^NP~NP", "PP^NP~VP", "VBZ^VP~SINV", "VBZ^VP~VP", "VBZ^VP~S", "QP^NP~PP", "QP^NP~S", "QP^NP~VP", "QP^NP~NP", "QP^NP~ADVP", "WRB^WHADVP~SBARQ", "VP^S~VP", "VP^S~ROOT", "VP^S~PP", "VP^S~SBAR", "VP^S~NP", "VP^S~S", "VP^S~ADJP", "VP^S~PRN", "VP^S~SINV", "VP^VP~VP", "VP^VP~S", "VP^VP~NP", ":^NP~NP", ":^NP~S", ":^NP~ROOT", "VBD^VP~VP", "VBD^VP~SINV", "VBD^VP~NP", "VBD^VP~S", "CC^S~SBAR", "NNP^NP~PP", "NNP^NP~S", "NNP^NP~NP", "NNP^NP~VP", "NNP^NP~SINV", "NNP^NP~ROOT", "NNP^NP~PRN", "NNP^NP~ADJP", "NNP^NP~FRAG", "NNP^NP~SQ", "SBAR^NP~VP", "SBAR^NP~S", "NP^SINV~SBAR", "NN^NP~PP", "NN^NP~S", "NN^NP~NP", "NN^NP~VP", "NN^NP~ADVP", "NN^NP~SINV", "NN^NP~ROOT", "NN^NP~PRN", "NN^NP~FRAG", "NN^NP~ADJP", "NN^NP~SQ", "NN^NP~UCP", "ADJP^NP~NP", "ADJP^NP~S", "ADJP^NP~VP", "ADJP^NP~PP", "VBG^VP~NP", "VBG^VP~VP", "VBG^VP~S", "VBG^VP~SINV", "ADJP^VP~VP", "PRP$^NP~S", "PRP$^NP~PP", "PRP$^NP~VP", "NP^S~SBAR", "NP^S~ROOT", "NP^S~S", "NP^S~VP", "NP^S~SINV", "NP^S~PRN", "NP^S~PP", "VBN^ADJP~VP", "VBN^ADJP~NP", "CD^NP~PP", "CD^NP~NP", "CD^NP~VP", "CD^NP~ADVP", "CD^NP~S", "CD^NP~ADJP", "CD^NP~PRN", "CD^NP~ROOT", "CD^NP~SINV", "PP^VP~S", "PP^VP~NP", "PP^VP~VP", "WP^WHNP~SBARQ"};
  private static String[] splitters3 = new String[] {"RB^VP", "RB^ADVP", "RB^ADJP", "RB^QP", "RB^NP", "RB^CONJP", "RB^PP", "RB^SBAR", "RB^S", "ADJP^NP", "ADJP^VP", "ADJP^ADJP", "ADJP^S", "ADJP^PP", "NX^NP", "NX^NX", "JJ^ADJP", "JJ^NP", "JJ^PP", "JJ^ADVP", "JJ^QP", "JJ^VP", "JJ^NX", "JJ^UCP", "NP^S", "NP^NP", "NP^PP", "NP^VP", "NP^ROOT", "NP^SINV", "NP^ADVP", "NP^ADJP", "NP^PRN", "NP^FRAG", "NP^SQ", "PP^NP", "PP^VP", "PP^S", "PP^ADVP", "PP^PP", "PP^ADJP", "NN^ADJP", "NN^QP", "NN^NX", "NN^VP", "NN^UCP", "NN^ADVP", "NN^WHNP", "NN^NP", "NN^SBAR", "S^ROOT", "S^SBAR", "S^VP", "S^S", "S^PP", "S^SINV", "S^NP", "S^ADJP", "S^PRN", "VBN^NP", "VBN^ADJP", "VBN^PP", "VBN^VP", ":^PRN", "VBZ^NP", "DT^SBAR", "DT^ADVP", "DT^QP", "NNS^NX", "NNS^QP", "NNS^VP", "NNS^WHNP", "NNP^NAC", "NNP^NX", "NNP^ADJP", "NNP^NP", "NNP^VP", "JJR^QP", "VP^VP", "VP^S", "VP^NP", "VP^SINV", "VP^SQ", "CD^QP", "CD^NP", "CD^ADJP", "VBG^PP", "VBG^NP", "VBG^VP", "VBG^ADJP", "IN^SBAR", "IN^PP", "IN^QP", "IN^ADVP", "IN^WHNP", "IN^NP", "IN^CONJP", "IN^S", "QP^ADJP", "PRN^S", "PRN^NP", "VB^NP", "SBAR^NP", "SBAR^VP", "SBAR^S", "ADVP^S", "ADVP^ADVP", "ADVP^NP", "ADVP^PP", "ADVP^VP", "ADVP^ADJP", "CC^S", "CC^NP", "CC^VP", ".^SBARQ", ".^SQ", "S^SBAR~NP", "S^SBAR~VP", "S^SBAR~S", "IN^PP~VP", "IN^PP~S", "IN^PP~NP", "IN^PP~ADJP", "IN^PP~ADVP", "IN^PP~PP", "IN^PP~SINV", "NP^VP~VP", "NP^VP~S", "NP^VP~NP", "JJ^NP~PP", "JJ^NP~S", "JJ^NP~VP", "JJ^NP~NP", "JJ^NP~ROOT", "JJ^NP~ADVP", "JJ^ADJP~NP", "JJ^ADJP~VP", "JJ^ADJP~S", "JJ^ADJP~ADJP", "JJ^ADJP~PP", "RB^ADJP~VP", "RB^ADJP~NP", "RB^ADJP~ADJP", "NP^PP~NP", "NP^PP~VP", "NP^PP~S", "NP^PP~PP", "NP^PP~ADJP", "NP^PP~ADVP", "NP^PP~NAC", "NNS^NP~S", "NNS^NP~PP", "NNS^NP~VP", "NNS^NP~NP", "NNS^NP~ADVP", "NNS^NP~ROOT", "NNS^NP~ADJP", "PRP^NP~VP", "PRP^NP~PP", "PRP^NP~S", "VB^VP~S", "VB^VP~VP", "VBN^VP~NP", "VBN^VP~S", "VBN^VP~VP", "DT^NP~S", "DT^NP~PP", "DT^NP~VP", "DT^NP~NP", "DT^NP~ADVP", "NNPS^NP~PP", "NNPS^NP~S", "RB^ADVP~S", "RB^ADVP~VP", "RB^ADVP~NP", "RB^ADVP~PP", "RB^ADVP~ADVP", "RB^ADVP~SINV", "RB^ADVP~SBAR", "RB^ADVP~ADJP", "VBP^VP~VP", "IN^SBAR~S", "IN^SBAR~VP", "IN^SBAR~NP", "NP^NP~VP", "NP^NP~SINV", "NP^NP~NP", "NP^NP~S", "NP^NP~PP", "NP^NP~ROOT", "NP^NP~PRN", "NP^NP~FRAG", "SBAR^VP~VP", "CD^QP~ADJP", "VBZ^VP~SINV", "VBZ^VP~VP", "VBZ^VP~S", "QP^NP~PP", "VP^S~VP", "VP^S~ROOT", "VP^S~PP", "VP^S~SBAR", "VP^S~NP", "VP^S~S", "VP^S~ADJP", "VP^S~PRN", "VP^S~SINV", "VP^VP~VP", "VP^VP~S", "VP^VP~NP", "VBD^VP~VP", "VBD^VP~SINV", "VBD^VP~NP", "NNP^NP~PP", "NNP^NP~S", "NNP^NP~NP", "NNP^NP~VP", "NNP^NP~SINV", "NNP^NP~ROOT", "NNP^NP~PRN", "NNP^NP~ADJP", "NNP^NP~FRAG", "NN^NP~PP", "NN^NP~S", "NN^NP~NP", "NN^NP~VP", "NN^NP~ADVP", "NN^NP~SINV", "NN^NP~ROOT", "NN^NP~PRN", "NN^NP~FRAG", "NN^NP~ADJP", "ADJP^NP~NP", "ADJP^NP~S", "VBG^VP~NP", "VBG^VP~VP", "VBG^VP~S", "NP^S~SBAR", "NP^S~ROOT", "NP^S~S", "NP^S~VP", "NP^S~SINV", "NP^S~PRN", "CD^NP~PP", "CD^NP~NP", "CD^NP~VP", "CD^NP~ADVP", "CD^NP~S", "CD^NP~ADJP", "CD^NP~PRN", "PP^VP~S", "PP^VP~NP"};
  private static String[] splitters4 = new String[] {"RB^VP", "RB^ADVP", "RB^ADJP", "RB^QP", "RB^NP", "RB^CONJP", "RB^PP", "RB^SBAR", "ADJP^NP", "ADJP^VP", "ADJP^ADJP", "JJ^ADJP", "JJ^NP", "JJ^PP", "JJ^ADVP", "JJ^QP", "JJ^VP", "JJ^NX", "NP^S", "NP^NP", "NP^PP", "NP^VP", "NP^ROOT", "NP^SINV", "NP^ADVP", "NP^ADJP", "NP^PRN", "PP^NP", "PP^VP", "PP^S", "PP^ADVP", "PP^PP", "PP^ADJP", "NN^ADJP", "NN^QP", "NN^NX", "NN^VP", "NN^UCP", "S^ROOT", "S^SBAR", "S^VP", "S^S", "S^PP", "S^SINV", "S^NP", "S^ADJP", "S^PRN", "VBN^NP", "VBN^ADJP", "VBN^PP", "VBN^VP", ":^PRN", "DT^SBAR", "NNS^NX", "NNP^NAC", "NNP^NX", "NNP^ADJP", "VP^VP", "VP^S", "VP^NP", "VP^SINV", "CD^QP", "CD^NP", "CD^ADJP", "VBG^PP", "VBG^NP", "VBG^VP", "IN^SBAR", "IN^PP", "IN^QP", "IN^ADVP", "IN^WHNP", "IN^NP", "SBAR^NP", "SBAR^VP", "SBAR^S", "ADVP^S", "ADVP^ADVP", "CC^S", "CC^NP", "CC^VP", ".^SBARQ", ".^SQ", "S^SBAR~NP", "S^SBAR~VP", "IN^PP~VP", "IN^PP~S", "IN^PP~NP", "IN^PP~ADJP", "IN^PP~ADVP", "NP^VP~VP", "NP^VP~S", "NP^VP~NP", "JJ^NP~PP", "JJ^NP~S", "JJ^NP~VP", "JJ^NP~NP", "JJ^ADJP~NP", "JJ^ADJP~VP", "JJ^ADJP~S", "JJ^ADJP~ADJP", "JJ^ADJP~PP", "NP^PP~NP", "NP^PP~VP", "NP^PP~S", "NP^PP~PP", "NP^PP~ADJP", "NP^PP~ADVP", "NNS^NP~S", "NNS^NP~PP", "NNS^NP~VP", "NNS^NP~NP", "NNS^NP~ADVP", "NNS^NP~ROOT", "PRP^NP~VP", "PRP^NP~PP", "PRP^NP~S", "VB^VP~S", "VBN^VP~NP", "VBN^VP~S", "VBN^VP~VP", "DT^NP~S", "DT^NP~PP", "DT^NP~VP", "DT^NP~NP", "DT^NP~ADVP", "RB^ADVP~S", "RB^ADVP~VP", "RB^ADVP~NP", "RB^ADVP~PP", "RB^ADVP~ADVP", "IN^SBAR~S", "IN^SBAR~VP", "NP^NP~VP", "NP^NP~SINV", "NP^NP~NP", "NP^NP~S", "NP^NP~PP", "NP^NP~ROOT", "SBAR^VP~VP", "CD^QP~ADJP", "VBZ^VP~SINV", "VBZ^VP~VP", "VP^S~VP", "VP^S~ROOT", "VP^S~PP", "VP^S~SBAR", "VP^S~NP", "VP^S~S", "VP^S~ADJP", "VP^S~PRN", "VP^S~SINV", "VP^VP~VP", "VP^VP~S", "VBD^VP~VP", "VBD^VP~SINV", "VBD^VP~NP", "NNP^NP~PP", "NNP^NP~S", "NNP^NP~NP", "NNP^NP~VP", "NNP^NP~SINV", "NNP^NP~ROOT", "NNP^NP~PRN", "NNP^NP~ADJP", "NN^NP~PP", "NN^NP~S", "NN^NP~NP", "NN^NP~VP", "NN^NP~ADVP", "NN^NP~SINV", "NN^NP~ROOT", "VBG^VP~NP", "VBG^VP~VP", "VBG^VP~S", "NP^S~SBAR", "NP^S~ROOT", "NP^S~S", "NP^S~VP", "CD^NP~PP", "CD^NP~NP", "CD^NP~VP", "CD^NP~ADVP", "CD^NP~S", "CD^NP~ADJP"};


  public String[] splitters() {
    switch (EnglishTrain.selectiveSplitLevel) {
    case 1:
      return splitters1;
    case 2:
      return splitters2;
    case 3:
      return splitters3;
    case 4:
      return splitters4;
    default:
      return new String[0];
    }
  }

  // Automatically generated by SisterAnnotationStats -- preferably don't edit
  private static String[] sisterSplit1 = new String[] {"ADJP=l=VBD","ADJP=l=VBP","NP=r=RBR","PRN=r=.","ADVP=l=PP","PP=l=JJ","PP=r=NP","SBAR=l=VB","PP=l=VBG","ADJP=r=,","ADVP=r=.","ADJP=l=VB","FRAG=l=FRAG","FRAG=r=:","PP=r=,","ADJP=l=,","FRAG=r=FRAG","FRAG=l=:","PRN=r=VP","PP=l=RB","S=l=ADJP","SBAR=l=VBN","NP=r=NX","SBAR=l=VBZ","SBAR=l=ADVP","QP=r=JJ","SBAR=l=PP","SBAR=l=ADJP","NP=r=VBG","VP=r=:","VP=l=ADJP","SBAR=l=VBP","ADVP=r=NP","PP=l=VB","VP=r=PP","ADJP=r=SBAR","NP=r=JJR","SBAR=l=NN","S=l=RB","S=l=NNS","S=r=SBAR","S=l=WHPP","VP=l=:","ADVP=l=NP","ADVP=r=PP","ADJP=l=JJ","NP=r=VBN","NP=l=PRN","VP=r=S","NP=r=NNPS","NX=r=NX","ADJP=l=PRP$","SBAR=l=CC","SBAR=l=S","S=l=PRT","ADVP=l=VB","ADVP=r=JJ","NP=l=DT"};
  private static String[] sisterSplit2 = new String[] {"S=r=PP","NP=r=JJS","ADJP=r=NNP","NP=l=PRT","ADJP=r=PP","ADJP=l=VBZ","PP=r=VP","NP=r=CD","ADVP=l=IN","ADVP=l=,","ADJP=r=JJ","ADVP=l=VBD","PP=r=.","S=l=ADVP","S=l=DT","PP=l=NP","VP=l=PRN","NP=r=IN","NP=r=``"};
  private static String[] sisterSplit3 = new String[] {"PP=l=VBD","ADJP=r=NNS","S=l=:","NP=l=ADVP","NP=r=PRN","NP=r=-RRB-","NP=l=-LRB-","NP=l=JJ","SBAR=r=.","S=r=:","ADVP=r=VP","NP=l=RB","NP=r=RB","S=l=VBP","SBAR=r=,","VP=r=,","PP=r=PP","NP=r=S","ADJP=l=NP","VP=l=VBG","PP=l=PP"};
  private static String[] sisterSplit4 = new String[] {"VP=l=NP","NP=r=NN","NP=r=VP","VP=r=.","NP=r=PP","VP=l=TO","VP=l=MD","NP=r=,","NP=r=NP","NP=r=.","NP=l=IN","NP=l=NP","VP=l=,","VP=l=S","NP=l=,","VP=l=VBZ","S=r=.","NP=r=NNS","S=l=IN","NP=r=JJ","NP=r=NNP","VP=l=VBD","S=l=WHNP","VP=r=NP","VP=l=''","VP=l=VBP","NP=l=:","S=r=,","VP=l=``","VP=l=VB","NP=l=S","NP=l=VP","NP=l=VB","NP=l=VBD","NP=r=SBAR","NP=r=:","VP=l=PP","NP=l=VBZ","NP=l=CC","NP=l=''","S=r=NP","S=r=S","S=l=VBN","NP=l=``","ADJP=r=NN","S=r=VP","NP=r=CC","VP=l=RB","S=l=S","S=l=NP","NP=l=TO","S=l=,","S=l=VBD","S=r=''","S=l=``","S=r=CC","PP=l=,","S=l=CC","VP=l=CC","ADJP=l=DT","NP=l=VBG","VP=r=''","SBAR=l=NP","VP=l=VP","NP=l=PP","S=l=VB","SBAR=l=VBD","VP=l=ADVP","VP=l=VBN","NP=r=''","VP=l=SBAR","SBAR=l=,","S=l=WHADVP","VP=r=VP","NP=r=ADVP","QP=r=NNS","NP=l=VBP","S=l=VBZ","NP=l=VBN","S=l=PP","VP=r=CC","NP=l=SBAR","SBAR=r=NP","S=l=VBG","SBAR=r=VP","NP=r=ADJP","S=l=JJ","S=l=NN","QP=r=NN"};

  public String[] sisterSplitters() {
    switch (EnglishTrain.sisterSplitLevel) {
    case 1:
      return sisterSplit1;
    case 2:
      return sisterSplit2;
    case 3:
      return sisterSplit3;
    case 4:
      return sisterSplit4;
    default:
      return new String[0];
    }
  }



  static class EnglishTrain {
    /* THESE OPTIONS ARE ENGLISH-SPECIFIC AND AFFECT ONLY TRAIN TIME */

    /** Annotate prepositions with a S parent (putative subordinating
     *  conjunctions) differently from others (real prepositions). OK.
     */
    public static boolean splitIN = false;
    /** Annotate prepositions 3 ways: S* parent, N* parent or rest
     *  (generally predicative ADJP, VP). Better than sIN.  Good. */
    public static boolean splitIN2 = false;
    /** Annotate prepositions 6 ways: real feature engineering.  Great. */
    public static boolean splitIN3 = false;//true;

    /** Mark the nouns that are percent signs.  Slightly good. */
    public static boolean splitPercent = false;

    /** Join pound with dollar. */
    public static boolean joinPound = false;
    /** Joint comparative and superlative adjective with positive. */
    public static boolean joinJJ = false;

    /** A special test for "such" mainly ("such as Fred"). A wash, so omit */
    public static boolean splitPPJJ = false;

    /** Put a special tag on 'transitive adjectives' with NP complement, like
     *  'due May 15' -- it also catches 'such' in 'such as NP', which may
     *  be a good.  Matches 658 times in 2-21 training corpus. Wash. */
    public static boolean splitTRJJ = false;

    /** Put a special tag on 'adjectives with complements'.  This acts as a 
     *  general subcat feature for adjectives. */
    public static boolean splitJJCOMP = false;

    /** "Intransitive" DT.  Good. */
    public static boolean unaryDT = false;//true;
    /** "Intransitive" RB.  Good. */
    public static boolean unaryRB = false;//true;
    /** "Intransitive" PRP. Wash */
    public static boolean unaryPRP = false;
    /** "Intransitive" IN. Bad */
    public static boolean unaryIN = false;

    /** Annotates "and", and "or" specially - the point of this isn't
     *  so clear to cdm since this is most conjunctions.  "But" different? */
    public static boolean splitCC = false;
    /** Alternative: separate off [Bb]ut and &amp;.  Exclusive with splitCC.
	Marginally better than splitCC. Good. */
    public static boolean splitCC2 = false;//true;

    /** Annotates forms of "not" specially as tag "NOT". BAD */
    public static boolean splitNOT = false;
    /** Split modifier (NP, AdjP) adverbs from others.  
     *  This does nothing if you're already doing tagPA. 
     */
    public static boolean splitRB = false;

    /** Make special tags for forms of BE and HAVE.  Positive PCFG effect,
     *  but neutral to negative in Combo, and impossible if use gPA.  */
    public static boolean splitAux = false;//true;

    /** Pitiful attempt at marking V* preterms with their surface subcat 
     *  frames.  Bad so far. */
    public static boolean vpSubCat = false;

    /** Add (verbal) tags to VPs */
    public static boolean splitVP = false;
    /** Add (verbal) tags to VPs, but collapse "finite" ones together. Good. */
    public static boolean splitVP2 = true;
    /** Mark S nodes according to verbal tag.  Bad. */
    public static boolean splitSTag = false;

    /** Mark phrases which are conjunctions.  Seems marginal */
    public static boolean markCC = false;
    /** Mark phrases which are conjunctions, more like Charniak. 
	Not yet implemented.  Need to annotate _before_ annotate children!
	Charniak: np or vp with two or more np/vp children, a comma, cc or 
	conjp, and nothing else. */
    public static boolean markCC2 = false;

    /** Mark specially S nodes with "gapped" subject (control, raising).
     *  1 is basic version.  2 is better mark S nodes with "gapped" subject.
     *  3 seems best on small training set, but all of these are too similar;
     *  4 can't be differentiated.
     *  5 is done on tree before empty splitting. (Bad!?)
     */
    public static int splitSGapped = 3;

    /** Mark "numeric NPs".  Probably bad? */
    public static boolean splitNumNP = false;

    /** Give a special tag to NPs which are possessive NPs (end in 's) */
    public static boolean splitPoss = false;
    /** Mark base NPs.  Good. */
    public static boolean splitBaseNP = false;
    
    /** Mark base NPs _and_ add a NP node if alone, as in Collins */
    public static boolean collinsBaseNP = false;

    /** Retain NP-TMP annotation.  Good. */
    public static boolean splitNPTMP = true;

    /** Retain all -TMP annotation. */
    public static boolean splitTMP = false;


    /** Set the support * KL cutoff level (1-4) -- use of 2 or 3 is good. */
    public static int selectiveSplitLevel = 1;


    /** Set the support * KL cutoff level (1-4) for sister splitting
     * -- don't use it, as far as we can tell so far  */
    public static int sisterSplitLevel = 1;


    public static void display() {
      String englishParams = "Using EnglishTreebankParserParams"
	+ " splitIN=" + splitIN
	+ " splitIN2=" + splitIN2
	+ " splitIN3=" + splitIN3
	+ " sPercent=" + splitPercent
	+ " j#=" + joinPound
	+ " jJJ=" + joinJJ
	+ " sPPJJ=" + splitPPJJ
	+ " sTRJJ=" + splitTRJJ
	+ " sJJCOMP=" + splitJJCOMP
	+ " unaryDT=" + unaryDT
	+ " unaryRB=" + unaryRB
	+ " unaryPRP=" + unaryPRP
	+ " unaryIN=" + unaryIN
	+ " sCC=" + splitCC
	+ " sCC2=" + splitCC2
	+ " sNT=" + splitNOT
	+ " sRB=" + splitRB
	+ " sAux=" + splitAux
	+ " vpSubCat=" + vpSubCat
	+ " sVP=" + splitVP
	+ " sVP2=" + splitVP2
	+ " sSTag=" + splitSTag
	+ " mCC=" + markCC
	+ " mCC2=" + markCC2
	+ " sSGapped=" + splitSGapped
	+ " numNP=" + splitNumNP
	+ " sPoss=" + splitPoss
	+ " baseNP=" + splitBaseNP
	+ " collinsNP=" + collinsBaseNP
	+ " selSplitLevel=" + selectiveSplitLevel
	+ " sNPTMP=" + splitNPTMP
	+ " sTMP=" + splitTMP
	;
      System.out.println(englishParams);
    }

  }


  /** transformTree does language-specific tree transformations such
     * as splicing. Any parameterizations should be inside the specific
     * TreebankLangParserParams class */
  public TreeHeadPair transformTree(Tree t, Tree root, TreeHeadPair thp) {
    Tree parent;
    Tree grandParent;
    String parentStr;
    String grandParentStr;
    if (root == null || t.equals(root)) {
      parent = null;
      parentStr = "";
    } else {
      parent = t.parent(root);
      parentStr = parent.label().value();
    }
    if(parent == null || parent.equals(root)) {
      grandParent = null;
      grandParentStr = "";
    } else {
      grandParent = parent.parent(root);
      grandParentStr = grandParent.label().value();
    }
    String baseParentStr = tlp.basicCategory(parentStr);
    String baseGrandParentStr = tlp.basicCategory(grandParentStr);
    String word = thp.head.word();
    // Crucially get this from thp to get annotations already done
    String cat = thp.tree.label().value();
    String baseCat = tlp.basicCategory(cat);

    if (t.isPreTerminal()) {

      if (EnglishTrain.splitIN && baseCat.equals("IN") && 
	  parentStr.charAt(0) == 'S') {
	cat = cat+"^S";
      } else if (EnglishTrain.splitIN2 && baseCat.equals("IN")) {
	if (parentStr.charAt(0) == 'S') {
	  cat = cat+"^S";
	} else if (grandParentStr.charAt(0) == 'N') {
	  cat = cat+"^N";
	}
      } else if (EnglishTrain.splitIN3 && baseCat.equals("IN")) {
	// 6 classes seems good!
	// but have played with joining first two, splitting out ADJP/ADVP,
	// and joining two SC cases
	if (grandParentStr.charAt(0) == 'N' && 
	    (parentStr.charAt(0) == 'P' || parentStr.charAt(0) == 'A')) {
	  // noun postmodifier PP (or so-called ADVP like "outside India")
	  cat = cat+"^N";
	} else if (grandParentStr.charAt(0) == 'N' && 
		   parentStr.charAt(0) == 'Q') {
	  // about, than, between, etc. in a QP preceding head of NP
	  cat = cat+"^Q";
	} else if (grandParentStr.equals("S")) {
	  if (baseParentStr.equals("SBAR")) {
	    // sentential subordinating conj: although, if, until, as, while
	    cat = cat+"^SCC";
	  } else {
	    // PP adverbial clause: among, in, for, after
	    cat = cat+"^SC";
	  }
	} else if (baseParentStr.equals("SBAR") || 
		   baseParentStr.equals("WHNP")) {
	  // that-clause complement of VP or NP (or whether, if complement)
	  // but also VP adverbial because, until, as, etc.
	  cat = cat+"^T";
	}
	// all the rest under VP, PP, ADJP, ADVP, etc. are basic case
      }
      if (EnglishTrain.splitPercent && word.equals("%")) {
	// nearly all % are NN, a handful are JJ which we 'correct'
	cat = changeBaseCat(cat, "NN");
	cat += "-%";
      }
      if (EnglishTrain.joinPound && baseCat.equals("#")) {
	cat = changeBaseCat(cat, "$");
      }
      if (EnglishTrain.joinJJ && cat.startsWith("JJ")) {
	cat = changeBaseCat(cat, "JJ");
      }
      if (EnglishTrain.splitPPJJ && cat.startsWith("JJ") &&
	  parentStr.startsWith("PP")) {
	cat = cat+"^S";
      }
      if (EnglishTrain.splitTRJJ && cat.startsWith("JJ") && 
	  (parentStr.startsWith("PP") || parentStr.startsWith("ADJP")) &&
	   etbHeadFinder.determineHead(parent) == t) {
	// look for NP right sister of head JJ -- if so transitive adjective
	Tree[] kids = parent.children();
	boolean foundJJ = false; 
	int i = 0;
	for ( ; i < kids.length && ! foundJJ; i++) {
	  if (kids[i].label().value().startsWith("JJ")) {
	    foundJJ = true;
	  }
	}
	boolean foundNP = false;
	for (int j = i; j < kids.length; j++) {
	  if (kids[j].label().value().startsWith("NP")) {
	    cat = cat+"^T";
	    break;
	  }
	}
      }
      if (EnglishTrain.splitJJCOMP && cat.startsWith("JJ") && 
	  (parentStr.startsWith("PP") || parentStr.startsWith("ADJP")) &&
	  etbHeadFinder.determineHead(parent) == t) {
	// look for _anything_ right sister of JJ -- if so has comp (I guess)
	Tree[] kids = parent.children();
	boolean foundJJ = false; 
	int i = 0;
	for ( ; i < kids.length && ! foundJJ; i++) {
	  if (kids[i].label().value().startsWith("JJ")) {
	    break;
	  }
	}
	if (i < kids.length - 1) {
	  // there's still a complement to go.
	  cat = cat+"^CMPL";
	}
      }
      if (EnglishTrain.unaryDT && cat.startsWith("DT")) {
	if (parent.children().length == 1)
	  cat = cat+"^U";
      }
      if (EnglishTrain.unaryRB && cat.startsWith("RB")) {
	if (parent.children().length == 1)
	  cat = cat+"^U";
      }
      if (EnglishTrain.unaryPRP && cat.startsWith("PRP")) {
	if (parent.children().length == 1)
	  cat = cat+"^U";
      }
      if (EnglishTrain.unaryIN && cat.startsWith("IN")) {
	if (parent.children().length == 1)
	  cat = cat+"^U";
      }
      if (EnglishTrain.splitCC && baseCat.equals("CC") && 
	  (word.equals("and") || word.equals("or")))
	cat = cat+"-C";
      else if (EnglishTrain.splitCC2 && baseCat.equals("CC")) {
	if (word.equalsIgnoreCase("but")) {
	  cat = cat+"-B";
	} else if (word.equals("&")) {
	  cat = cat+"-A";
	}
      }
      if (EnglishTrain.splitNOT && baseCat.equals("RB") && 
	  (word.equalsIgnoreCase("n't") || word.equalsIgnoreCase("not") ||
	   word.equalsIgnoreCase("nt"))) {
	cat = cat+"-N";
      } else if (EnglishTrain.splitRB && baseCat.equals("RB") &&
		 (baseParentStr.equals("NP") || baseParentStr.equals("QP") ||
		  baseParentStr.equals("ADJP"))) {
	cat = cat+"^M";
      }
      if (EnglishTrain.splitAux && (baseCat.equals("VBZ") ||
			     baseCat.equals("VBP") ||
			     baseCat.equals("VBD") ||
			     baseCat.equals("VBN") ||
			     baseCat.equals("VBG") ||
			     baseCat.equals("VB"))) {
	if (word.equalsIgnoreCase("is") ||
	    word.equalsIgnoreCase("am") ||
	    word.equalsIgnoreCase("are") ||
	    word.equalsIgnoreCase("was") ||
	    word.equalsIgnoreCase("were") ||
	    word.equalsIgnoreCase("'m") ||
	    word.equalsIgnoreCase("'re") ||
	    word.equalsIgnoreCase("'s") ||     // imperfect -- could be (ha)s
	    word.equalsIgnoreCase("being") ||
	    word.equalsIgnoreCase("be") ||
	    word.equalsIgnoreCase("been"))
	  cat = cat+"-BE";
	if (word.equalsIgnoreCase("have") ||
	    word.equalsIgnoreCase("'ve") ||
	    word.equalsIgnoreCase("having") ||
	    word.equalsIgnoreCase("has") ||
	    word.equalsIgnoreCase("had") ||
	    word.equalsIgnoreCase("'d"))
	  cat = cat+"-HV";
      }
      if (EnglishTrain.vpSubCat && cat.startsWith("VB")) {
	cat = cat+subCatify(parent);
      }
    } else if (t.isPhrasal()) {
      Tree[] kids = t.children();
      // tag is always really "baseTag"
      String tag = tlp.basicCategory(thp.head.tag());

      if (EnglishTrain.splitVP && baseCat.equals("VP")) { 
	cat = cat+"-"+tag;
      } else if (EnglishTrain.splitVP2 && baseCat.equals("VP")) {
	if (tag.equals("VBZ") || tag.equals("VBD") || tag.equals("VBP") ||
	    tag.equals("MD")) {
	  cat = cat+"-VBF";
	} else {
	  cat = cat+"-"+tag;
	}
      }
      if (EnglishTrain.splitSTag && (baseCat.equals("S") ||
			     baseCat.equals("SINV") ||
			     baseCat.equals("SQ"))) {
	if (EnglishTrain.splitVP2 && (tag.equals("VBZ") || tag.equals("VBD") ||
			      tag.equals("VBP") || tag.equals("MD"))) {
	  cat = cat+"-VBF";
	} else {
	  cat = cat+"-"+tag;
	}
      }
      if (EnglishTrain.markCC) {
	boolean seenCC = false;
	for (int i = 0; i < kids.length; i++) {
	  String cat2 = kids[i].label().value();
	  if (cat2.startsWith("CC")) {
	    cat = cat+"-CC";
	    break;
	  }
	}
      }
      if (EnglishTrain.splitSGapped == 1 && baseCat.equals("S") &&
          ! kids[0].label().value().startsWith("NP")) {
        // this doesn't handle predicative NPs right yet
        // to do that, need to intervene before tree normalization
        cat = cat+"-G";
      } else if (EnglishTrain.splitSGapped == 2 && baseCat.equals("S")) {
	// better version: you're gapped if there is no NP, or there is just
	// one (putatively predicative) NP with no VP, ADJP, NP, PP, or UCP
	boolean seenPredCat = false;
	int seenNP = 0;
	for (int i = 0; i < kids.length; i++) {
	  String cat2 = kids[i].label().value();
	  if (cat.startsWith("NP")) {
	    seenNP++;
	  } else if (cat.startsWith("VP") || cat.startsWith("ADJP") || 
		     cat.startsWith("PP") || cat.startsWith("UCP")) {
	    seenPredCat = true;
	  }
	}
	if (seenNP == 0 || (seenNP == 1 && ! seenPredCat)) {
	  cat = cat+"-G";
	}
      } else if (EnglishTrain.splitSGapped == 3 && baseCat.equals("S")) {
	// better version: you're gapped if there is no NP, or there is just
	// one (putatively predicative) NP with no VP, ADJP, NP, PP, or UCP
	// NEW: but you're not gapped if you have an S and CC daughter (coord)
	boolean seenPredCat = false;
	boolean seenCC = false;
	boolean seenS = false;
	int seenNP = 0;
	for (int i = 0; i < kids.length; i++) {
	  String cat2 = kids[i].label().value();
	  if (cat2.startsWith("NP")) {
	    seenNP++;
	  } else if (cat2.startsWith("VP") || cat2.startsWith("ADJP") || 
		     cat2.startsWith("PP") || cat2.startsWith("UCP")) {
	    seenPredCat = true;
	  } else if (cat2.startsWith("CC")) {
	    seenCC = true;
	  } else if (cat2.startsWith("S")) {
	    seenS = true;
	  }
	}
	if ((! (seenCC && seenS)) && 
	    (seenNP == 0 || (seenNP == 1 && ! seenPredCat))) {
	  cat = cat+"-G";
	}
      } else if (EnglishTrain.splitSGapped == 4 && baseCat.equals("S")) {
	// better version: you're gapped if there is no NP, or there is just
	// one (putatively predicative) NP with no VP, ADJP, NP, PP, or UCP
	// But: not gapped if S(BAR)-NOM-SBJ constituent
	// But: you're not gapped if you have two /^S/ daughters
	boolean seenPredCat = false;
	boolean sawSBeforePredCat = false;
	int seenS = 0;
	int seenNP = 0;
	for (int i = 0; i < kids.length; i++) {
	  String cat2 = kids[i].label().value();
	  if (cat2.startsWith("NP")) {
	    seenNP++;
	  } else if (cat2.startsWith("VP") || cat2.startsWith("ADJP") || 
		     cat2.startsWith("PP") || cat2.startsWith("UCP")) {
	    seenPredCat = true;
	  } else if (cat2.startsWith("S")) {
	    seenS++;
	    if (!seenPredCat) {
	      sawSBeforePredCat = true;
	    }
	  }
	}
	if ((seenS < 2) && ( ! (sawSBeforePredCat && seenPredCat)) &&
	    (seenNP == 0 || (seenNP == 1 && ! seenPredCat))) {
	  cat = cat+"-G";
	}
      }
      if (EnglishTrain.splitNumNP && baseCat.equals("NP")) {
	boolean seenNum = false;
	for (int i = 0; i < kids.length; i++) {
	  String cat2 = kids[i].label().value();
	  if (cat2.startsWith("QP") || cat2.startsWith("CD") ||
	      cat2.startsWith("$") || cat2.startsWith("#") ||
	      (cat2.startsWith("NN") && cat2.indexOf("-%") >= 0 )) {
	    seenNum = true;
	    break;
	  }
	}
	if (seenNum) {
	  cat += "-NUM";
	}
      }
      if (EnglishTrain.splitPoss && baseCat.equals("NP") &&
	  kids[kids.length - 1].label().value().startsWith("POS")) {
	cat = cat+"-P";
      }
      if (EnglishTrain.collinsBaseNP && baseCat.equals("NP")) {
	if (t.isPrePreTerminal()) {
	  if (parentStr.startsWith("NP")) { // already got one above us
	    cat = cat+"-B";
	  } else {
	    // special case splice in a new node!  Do it all here
	    Label labelBot = new CategoryWordTag("NP^NP-B", word, tag);
	    thp.tree.setLabel(labelBot);
	    TreeHeadPair resultTop = new TreeHeadPair();
	    resultTop.head = thp.head;
	    Label labelTop = new CategoryWordTag(cat, word, tag);
	    List newerChildren = new ArrayList(1);
	    newerChildren.add(thp.tree);
	    resultTop.tree = tf.newTreeNode(labelTop, newerChildren);
	    return resultTop;
	  }
	}
      } else if (EnglishTrain.splitBaseNP && baseCat.equals("NP")) {
	if (t.isPrePreTerminal()) {
	  cat = cat+"-B";
	}
      }
    }

    // update to new value
    if (t.isPreTerminal()) {
      thp.head.setTag(cat);
    }
    Label label = new CategoryWordTag(cat, thp.head.word(), thp.head.tag());
    thp.tree.setLabel(label);
    return thp;
  }

 
  private static String subCatify(Tree t) {
    StringBuffer sb = new StringBuffer("^a");
    boolean n = false;
    boolean s = false;
    boolean p = false;
    for (int i=0; i<t.children().length; i++) {
      String childStr = t.children()[i].label().value();
      n = (n || childStr.startsWith("NP"));
      s = (s || childStr.startsWith("S"));
      p = (p || childStr.startsWith("PP"));
    }
    n = false;
    if (n)
      sb.append("N");
    if (p)
      sb.append("P");
    if (s)
      sb.append("S");
    return sb.toString();
  }


  private String changeBaseCat(String cat, String newBaseCat) {
    int i = 1;  // not 0 in case tag is annotation introducing char
    int leng = cat.length();
    for ( ; (i < leng); i++) {
      if (tlp.isLabelAnnotationIntroducingCharacter(cat.charAt(i))) {
	break;
      }
    }
    if (i < leng) {
      return newBaseCat + cat.substring(i);
    } else {
      return newBaseCat;
    }
  }


  public void display() {
    EnglishTrain.display();
  }


  /** Set language-specific options according to flags.
   *  This routine should process the option starting in args[i] (which
   *  might potentially be several arguments long if it takes arguments).
   *  It should return the index after the last index it consumed in
   *  processing.  In particular, if it cannot process the current option,
   *  the return value should be i.
   */
  public int setOptionFlag(String[] args, int i) {
    if (args[i].equalsIgnoreCase("-splitIN")) {
      EnglishTrain.splitIN3 = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-splitPercent")) {
      EnglishTrain.splitPercent = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-splitTRJJ")) {
      EnglishTrain.splitTRJJ = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-splitJJCOMP")) {
      EnglishTrain.splitJJCOMP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-unaryDT")) {
      EnglishTrain.unaryDT = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-unaryRB")) {
      EnglishTrain.unaryRB = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-splitCC")) {
      EnglishTrain.splitCC2 = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-splitRB")) {
      EnglishTrain.splitRB = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-splitAux")) {
      EnglishTrain.splitAux = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-splitVP")) {
      EnglishTrain.splitVP2 = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-splitSG")) {
      EnglishTrain.splitSGapped = 4;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-splitNPTMP")) {
      EnglishTrain.splitNPTMP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-splitTMP")) {
      EnglishTrain.splitTMP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-possNP")) {
      EnglishTrain.splitPoss = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-baseNP")) {
      EnglishTrain.splitBaseNP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-baseNPc")) {
      EnglishTrain.collinsBaseNP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-vSelSplitLevel") &&
	       (i+1 < args.length)) {
      EnglishTrain.selectiveSplitLevel = Integer.parseInt(args[i+1]);
      i+=2;
    } else if (args[i].equalsIgnoreCase("-acl03pcfg")) {
      EnglishTrain.splitIN3 = true;
      EnglishTrain.splitPercent = true;
      EnglishTrain.splitPoss = true;
      EnglishTrain.splitCC2 = true;
      EnglishTrain.unaryDT = true;
      EnglishTrain.unaryRB = true;
      EnglishTrain.splitAux = true;
      EnglishTrain.splitVP2 = true;
      EnglishTrain.splitSGapped = 4;
      EnglishTrain.splitNPTMP = true;
      EnglishTrain.selectiveSplitLevel = 2;
      EnglishTrain.splitBaseNP = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-ijcai03")) {
      EnglishTrain.splitIN3 = true;
      EnglishTrain.splitPercent = true;
      EnglishTrain.splitPoss = true;
      EnglishTrain.splitCC2 = true;
      EnglishTrain.unaryDT = false;
      EnglishTrain.unaryRB = false;
      EnglishTrain.splitAux = false;
      EnglishTrain.splitVP2 = true;
      EnglishTrain.splitSGapped = 4;
      EnglishTrain.splitNPTMP = true;
      EnglishTrain.selectiveSplitLevel = 2;
      EnglishTrain.splitBaseNP = true;
      i+=1;
    }
    return i;
  }

}
