package edu.stanford.nlp.parser.lexparser;

import java.util.*;
import java.util.regex.*;

import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.util.Numberer;

class FilterScorer implements Scorer {

  protected int MAX_LENGTH;

  protected Lexicon lex;
  protected PCFGGrammar grammar;

  protected List words;

  protected int[][] filters; // state -> left completion requirements

  protected int[] tagOf;

  protected boolean[][] tags; // position x tag -> present
  
  protected double max(double x, double y) { return (x > y ? x : y); }

  protected void buildFilters() {
    Numberer stateNumberer = grammar.stateNumberer();
    Numberer tagNumberer = lex.tagNumberer();
    tagOf = new int[stateNumberer.total()];
    Pattern p = Pattern.compile("@[^:]+:\\s*(.*)\\s*\\.\\.\\.\\s*(.*)");
    filters = new int[stateNumberer.total()][];
    for (int state = 0; state < stateNumberer.total(); state++) {
      String stateStr = (String)stateNumberer.object(state);
      if (tagNumberer.hasSeen(stateStr)) {
	tagOf[state] = tagNumberer.number(stateStr);
      } else {
	tagOf[state] = -1;
      }
      if (stateStr.length() > 1 &&
	  stateStr.charAt(0) == '@' &&
	  stateStr.charAt(1) != '?') {
	Matcher m = p.matcher(stateStr);
	if (m.matches()) {
	  String rightThings = m.group(2);
	  String[] things = rightThings.split(" ");
	  //System.out.println("Going after "+rightThings+" from "+stateStr);
	  filters[state] = new int[things.length];
	  for (int i=0; i<things.length; i++) {
	    int tag = -1;
	    if (tagNumberer.hasSeen(things[i])) {
	      tag = tagNumberer.number(things[i]);
	    }
	    filters[state][i] = tag;
	  }
	}
      } else {
	filters[state] = null;
      }
    }
  }
  
  protected boolean filter(int state, int loc) {
    int[] filter = filters[state];
    if (filter == null)
      return true;
    // crude version, fix later
    return matches(filter, 0, loc, true);
  }

  protected boolean looseFilter(int state, int loc) {
    int[] filter = filters[state];
    if (filter == null)
      return true;
    // crude version, fix later
    return matches(filter, 0, loc, false);
  }

  protected boolean matches(int[] filter, int fPos, int wPos, boolean locked) {
    if (fPos >= filter.length)
      return true;
    if (wPos >= tags.length)
      return false;
    int tag = filter[fPos];
    if (tag == -1) {
      return matches(filter, fPos+1, wPos+1, false);
    }
    if (locked) {
      if (!tags[wPos][tag])
	return false;
      return matches(filter, fPos+1, wPos+1, true);
    }
    boolean ok = false;
    if (tags[wPos][tag])
      ok = (ok || matches(filter, fPos+1, wPos+1, true));
    ok = (ok || matches(filter, fPos, wPos+1, false));
    return ok;
  }

  protected boolean filterFast(int state, int loc) {
    int[] filter = filters[state];
    if (filter == null)
      return true;
    // crude version, fix later
    if (filter[0] != -1 && ! tags[loc][filter[0]])
      return false;
    int matching = 0;
    for (int pos = loc; pos < tags.length; pos++) {
      int f = filter[matching];
      if (f == -1 || tags[pos][f]) {
	matching++;
	if (matching >= filter.length)
	  return true;
      }
    }
    return false;
  }

  protected boolean looseFilterFast(int state, int loc) {
    int[] filter = filters[state];
    if (filter == null)
      return true;
    // crude version, fix later
    int matching = 0;
    for (int pos = loc; pos < tags.length; pos++) {
      int f = filter[matching];
      if (f == -1 || tags[pos][f]) {
	matching++;
	if (matching >= filter.length)
	  return true;
      }
    }
    return false;
  }

  protected void buildTaggings() {
    tags = new boolean[words.size()][lex.tagNumberer().total()];
    for (int pos=0; pos<words.size(); pos++) {
      for (Iterator tI = lex.ruleIteratorByWord(Numberer.getGlobalNumberer("words").number(words.get(pos).toString()), pos); tI.hasNext();) {
	IntTaggedWord ur = (IntTaggedWord)tI.next();
	//System.out.println(ur);
	tags[pos][ur.tag] = true;
      }
    }
  }

  protected boolean filterPostHook(int subState, int state, int end) {
    int tag = tagOf[subState];
    if (tag != -1 && ! tags[end][tag])
      return false;
    if (tag != -1 && ! filter(state, end+1))
      return false;
    if (tag == -1 && ! looseFilter(state, end+1))
      return false;
    return true;
  }


  public double oScore(Edge edge) {
    //if (!filter(edge.state, edge.end) &&
    //	filterFast(edge.state, edge.end))
    //  System.out.println("Blocking "+edge);
    return (filter(edge.state, edge.end) ? 0.0 : Double.NEGATIVE_INFINITY);
  }

  public double iScore(Edge edge) {
    return 0.0;
  }

  public boolean iPossible(Hook hook) {
    return true;
  }

  public boolean oPossible(Hook hook) {
    if (hook.isPostHook()) {
      return filterPostHook(hook.subState, hook.state, hook.end);
    }
    return true;
  }

  public boolean parse(List words) {
    this.words = words;
    buildTaggings();
    return true;
  }

  public FilterScorer(Lexicon lex, PCFGGrammar grammar) {
    this.lex = lex;
    this.grammar = grammar;
    buildFilters();
  }

}
