// StanfordLexicalizedParser -- a probabilistic lexicalized NL CFG parser
// Copyright (c) 2002, 2003 Leland Stanford Junior University
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// For more information, bug reports, fixes, contact:
//    Christopher Manning
//    Dept of Computer Science, Gates 4A
//    Stanford CA 94305-9040
//    USA
//    manning@cs.stanford.edu
//    http://nlp.stanford.edu/downloads/lex-parser.shtml

package edu.stanford.nlp.parser.lexparser;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.zip.*;

import edu.stanford.nlp.parser.*;
import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.util.*;
import edu.stanford.nlp.io.NumberRangeFileFilter;
import edu.stanford.nlp.process.*;
import edu.stanford.nlp.dbm.*;

/** A reasonably good lexicalized PCFG parser.  It does a product-of-experts
 *  model of plain PCFG parsing and lexicalized dependency parsing.  Or it
 *  can do unlexicalized PCFG parsing by using just that component parser.
 *  Note that training requires a lot of memory to run.  Try -mx1500m.
 *  See the package documentation for more details and examples of use.
 *  See the main method documentation for details of invoking the parser.
 *
 *  @author Dan Klein
 *  @author Christopher Manning (expanded the API)
 *  @author Teg Grenager (made small interface changes)
 */
public class LexicalizedParser implements ViterbiParser, Appliable {

  private boolean fallbackToPCFG = false;

  private ExhaustivePCFGParser pparser;
  private ExhaustiveDependencyParser dparser;
  private BiLexPCFGParser bparser;
  private TreeTransformer debinarizer;
  private SubcategoryStripper subcategoryStripper;

  private transient ParserData pd;
  private TreebankLangParserParams tlpParams =
    new EnglishTreebankParserParams();;


  /** Allows the caller to specify a TreebankLangParserParams to use. 
   *  @param tlpp The one to use
   */
  public void setTreebankLangParserParams(TreebankLangParserParams tlpp) {
    tlpParams = tlpp;
  }


  /** Converts a Sentence/List into a Tree.
   *  If it can't be parsed, it is made into a trivial tree in which each
   *  word is attached to a start nonterminal.
   *  @param in The input Sentence/List
   *  @return A Tree that is the parse tree for the sentence.  If the parser
   *          fails, a new Tree is synthesized which attaches all words to the
   *          root.
   *  @throws IllegalArgumentException If argument isn't a List
   */
  public Object apply(Object in) {
    if ( ! (in instanceof List)) {
      throw new IllegalArgumentException("Can only parse Sentence/List");
    }
    try {
      if (parse((List) in)) {
        return getBestParse();
      }
    } catch (Exception e) {
      // if can't parse or exception, fall through
    }
    TreeFactory lstf = new LabeledScoredTreeFactory();
    String root = Options.get().langpack.startSymbol();
    return lstf.newTreeNode(root, (List) in);
  }


  /** Parse a Sentence.
   *  @return true iff it could be parsed
   */
  public boolean parse(Sentence sentence) {
    return parse((List) sentence);
  }


  /** Parse a Sentence.
   *  This hasn't yet been implemented.  At present the goal is ignored.
   *  @return true iff it could be parsed
   */
  public boolean parse(Sentence sentence, String goal) {
    return parse((List) sentence);
  }


  private static boolean parseSucceeded = false;

  /** Parse a sentence represented as a List.
   *  @param sentence The sentence to parse
   *  @return true Iff the sentence was accepted by the grammar
   *  @throws UnsupportedOperationException If the Sentence is too long or
   *        otherwise fails for resource reasons
   */
  public boolean parse(List sentence) {
    parseSucceeded = false;
    Sentence sentenceB = new Sentence(sentence);
    sentenceB.add(Lexicon.BOUNDARY);
    int leng = sentenceB.length();
    if (leng > Test.maxLength + 1) {  // + 1 for boundary symbol
      throw new UnsupportedOperationException("Sentence too long");
    }
    if (Options.get().doPCFG) {
      if (! pparser.parse(sentenceB))
        return parseSucceeded;
      if (Test.verbose && Options.get().doDep) {
        System.out.println("PParser output");
        debinarizer.transformTree(pparser.getBestParse(), null).pennPrint();
        //(subcategoryStripper.transformTree(debinarizer.transformTree(pparser.getBestParse(),null),null)).pennPrint();
      }
    }
    if (Options.get().doDep) {
      dparser.parse(sentenceB);
      if (Test.verbose) {
        System.out.println("DParser output");
        dparser.getBestParse().pennPrint();
      }
    }
    if (Options.get().doPCFG &&
        Options.get().doDep) {
      if (! bparser.parse(sentenceB))
        return parseSucceeded;
      else
        parseSucceeded = true;
    }
    return true;//parseSucceeded;
  }


  /** Return the best parse of the sentence most recently parsed.
   *  @return The best tree
   *  @throws NoSuchElementException If no previously successfully parsed
   *            sentence
   */
  public Tree getBestParse() {
    if (! parseSucceeded || bparser == null) {
      if (! pparser.hasParse() ||
          ! fallbackToPCFG)
        throw new NoSuchElementException();
      else
        return getBestPCFGParse();
    }
    Tree binaryTree = bparser.getBestParse();
    return subcategoryStripper.transformTree(debinarizer.transformTree(
                binaryTree, null), null);
  }


  public Tree getBestPCFGParse() {
    Tree binaryTree = pparser.getBestParse();
    return subcategoryStripper.transformTree(debinarizer.transformTree(
                binaryTree, null), null);
  }


  public Tree getBestDependencyParse() {
    return (dparser != null ? dparser.getBestParse() : null);
  }

  /**
   * Construct a new LexicalizedParser object from a previously assembled
   * grammar read from a property
   * <code>edu.stanford.nlp.SerializedLexicalizedParser</code>,
   * or a default place.
   * @param in The ObjectInputStream
   */
  public LexicalizedParser() {
    String source =
      System.getProperty("edu.stanford.nlp.SerializedLexicalizedParser");
    if (source == null) {
      source = "/u/nlp/data/serializedParser.gz";
    }
    ParserData pd = deserializeParser(source);
    if (pd != null) {
      init2(pd);
    }
  }


  /**
   * Construct a new LexicalizedParser.  This loads a grammar that
   * was previously assembled and stored.
   * @param treebankPath a <code>String</code> value
   * @throws IllegalArgumentException If parser data cannot be loaded
   */
  public LexicalizedParser(String serializedFileOrUrl) {
    ParserData pd = deserializeParser(serializedFileOrUrl);
    if (pd == null) {
      throw new IllegalArgumentException("Error loading parser data");
    }
    init2(pd);
  }


  /**
   * Construct a new LexicalizedParser.  This loads a grammar that
   * was previously assembled and stored.
   * @param treebankPath a <code>String</code> value
   * @param maxLeng Maximum sentence length that you want the parser to be 
   *    able to parse (this effects memory consumption)
   * @throws IllegalArgumentException If parser data cannot be loaded
   */
  public LexicalizedParser(String serializedFileOrUrl, int maxLeng) {
    Test.maxLength = maxLeng;
    ParserData pd = deserializeParser(serializedFileOrUrl);
    if (pd == null) {
      throw new IllegalArgumentException("Error loading parser data");
    }
    init2(pd);
  }


  /**
   * Construct a new LexicalizedParser object from a previously assembled
   * grammar.
   * @param pd A <code>ParserData</code> object (not <code>null</code>)
   */
  public LexicalizedParser(ParserData pd) {
    init2(pd);
  }


  /**
   * Construct a new LexicalizedParser object from a previously assembled
   * grammar read from an InputStream.
   * One (ParserData) object is read from the stream.  It isn't closed.
   * @param in The ObjectInputStream
   */
  public LexicalizedParser(ObjectInputStream in) throws Exception {
    this((ParserData) in.readObject());
  }


  /**
   * Construct a new LexicalizedParser object from a previously assembled
   * grammar read from an InputStream.
   * One (ParserData) object is read from the stream.  It isn't closed.
   * @param in The ObjectInputStream
   * @param maxLeng Maximum sentence length that you want the parser to be 
   *    able to parse (this effects memory consumption)
   */
  public LexicalizedParser(ObjectInputStream in, int maxLeng)
      throws Exception {
    Test.maxLength = maxLeng;
    ParserData pd = (ParserData) in.readObject();
    init2(pd);
  }


  /**
   * Construct a new LexicalizedParser by training from treebank files.
   * @param treebankPath a <code>String</code> value
   * @param filt a <code>FileFilter</code> value. This may be
   *       <code>null</code> if no filtering of selected files is needed.
   */
  public LexicalizedParser(String treebankPath, FileFilter filt,
                           TreebankLangParserParams tlpParams) {
    this.tlpParams = tlpParams;
    init(treebankPath, filt);
  }


  /**
   * Construct a new LexicalizedParser by training from treebank files.
   * @param treebankPath a <code>String</code> value
   */
  public LexicalizedParser(String treebankPath,
                           TreebankLangParserParams tlpParams) {
    this(treebankPath, null, tlpParams);
  }


  /**
   * Construct a new LexicalizedParser.
   * @param treebankPath a <code>String</code> value
   * @param filt a <code>FileFilter</code> value
   * @param maxLeng The maximum length sentences to be able to parser.
   *           A large value for this requires a great deal of memory (and
   *           time) for parsing, but allows parsing longer sentences.
   * @param tlpParams The Treebank parameters class for different languages
   */
  public LexicalizedParser(String treebankPath, FileFilter filt, int maxLeng,
                           TreebankLangParserParams tlpParams) {
    this.tlpParams = tlpParams;
    Test.maxLength = maxLeng;
    init(treebankPath, filt);
  }


  /**
   * Construct a new LexicalizedParser.
   * @param treebankPath a <code>String</code> value
   * @param filt a <code>FileFilter</code> value
   * @param maxLeng The maximum length sentences to be able to parser.
   *           A large value for this requires a great deal of memory (and
   *           time) for parsing, but allows parsing longer sentences.
   */
  public LexicalizedParser(String treebankPath, FileFilter filt, int maxLeng) {
    Test.maxLength = maxLeng;
    init(treebankPath, filt);
  }


  private final void init(String treebankPath, FileFilter filt) {
    pd = init1(treebankPath, filt);
    init2(pd);
  }

  private ParserData parserData() {
    return pd;
  }

  static void serializeParser(ParserData pd, String filename) {
    try {
      ObjectOutputStream out;
      if (filename.endsWith(".gz")) {
	// it's faster to do the buffering _outside_ the gzipping as here
        out = new ObjectOutputStream(new BufferedOutputStream(
                new GZIPOutputStream(new FileOutputStream(filename))));
      } else {
        out = new ObjectOutputStream(
                new BufferedOutputStream(new FileOutputStream(filename)));
      }
      out.writeObject(pd);
      out.close();
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }

  protected static ParserData deserializeParser(String filenameOrUrl) {
    ParserData pd = null;
    try {
      ObjectInputStream in;
      InputStream is;
      if (filenameOrUrl.startsWith("http://")) {
        URL u = new URL(filenameOrUrl);
        URLConnection uc = u.openConnection();
        is = uc.getInputStream();
      } else {
        is = new FileInputStream(filenameOrUrl);
      }
      if (filenameOrUrl.endsWith(".gz")) {
	// it's faster to do the buffering _outside_ the gzipping as here
        in = new ObjectInputStream(new BufferedInputStream(new GZIPInputStream(
                                              is)));
      } else {
        in = new ObjectInputStream(new BufferedInputStream(is));
      }
      pd = (ParserData) in.readObject();
      in.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    return pd;
  }


  private final ParserData init1(String treebankPath, FileFilter filt) {
    System.out.println("Currently "+new Date());
    // setup tree transforms
    Treebank trainTreebank = tlpParams.memoryTreebank();
    System.out.print("Reading trees...");
    if (filt == null) {
      trainTreebank.loadPath(treebankPath);
    } else {
      trainTreebank.loadPath(treebankPath, filt);
    }
    Timing.tick("done.");
    System.out.print("Binarizing trees...");
    TreeTransformer binarizer = BinarizerFactory.getHeadRespectingBinarizer(
                        tlpParams.headFinder(), !Train.outsideFactor());
    List binaryTrainTrees = new ArrayList();
    if (Train.hSelSplit) {
      ((BinarizerFactory.HeadRespectingBinarizer)binarizer).countStates = true;
      for (Iterator tI=trainTreebank.iterator(); tI.hasNext();) {
        Tree tree = (Tree)tI.next();
        if (Train.collinsPunc)
          tree = CollinsPuncTransformer.transformTree(tree);
        tree = binarizer.transformTree(tree, null);
        //binaryTrainTrees.add(tree);
      }
      ((BinarizerFactory.HeadRespectingBinarizer)binarizer).countStates = false;
    }
    for (Iterator tI=trainTreebank.iterator(); tI.hasNext();) {
      Tree tree = (Tree)tI.next();
      if (Train.collinsPunc)
	tree = CollinsPuncTransformer.transformTree(tree);
      tree = binarizer.transformTree(tree, null);

      binaryTrainTrees.add(tree);
    }
    Timing.tick("done.");
    // extract grammars
    //Extractor bgExtractor = new SmoothedBinaryGrammarExtractor();
    Extractor bgExtractor = new BinaryGrammarExtractor();
    // Extractor lexExtractor = new LexiconExtractor();
    Extractor dgExtractor = new DependencyGrammarExtractor();
    //TreeExtractor uwmExtractor = new UnknownWordModelExtractor(binaryTrainTrees.size());
    Extractor cdExtractor = new CorrectDependencyExtractor();
    BinaryGrammar bg = null;
    UnaryGrammar ug = null;
    System.out.print("Extracting PCFG...");
    Pair bgug = (Pair)bgExtractor.extract(binaryTrainTrees);
    bg = (BinaryGrammar)bgug.second;
    bg.splitRules();
    ug = (UnaryGrammar)bgug.first;
    ug.purgeRules();
    Timing.tick("done.");
    System.out.print("Extracting Lexicon...");
    Lexicon lex = new Lexicon();
    lex.train(binaryTrainTrees);
    Timing.tick("done.");
    System.out.print("Tuning Lexicon...");
    lex.tune(binaryTrainTrees);
    Timing.tick("done.");

    DependencyGrammar dg = null;
    if (Options.get().doDep) {
      System.out.print("Extracting Dependencies...");
      dg = (DependencyGrammar)dgExtractor.extract(binaryTrainTrees);
      Timing.tick("done.");
      //System.out.println(dg);
      //System.out.print("Extracting Unknown Word Model...");
      //UnknownWordModel uwm = (UnknownWordModel)uwmExtractor.extract(binaryTrainTrees);
      //Timing.tick("done.");
      System.out.print("Extracting Tune Dependencies...");
      List tuneDeps = (List)cdExtractor.extract(binaryTrainTrees);
      Timing.tick("done.");
      System.out.print("Tuning Dependency Model...");
      dg.tune(tuneDeps);
      Timing.tick("done.");
    }
    Map numbs = Numberer.getNumberers();
    return new ParserData(lex, bg, ug, dg, numbs, Options.get());
  }


  private final void init2(ParserData pd) {
    BinaryGrammar bg = pd.bg;
    UnaryGrammar ug = pd.ug;
    Lexicon lex = pd.lex;
    DependencyGrammar dg = pd.dg;
    Numberer.setNumberers(pd.numbs);
    Options.set(pd.pt);

    pparser = new ExhaustivePCFGParser(bg, ug, lex);
    if (Options.get().doDep) {
      dparser = new ExhaustiveDependencyParser(dg, lex);

      Scorer scorer = new TwinScorer(pparser, dparser);
      //Scorer scorer = parser;
      bparser = new BiLexPCFGParser(scorer, pparser, dparser, bg, ug, dg, lex);
      bparser.setMaxItems(200000);
    }
    fallbackToPCFG = true;

    debinarizer = BinarizerFactory.getDebinarizer();
    subcategoryStripper = new SubcategoryStripper();
  }


  /** returns the input sentence for the parser */
  static Sentence getInputSentence(Tree t) {
    if(Test.forceTags) { 
      if(Test.preTag) // doesn't yet support 
	throw new RuntimeException("Sorry -- haven't implemented support for external tagger yet.");
      else
	return t.taggedYield();
    }
    else
      return t.yield();
  }


  /** A simple main program for using the parser.
   *  This provides three modes of usage: one for building and serializing
   *  a parser from treebank data, one for parsing sentences from a file
   *  or URL based
   *  on a serialized parser, and a third (mainly for parser quality testing)
   *  for training and testing a parser all in one go. <p>
   *  Usages: <br><code>
   *  java edu.stanford.nlp.parser.lexparser.LexicalizedParser [-v] -train
   *      trainFilesPath [start stop] serializedParserFilename<p>
   *  java edu.stanford.nlp.parser.lexparser.LexicalizedParser [-v] -validate
   *      trainFilesPath [start stop
   *      -treebank [testFilePath [start stop]]]<p>
   *  java -mx512m edu.stanford.nlp.parser.lexparser.LexicalizedParser
   *      [-v] serializedParserFilename filename+<p>
   *  java -mx512m edu.stanford.nlp.parser.lexparser.LexicalizedParser
   *      [-v] serializedParserFilename -treebank testFilePath [start stop]
   *  </code><p>
   *  If the <code>serializedParserFilename</code> ends in <code>.gz</code>,
   *  then the serialization data is written and read compressed (GZip).
   *  The argument <code>filename</code> may be a URL, starting with
   *  <code>http://</code>.
   *  If no files are supplied in the third usage, then a hardwired sentence
   *  is parsed.
   *  All final arguments are passed to FactoredParser.<p>
   *  In the same position as the verbose flag (<code>-v</code>), many other
   *  options can be specified.  The most useful to an end user are:
   *  <UL>
   *  <LI><code>-tLPP class</code> Specify a different 
   *     TreebankLangParserParams, for when using a different language or
   *     treebank (the default is English Penn Treebank)</LI>
   *  <LI><code>-encoding charset</code> Specify the character encoding of the
   *     input files</LI>
   *  <LI><code>-tokenized</code> Says that the input is already separated
   *     into whitespace-delimited tokens</LI>
   *  <LI><code>-tokenizer class</code> Specifies a Tokenizer class to be used
   *     for tokenization</LI>
   *  <LI><code>-sentences token </code> Specifies a token that marks sentence
   *     bounaries, or "sDelimited" or "newline", which have special
   *     interpretations (see package documentation)</LI>
   *  <LI><code>-tagSeparator char</code> Specifies to look for tags on words
   *     separated by a reserved character char.</LI>
   *  <LI><code>-maxLength leng</code> Specify the longest sentence that 
   *     will be parsed (and hence indirectly the amount of memory needed.</LI>
   *  <LI><code>-outputTreeFormat style</code> Choose the style of output
   *     sentences: <code>penn</code> for prettyprinting as in the Penn
   *     treebank files, or <code>oneline</code> for printing sentences one
   *     per line.</LI>
   *  </UL>
   *  @param args Command line arguments, as above
   */
  public static void main(String[] args) {
    TreebankLangParserParams tlpParams = new EnglishTreebankParserParams();
    //TreebankLangParserParams tlpParams = new ChineseTreebankParserParams();
    boolean train = false;
    boolean serialize = false;
    Tokenizer tokenizer = null;
    boolean tokenized = false;
    char tagSeparator = 0;
    boolean sentencesSplit = false;
    String sentenceBoundary = null;
    String suffix = "mrg";
    String inputFormat;
    int argIndex = 0;
    if (args.length < 1) {
      System.err.println("usage: " +
              "java edu.stanford.nlp.parser.lexparser.LexicalizedParser " +
              "serializedParserFilename filename*");
      // System.err.println("usage: " +
      //      "java edu.stanford.nlp.parser.lexparser.LexicalizedParser " +
      //       "-validate trainFilesPath [start stop [testFilePath " +
      //      "[start stop [maxLeng]]]]");
      System.exit(1);
    }

    // while loop through option arguments
    while (argIndex < args.length && args[argIndex].startsWith("-")) {
      if (args[argIndex].equalsIgnoreCase("-train")) {
        train = true;
        serialize = true;
        argIndex++;
      } else if (args[argIndex].equalsIgnoreCase("-tLPP") && (argIndex+1 < args.length)) {
        try {
          tlpParams =
            (TreebankLangParserParams) Class.forName(args[argIndex+1]).newInstance();
        } catch (Exception e) {
          System.err.println("Couldn't instantiate TreebankLangParserParams " + args[argIndex+1]);
        }
        argIndex+=2;
      } else if (args[argIndex].equalsIgnoreCase("-encoding")) { // sets encoding for TreebankLangParserParams

	tlpParams.setInputEncoding(args[argIndex+1]);
	tlpParams.setOutputEncoding(args[argIndex+1]);
	argIndex+=2;
      }	else if (args[argIndex].equalsIgnoreCase("-validate")) {
	train = true;
	argIndex++;
      } else if (args[argIndex].equalsIgnoreCase("-suffix")) {
	suffix = args[argIndex+1];
	argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-tokenized")) {
	  tokenized = true;
	argIndex += 1;
      } else if (args[argIndex].equalsIgnoreCase("-tokenizer")) {
	  try {
	      tokenizer = (Tokenizer) Class.forName(args[argIndex+1]).newInstance();
	  } catch (Exception e) {
	      System.err.println("Couldn't instantiate Tokenizer " + args[argIndex+1]);
	  }
	  argIndex+=2;
      } else if (args[argIndex].equalsIgnoreCase("-sentences")) {
	  sentencesSplit = true;
	  sentenceBoundary = args[argIndex+1];
	  argIndex+=2;
      } else if (args[argIndex].equalsIgnoreCase("-tagSeparator")) {
	  tagSeparator = args[argIndex+1].charAt(0);
	  argIndex+=2;
      } else if (args[argIndex].equalsIgnoreCase("-binary")) {
	  // load the parser from a binary serialized file
	  // in progress
      } else if (args[argIndex].equalsIgnoreCase("-gramlex")) {
	  // load the parser from ascii declarative grammars
	  // in progress
      } else if (args[argIndex].equalsIgnoreCase("-input")) {
	  // next argument is the format of the input
	  inputFormat = args[argIndex+1];
	  argIndex += 2;
      } else {
        int j = FactoredParser.setOptionFlag(args, argIndex);
        if (j == argIndex) {
          j = tlpParams.setOptionFlag(args, argIndex);
        }
        if (j == argIndex) {
          System.err.println("Unknown option ignored: " + args[argIndex]);
          j++;
        }
        argIndex = j;
      }
    } // end while loop through arguments


    if (Test.verbose) {
      System.out.println("Currently " + new Date());
      System.out.print("FactoredParser invoked with arguments:");
      for(int i = 0; i < args.length; i++) {
        System.out.print(" " + args[i]);
      }
      System.out.println();
      System.out.println("ParserPack is " + tlpParams);
    }

    TreePrint treePrint = Test.treePrint();
    treePrint.setHeadFinder(tlpParams.headFinder());

    Options.get().langpack = tlpParams.treebankLanguagePack();
    Train.splitters = new HashSet(Arrays.asList(tlpParams.splitters()));
    Train.sisterSplitters = new HashSet(Arrays.asList(tlpParams.sisterSplitters()));
    BinarizerFactory.HeadRespectingBinarizer.setTreebankLang(tlpParams);
    PrintWriter pwOut = tlpParams.pw();
    PrintWriter pwErr = tlpParams.pw(System.err);

    LexicalizedParser lp = null;
    if (train) {
	// the next three arguments must be the treebank path and the range for training
	// so we train a parser using the treebank
      if (args.length < argIndex + 3) { 
        lp = new LexicalizedParser(args[argIndex], tlpParams);
        argIndex++;
      } else {
        int low = Integer.parseInt(args[argIndex+1]);
        int high = Integer.parseInt(args[argIndex+2]);
        lp = new LexicalizedParser(args[argIndex],
                                 new NumberRangeFileFilter(low, high, true),
                                 tlpParams);
        argIndex += 3;
      }
    } else {
	// the next argument must be the path to the serialized parser
	// so we load a serialized parser
      System.err.println("Loading serialized parser from " + args[argIndex]);
      try {
        lp = new LexicalizedParser(args[argIndex]);
      } catch (IllegalArgumentException e) {
        System.err.println("Error loading parser, exiting...");
        System.exit(0);
      }
      System.err.println("Done loading parser.");
      argIndex++;
    }

    if (serialize) {
	// the next argument must be the path to serialize to
      ParserData pd = lp.parserData();
      if (argIndex < args.length) {
        serializeParser(pd, args[argIndex]);
        argIndex++;
      } else {
        System.err.println("usage: " +
              "java edu.stanford.nlp.parser.lexparser.LexicalizedParser " +
              "-train trainFilesPath [start stop] serializedParserFilename");
      }
    } else if (argIndex >= args.length) {
	// no more arguments, so we just parse our own test sentence
      String[] sentence = { "This", "is", "just", "a", "test", "." };
      // String[] sentence = { "", "", "ѧУ", "", "ѧϰ", ""};
      lp.parse(Arrays.asList(sentence));
      treePrint.printTree(lp.getBestParse(),pwOut);
    } else if (args[argIndex].equalsIgnoreCase("-treebank")) {
	// the next argument is the treebank path and range for testing
      Treebank testTreebank = tlpParams.testMemoryTreebank();
      if (args.length < argIndex + 4) {
        testTreebank.loadPath(args[argIndex + 1]);
      } else {
        int low = Integer.parseInt(args[argIndex + 2]);
        int high = Integer.parseInt(args[argIndex + 3]);
        testTreebank.loadPath(args[argIndex + 1],
                              new NumberRangeFileFilter(low, high, true));
      }
      if (Test.verbose) {
        Test.display();
        Options.get().display();
        tlpParams.display();
      }
      LabeledConstituentEval pcfgPE = new LabeledConstituentEval("pcfg LP/LR");
      DependencyEval depDE = new DependencyEval("depend DE");
      LabeledConstituentEval comboPE = new LabeledConstituentEval("factor LP/LR");
      TaggingEval comboTE = new TaggingEval("Tag");
      TreeTransformer binarizerOnly = BinarizerFactory.getHeadRespectingBinarizer(tlpParams.headFinder(), false, false);
      for (Iterator it = testTreebank.iterator(); it.hasNext(); ) {
        Tree goldTree = (Tree) it.next();
	//System.out.println(goldTree.label().getClass());//debugging
	//System.out.println(Test.forceTags);
        Sentence s = getInputSentence(goldTree);
	//System.out.println("is it a taggedword?" + (s.get(0) instanceof TaggedWord));
        long time = System.currentTimeMillis();
	// REMOVE
	//if (goldTree.yield().length() != Test.maxLength)
	//  continue;
        pwErr.println("Parsing [len. " + s.length() + "]: " + s);
        Tree tree = null;
        try {
          if ( ! lp.parse(s)) {
            pwErr.println("Sentence couldn't be parsed by grammar.");
          } else {
            tree = lp.getBestParse();
	    if (Test.verbose) {
	      pwOut.println("ComboParser best");
	    }
            treePrint.printTree(tree,pwOut); //combo parse goes to System.out
	    // if (lp.pparser != null)
	    //   pwErr.println("Correct parse\tScore: " + lp.pparser.scoreBinarizedTree(binarizer.transformTree(goldTree,null),true));
	    // else
	    if (Test.verbose) {
	      pwOut.println("Correct parse");
	      treePrint.printTree(goldTree,pwOut);
	    }
          }
        } catch (UnsupportedOperationException uoe) {
          pwErr.println("Sentence too long");
        }
        long elapsed = System.currentTimeMillis() - time;
        pwErr.println("Time: "+ (elapsed/1000) +" sec.");
        if (tree != null) {
          TreeTransformer tc = tlpParams.collinizer();
          Tree transGoldTree = tc.transformTree(goldTree,null);
          Tree treePCFG = lp.getBestPCFGParse();
          pcfgPE.evaluate(tc.transformTree(treePCFG,null), transGoldTree,pwErr);
          Tree treeDep = lp.getBestDependencyParse();
          if (treeDep != null)
            depDE.evaluate(treeDep, binarizerOnly.transformTree(goldTree, null),pwErr);
          comboPE.evaluate(tc.transformTree(tree,null), transGoldTree,pwErr);
          comboTE.evaluate(tree, goldTree,pwErr);
        }
        pwErr.println();
      } // for tree iterator
    } else {
	// the rest of the arguments must be names of files to parse
	// so we parse through the various files left on the command line

	// choose a tokenizer
	if (tokenizer == null) {
	    if (!tokenized) {
		// no Tokenizer was specified on the command line
		// so let's see if there is a recommended one in the language pack
		tokenizer = tlpParams.treebankLanguagePack().getTokenizer();
		if (tokenizer == null) {
		    // none was specified in the language pack, so we use PTBTokenizer as default
		    tokenizer = new PTBTokenizer();
		}
	    } else {
		// it was already tokenized, so we just split on whitespace
		tokenizer = new SimpleTokenizer();
	    }
	}

	// choose a sentence splitter
	Processor wordToSent;
	if (sentencesSplit) {
	    if (sentenceBoundary.equalsIgnoreCase("sDelimited")) {
		wordToSent = new WordToSentenceProcessor(new HashSet(), new HashSet(), new HashSet(Arrays.asList(new String[] {"<s>", "</s>", "\n"})));
	    } else if (sentenceBoundary.equalsIgnoreCase("newline")) {
		Set s = new HashSet();
		s.add("\n");
		wordToSent = new WordToSentenceProcessor(new HashSet(), new HashSet(), s);
	    } else {
		Set s = new HashSet();
		s.add(sentenceBoundary);
		wordToSent = new WordToSentenceProcessor(new HashSet(), new HashSet(), s);
	    }
	} else {
	    wordToSent = new WordToSentenceProcessor(new HashSet(Arrays.asList(tlpParams.treebankLanguagePack().sentenceFinalPunctuationWords())));
	}

	// go through the documents, process them and parse them
	Document d1, d2, d3;
	Processor tagProcessor;
      for (int i = argIndex; i < args.length; i++) {
        String filename = args[i];
	d1 = null;
        try {
          if (filename.startsWith("http://")) {
            Document dpre = new BasicDocument(tokenizer).init(new URL(filename));
            Processor notags = new StripTagsProcessor();
            d1 = notags.process(dpre);
          } else {
	      d1 = new BasicDocument(tokenizer).init(new InputStreamReader(new FileInputStream(filename), tlpParams.treebankLanguagePack().getEncoding()));
          }
        } catch (Exception e) {
          e.printStackTrace();
        }
        if (d1 != null) {
	    tagProcessor = new WordToTaggedWordProcessor(tagSeparator);
	  d2 = tagProcessor.process(d1);
	  d3 = wordToSent.process(d2);
          for (Iterator it = d3.iterator(); it.hasNext(); ) {
            Sentence s = (Sentence) it.next();
            pwErr.println("Parsing [len. " + s.length() + "]: " + s);
            try {
              lp.parse(s);
              lp.getBestParse().pennPrint(pwOut);
              // lp.getBestParse().pennPrint(pwErr);
            } catch (UnsupportedOperationException uoe) {
              pwErr.println("Whoops, sentence too long.");
            }
          }
        }
      }
    }
  }


  /** Stores the serialized material for a parser */
  static class ParserData implements Serializable {

    Lexicon lex;
    BinaryGrammar bg;
    UnaryGrammar ug;
    DependencyGrammar dg;
    Map numbs;
    Options pt;

    public ParserData(Lexicon lex, BinaryGrammar bg, UnaryGrammar ug,
               DependencyGrammar dg, Map numbs, Options pt) {
      this.lex = lex;
      this.bg = bg;
      this.ug = ug;
      this.dg = dg;
      this.numbs = numbs;
      this.pt = pt;
    }

    private static final long serialVersionUID = 1;

  }

}
