package edu.stanford.nlp.parser.lexparser;

import java.util.*;
import java.io.Reader;
import java.text.NumberFormat;

import edu.stanford.nlp.util.*;
import edu.stanford.nlp.trees.*;

/** See what parent annotation helps in treebank, based on support and 
 *  KL divergence.
 *  @author Christopher Manning
 *  @version 2003/01/04
 */
public class ParentAnnotationStats implements TreeProcessor {

    public static boolean DO_TAGS = false;

    private Map nodeRules = new HashMap();
    private Map pRules = new HashMap();
    private Map gPRules = new HashMap();

    /** Minimum support * KL to be included in output and as feature */
    public static final double[] CUTOFFS = { 100.0, 200.0, 500.0, 1000.0 };

    /** Minimum support of parent annotated node for grandparent to be
     *  studied.  Just there to reduce runtime and printout size. 
     */
    public static final double SUPPCUTOFF = 100.0;

    /** Does whatever one needs to do to a particular parse tree */
    public void processTree(Tree t) {
	processTreeHelper("TOP", "TOP", t);
    }

    public static List kidLabels(Tree t) {
	Tree[] kids = t.children();
	List l = new ArrayList(kids.length);
	for (int i = 0; i < kids.length; i++) {
	    l.add(kids[i].label().value());
	}
	return l;
    }

    public void processTreeHelper(String gP, String p, Tree t) {
	if ( ! t.isLeaf() && 
	     ( DO_TAGS || ! t.isPreTerminal())) { // stop at words/tags
	    String n = t.label().value();
	    List kidn = kidLabels(t);
	    Counter cntr = (Counter) nodeRules.get(n);
	    if (cntr == null) {
		cntr = new Counter();
		nodeRules.put(n, cntr);
	    }
	    cntr.increment(kidn);
	    List pairStr = new ArrayList(2);
	    pairStr.add(n);
	    pairStr.add(p);
	    cntr = (Counter) pRules.get(pairStr);
	    if (cntr == null) {
		cntr = new Counter();
		pRules.put(pairStr, cntr);
	    }
	    cntr.increment(kidn);
	    List tripleStr = new ArrayList(3);
	    tripleStr.add(n);
	    tripleStr.add(p);
	    tripleStr.add(gP);
	    cntr = (Counter) gPRules.get(tripleStr);
	    if (cntr == null) {
		cntr = new Counter();
		gPRules.put(tripleStr, cntr);
	    }
	    cntr.increment(kidn);
	    Tree[] kids = t.children();
	    for (int i = 0; i < kids.length; i++) {
		processTreeHelper(p, n, kids[i]);
	    }
	}
    }
				  

    public void printStats() {
	NumberFormat nf = NumberFormat.getNumberInstance();
	nf.setMaximumFractionDigits(2);
	// System.out.println("Node rules");
	// System.out.println(nodeRules);
	// System.out.println("Parent rules");
	// System.out.println(pRules);
	// System.out.println("Grandparent rules");
	// System.out.println(gPRules);

	// Store java code for selSplit
	StringBuffer[] javaSB = new StringBuffer[CUTOFFS.length];
	for (int i = 0; i < CUTOFFS.length; i++) {
	    javaSB[i] = new StringBuffer("  private static String[] splitters"
					 + (i+1) + " = new String[] {");
	}

	// do value of parent
	for (Iterator it = nodeRules.keySet().iterator(); it.hasNext(); ) {
	    ArrayList answers = new ArrayList();
	    String node = (String) it.next();
	    Counter cntr = (Counter) nodeRules.get(node);
	    double support = cntr.total();
	    System.out.println("Node " + node + " support is " + support);
	    for (Iterator it2 = pRules.keySet().iterator(); it2.hasNext(); ) {
		List key = (List) it2.next();
		if (key.get(0).equals(node)) {   // only do it if they match
		    Counter cntr2 = (Counter) pRules.get(key);
		    double support2 = cntr2.total();
		    double kl = cntr2.klDivergence(cntr);
		    System.out.println("KL(" + key + "||" + node + ") = " +
				       nf.format(kl) + "\t" + "support(" + 
				       key + ") = " + support2);
		    answers.add(new Pair(key, new Double(kl * support2)));
		}
	    }
	    System.out.println("----");
	    System.out.println("Sorted descending support * KL");
	    Collections.sort(answers,
			     new Comparator() {
				 public int compare(Object o1, Object o2) {
				     Pair p1 = (Pair) o1;
				     Pair p2 = (Pair) o2;
				     Double p12 = (Double) p1.second();
				     Double p22 = (Double) p2.second();
				     return p22.compareTo(p12);
				 }
			     });
	    for (int i = 0, size = answers.size(); i < size; i++) {
		Pair p = (Pair) answers.get(i);
		double psd = ((Double) p.second()).doubleValue();
		System.out.println(p.first() + ": " + nf.format(psd));
		if (psd >= CUTOFFS[0]) {
		    List lst = (List) p.first();
		    String nd = (String) lst.get(0);
		    String par = (String) lst.get(1);
		    for (int j=0; j < CUTOFFS.length; j++) {
			if (psd >= CUTOFFS[j]) {
			    javaSB[j].append("\"").append(nd).append("^");
			    javaSB[j].append(par).append("\", ");
			}
		    }
		}
	    }
	    System.out.println();
	}

	// do value of parent with info gain -- yet to finish this
	for (Iterator it = nodeRules.entrySet().iterator(); it.hasNext(); ) {
	    Map.Entry pair = (Map.Entry) it.next();
	    String node = (String) pair.getKey();
	    Counter cntr = (Counter) pair.getValue();
	    double support = cntr.total();
	    System.out.println("Node " + node + " support is " + support);
	    ArrayList dtrs = new ArrayList();
	    for (Iterator it2 = pRules.entrySet().iterator(); it2.hasNext();) {
		HashMap annotated = new HashMap();
		Map.Entry pair2 = (Map.Entry) it2.next();
		List node2 = (List) pair2.getKey();
		Counter cntr2 = (Counter) pair2.getValue();
		if (node2.get(0).equals(node)) {   // only do it if they match
		    annotated.put(node2, cntr2);
		}
	    }
		    
	    // upto
		    
	    List answers = new ArrayList();
	    System.out.println("----");
	    System.out.println("Sorted descending support * KL");
	    Collections.sort(answers,
			     new Comparator() {
				 public int compare(Object o1, Object o2) {
				     Pair p1 = (Pair) o1;
				     Pair p2 = (Pair) o2;
				     Double p12 = (Double) p1.second();
				     Double p22 = (Double) p2.second();
				     return p22.compareTo(p12);
				 }
			     });
	    for (int i = 0, size = answers.size(); i < size; i++) {
		Pair p = (Pair) answers.get(i);
		double psd = ((Double) p.second()).doubleValue();
		System.out.println(p.first() + ": " + nf.format(psd));
		if (psd >= CUTOFFS[0]) {
		    List lst = (List) p.first();
		    String nd = (String) lst.get(0);
		    String par = (String) lst.get(1);
		    for (int j=0; j < CUTOFFS.length; j++) {
			if (psd >= CUTOFFS[j]) {
			    javaSB[j].append("\"").append(nd).append("^");
			    javaSB[j].append(par).append("\", ");
			}
		    }
		}
	    }
	    System.out.println();
	}

	// do value of grandparent
	for (Iterator it = pRules.keySet().iterator(); it.hasNext(); ) {
	    ArrayList answers = new ArrayList();
	    List node = (List) it.next();
	    Counter cntr = (Counter) pRules.get(node);
	    double support = cntr.total();
	    if (support < SUPPCUTOFF) {
		continue;
	    }
	    System.out.println("Node " + node + " support is " + support);
	    for (Iterator it2 = gPRules.keySet().iterator(); it2.hasNext(); ) {
		List key = (List) it2.next();
		if (key.get(0).equals(node.get(0)) && 
		    key.get(1).equals(node.get(1))) {  // only do it if they match
		    Counter cntr2 = (Counter) gPRules.get(key);
		    double support2 = cntr2.total();
		    double kl = cntr2.klDivergence(cntr);
		    System.out.println("KL(" + key + "||" + node + ") = " +
				       nf.format(kl) + "\t" + "support(" + 
				       key + ") = " + support2);
		    answers.add(new Pair(key, new Double(kl * support2)));
		}
	    }
	    System.out.println("----");
	    System.out.println("Sorted descending support * KL");
	    Collections.sort(answers,
			     new Comparator() {
				 public int compare(Object o1, Object o2) {
				     Pair p1 = (Pair) o1;
				     Pair p2 = (Pair) o2;
				     Double p12 = (Double) p1.second();
				     Double p22 = (Double) p2.second();
				     return p22.compareTo(p12);
				 }
			     });
	    for (int i = 0, size = answers.size(); i < size; i++) {
		Pair p = (Pair) answers.get(i);
		double psd = ((Double) p.second()).doubleValue();
		System.out.println(p.first() + ": " + nf.format(psd));
		if (psd >= CUTOFFS[0]) {
		    List lst = (List) p.first();
		    String nd = (String) lst.get(0);
		    String par = (String) lst.get(1);
		    String gpar = (String) lst.get(2);
		    for (int j=0; j < CUTOFFS.length; j++) {
			if (psd >= CUTOFFS[j]) {
			    javaSB[j].append("\"").append(nd).append("^");
			    javaSB[j].append(par).append("~");
			    javaSB[j].append(gpar).append("\", ");
			}
		    }
		}
	    }
	    System.out.println();
	}
	System.out.println();
	System.out.println("  // Automatically generated by ParentAnnotationStats -- preferably don't edit");
	for (int i = 0; i < CUTOFFS.length; i++) {
	    int len = javaSB[i].length();
	    javaSB[i].replace(len - 2, len, "};");
	    System.out.println(javaSB[i]);
	}
	System.out.print("  public static HashSet splitters = new HashSet(Arrays.asList(");
	for (int i = CUTOFFS.length; i > 0; i--) {
	    if (i == 1) {
		System.out.print("splitters1");
	    } else {
		System.out.print("selectiveSplit" + i + " ? splitters" + i +
				 " : (");
	    }
	}
	// need to print extra one to close other things open
	for (int i = CUTOFFS.length; i >= 0; i--) {
	    System.out.print(")");
	}
	System.out.println(";");
    }


    /** Calculate parent annotation statistics suitable for doing
     *  selective parent splitting in the PCFGParser inside the 
     *  FactoredParser.  <p>
     *  Usage: java edu.stanford.nlp.parser.lexparser.ParentAnnotationStats 
     *       [-tags] treebankPath
     *  @param args One argument: path to the Treebank
     */
    public static void main(String[] args) {
	if (args.length < 1) {
	    System.out.println("Usage: java edu.stanford.nlp.parser.lexparser.ParentAnnotationStats [-tags] treebankPath");
	} else {
	    int i = 0;
	    if (args[i].equals("-tags")) {
		DO_TAGS = true;
		i++;
	    }
	    ParentAnnotationStats pas = new ParentAnnotationStats();
	    Treebank treebank = new DiskTreebank(new TreeReaderFactory() {
		    public TreeReader newTreeReader(Reader in) {
			return new PennTreeReader(in, 
				    new LabeledScoredTreeFactory(
						 new StringLabelFactory()),
				    new BobChrisTreeNormalizer());
		    }
		});
	    treebank.loadPath(args[i]);
	    treebank.apply(pas);
	    pas.printStats();
	}
    }

}
