package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.trees.*;
import java.util.*;

class SXOScorer implements Scorer {

  protected int MAX_LENGTH = 41;

  protected PCFGGrammar grammar;

  protected List words;

  protected double[][] iScoreSX;
  protected double[][][] oScoreSX;
  
  protected double max(double x, double y) { return (x > y ? x : y); }
  
  protected void buildIScoreSX() {
    // allocate and initialize
    iScoreSX = new double[MAX_LENGTH+1][grammar.numStates()];
    for (int iSpan=0; iSpan<=MAX_LENGTH; iSpan++) {
      Arrays.fill(iScoreSX[iSpan], Double.NEGATIVE_INFINITY);
    }
    // base case: all single spans have score one
    for (int state=0; state<grammar.numStates(); state++) {
      if (grammar.isTerminal(state))
	iScoreSX[1][state] = 0.0;
    }
    // unary
    for (Iterator i=grammar.closedUnaryRuleIterator(); i.hasNext();) {
      UnaryRule rule = (UnaryRule)i.next();
      iScoreSX[1][rule.parent] = max(iScoreSX[1][rule.parent],
				     rule.score()+
				     iScoreSX[1][rule.child]);
    }
    for (int iSpan=2; iSpan<=MAX_LENGTH; iSpan++) {
      // binary
      for (int state=0; state<grammar.numStates(); state++) {
	Iterator rI = grammar.binaryRuleIteratorByParent(state);
	while(rI.hasNext()) {
	  BinaryRule rule = (BinaryRule)rI.next();
	  for (int lSpan=1; lSpan<iSpan; lSpan++) {
	    int rSpan = iSpan-lSpan;
	    iScoreSX[iSpan][state] = max(iScoreSX[iSpan][state],
					 rule.score()+
					 iScoreSX[lSpan][rule.leftChild]+
					 iScoreSX[rSpan][rule.rightChild]);
	  }
	}
      }
      // unary
      for (Iterator i=grammar.closedUnaryRuleIterator(); i.hasNext();) {
	UnaryRule rule = (UnaryRule)i.next();
	iScoreSX[iSpan][rule.parent] = max(iScoreSX[iSpan][rule.parent],
					      rule.score()+
					      iScoreSX[iSpan][rule.child]);
      }
    }
  }

  protected void buildOScoreSX() {
    // allocate and initialize
    oScoreSX = new double[MAX_LENGTH+1][][];
    for (int lSpan=0; lSpan<=MAX_LENGTH; lSpan++) {
      oScoreSX[lSpan] = new double[MAX_LENGTH+1-lSpan][grammar.numStates()];
      for (int rSpan=0; rSpan<=MAX_LENGTH-lSpan; rSpan++) {
	Arrays.fill(oScoreSX[lSpan][rSpan], Double.NEGATIVE_INFINITY);
      }
    }
    // base case: goal state with oSpan zero has score one
    oScoreSX[0][0][grammar.goalState()] = 0.0;
    for (int oSpan=0; oSpan<=MAX_LENGTH; oSpan++) {
      // unary
      for (Iterator i=grammar.closedUnaryRuleIterator(); i.hasNext();) {
	UnaryRule rule = (UnaryRule)i.next();
	for (int lSpan=0; lSpan<=oSpan; lSpan++) {
	  int rSpan = oSpan-lSpan;
	  oScoreSX[lSpan][rSpan][rule.child] = max(oScoreSX[lSpan][rSpan][rule.child],
						     rule.score()+
						     oScoreSX[lSpan][rSpan][rule.parent]);
	}
      }
      // binary
      for (int state=0; state<grammar.numStates(); state++) {
	Iterator rI = grammar.binaryRuleIteratorByParent(state);
	while(rI.hasNext()) {
	  BinaryRule rule = (BinaryRule)rI.next();
	  for (int lSpan=0; lSpan<=oSpan; lSpan++) {
	    int rSpan = oSpan-lSpan;
	    int remainder = MAX_LENGTH-oSpan;
	    for (int split=1; split<=remainder; split++) {
	      oScoreSX[lSpan+split][rSpan][rule.rightChild] = max(oScoreSX[lSpan+split][rSpan][rule.rightChild],
								    rule.score()+
								    oScoreSX[lSpan][rSpan][rule.parent]+
								    iScoreSX[split][rule.leftChild]);
	      oScoreSX[lSpan][rSpan+split][rule.leftChild] = max(oScoreSX[lSpan][rSpan+split][rule.leftChild],
								   rule.score()+
								   oScoreSX[lSpan][rSpan][rule.parent]+
								   iScoreSX[split][rule.rightChild]);
	    }
	  }
	}
      }
    }
  }

  public double oScore(Edge edge) {
    return oScoreSX[edge.start][words.size()-edge.end][edge.state];
  }

  public double iScore(Edge edge) {
    return iScoreSX[edge.end-edge.start][edge.state];
  }

  public boolean iPossible(Hook hook) {
    return true;
  }

  public boolean oPossible(Hook hook) {
    return true;
  }

  public boolean parse(List words) {
    this.words = words;
    return true;
  }

  public SXOScorer(PCFGGrammar grammar, int MAX_LENGTH) {
    this.grammar = grammar;
    this.MAX_LENGTH = MAX_LENGTH;
    iScoreSX = new double[MAX_LENGTH+1][grammar.numStates()];
    oScoreSX = new double[MAX_LENGTH+1][MAX_LENGTH+1][grammar.numStates()];
    //Timing.tick("\nBuilding iScores...");
    buildIScoreSX();
    //System.out.println("ROOT 4: "+iScoreSX[4][grammar.goalState()]);
    //Timing.tick("done.\n");
    //System.out.println("Building oScores...");
    buildOScoreSX();
    //System.out.println("NP 0-5: "+oScoreSX[0][5][grammar.stateNumberer().number("NP")]);
    //Timing.tick("done.\n");
  }

}


