package edu.stanford.nlp.process;


import java.io.*;
import java.util.*;

import edu.stanford.nlp.trees.Word;

/**
 * Tokenizer implementation that conforms to the Penn Treebank tokenization
 * conventions.
 * This tokenizer is a Java implementation of Professor Chris Manning's Flex
 * tokenizer, pgtt-treebank.l.  It reads raw text and outputs
 * tokens as edu.stanford.nlp.trees.Words in the Penn treebank format. It can 
 * optionally return carriage returns as tokens.
 * 
 * @author Teg Grenager (grenager@stanford.edu)
 */
public class PTBTokenizer extends AbstractTokenizer  {

    // whether carriage returns should be returned as tokens
    private boolean tokenizeCRs; 
    // the underlying lexer
    PTBLexer lexer;
    // the next token to be returned
    Word nextToken;

    /** 
     * Constructs a new PTBTokenizer that treats carriage returns as normal whitespace. 
     * No source is specified, so hasNext() will return false. 
     */
    public PTBTokenizer() { 
	this(false); 
    }

    /** 
     * Constructs a new PTBTokenizer that optionally returns carriage returns
     * as their own token. CRs come back as Words whose text is 
     * {@link PTBLexer#cr}.
     */
    public PTBTokenizer(boolean tokenizeCRs) { 
	this.tokenizeCRs=tokenizeCRs; 
    }

    /** 
     * Constructs a new PTBTokenizer that treats carriage returns as normal whitespace. 
     */
    public PTBTokenizer(Reader r) { 
	this(r, false); 
    }
    
    /** 
     * Constructs a new PTBTokenizer that optionally returns carriage returns
     * as their own token. CRs come back as Words whose text is 
     * {@link PTBLexer#cr}.
     */
    public PTBTokenizer(Reader r, boolean tokenizeCRs) { 
	this.tokenizeCRs=tokenizeCRs; 
	setSource(r);
    }
    
    /**
     * Get the next valid Word from the lexer if possible.
     */
    private void getNext() {
	if (lexer == null) {
	    nextToken = null;
	    return;
	}
	try {
	    nextToken = lexer.next();
	    // get rid of CRs if necessary
	    while (!tokenizeCRs && PTBLexer.cr.equals(nextToken.word()))
		nextToken = lexer.next();
	} catch (Exception e) {
	    nextToken = null;
	}
    }

    /**
     * Returns true if this Tokenizer has more elements.
     */
    public boolean hasNext() {
	return nextToken != null;
    }
    
    /**
     * Returns the next Word token, or null if there is none.
     */
    public Object next() {
	Word result = nextToken;
	getNext();
	return result;
    }


    /** Reads a file from the argument and prints its tokens one per line.
     *  This is mainly as a testing aid, but it can also be quite useful 
     *  standalone to turn a corpus into a one token per line file of tokens.
     *  <p>
     *  Usage: <code>java edu.stanford.nlp.process.PTBTokenizer filename
     *  </code>
     *  @param args Command line arguments
     */
    public static void main (String[] args) throws IOException {
    if (args.length < 1) {
        System.err.println("usage: java edu.stanford.nlp.process." +
                   "PTBTokenizer [-cr] filename");
        return;
    }
    PTBTokenizer tokenizer = new PTBTokenizer(new FileReader(args[args.length-1]), "-cr".equals(args[0]));
    List words = tokenizer.tokenize();
    for (int i=0; i < words.size(); i++)
        System.out.println(words.get(i));
    }

    /**
     * Sets the source of this Tokenizer to be the Reader r.
     */
    public void setSource(Reader r) {
	lexer = new PTBLexer(r);
	getNext();
    }

}

