package edu.stanford.nlp.process;

import java.util.*;
import java.io.*;
import edu.stanford.nlp.trees.Word;

/**
 * Simple tokenizer implementation that wraps a StringTokenizer.
 * Word delimiter chars are space and tab only. Newlines are returned as tokens.
 * The tokens returned are edu.stanford.nlp.trees.Word objects.
 *
 * @author Joseph Smarr (jsmarr@stanford.edu)
 * @author Teg Grenager (grenager@stanford.edu)
 */
public class SimpleTokenizer extends AbstractTokenizer {

    StringTokenizer st;

    /** Word delimiter characters used to tokenize text: {@value} */
    protected static final String delims=" \t\n";
    
    /**
     * Constructs a new SimpleTokenizer. 
     * No source is specified, so hasNext() will return false. 
     */
     public SimpleTokenizer() {}


    /**
     * Constructs a new SimpleTokenizer with the Reader r as its source.
     */
    public SimpleTokenizer(Reader r) {
	setSource(r);
    }

    /**
     * Returns true if this Tokenizer has more elements.
     */
    public boolean hasNext() {
	if (st==null)
	    return false;
	return st.hasMoreTokens();
    }

    /**
     * Returns the next Word token, or null if there is none.
     */
    public Object next() {
	if (st==null)
	    return null;
	String token = st.nextToken();
	while (token.charAt(0) == '\t' || token.charAt(0) == ' ')
	    token = st.nextToken();
	return new Word(token);
    }

    /**
     * Sets the source of this Tokenizer to be the Reader r.
     */
    public void setSource(Reader r) {
	String text=getText(r);
	st=new StringTokenizer(text,delims, true);
    }

    
    private static String getText(Reader r) {
	try {
        // returns null if the reader is null
        if(r==null) return(null);
        
        // ensures the reader is buffered
        if(!(r instanceof BufferedReader)) r=new BufferedReader(r);
        BufferedReader br=(BufferedReader)r;
        
        // reads all the chars into a buffer
        StringBuffer sb=new StringBuffer();
        while(true)
        {
            int c=br.read();
            if(c==-1) break;
            sb.append((char)c);
        }
        
        return(sb.toString());
	} catch (IOException e) {
	    return null;
	}
    }
}

