package edu.stanford.nlp.process;

import java.util.*;
import java.net.URL;
import java.io.File;

import edu.stanford.nlp.dbm.*;
import edu.stanford.nlp.trees.*;

/**
 * Transforms a Document of Words into a Document of Sentences by grouping the
 * Words. 
 *
 * @author Joseph Smarr (jsmarr@stanford.edu)
 * @author Christopher Manning
 * @author Teg Grenager (grenager@stanford.edu)
 */
public class WordToSentenceProcessor implements Processor {

    /** Set of tokens (Strings) that qualify as sentence-final tokens. */
    private Set sentenceBoundaryTokens; 

    
    /** Set of tokens (Strings) that qualify as sentence-final tokens. */
    private Set sentenceBoundaryFollowers; 

    /** Set of tokens (Strings) that are sentence boundaries to be discarded.
    */
    private Set sentenceBoundaryToDiscard;

    /**
     * Returns a new Document where each element is a Sentence built from a run
     * of Words in the input Document. Specifically, reads through each word in
     * the input document and breaks off a sentence after finding a valid
     * setence boundary token or end of file.
     * Note that for this to work, the words in the
     * input document must have been tokenized with a tokenizer that makes
     * sentence boundary tokens
     * their own tokens (e.g., PTBTokenizer).
     * @see #sentenceBoundaryTokens
     * @see Sentence
     * @see PTBTokenizer
     */
    public Document process(Document words)
    {
        Document sentences= new SentenceDocument();

        Sentence currentSentence=null;
	Sentence lastSentence = null;
	HasWord h;
	String w;
        for (Iterator iter=words.iterator(); iter.hasNext(); )
        {
            h = (HasWord) iter.next();
	    w = h.word();
            if (currentSentence==null) 
		currentSentence=new Sentence();
	    if (sentenceBoundaryFollowers.contains(w) && 
		lastSentence != null && currentSentence.length() == 0) {
		lastSentence.add(h);
	    } else {
		// doesn't actually add these, just uses them to detect
		// sentence boundaries
		if (!sentenceBoundaryToDiscard.contains(w))
		    currentSentence.add(h);
		if ((sentenceBoundaryTokens.contains(w) || 
		     sentenceBoundaryToDiscard.contains(w)) &&
		    currentSentence.size()>0) {
		    sentences.add(currentSentence);
		    // adds this sentence now that it's complete
		    lastSentence = currentSentence;
		    currentSentence=null; // clears the current sentence
		}
            }
        }

	// add any words at the end, even if there isn't a sentence
	// terminator at the end of file
	if (currentSentence != null && currentSentence.size() > 0) {
	    sentences.add(currentSentence); // adds last sentence
	}
        return(sentences);
    }


    /** Creat a <code>WordToSentenceProcessor</code> using a sensible default
     *  list of tokens to split on.
     */
    public WordToSentenceProcessor() {
	this(new HashSet(Arrays.asList(new String[] {".","?","!"})));
    }

    /** Flexibly set the set of acceptable sentence boundary tokens. */
    public WordToSentenceProcessor(Set boundaryTokens) {
	this(boundaryTokens, new HashSet(Arrays.asList(
	        new String[] {")","\"","\'", "''", "-RRB-", "-RSB-"})));
    }

    /** Flexibly set the set of acceptable sentence boundary tokens and
     * also the set of tokens commonly following sentence boundaries. 
     */
    public WordToSentenceProcessor(Set boundaryTokens, Set boundaryFollowers) {
	this(boundaryTokens, boundaryFollowers, new HashSet(Arrays.asList(
	        new String[] {"\n"})));
    }

    /** 
     * Flexibly set the set of acceptable sentence boundary tokens,
     * the set of tokens commonly following sentence boundaries, and also
     * the set of tokens that are sentences boundaries that should be discarded.
     */
    public WordToSentenceProcessor(Set boundaryTokens, Set boundaryFollowers, Set boundaryToDiscard) {
	sentenceBoundaryTokens = boundaryTokens;
	sentenceBoundaryFollowers = boundaryFollowers;
	sentenceBoundaryToDiscard = boundaryToDiscard;
	//printSet(boundaryTokens);
	//printSet(boundaryFollowers);
	//printSet(boundaryToDiscard);
    }

    // for testing only
    private void printSet(Set s) {
	for (Iterator i = s.iterator(); i.hasNext();) {
	    System.out.print(i.next() + " ");
	}
	System.out.println();
    }


    /** This will print out as sentences some text.  It can be used to 
     *  test sentence division.  <br>
     *  Usage: java edu.stanford.nlp.process.WordToSentenceProcessor fileOrUrl
     *  @param args Command line argument: a file or URL
     */
    public static void main(String[] args) {
	if (args.length != 1) {
	    System.out.println("usage: java edu.stanford.nlp.process.WordToSentenceProcessor fileOrUrl");
	    System.exit(0);
	}
	String filename = args[0];
	try {
	    Document d = null;
	    if (filename.startsWith("http://")) {
		Document dpre = new BasicDocument().init(new URL(filename));
		Processor notags = new StripTagsProcessor();
		d = notags.process(dpre);
	    } else {
		d = new BasicDocument().init(new File(filename));
	    }
	    Processor proc = new WordToSentenceProcessor();
	    Document sentd = proc.process(d);
	    // System.out.println(sentd);
	    int i = 0;
	    for (Iterator it = sentd.iterator(); it.hasNext(); ) {
		Sentence sent = (Sentence) it.next();
		System.out.println(i + ": " + sent);
		i++;
	    }
	} catch (Exception e) {
	  e.printStackTrace();
	}
    }

}
