package edu.stanford.nlp.process;

import java.util.*;
import java.net.URL;
import java.io.File;

import edu.stanford.nlp.dbm.*;
import edu.stanford.nlp.trees.*;

/**
 * Transforms a Document of Words into a document all or partly of 
 * TaggedWords by breaking words on a tag divider character.
 *
 * @author Teg Grenager (grenager@stanford.edu)
 * @author Christopher Manning
 */
public class WordToTaggedWordProcessor implements Processor {

    /** The char that we will split on. */
    protected char splitChar;
    
    /**
     * Returns a new Document where each Word with a tag has been converted
     * to a TaggedWord.  Things in the input which don't implement HasWord
     * will be deleted in the output.  Things which do will be scanned for 
     * being word + splitChar + tag.  If they are, they are split up and 
     * inserted as TaggedWords, otherwise they are added to the document
     * with their current type.  More precisely, they will be split on the
     * last instance of splitChar with index above 0.  This will give the
     * correct split, providing tags don't include the splitChar, regardless
     * of escaping, and will not allow an empty or null word - you can think
     * of the first character as always being escaped.
     * @param words The input Document (should be of HasWords)
     * @return A new Document, perhaps with some of the things TaggedWords
     */
    public Document process(Document words) {
	Document result = new BasicDocument();
        for (Iterator iter=words.iterator(); iter.hasNext(); ) {
	    Object o = iter.next();
	    if (! (o instanceof HasWord))
		continue;
	    result.add(splitTag((HasWord) o));
        }
        return result;
    }

    /**
     * Splits the Word w on the character splitChar.
     */
    private HasWord splitTag(HasWord w) {
	if (splitChar == 0) {
	    return w;
	}
	String s = w.word();
	int split = s.lastIndexOf(splitChar);
	if (split <= 0) {    // == 0 isn't allowed - no empty words!
	    return w;
	}
	String word = s.substring(0, split);
	String tag = s.substring(split+1, s.length());
	return new TaggedWord(word, tag);
    }


    /** 
     * Create a <code>WordToTaggedWordProcessor</code> using the default 
     * forward slash character to split on.
     */
    public WordToTaggedWordProcessor() {
	this('/');
    }

    /** Flexibly set the tag splitting chars.  A splitChar of 0 is 
     *  interpreted to mean never split off a tag.
     *  @param splitChar The character at which to split
     */
    public WordToTaggedWordProcessor(char splitChar) {
	this.splitChar = splitChar;
    }

    /** for testing use only */
    private static void printSet(Set s) {
	for (Iterator i = s.iterator(); i.hasNext();) {
	    System.out.print(i.next() + " ");
	}
	System.out.println();
    }


    /** This will print out some text, recognizing tags.  It can be used to 
     *  test tag breaking.  <br>  Usage: <code>
     *  java edu.stanford.nlp.process.WordToTaggedWordProcessor fileOrUrl
     *  </code>
     *  @param args Command line argument: a file or URL
     */
    public static void main(String[] args) {
	if (args.length != 1) {
	    System.out.println("usage: java edu.stanford.nlp.process.WordToTaggedWordProcessor fileOrUrl");
	    System.exit(0);
	}
	String filename = args[0];
	try {
	    Document d = null;
	    if (filename.startsWith("http://")) {
		Document dpre = new BasicDocument().init(new URL(filename));
		Processor notags = new StripTagsProcessor();
		d = notags.process(dpre);
	    } else {
		d = new BasicDocument().init(new File(filename));
	    }
	    Processor proc = new WordToTaggedWordProcessor();
	    Document sentd = proc.process(d);
	    // System.out.println(sentd);
	    int i = 0;
	    for (Iterator it = sentd.iterator(); it.hasNext(); ) {
		HasWord w = (HasWord) it.next();
		System.out.println(i + ": " + w);
		i++;
	    }
	} catch (Exception e) {
	  e.printStackTrace();
	}
    }

}
