package edu.stanford.nlp.trees;

/** An <code>AbstractLabel</code> object acts as a Label with linguistic
 *  attributes.  This is an abstract class, which doesn't actually store
 *  or return anything.  It returns <code>null</code> to any requests.
 *  Designed to be extended.
 *  @author Christopher Manning
 */
public abstract class AbstractLabel implements Label {

    protected AbstractLabel() {
    }


    /** Return the value of the label (or null if none).
     *  The default value returned by an <code>AbstractLabel</code> is
     *  always <code>null</code>
     *  @return the value for the label
     */
    public String value() {
	return null;
    }


    /** Set the value for the label (if one is stored).
     *  @param value - the value for the label
     */
    public void setValue(String value) {
    }


    /** Return a string representation of the label.  This will just
     *  be the <code>value()</code> if it is non-<code>null</code>,
     *  and the empty string otherwise.
     *  @return the string representation
     */
    public String toString() {
	String val = value();
	return (val == null) ? "": val;
    }


    public void setFromString(String labelStr) {
	throw new UnsupportedOperationException();
    }


    /** Equality for <code>Label</code>s is defined in the first instance
     *  as equality of their <code>String</code> <code>value()</code>.
     *  Now rewritten to correctly enforce the contract of equals in Object.
     *  Equality for a <code>Label</code> is determined simply by String
     *  equality of its <code>value()</code>.  Subclasses should not redefine
     *  this to include other aspects of the <code>Label</code>, or the
     *  contract for <code>equals()</code> is broken.
     *  @param obj the object against which equality is to be checked
     *  @return true if <code>this</code> and <code>obj</code> are equal
     */
    public boolean equals(Object obj) {
        String val = value();
	return (obj instanceof Label) &&
	  (val == null ? ((Label) obj).value() == null :
	                 val.equals(((Label) obj).value()));
    }


    /** Return the hashCode of the String value providing there is one.
     *  Otherwise, returns an arbitrary constant for the case of
     *  <code>null</code>.
     */
    public int hashCode() {
	String val = value();
	return val == null ? 3 : val.hashCode();
    }


    /** Implementation of comparison for <code>Label</code>.
     *  Orders by <code>value</code>'s lexicographic ordering.
     *  @param obj object to compare to
     *  @return result (positive if this is greater than obj)
     */
    public int compareTo(Object obj) {
	// will throw an error if cast doesn't work, as is specified
	// in the contract for Comparable
	Label that = (Label) obj;
	return value().compareTo(that.value());
    }


    /** Returns a factory that makes Labels of the appropriate sort.
     *  @return the <code>LabelFactory</code>
     */
    public abstract LabelFactory labelFactory();

}
