package edu.stanford.nlp.trees;

import edu.stanford.nlp.util.Filter;
import edu.stanford.nlp.util.Filters;

/** This provides an implementation of parts of the TreebankLanguagePack
 *  API to reduce the load on fresh implementations.  Only the abstract 
 *  methods below need to be implemented to give a reasonable solution for
 *  a new language.
 * 
 *  @author Christopher Manning
 *  @version 1.1
 */
public abstract class AbstractTreebankLanguagePack
                         implements TreebankLanguagePack {

    /** Use this as the default encoding for Readers and Writers of
     *	Treebank data.
     */
    public static final String DEFAULT_ENCODING = "UTF-8";


    /** Gives a handle to the TreebankLanguagePack */
    public AbstractTreebankLanguagePack() {}


    /** Returns a String array of punctuation tags for this treebank/language.
     *  @return The punctuation tags
     */
    public abstract String[] punctuationTags();

    /** Returns a String array of punctuation words for this treebank/language.
     *  @return The punctuation words
     */
    public abstract String[] punctuationWords();

    /** Returns a String array of sentence final punctuation tags for this
     *  treebank/language.
     *  @return The sentence final punctuation tags
     */
    public abstract String[] sentenceFinalPunctuationTags();

    /** Returns a String array of punctuation tags that EVALB-style evaluation
     *  should ignore for this treebank/language.
     *  Traditionally, EVALB has ignored a subset of the total set of 
     *  punctuation tags in the English Penn Treebank (quotes and 
     *  period, comma, colon, etc., but not brackets)
     *  @return Whether this is a EVALB-ignored punctuation tag
     */
    public String[] evalBIgnoredPunctuationTags() {
	return punctuationTags();
    }


    /** Accepts a String that is a punctuation 
     *  tag name, and rejects everything else.
     *  @return Whether this is a punctuation tag
     */
    public boolean isPunctuationTag(String str) {
	return punctTagStringAcceptFilter.accept(str);
    }


    /** Accepts a String that is a punctuation 
     *  word, and rejects everything else.
     *  If one can't tell for sure (as for ' in the Penn Treebank), it
     *  maks the best guess that it can.
     *  @return Whether this is a punctuation word
     */
    public boolean isPunctuationWord(String str) {
	return punctWordStringAcceptFilter.accept(str);
    }


    /** Accepts a String that is a sentence end
     *  punctuation tag, and rejects everything else.
     *  @return Whether this is a sentence final punctuation tag
     */
    public boolean isSentenceFinalPunctuationTag(String str) {
	return sFPunctTagStringAcceptFilter.accept(str);
    }


    /** Accepts a String that is a punctuation 
     *  tag that should be ignored by EVALB-style evaluation,
     *  and rejects everything else.
     *  Traditionally, EVALB has ignored a subset of the total set of 
     *  punctuation tags in the English Penn Treebank (quotes and 
     *  period, comma, colon, etc., but not brackets)
     *  @return Whether this is a EVALB-ignored punctuation tag
     */
    public boolean isEvalBIgnoredPunctuationTag(String str) {
	return eIPunctTagStringAcceptFilter.accept(str);
    }


    /** Return a filter that accepts a String that is a punctuation 
     *  tag name, and rejects everything else.
     *  @return The filter
     */
    public Filter punctuationTagAcceptFilter() {
	return punctTagStringAcceptFilter;
    }


    /** Returns a filter that accepts a String that is a punctuation 
     *  word, and rejects everything else.
     *  If one can't tell for sure (as for ' in the Penn Treebank), it
     *  maks the best guess that it can.
     *  @return The Filter
     */
    public Filter punctuationWordAcceptFilter() {
	return punctWordStringAcceptFilter;
    }


    /** Returns a filter that accepts a String that is a sentence end
     *  punctuation tag, and rejects everything else.
     *  @return The Filter
     */
    public Filter sentenceFinalPunctuationTagAcceptFilter() {
	return sFPunctTagStringAcceptFilter;
    }


    /** Returns a filter that accepts a String that is a punctuation 
     *  tag that should be ignored by EVALB-style evaluation,
     *  and rejects everything else.
     *  Traditionally, EVALB has ignored a subset of the total set of 
     *  punctuation tags in the English Penn Treebank (quotes and 
     *  period, comma, colon, etc., but not brackets)
     *  @return The Filter
     */
    public Filter evalBIgnoredPunctuationTagAcceptFilter() {
	return eIPunctTagStringAcceptFilter;
    }


    /** Return the input Charset encoding for the Treebank.
     *  See documentation for the <code>Charset</code> class.
     *  @return Name of Charset
     */
    public String getEncoding() {
	return DEFAULT_ENCODING;
    }



    /** Return an array of characters at which a String should be
     *  truncated to give the basic syntactic category of a label.
     *  The idea here is that Penn treebank style labels follow a syntactic
     *  category with various functional and crossreferencing information
     *  introduced by special characters (such as "NP-SBJ=1").  This would
     *  be truncated to "NP" by the array containing '-' and "=".
     *  @return An array of characters that set off label name suffixes
     */
    public char[] labelAnnotationIntroducingCharacters() {
	return new char[0];
    }


    /** Returns the basic syntactic category of a String by truncating
     *  stuff after a (non-word-initial) occurrence of one of the
     *  <code>labelAnnotationIntroducingCharacters()</code>.
     *  @param category The whole String name of the label
     *  @return The basic category of the String
     */
    public String basicCategory(String category) {
	if (category != null) {
	    char[] cutChars = labelAnnotationIntroducingCharacters();
	    for (int i = 0; i < cutChars.length; i++) {
		// a char  at the beginning of label is okay (punctuation tag!)
		int k = category.indexOf(cutChars[i]);
		if (k > 0) {
		    category = category.substring(0, k);
		}
	    }
	}
	return category;
    }


    /** Say whether this character is an annotation introducing
     *  character.
     *  @param ch The character to check
     *  @return Whether it is an annotation introducing character
     */
    public boolean isLabelAnnotationIntroducingCharacter(char ch) {
	char[] cutChars = labelAnnotationIntroducingCharacters();
	for (int i = 0; i < cutChars.length; i++) {
	    if (ch == cutChars[i]) {
		return true;
	    }
	}
	return false;
    }


    /** Accepts a String that is a start symbol of the treebank.
     *  @return Whether this is a start symbol
     */
    public boolean isStartSymbol(String str) {
	return startSymbolAcceptFilter.accept(str);
    }


    /** Return a filter that accepts a String that is a start symbol
     *  of the treebank, and rejects everything else.
     *  @return The filter
     */
    public Filter startSymbolAcceptFilter() {
	return startSymbolAcceptFilter;
    }


    /** Returns a String array of treebank start symbols.
     *  @return The start symbols
     */
    public abstract String[] startSymbols();


    /** Returns a String which is the first (perhaps unique) start symbol
     *  of the treebank, or null if none is defined.
     *  @return The start symbol
     */
    public String startSymbol() {
	String[] ssyms = startSymbols();
	if (ssyms == null || ssyms.length == 0) {
	    return null;
	}
	return ssyms[0];
    }


    private final Filter punctTagStringAcceptFilter = 
	    Filters.collectionAcceptFilter(punctuationTags());

    private final Filter punctWordStringAcceptFilter = 
	    Filters.collectionAcceptFilter(punctuationWords());

    private final Filter sFPunctTagStringAcceptFilter =
	    Filters.collectionAcceptFilter(sentenceFinalPunctuationTags());

    private final Filter eIPunctTagStringAcceptFilter =
	    Filters.collectionAcceptFilter(evalBIgnoredPunctuationTags());

    private final Filter startSymbolAcceptFilter = 
	    Filters.collectionAcceptFilter(startSymbols());

}
