package edu.stanford.nlp.trees;

import java.util.*;

/** Performs collinization operations on Chinese trees similar to
 *  those for English Namely: <ul>
 *  <li> strips all functional &amp; automatically-added tags
 *  <li> strips all punctuation
 *  <li> merges PRN and ADVP
 *  <li> eliminates ROOT (note that there are a few non-unary ROOT nodes;
 *       these are not eliminated)
 *  </ul>
 *  @author Roger Levy
 */
public class ChineseCollinizer implements TreeTransformer {

  private final boolean deletePunct;

  protected TreeFactory tf = new LabeledScoredTreeFactory();


  public ChineseCollinizer() {
    this(true);
  }

  public ChineseCollinizer(boolean deletePunct) {
    this.deletePunct = deletePunct;
  }


  public Tree transformTree(Tree tree, TreeFactory dummy) {
    String label = tree.label().value();

    //System.out.println("Node label is " + label);

    if(tree.isLeaf()) {
      //System.out.println("Node is leaf");
      return tf.newLeaf(new StringLabel(label));
    }
    if(tree.isPreTerminal() && deletePunct && label.matches("^PU.*")) {
      // System.out.println("Deleting punctuation");
      return null;
    }
    List children = new ArrayList();

    if(label.matches("^ROOT") && tree.children().length == 1){ // keep non-unary roots for now
      return transformTree(tree.children()[0],null);
    }

    //System.out.println("Enhanced label is " + label);

    // remove all functional and machine-generated annotations
    label = label.replaceFirst("[^A-Z].*$","");
    // merge parentheticals with adverb phrases
    label = label.replaceFirst("PRN","ADVP");

    //System.out.println("New label is " + label);

    for (int cNum=0; cNum<tree.children().length; cNum++) {
      Tree child = tree.children()[cNum];
      Tree newChild = transformTree(child,null);
      if (newChild != null)
        children.add(newChild);
    }
    if (children.size() == 0)
      return null;
    return tf.newTreeNode(new StringLabel(label), children);
  }

}
