package edu.stanford.nlp.trees;

import java.util.HashMap;
import java.util.HashSet;
import java.io.Reader;

import edu.stanford.nlp.util.Dbg;

/** Implements a 'semantic head' variant of the the HeadFinder found
 *  in Michael Collins' 1999 thesis.
 *  This version chooses the semantic head verb rather than the verb form
 *  for cases with verbs.  Should remember auxiliaries to differentiate past
 *  and passive, though.
 *  Except: we've added a head rule for NX (returns rightmost), whereas
 *  no rule for the head of NX is found in any of the versions of
 *  Collins' head table that we have (did he perhaps use the NP rules
 *  for NX?
 *  @author John Rappaport
 */
public class CollinsSemanticHeadFinder implements HeadFinder {

    /** Determine which daughter of the current parse tree is the
     *  head.  It assumes that the daughters alread have had their
     *  heads determined. 
     *  @param pt The parse tree to examine the daughters of.
     *  This is assumed to never be a leaf
     *  @return The parse tree that is the head
     */
    public Tree determineHead(Tree t) {
	if (t.isLeaf()) {
	    return null;
	}
	Tree[] kids = t.children();
	String motherCat = t.label().value();
	// if the node is a unary, then that kid must be the head
	// it used to special case preterminal and ROOT/TOP case
	// but that seemed bad (especially hardcoding string "ROOT")
	if (kids.length == 1) {
	    return kids[0];
	} else {
	    if (Dbg.ONE) {
		Dbg.print("looking for head of " + t.label());
	    }
	    // We know we have nonterminals underneath
	    // (a bit of a Penn Treebank assumption, but)
	    // look at label.
	    Tree theHead;
	    if (motherCat.equals("NP")) {
		theHead = determineNPHead(t);
	    } else if (motherCat.equals("VP")) {
		theHead = determineVPHead(t);
	    } else {
		theHead = traverseLocate(kids, motherCat, true);
	    }
	    if (Dbg.ONE) {
		Dbg.print("  head was " + motherCat + "[" +
			  /* theHead.headWord + */ "]");
	    }
	    return theHead;
	}
    }


    private Tree determineVPHead(Tree mum) {
	Tree[] kids = mum.children();
	// try to find if there is an auxiliary verb
	boolean hasAux = false;
	for (int i = 0; i < kids.length; i++) {
	    if (isVerbalAuxiliary(kids[i])) {
		hasAux = true;
	    }
	}
	Tree pti;
	if (hasAux) {
	    pti = traverseLocate(kids, "!!VP1", false);
	    if (pti != null) {
		return pti;
	    }
	}
	return traverseLocate(kids, "VP", true);
    }



    private Tree determineNPHead(Tree hptr) {
	// a total special case....
	// first look for POS tag at end
	Tree lastDtr = hptr.lastChild();
	if (lastDtr.label().value().equals("POS")) {
	    return lastDtr;
	} else {
	    Tree pti;
	    // search from R to L for first NN, NNP, NNPS, NNS, NX,
	    // POS or JJR
	    // CDM: why does one need that first rule then.  This
	    // second rule would always cover the first!!
	    pti = traverseLocate(hptr.children(), "!!NP1", false);
	    if (pti != null) {
		return pti;
	    }
	    pti = traverseLocate(hptr.children(), "!!NP2", false);
	    if (pti != null) {
		return pti;
	    }
	    pti = traverseLocate(hptr.children(), "!!NP3", false);
	    if (pti != null) {
		return pti;
	    }
	    pti = traverseLocate(hptr.children(), "!!NP4", false);
	    if (pti != null) {
		return pti;
	    }
	    // if the things in !!NP5 aren't found, the last word
	    // will be returned, which is Collins' final clause
	    return traverseLocate(hptr.children(), "!!NP5", true);
	}
    }


    /** Go through daughterTrees looking for things from how, and if
     *	you do not find one, take leftmost or rightmost thing iff
     *	deflt is true
     */
    private Tree traverseLocate(Tree[] daughterTrees,
					String motherKey, boolean deflt) {

	String[] how = (String[]) nonTerminalInfo.get(motherKey);
	Tree theHead = null;
	int headIdx = 0;
	String childCat;
	boolean found = false;

	if (how == null) {
	    return null;
	}
	if (how[0].equals("left")) {
	    twoloop:
	    for (int i = 1; i < how.length; i++) {
		for (headIdx = 0; headIdx < daughterTrees.length; headIdx++) {
		    childCat = daughterTrees[headIdx].label().value();
		    if (how[i] == childCat) {
			theHead = daughterTrees[headIdx];
			found = true;
			break twoloop;
		    }
		}
	    }
	    if ( ! found) {
		// none found by tag, so return first or null
		if (deflt) {
		    headIdx = 0;
		    theHead = daughterTrees[headIdx];
		} else {
		    return null;
		}
	    }
	} else if (how[0].equals("rightdis")) {
	    // from right, but search for any, not in turn
	    twoloop:
	    for (headIdx = daughterTrees.length - 1; headIdx >= 0; headIdx--) {
		childCat = daughterTrees[headIdx].label().value();
		for (int i = 1; i < how.length; i++) {
		    if (Dbg.ONE) {
			Dbg.print("Testing for whether " + how[i] + " == " +
				  childCat + ": " + ((how[i].equals(childCat))?
				  "true": "false"));
		    }
		    if (how[i].equals(childCat)) {
			theHead = daughterTrees[headIdx];
			found = true;
			break twoloop;
		    }
		}
	    }
	    if ( ! found) {
		// none found by tag, so return last or null
		if (deflt) {
		    headIdx = daughterTrees.length - 1;
		    theHead = daughterTrees[headIdx];
		} else {
		    return null;
		}
	    }
	} else {
	    // from right
	    twoloop:
	    for (int i = 1; i < how.length; i++) {
		for (headIdx = daughterTrees.length - 1; headIdx >= 0;
								headIdx--) {
		    childCat = daughterTrees[headIdx].label().value();
		    if (how[i].equals(childCat)) {
			theHead = daughterTrees[headIdx];
			found = true;
			break twoloop;
		    }
		}
	    }
	    if ( ! found) {
		// none found by tag, so return last or null
		if (deflt) {
		    headIdx = daughterTrees.length - 1;
		    theHead = daughterTrees[headIdx];
		} else {
		    return null;
		}
	    }
	}
	// fix for coordination so left of pair is alway head (p.237). Messy
	if (headIdx >= 2) {
	    String prevLab = daughterTrees[headIdx - 1].label().toString();
	    if (prevLab.equals("CC")) {
		headIdx -= 2;
		theHead = daughterTrees[headIdx];
	    }
	}
	if (Dbg.ONE) {
	    Dbg.print("Head for " + motherKey +
		      " chose index " + headIdx + " " +
		      theHead.label().toString() +
		      "[" + /* theHead.headWord + */ "]");
	}
	return theHead;
    }


    public boolean isVerbalAuxiliary(Tree t) {
	String tag = ((HasTag) t.label()).tag();
	String word = ((HasWord) t.label()).word();
	// System.out.print("Asking whether " + word + "/" + tag + " is an aux.");
	if (verbalTags.contains(tag) && verbalAuxiliaries.contains(word)) {
	    // System.out.println(" Yes.");
	    return true;
	} else {
	    // System.out.println(" No.");
	    return false;
	}
    }


    private HashMap nonTerminalInfo;

    // Initialization block
    {
	nonTerminalInfo = new HashMap();
	// This version from Collins' diss pp. 236-238
	nonTerminalInfo.put("ADJP",
			    new String[] {"left", "NNS", "QP", "NN", "$",
					  "ADVP", "JJ",
					  "VBN", "VBG", "ADJP", "JJR", "NP",
					  "JJS", "DT", "FW", "RBR", "RBS",
					  "SBAR", "RB"});
	nonTerminalInfo.put("ADVP",
			    new String[] {"right", "RB", "RBR", "RBS", "FW",
					  "ADVP",
					  "TO", "CD", "JJR", "JJ", "IN",
					  "NP", "JJS", "NN" });
	nonTerminalInfo.put("CONJP",
			    new String[] {"right", "CC", "RB", "IN"});
	nonTerminalInfo.put("FRAG",
			    new String[] {"right"}); // crap
	nonTerminalInfo.put("INTJ",
			    new String[] {"left"});
	nonTerminalInfo.put("LST",
			    new String[] {"right", "LS", ":"});
	nonTerminalInfo.put("NAC",
			    new String[] {"left", "NN", "NNS", "NNP", "NNPS",
					  "NP", "NAC", "EX", "$", "CD", "QP",
					  "PRP", "VBG", "JJ", "JJS", "JJR",
					  "ADJP", "FW"});
	nonTerminalInfo.put("NX",
			    new String[] {"right"}); // crap
	nonTerminalInfo.put("PP",
			    new String[] {"right", "IN", "TO", "VBG", "VBN",
					      "RP", "FW"});
	// should prefer JJ? (PP (JJ such) (IN as) (NP (NN crocidolite)))
	nonTerminalInfo.put("PRN",
			    new String[] {"left"});
	nonTerminalInfo.put("PRT",
			    new String[] {"right", "RP"});
	nonTerminalInfo.put("QP",
			    new String[] {"left", "$", "IN", "NNS", "NN",
					      "JJ", "RB", "DT", "CD", "NCD",
					      "QP", "JJR", "JJS"});
	nonTerminalInfo.put("RRC",
			    new String[] {"right", "VP", "NP", "ADVP", "ADJP",
					      "PP"});
	nonTerminalInfo.put("S",
			    new String[] {"left", "TO", "IN", "VP", "S",
					      "SBAR", "ADJP", "UCP", "NP"});
	nonTerminalInfo.put("SBAR",
			    new String[] {"left", "WHNP", "WHPP", "WHADVP",
					      "WHADJP", "IN", "DT", "S",
					      "SQ", "SINV", "SBAR", "FRAG"});
	nonTerminalInfo.put("SBARQ",
			    new String[] {"left", "SQ", "S", "SINV", "SBARQ",
					      "FRAG"});
	nonTerminalInfo.put("SINV",
			    new String[] {"left", "VBZ", "VBD", "VBP",
					      "VB", "MD", "VP", "S", "SINV",
					      "ADJP", "NP"});
	nonTerminalInfo.put("SQ",
			    new String[] {"left", "VBZ", "VBD", "VBP",
					      "VB", "MD", "VP", "SQ"});
	nonTerminalInfo.put("UCP",
			    new String[] {"right"});
	nonTerminalInfo.put("VP",
			    new String[] {"left", "TO", "VBD", "VBN", "MD",
					      "VBZ", "VB", "VBG", "VBP",
					      "VP", "ADJP", "NN",
					      "NNS", "NP"});
	nonTerminalInfo.put("WHADJP",
			    new String[] {"left", "CC", "WRB", "JJ", "ADJP"});
	nonTerminalInfo.put("WHADVP",
			    new String[] {"right", "CC", "WRB"});
	nonTerminalInfo.put("WHNP",
			    new String[] {"left", "WDT", "WP", "WP$",
					      "WHADJP", "WHPP", "WHNP"});
	nonTerminalInfo.put("WHPP",
			    new String[] {"right", "IN", "TO", "FW"});
	nonTerminalInfo.put("X",
			    new String[] {"right"}); // crap
	nonTerminalInfo.put("!!NP1",
			    new String[] {"rightdis", "NN", "NNP", "NNPS",
					      "NNS", "NX", "POS", "JJR"});
	nonTerminalInfo.put("!!NP2",
			    new String[] {"left", "NP"});
	nonTerminalInfo.put("!!NP3",
			    new String[] {"rightdis", "$", "ADJP", "PRN"});
	nonTerminalInfo.put("!!NP4",
			    new String[] {"right", "CD"});
	nonTerminalInfo.put("!!NP5",
			   new String[] {"rightdis", "JJ", "JJS", "RB", "QP"});
	//	nonTerminalInfo.put("!!VP1", new String[] {"left", "VP",
	//						"ADJP", "NP"});
	// Including NP etc seems okay for copular sentences but is 
	// problematic for other auxiliaries, like 'he has an answer'
	nonTerminalInfo.put("!!VP1", new String[] {"left", "VP"});
    }


    private HashSet verbalAuxiliaries;
    private HashSet verbalTags;

    // Initialization block
    {
	verbalAuxiliaries = new HashSet();
	verbalAuxiliaries.add("will");
	verbalAuxiliaries.add("wo");	// for wo n't
	verbalAuxiliaries.add("shall");
	verbalAuxiliaries.add("may");
	verbalAuxiliaries.add("might");
	verbalAuxiliaries.add("should");
	verbalAuxiliaries.add("would");
	verbalAuxiliaries.add("can");
	verbalAuxiliaries.add("ca");	// for ca n't
	verbalAuxiliaries.add("could");
	verbalAuxiliaries.add("must");
	verbalAuxiliaries.add("has");
	verbalAuxiliaries.add("have");
	verbalAuxiliaries.add("had");
	verbalAuxiliaries.add("having");
	verbalAuxiliaries.add("am");
	verbalAuxiliaries.add("are");
	verbalAuxiliaries.add("is");
	verbalAuxiliaries.add("was");
	verbalAuxiliaries.add("were");
	verbalAuxiliaries.add("be");
	verbalAuxiliaries.add("being");
	verbalAuxiliaries.add("been");
	verbalAuxiliaries.add("get");
	verbalAuxiliaries.add("gets");
	verbalAuxiliaries.add("getting");
	verbalAuxiliaries.add("got");
	verbalAuxiliaries.add("gotten");
	verbalAuxiliaries.add("do");
	verbalAuxiliaries.add("does");
	verbalAuxiliaries.add("did");
	verbalAuxiliaries.add("to");
	verbalAuxiliaries.add("'m");
	verbalAuxiliaries.add("'re");
	verbalAuxiliaries.add("'s");
	verbalAuxiliaries.add("'ve");
	verbalAuxiliaries.add("'d");
	verbalAuxiliaries.add("'ll");

	verbalTags = new HashSet();
	verbalTags.add("TO");
	verbalTags.add("MD");
	verbalTags.add("VB");
	verbalTags.add("VBD");
	verbalTags.add("VBP");
	verbalTags.add("VBD");
	verbalTags.add("VBZ");
	verbalTags.add("VBG");
	verbalTags.add("VBN");
	// include Charniak tags so can do BLLIP right
	verbalTags.add("AUX");
	verbalTags.add("AUXG");
    }


    /** Go through trees and determine their heads and print them.
     *  Just for debuggin'
     *  Use: java CollinsHeadFinder <path>/wsj/07
     */
    public static void main(String[] args) {
	Treebank treebank = new DiskTreebank(new TreeReaderFactory() {
	    public TreeReader newTreeReader(Reader in) {
	        return new PennTreeReader(in,
			  new LabeledScoredTreeFactory(
			  new CategoryWordTagFactory()),
			  new BobChrisTreeNormalizer());
	    }
	});
	treebank.loadPath(args[0]);
	final HeadFinder chf = new CollinsHeadFinder();
	treebank.apply(new TreeProcessor() {
	    public void processTree(Tree pt) {
	        pt.percolateHeads(chf);
		pt.pennPrint();
		System.out.println();
	    }
	});
    }

}
