package edu.stanford.nlp.trees;

import java.io.*;
import java.util.*;

import edu.stanford.nlp.util.*;

/** A <code>DiskTreebank</code> object stores merely the information to
 *  get at a corpus of trees that is stored on disk.  Access is usually
 *  via applying a TreeProcessor to an iteration over the trees.
 * 
 *  @author Christopher Manning
 */
public final class DiskTreebank extends Treebank {

    private String encoding = TreebankLanguagePack.DEFAULT_ENCODING;

    private ArrayList filePaths = new ArrayList();
    private ArrayList fileFilters = new ArrayList();

    /** Maintains as a class variable the <code>File</code> from which
     *  trees are currently being read.
     */
    private File currentFile = null;

    /** If this is true, the system will retry opening files a few times
     *  before concluding that there is really a problem. This seems to
     *  be necessary with NFS on Linux boxes -- at least the DB ones.
     */
    private static final boolean BROKEN_NFS = true;


    /** Create a new DiskTreebank.
     *  The trees are made with a <code>SimpleTreeReaderFactory</code>.
     */
    public DiskTreebank() {
	this(new SimpleTreeReaderFactory());
    }

    /** Create a new tree bank, set the encoding for file access.
     *  @param encoding the encoding to use for file access.
     */
    public DiskTreebank(String encoding) {
	this.encoding = encoding;
    }

    /** Create a new DiskTreebank.
     *	@param trf the factory class to be called to create a new
     *  <code>TreeReader</code>
     */
    public DiskTreebank(TreeReaderFactory trf) {
	super(trf);
    }

    /** Create a new DiskTreebank.
     *	@param trf the factory class to be called to create a new
     *  <code>TreeReader</code>
     *  @param encoding the encoding to use for file access.
     */
    public DiskTreebank(TreeReaderFactory trf, String encoding) {
	super(trf);
	this.encoding = encoding;
    }

    /** Create a new Treebank.
     *  The trees are made with a <code>SimpleTreeReaderFactory</code>.
     *	@param initialCapacity The initial size of the underlying Collection.
     *  For a <code>DiskTreebank</code>, this parameter is ignored.
     */
    public DiskTreebank(int initialCapacity) {
	this(initialCapacity, new SimpleTreeReaderFactory());
    }


    /** Create a new Treebank.
     *	@param initialCapacity The initial size of the underlying Collection,
     *  For a <code>DiskTreebank</code>, this parameter is ignored.
     *	@param trf the factory class to be called to create a new
     *  <code>TreeReader</code>
     */
    public DiskTreebank(int initialCapacity, TreeReaderFactory trf) {
	this(trf);
    }


    /** Empty a <code>Treebank</code>.
     */
    public void clear() {
	filePaths.clear();
	fileFilters.clear();
    }


    /** Load trees from given directory.  This version just records
     *  the paths to be processed, and actually processes them at apply time.
     * @param path  file or directory to load from
     * @param filt a FilenameFilter of files to load
     */
    public void loadPath(File path, FileFilter filt) {
	filePaths.add(path);
	fileFilters.add(filt);
    }


    /** Applies the TreeProcessor to to all trees in the Treebank.
     *  @param tp A class that can process trees.
     */
    public void apply(final TreeProcessor tp) {
	DiskTreebankFileProcessor dtfp = new DiskTreebankFileProcessor(tp);
	int numPaths = filePaths.size();
	for (int i = 0; i < numPaths; i++) {
	    FilePathProcessor.processPath((File) filePaths.get(i),
					  (FileFilter) fileFilters.get(i),
					  dtfp);
	}
    }


    private final class DiskTreebankFileProcessor implements FileProcessor {

	/** The <code>TreeProcessor</code> to apply to trees in the file
	 */
	final TreeProcessor tp;

	DiskTreebankFileProcessor(TreeProcessor tp) {
	    this.tp = tp;
	}


	/** Load a collection of parse trees from the file of given name.
	 *  Each tree may optionally be encased in parens to allow for Penn
	 *  Treebank style trees.
	 *  @param file <code>File</code> to load trees from
	 */
	public void processFile(File file) {
	    TreeReader tr = null;
	    Tree pt;

	    // maybe print file name to stdout to get some feedback
	    if (Dbg.PROGRESS) {
		System.err.println(file);
	    }
	    // save the current file so you can get at it
	    currentFile = file;
	    try {
		// could throw an IO exception if can't open for reading
		if (BROKEN_NFS) {
		    // try a few times to open a file: NFS is broken
		    int numTries = 0;
		    while (tr == null) {
			try {
			    tr = treeReaderFactory().newTreeReader(
				    new BufferedReader(new InputStreamReader(new FileInputStream(file),encoding)));
			} catch (IOException e) {
			    if (Dbg.ERROR) e.printStackTrace(System.err);
			    // if we've hit the max number of retries, throw e
			    if (numTries >= 10)
				throw e;
			    numTries++;
			    // sleep a bit so that NFS gets a chance to react
			    try {
				Thread.sleep((int)Math.exp(numTries)/2);
			    } catch (InterruptedException iE) {}
			}
		    }
		} else {
		    // could throw an IOException if can't open for reading
		    tr = treeReaderFactory().newTreeReader(
				    new BufferedReader(new InputStreamReader(new FileInputStream(file),encoding)));
		}
		while ((pt = tr.readTree()) != null) {
		    tp.processTree(pt);
		}
	    } catch (IOException e) {
		System.err.println("loadTree IO Exception: " + e + " in " +
			getCurrentFile());
	    } finally {
		currentFile = null;
		try {
		    if (tr != null) {
			tr.close();  // important: closes file even if error!
		    }
		} catch (IOException e) {
		}
	    }
	} // end processFile()

    } // end private class DiskTreebankFileProcessor


    /** Return the <code>File</code> from which trees are currently being
     *  read by <code>apply()</code>, and pased to a
     *  <code>TreePprocessor</code>.  This is useful if one wants to map
     *  the original file and
     *  directory structure over to a set of modified trees.
     *  @return the file that trees are currently being read from, or
     *  <code>null</code> if no file is currently open
     */
    public File getCurrentFile() {
	return currentFile;
    }


    private class DiskTreebankIterator implements Iterator {

	private int fileUpto = -1; // before starting on index array 0
	private int treeUpto; // = 0
	private List files;
	private MemoryTreebank currentFileTrees;
	private boolean hasNext;

	private DiskTreebankIterator() {
	    files = new ArrayList();
	    // get the list of files in the Treebank via a new
	    // FilePathProcessor()
	    FileProcessor dtifp = new FileProcessor() {
		    public void processFile(File file) {
			files.add(file.toString());
		    };
		};
	    int numPaths = filePaths.size();
	    for (int i = 0; i < numPaths; i++) {
	        FilePathProcessor.processPath((File) filePaths.get(i),
					  (FileFilter) fileFilters.get(i),
					  dtifp);
	    }
	    currentFileTrees = new MemoryTreebank(treeReaderFactory());
	    // we're now all setup to read a new file on the priming call
	    // first time treeUpto = currentFileTrees.size() = 0
	    hasNext = primeNextFile();
	}

	private boolean primeNextFile() {
	    while (fileUpto < files.size()) {
		if (treeUpto < currentFileTrees.size()) {
		    return true;
		} else {
		    // load the next file
		    currentFileTrees.clear();
		    fileUpto++;
		    treeUpto = 0;
		    if (fileUpto < files.size()) {
			currentFileTrees.loadPath((String)files.get(fileUpto));
		    }
		}
	    }
	    // there's nothing left;
	    return false;
	}

	/** Returns true if the iteration has more elements. */
	public boolean hasNext() {
	    return hasNext;
	}

	/** Returns the next element in the iteration. */
	public Object next() {
	    Object ret = currentFileTrees.get(treeUpto++);
	    hasNext = primeNextFile();
	    return ret;
	}

	/** Not supported */
	public void remove() {
	    throw new UnsupportedOperationException();
	}

    }


    /** Return an Iterator over Trees in the Treebank.  This is implemented
     *  by building per-file MemoryTreebanks for the files in the
     *  DiskTreebank.  As such, it isn't as efficient as using
     *  <code>apply()</code>.
     */
    public Iterator iterator() {
	return new DiskTreebankIterator();
    }


    /** Loads treebank from first argument and prints it.
     *  All files below the designated <code>filePath</code> with the
     *  given extension (or <code>.mrg</code> by default) are loaded.
     *  Usage: <code>
     *  java edu.stanford.nlp.trees.DiskTreebank filePath [ext]
     *  </code>
     *  @param args Array of command-line arguments
     */
    public static void main(String[] args) {
	Timing.startTime();
	Treebank treebank = new DiskTreebank(new TreeReaderFactory() {
	    public TreeReader newTreeReader(Reader in) {
	        return new PennTreeReader(in, new LabeledScoredTreeFactory());
	    }
	});
	if (args.length > 1) {
	    treebank.loadPath(args[0], args[1], true);
	} else {
	    treebank.loadPath(args[0]);
	}
	treebank.apply(new TreeProcessor() {
	    public void processTree(final Tree t) {
		System.out.println(t);
	    }
	});
	System.out.println();
	Timing.endTime("traversing corpus, printing trees 1-by-1");

	System.out.println();
	Timing.startTime();
	System.out.println(treebank);
	Timing.endTime("traversing corpus, appending all files");

	System.out.println();
	Timing.startTime();
	for (Iterator iter = treebank.iterator(); iter.hasNext(); ) {
	    Tree t = (Tree) iter.next();
	    System.out.println(t);
	}
	Timing.endTime("traversing corpus, with iterator");

	System.out.println();
	Timing.startTime();
	System.out.println("This treebank contains " + treebank.size() +
			   " trees.");
	Timing.endTime("size of corpus");
    }

}
