package edu.stanford.nlp.trees;

/** Something that implements the <code>Label</code> interface can act as a
 *  constituent, node, or word label with linguistic attributes.
 *  A <code>Label</code> is expected to have a "primary" <code>String</code>
 *  <code>value()</code> (although this may be null).  This is referred to as
 *  its 
 *  <code>value</code>.  Equality of Labels is standardly defined by the
 *  String equality of its value.  
 *  <p>
 *  <i>Implementors notes:</i>
 *  In general, you should live with the definition of equality, and see
 *  other fields
 *  of a Label subclass as secondary facets (in theory subclasses may override
 *  this behavior, but in practice it is almost impossible to do this in 
 *  a useful way while observing the contract for equality defined for Object,
 *  in particular, that equality must by symmetric).
 *  A subclass that extends another Label class <i>should</i> override
 *  the definition of <code>labelFactory()</code>, since the contract for
 *  this method is that it should return a factory for labels of the 
 *  exact same object type.
 *
 *  @author Christopher Manning
 */
public interface Label extends Comparable {

    /** Return a String representation of just the "main" value of this label.
     *  @return the "value" of the label
     */
    public String value();


    /** Set the value for the label (if one is stored).
     *  @param value - the value for the label
     */
    public void setValue(String value);


    /** Return a String representation of the label.  For a multipart label,
     *  this will return all parts.  The <code>toString()</code> method
     *  causes a label to spill its guts.  It should always return an
     *  empty string rather than <code>null</code> if there is no value.
     *  @return a text representation of the full label contents
     */
    public String toString();


    /** Set the contents of this label to this <code>String</code>
     *  representing the
     *  complete contents of the label.  A class implementing label may
     *  throw an <code>UnsupportedOperationException</code> for this
     *  method (only).  Typically, this method would do
     *  some appropriate decoding of the string in a way that sets
     *  multiple fields in an inverse of the <code>toString()</code>
     *  method.
     *  @param labelStr the String that translates into the content of the
     *           label
     */
    public void setFromString(String labelStr);

    /** Implementation of comparison for <code>Label</code>.
     *  By default this orders by value's string lexicographic ordering.
     *  @param obj object to compare to
     *  @return result (positive if <code>this</code> is greater than
     *       <code>obj</code>, 0 if equal, negative otherwise)
     */
    public int compareTo(Object obj);

    /** Returns a factory that makes labels of the same type as this one.
     *  May return <code>null</code> if no appropriate factory is known.
     *  @return the LabelFactory for this kind of label
     */
    public LabelFactory labelFactory();

}
