package edu.stanford.nlp.trees;

import java.io.*;
import java.util.*;

import edu.stanford.nlp.util.*;

/**
 * A <code>MemoryTreebank</code> object stores a corpus of examples with
 * given tree structures in memory (as a <code>List</code>).
 * @author Christopher Manning
 * @author Roger Levy (added encoding variable and method)
 * @version 2003/26/01
 */
public final class MemoryTreebank extends Treebank implements FileProcessor,
                                                              List {

    private String encoding = TreebankLanguagePack.DEFAULT_ENCODING;

    /** Returns the encoding in use for bytestream access.
     */
    public String getEncoding() {
        return encoding;
    }


    /** The collection of parse trees.
     */
    private ArrayList parseTrees;


    /** Create a new tree bank.
     */
    public MemoryTreebank() {
        this(new SimpleTreeReaderFactory());
    }

    /** Create a new tree bank, set the encoding for file access
     * @param encoding the encoding to use for file access.
     */
    public MemoryTreebank(String encoding) {
        this(new SimpleTreeReaderFactory());
        this.encoding = encoding;
    }

    /** Create a new tree bank.
     *  @param trf the factory class to be called to create a new
     *  <code>TreeReader</code>
     */
    public MemoryTreebank(TreeReaderFactory trf) {
        super(trf);
        parseTrees = new ArrayList();
    }


    /** Create a new tree bank.
     *  @param trf the factory class to be called to create a new
     *  <code>TreeReader</code>
     * @param encoding the encoding to use for file access.
     */
    public MemoryTreebank(TreeReaderFactory trf, String encoding) {
        super(trf);
        this.encoding = encoding;
        parseTrees = new ArrayList();
    }

    /** Create a new Treebank.
     *  @param initialCapacity The initial size of the underlying Collection,
     *  (if a Collection-based storage mechanism is being provided)
     */
    public MemoryTreebank(int initialCapacity) {
        this(initialCapacity, new SimpleTreeReaderFactory());
    }


    /** Create a new tree bank.
     *  @param initialCapacity The initial size of the underlying Collection
     *  @param trf the factory class to be called to create a new
     *  <code>TreeReader</code>
     */
    public MemoryTreebank(int initialCapacity, TreeReaderFactory trf) {
        super(initialCapacity, trf);
        parseTrees = new ArrayList(initialCapacity);
    }


    /** Empty a <code>Treebank</code>.
     */
    public void clear() {
        parseTrees.clear();
    }


    /** Load trees from given directory.
     * @param path  file or directory to load from
     * @param filt a FilenameFilter of files to load
     */
    public void loadPath(File path, FileFilter filt) {
        FilePathProcessor.processPath(path, filt, this);
    }


    /** Load a collection of parse trees from the file of given name.
     *  Each tree may optionally be encased in parens to allow for Penn
     *  Treebank style trees.
     *  This methods implements the <code>FileProcessor</code> interface.
     *  @param file file to load a tree from
     */
    public void processFile(File file) {
        TreeReader tr = null;
        Tree pt;

        try {
            // maybe print file name to stdout to get some feedback
            if (Dbg.PROGRESS) {
                System.err.println(file);
            }
            // could throw an IO exception if can't open for reading
            tr = treeReaderFactory().newTreeReader(new BufferedReader(
                                                    new InputStreamReader(new FileInputStream(file),encoding)));
            while ((pt = tr.readTree()) != null) {
                parseTrees.add(pt);
            }
        } catch (IOException e) {
            System.err.println("loadTree IO Exception: " + e);
        } finally {
            try {
                if (tr != null) {
                    tr.close();  // important: closes file even if error!
                }
            } catch (IOException e) {
            }
        }
    }


    /** Get a tree by index from the Treebank.
     *  This operation isn't in the <code>Treebank</code> feature set, and
     *  so is only available with a <code>MemoryTreebank</code>, but is
     *  useful in allowing the latter to be used as an <code>ArrayList</code>.
     *  @param i The integer (counting from 0) index of the tree
     *  @return A tree
     */
    public Object get(int i) {
        return parseTrees.get(i);
    }


    /** Apply the TreeProcessor tp to all trees in the Treebank.
     *  @param tp A class that implements the TreeProcessor interface
     */
    public void apply(TreeProcessor tp) {
        for (int i = 0, size = parseTrees.size(); i < size; i++) {
            tp.processTree((Tree) parseTrees.get(i));
        }
        // or could do as Iterator but slower
        // Iterator iter = parseTrees.iterator();
        // while (iter.hasNext()) {
        //    tp.processTree((Tree) iter.next());
        // }
    }


    /** Return an Iterator over Trees in the Treebank.
     *  @return The iterator
     */
    public Iterator iterator() {
        return parseTrees.iterator();
    }


    /** Returns the size of the Treebank.
     *  Provides a more efficient implementation than the one for a
     *  generic <code>Treebank</code>
     *  @return the number of trees in the Treebank
     */
    public int size() {
        return parseTrees.size();
    }


    // Extra stuff to implement List interface

    public void add(int index, Object element) {
        parseTrees.add(index, element);
    }

    public boolean add(Object element) {
        return parseTrees.add(element);
    }

    public boolean addAll(int index, Collection c) {
        return parseTrees.addAll(index, c);
    }

    public int indexOf(Object o) {
        return parseTrees.indexOf(o);
    }

    public int lastIndexOf(Object o) {
        return parseTrees.lastIndexOf(o);
    }

    public Object remove(int index) {
        return parseTrees.remove(index);
    }

    public Object set(int index, Object element) {
        return parseTrees.set(index, element);
    }

    public ListIterator listIterator() {
        return parseTrees.listIterator();
    }

    public ListIterator listIterator(int index) {
        return parseTrees.listIterator(index);
    }

    public List subList(int fromIndex, int toIndex) {
        return parseTrees.subList(fromIndex, toIndex);
    }


    /** Loads treebank grammar from first argument and prints it.
     *  Just a demonstration of functionality. <br>
     *  <code>usage: java MemoryTreebank treebankFilesPath</code>
     *  @param args array of command-line arguments
     */
    public static void main(String[] args) {
        Timing.startTime();
        Treebank treebank = new MemoryTreebank(new TreeReaderFactory() {
            public TreeReader newTreeReader(Reader in) {
                return new PennTreeReader(in, new LabeledScoredTreeFactory());
            }
        });
        treebank.loadPath(args[0]);
        Timing.endTime();
        System.out.println(treebank);
    }

}
