package edu.stanford.nlp.trees;

import java.util.HashMap;
import java.io.Reader;

import edu.stanford.nlp.util.Dbg;

/** Implements a variant on the HeadFinder found in Michael Collins' 1999  
 *  thesis. 
 *  This starts with Collins' head finder.
 *  As in <code>CollinsHeadFinder.java</code>, we've added a head rule for
 *  NX (returns rightmost), whereas
 *  no rule for the head of NX is found in any of the versions of
 *  Collins' head table that we have (did he perhaps use the NP rules
 *  for NX?).
 *  Changes:
 *  <ol>
 *  <li>The PRN rule used to just take the leftmost thing, we now have it 
 *  choose the leftmost lexical category (not the common punctuation etc.)
 *  <li>Delete IN as a possible head of S, and add FRAG (low priority)
 *  <li>Place NN before QP in ADJP head rules (more to do for ADJP!)
 *  </ol>
 *  These rules are suitable for the Penn Treebank.
 *  This code assumes that tree category labels are interned and
 *  does comparisons with equality (==).  It won't work if the category labels
 *  are not interned.  In particular, this also means that functional tags
 *  should have been stripped.
 *  
 *  @author Christopher Manning
 */
public class ModCollinsHeadFinder implements HeadFinder {

    /** Determine which daughter of the current parse tree is the head.
     *  It assumes that category Strings for trees (as returned by
     *  <code>t.label().value()</code>) have been interned and can be
     *  compared with Penn Treebank label strings via ==.
     *  @param t The parse tree to examine the daughters of.
     *            If this is a leaf, <code>null</code> is returned
     *  @return The daughter parse tree that is the head of <code>t</code>
     *  @see Tree#percolateHeads(HeadFinder)
     *       for a routine to call this and spread heads throughout a tree
     */
    public Tree determineHead(Tree t) {
	if (t.isLeaf()) {
	    return null;
	}
	Tree[] kids = t.children();
	String motherCat = t.label().value();
	// if the node is a unary, then that kid must be the head
	// it used to special case preterminal and ROOT/TOP case
	// but that seemed bad (especially hardcoding string "ROOT")
	if (kids.length == 1) {
	    return kids[0];
	} else {
	    if (Dbg.ONE) {
		Dbg.print("looking for head of " + t.label());
	    }
	    // We know we have nonterminals underneath
	    // (a bit of a Penn Treebank assumption, but)
	    // look at label. We assume it was interned
	    Tree theHead;
	    if (motherCat.equals("NP")) {
		theHead = determineNPHead(t);
	    } else {
		theHead = traverseLocate(kids, motherCat, true);
	    }
	    if (Dbg.ONE) {
		Dbg.print("  head was " + motherCat + "[" +
			  /* theHead.headWord + */ "]");
	    }
	    return theHead;
	}
    }


    private Tree determineNPHead(Tree hptr) {
	// a total special case....
	// first look for POS tag at end
	Tree lastDtr = hptr.lastChild();
	if (lastDtr.label().value().equals("POS")) {
	    return lastDtr;
	} else {
	    Tree pti;
	    // search from R to L for first NN, NNP, NNPS, NNS, NX,
	    // POS or JJR
	    // CDM: why does one need that first rule then.  This
	    // second rule would always cover the first!!
	    pti = traverseLocate(hptr.children(), "!!NP1", false);
	    if (pti != null) {
		return pti;
	    }
	    pti = traverseLocate(hptr.children(), "!!NP2", false);
	    if (pti != null) {
		return pti;
	    }
	    pti = traverseLocate(hptr.children(), "!!NP3", false);
	    if (pti != null) {
		return pti;
	    }
	    pti = traverseLocate(hptr.children(), "!!NP4", false);
	    if (pti != null) {
		return pti;
	    }
	    // if the things in !!NP5 aren't found, the last word
	    // will be returned, which is Collins' final clause
	    return traverseLocate(hptr.children(), "!!NP5", true);
	}
    }


    /** Attempt to locate head daughter tree from among daughters.
     *  Go through daughterTrees looking for things from a set found by
     *  looking up the motherkey specifier in a hash map, and if
     *	you do not find one, take leftmost or rightmost thing iff
     *	deflt is true, otherwise return <code>null</code>.
     */
    private Tree traverseLocate(Tree[] daughterTrees,
					String motherKey, boolean deflt) {

	String[] how = (String[]) nonTerminalInfo.get(motherKey);
	Tree theHead = null;
	int headIdx = 0;
	String childCat;
	boolean found = false;

	if (how == null) {
	    if (Dbg.ONE) {
		Dbg.print("Warning: No rule found for " + motherKey);
	    }
	    return null;
	}
	if (how[0] == "left") {
	    twoloop:
	    for (int i = 1; i < how.length; i++) {
		for (headIdx = 0; headIdx < daughterTrees.length; headIdx++) {
		    childCat = daughterTrees[headIdx].label().value();
		    if (how[i].equals(childCat)) {
			theHead = daughterTrees[headIdx];
			found = true;
			break twoloop;
		    }
		}
	    }
	    if ( ! found) {
		// none found by tag, so return first or null
		if (deflt) {
		    headIdx = 0;
		    theHead = daughterTrees[headIdx];
		} else {
		    return null;
		}
	    }
	} else if (how[0] == "rightdis") {
	    // from right, but search for any, not in turn
	    twoloop:
	    for (headIdx = daughterTrees.length - 1; headIdx >= 0; headIdx--) {
		childCat = daughterTrees[headIdx].label().value();
		for (int i = 1; i < how.length; i++) {
		    if (Dbg.ONE) {
			Dbg.print("Testing for whether " + how[i] + 
				  " equals " + childCat + ": " + 
				  ((how[i].equals(childCat)) ?
				  "true": "false"));
		    }
		    if (how[i].equals(childCat)) {
			theHead = daughterTrees[headIdx];
			found = true;
			break twoloop;
		    }
		}
	    }
	    if ( ! found) {
		// none found by tag, so return last or null
		if (deflt) {
		    headIdx = daughterTrees.length - 1;
		    theHead = daughterTrees[headIdx];
		} else {
		    return null;
		}
	    }
	} else {
	    // from right
	    twoloop:
	    for (int i = 1; i < how.length; i++) {
		for (headIdx = daughterTrees.length - 1; headIdx >= 0;
								headIdx--) {
		    childCat = daughterTrees[headIdx].label().value();
		    if (how[i].equals(childCat)) {
			theHead = daughterTrees[headIdx];
			found = true;
			break twoloop;
		    }
		}
	    }
	    if ( ! found) {
		// none found by tag, so return last or null
		if (deflt) {
		    headIdx = daughterTrees.length - 1;
		    theHead = daughterTrees[headIdx];
		} else {
		    return null;
		}
	    }
	}
	if (false) {
	    // Collins (1999: 237) describes a special rule that does a fix
	    // for coordination so left of pair is alway head. However, with
	    // the current head rules, this is unnecessary: the !!NP1 rule
	    // only matches tags, so for cases of coordinated phrases, the
	    // !!NP2 rule would always return the leftmost NP, and this code
	    // is never invoked.  I've tested the code with and without this
	    // piece, and the rules are identical!  So I've commented it out.
	    // [Chris Manning, Jun 2002]
	    // Something like this would be needed, and would need to be more
	    // complex if applied it to base level coordinations: it would
	    // need to scan let for a CC.  It's also not clear this is always
	    // a good idea, e.g., (NP Securities and Exchange Commission)
	    if (headIdx >= 2) {
	        String prevLab = daughterTrees[headIdx - 1].label().toString();
	        if (prevLab.equals("CC")) {
	            headIdx -= 2;
		    theHead = daughterTrees[headIdx];
	        }
	    }
        }
	if (Dbg.ONE) {
	    Dbg.print("Head for " + motherKey + " chose index " + headIdx +
	    	      " " + theHead.label()
	    	      /* + "[" + hptr.headWord + "]" */ );
	}
	return theHead;
    }


    private HashMap nonTerminalInfo;

    // Initialization block.
    // "left" means search left-to-right by category and then by position
    // "right" means search right-to-left by category and then by position
    // "rightdis" means search right-to-left by position and then by category
    // In all cases the first thing found is returned.
    {
	nonTerminalInfo = new HashMap();
	// This version from Collins' diss (1999: 236-238)
	// NNS, NN is actually sensible (money, etc.)!
	// QP early isn't; should prefer JJR NN RB
	// remove ADVP; it just shouldn't be there.
	nonTerminalInfo.put("ADJP",
			    new String[] {"left", "NNS", "NN", "$", "QP",
					  "JJ",
					  "VBN", "VBG", "ADJP", "JJR", "NP",
					  "JJS", "DT", "FW", "RBR", "RBS",
					  "SBAR", "RB"});
	nonTerminalInfo.put("ADVP",
			    new String[] {"right", "RB", "RBR", "RBS", "FW",
					  "ADVP",
					  "TO", "CD", "JJR", "JJ", "IN",
					  "NP", "JJS", "NN" });
	nonTerminalInfo.put("CONJP",
			    new String[] {"right", "CC", "RB", "IN"});
	nonTerminalInfo.put("FRAG",
			    new String[] {"right"}); // crap
	nonTerminalInfo.put("INTJ",
			    new String[] {"left"});
	nonTerminalInfo.put("LST",
			    new String[] {"right", "LS", ":"});
	nonTerminalInfo.put("NAC",
			    new String[] {"left", "NN", "NNS", "NNP", "NNPS",
					  "NP", "NAC", "EX", "$", "CD", "QP",
					  "PRP", "VBG", "JJ", "JJS", "JJR",
					  "ADJP", "FW"});
	nonTerminalInfo.put("NX",
			    new String[] {"right"}); // crap
	nonTerminalInfo.put("PP",
			    new String[] {"right", "IN", "TO", "VBG", "VBN",
					      "RP", "FW"});
	// should prefer JJ? (PP (JJ such) (IN as) (NP (NN crocidolite)))
	nonTerminalInfo.put("PRN",
			    new String[] {"left", "VP", "NP", "PP", "S", 
					      "SINV", "SBAR", "ADJP", "ADVP",
					      "INTJ", "WHNP", "NAC",
					      "VBP", "JJ", "NN", "NNP" });
	nonTerminalInfo.put("PRT",
			    new String[] {"right", "RP"});
	nonTerminalInfo.put("QP",
			    new String[] {"left", "$", "IN", "NNS", "NN",
					      "JJ", "RB", "DT", "CD", "NCD",
					      "QP", "JJR", "JJS"});
	nonTerminalInfo.put("RRC",
			    new String[] {"right", "VP", "NP", "ADVP", "ADJP",
					      "PP"});
	// delete IN -- go for main part of sentence; add FRAG
	nonTerminalInfo.put("S",
			    new String[] {"left", "TO", "VP", "S", "FRAG",
					      "SBAR", "ADJP", "UCP", "NP"});
	nonTerminalInfo.put("SBAR",
			    new String[] {"left", "WHNP", "WHPP", "WHADVP",
					      "WHADJP", "IN", "DT", "S",
					      "SQ", "SINV", "SBAR", "FRAG"});
	nonTerminalInfo.put("SBARQ",
			    new String[] {"left", "SQ", "S", "SINV", "SBARQ",
					      "FRAG"});
	nonTerminalInfo.put("SINV",
			    new String[] {"left", "VBZ", "VBD", "VBP",
					      "VB", "MD", "VP", "S", "SINV",
					      "ADJP", "NP"});
	nonTerminalInfo.put("SQ",
			    new String[] {"left", "VBZ", "VBD", "VBP",
					      "VB", "MD", "VP", "SQ"});
	nonTerminalInfo.put("UCP",
			    new String[] {"right"});
	nonTerminalInfo.put("VP",
			    new String[] {"left", "TO", "VBD", "VBN", "MD",
					      "VBZ", "VB", "VBG", "VBP",
					      "VP", "ADJP", "NN",
					      "NNS", "NP"});
	nonTerminalInfo.put("WHADJP",
			    new String[] {"left", "CC", "WRB", "JJ", "ADJP"});
	nonTerminalInfo.put("WHADVP",
			    new String[] {"right", "CC", "WRB"});
	nonTerminalInfo.put("WHNP",
			    new String[] {"left", "WDT", "WP", "WP$",
					      "WHADJP", "WHPP", "WHNP"});
	nonTerminalInfo.put("WHPP",
			    new String[] {"right", "IN", "TO", "FW"});
	nonTerminalInfo.put("X",
			    new String[] {"right"}); // crap rule
	nonTerminalInfo.put("!!NP1",
			    new String[] {"rightdis", "NN", "NNP", "NNPS",
					      "NNS", "NX", "POS", "JJR"});
	nonTerminalInfo.put("!!NP2",
			    new String[] {"left", "NP"});
	nonTerminalInfo.put("!!NP3",
			    new String[] {"rightdis", "$", "ADJP", "PRN"});
	nonTerminalInfo.put("!!NP4",
			    new String[] {"right", "CD"});
	nonTerminalInfo.put("!!NP5",
			   new String[] {"rightdis", "JJ", "JJS", "RB", "QP"});
    }


    /** Go through trees and determine their heads and print them.
     *  Just for debuggin'. <br>
     *  Usage: <code>
     *  java edu.stanford.nlp.trees.ModCollinsHeadFinder treebankPath/wsj/07
     *  </code>
     *  @param args Command line arguments
     */
    public static void main(String[] args) {
	Treebank treebank = new DiskTreebank(new TreeReaderFactory() {
	    public TreeReader newTreeReader(Reader in) {
	        return new PennTreeReader(in,
			  new LabeledScoredTreeFactory(
			      new CategoryWordTagFactory()),
			  new BobChrisTreeNormalizer());
	    }
	});
	treebank.loadPath(args[0]);
	final HeadFinder chf = new ModCollinsHeadFinder();
	treebank.apply(new TreeProcessor() {
	    public void processTree(Tree pt) {
	        pt.percolateHeads(chf);
		pt.pennPrint();
		System.out.println();
	    }
	});
    }

}
