package edu.stanford.nlp.trees;

/** Tree normalizer for Negra.  Just like BobChrisTreeNormalizer, but
 * empty elements are differently marked.
 */
public class NegraTreeNormalizer extends TreeNormalizer {

    protected final TreebankLanguagePack tlp;


    public NegraTreeNormalizer() {
        this(new PennTreebankLanguagePack());
    }

    public NegraTreeNormalizer(TreebankLanguagePack tlp) {
        this.tlp = tlp;
    }


    /** Normalizes a leaf contents.
     *  This implementation interns the leaf.
     */
    public String normalizeTerminal(String leaf) {
        // We could unquote * and / with backslash \ in front of them
        return leaf.intern();
    }


    /** Normalizes a nonterminal contents.
     *  This implementation strips functional tags, etc. and interns the
     *  nonterminal.
     */
    public String normalizeNonterminal(String category) {
        return cleanUpLabel(category).intern();
    }


    /** Normalize a whole tree -- one can assume that this is the
     *  root.  This implementation deletes empty elements (ones with
     *  nonterminal tag label starting with '*T') from the tree.  It
     *  does work for a null tree.
     */
    public Tree normalizeWholeTree(Tree tree, TreeFactory tf) {
        return tree.prune(new edu.stanford.nlp.util.Filter() {
            public boolean accept(Object obj) {
                Tree t = (Tree) obj;
                Tree[] kids = t.children();
                Label l = t.label();
                if ((l != null) &&
                    l.value() != null &&
                    (l.value().matches("^\\*T.*$")) &&
                    ! t.isLeaf() &&
                    kids.length == 1 && kids[0].isLeaf()) {
                    // Delete empty/trace nodes (ones marked '-NONE-')
                    return false;
                }
                return true;
            }
        }, tf).spliceOut(new edu.stanford.nlp.util.Filter() {
            public boolean accept(Object obj) {
              Tree t = (Tree) obj;
              if (t.isLeaf() ||
                  t.isPreTerminal() ||
                  t.children().length != 1)
                return true;
              if (t.label() != null &&
                  t.label().equals(t.children()[0].label()))
                return false;
              return true;
            }
          }, tf);
    }


    /** Remove things like hyphened functional tags and equals from the
     *  end of a node label.
     */
    protected String cleanUpLabel(String label) {
        if (label == null) {
            label = "ROOT";
            // String constants are always interned
        } else {
            label = tlp.basicCategory(label);
        }
        return label;
    }

}
