package edu.stanford.nlp.trees;

import java.io.*;
import java.util.*;
import edu.stanford.nlp.io.*;

/** A <code>PennTreeReader</code> is a <code>TreeReader</code> that
 *  reads in Penn Treebank-style files.  Example usage:
 *  <br>
 *  <code> TreeReader tr = new PennTreeReader(new BufferedReader(new
 *  InputStreamReader(new FileInputStream(file),"UTF-8")), *
 *  myTreeFactory); </code>
 *
 *  @author Christopher Manning (mod. Roger Levy 2003/01)
 *  @version   2000/01/14
 */
public class PennTreeReader implements TreeReader{

    private Reader           in;
    private edu.stanford.nlp.io.StreamTokenizer  st;
    private TreeNormalizer   tn;
    private TreeFactory      tf;


    /** Read parse trees from a <code>Reader</code>. 
     *  For the defaulted arguments, you get a 
     *  <code>SimpleTreeFactory</code>, no <code>TreeNormalizer</code>, and
     *  a <code>PennTreebankStreamTokenizer</code>.
     *  @param in The <code>Reader</code>  
     */
    public PennTreeReader(Reader in) {
	this(in, new SimpleTreeFactory(), null,
	     new PennTreebankStreamTokenizer(in));
    }


    /** Read parse trees from a <code>Reader</code>.
     * @param in the Reader
     * @param tf TreeFactory -- factory to create some kind of Tree
     */
    public PennTreeReader(Reader in, TreeFactory tf) {
	this(in, tf, null, new PennTreebankStreamTokenizer(in));
    }

    /** Read parse trees from a <code>Reader</code>.
     * @param in the Reader
     * @param st the StreamTokenizer
     */
    public PennTreeReader(Reader in, edu.stanford.nlp.io.StreamTokenizer st) {
	this(in, new SimpleTreeFactory(), null, st);
    }


    /** Read parse trees from a Reader.
     *  @param in Reader
     *  @param tf TreeFactory -- factory to create some kind of Tree
     *  @param tn the method of normalizing trees
     */
    public PennTreeReader(Reader in, TreeFactory tf, TreeNormalizer tn) {
	this(in, tf, tn, new PennTreebankStreamTokenizer(in));
    }


    /** Read parse trees from a Reader.
     *  @param in Reader
     *  @param tf TreeFactory -- factory to create some kind of Tree
     *  @param tn the method of normalizing trees
     *  @param st edu.stanford.nlp.io.StreamTokenizer that divides up InputStream
     */
    public PennTreeReader(Reader in, TreeFactory tf, TreeNormalizer tn,
		      edu.stanford.nlp.io.StreamTokenizer st) {
	this.in = in;
	this.tf = tf;
	this.tn = tn;
	this.st = st;
    }


    /** Reads a single tree in standard Penn Treebank format,
     *  with or without an additional set of parens around it (an unnamed
     *  ROOT node). 
     *  @return A single tree, or <code>null</code> at end of file.
     */
    public Tree readTree() throws IOException {
	Tree tr = readTree(st.next());
	if (tr == null || tn == null) {
	    return tr;
	} else {
	    return tn.normalizeWholeTree(tr, tf);
	}
    }


    private Tree readTree(String token) throws IOException {
	String name;
	// a paren starts new tree, a string is a leaf symbol,
	// o.w. IO exception
	if (token == null) {
	    return null;
	} else if (token.equals("(")) {
	    //System.out.println("Open braces...");
	    // gets next token if there is one
	    name = st.next();
	    //System.out.println(name);
	    // checks if it's a normal string and returns it as the label
	    if (name.equals("(") || name.equals(")")) {
		st.pushBack();
		name = null;
	    }
	    if (tn != null) {
		name = tn.normalizeNonterminal(name); // we used to .intern();
	    }
	    return (tf.newTreeNode(name, readTrees()));
	} else if (! (token.equals(null))) {
	    if (tn != null) {
		name = tn.normalizeTerminal(token); // was: .intern();
	    }
	    else {
		name = token;
	    }
	    return tf.newLeaf(name);
	} else {
	    throw(new IOException("expecting left paren or terminal to begin tree, found: " + token));
	}
    }


    /** Parse sequence of trees, followed by a single right paren.
     */
    private List readTrees() throws IOException {
	// allocate array list for temporarily storing trees
	ArrayList parseTrees = new ArrayList();
	// until a paren closes all subtrees, keep reading trees
	String nextToken = st.next();
	while (nextToken != null && ! nextToken.equals(")")) {
	    parseTrees.add(readTree(nextToken));
	    nextToken = st.next();
	}
	if (nextToken == null) {
	    throw(new IOException("Expecting right paren found eof"));
	}
	return parseTrees;
    }


    /** Close the Reader behind this <code>TreeReader</code>.
     */
    public void close() throws IOException {
	in.close();
    }


    /** Loads treebank data from first argument and prints it.
     * @param args Array of command-line arguments: specifies a filename
     */
    public static void main(String[] args) {
	try {
	    Tree t;
	    TreeReader tr = new PennTreeReader(new BufferedReader(new
				    InputStreamReader(new FileInputStream(args[0]),"UTF-8")));
	    t = tr.readTree();
	    while (t != null) {
		System.out.println(t + "\n");
		t = tr.readTree();
	    } 
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	}
    }

}
