package edu.stanford.nlp.trees;

import java.io.*;
import java.util.*;

import edu.stanford.nlp.io.ExtensionFileFilter;

/**
 * A <code>Sentencebank</code> object provides access to a corpus of 
 * sentences -- just plain sentences or tagged sentences, etc.
 * This class implements the Collection interface, but in a read-only way.
 * A typical way of creating and using a Sentencebank looks like:
 * <p><blockquote><pre>
 * SentenceReaderFactory srf = new AdwaitSentenceReaderFactory(true);
 * Sentencebank sb1 = new DiskSentencebank(srf);
 * sb1.loadPath(inputFile);

 * for (Iterator i1 = sb1.iterator(); i1.hasNext(); ) {
 *     Sentence s1 = (Sentence) i1.next();
 *     System.out.println(s1);
 * }
 * </pre></blockquote></p>
 *
 * @author Christopher Manning
 */
public abstract class Sentencebank extends AbstractCollection {

    /** Stores the <code>SentenceReaderFactory</code> that will be used to
     *  create a <code>SentenceReader</code> to process a file of sentences.
     */
    private SentenceReaderFactory srf;

    /** Create a new Sentencebank.
     */
    public Sentencebank() {
	this(new SimpleSentenceReaderFactory());
    }


    /** Create a new Sentencebank.
     *	@param srf the factory class to be called to create a new 
     *      <code>SentenceReader</code>
     */
    public Sentencebank(SentenceReaderFactory srf) {
	this.srf = srf;
    }


    /** Create a new Sentencebank.
     *	@param initialCapacity The initial size of the underlying Collection,
     *  (if a Collection-based storage mechanism is being provided)
     */
    public Sentencebank(int initialCapacity) {
	this(initialCapacity, new SimpleSentenceReaderFactory());
    }


    /** Create a new Sentencebank.
     *	@param initialCapacity The initial size of the underlying Collection,
     *  (if a Collection-based storage mechanism is being provided)
     *	@param trf the factory class to be called to create a new 
     *      <code>SentenceReader</code>
     */
    public Sentencebank(int initialCapacity, SentenceReaderFactory srf) {
	this.srf = srf;
    }


    /** Get the <code>SentenceReaderFactory</code> for a 
     *  <code>Sentencebank</code> --  this method is provided in order to
     *  make the 
     *  <code>SentenceReaderFactory</code> available to subclasses.
     */
    protected SentenceReaderFactory sentenceReaderFactory() {
	return srf;
    }


    /** Empty a <code>Sentencebank</code>.
     */
    public abstract void clear();


    /** Load a sequence of sentences from the given directory and its
     *  subdirectories.
     *  Sentences should reside in files with the suffix ".pos".
     *  Or: load a single file with the given pathName (including extension)
     *  @param pathName file or directory name
     */
    public void loadPath(String pathName) {
	loadPath(new File(pathName));
    }


    /** Load a sequence of sentences from given directory and its
     *  subdirectories.
     *  Sentences should reside in files with the suffix ".pos".
     *  @param path File specification
     */
    public void loadPath(File path) {
	loadPath(path, "pos", true);
    }


    /** Load sentences from given directory.
     *  @param path  file or directory to load from
     *  @param suffix suffix of files to load
     *  @param recursively descend into subdirectories as well
     */
    public void loadPath(File path, String suffix, boolean recursively) {
	loadPath(path, new ExtensionFileFilter(suffix, recursively));
    }


    /** Load sentences from given path specification.
     *  @param path  file or directory to load from
     *  @param filter a FilenameFilter of files to load
     */
    public abstract void loadPath(File path, FileFilter filt);


    /** Apply a SentenceVisitor to each sentence in the Sentencebank.
     *  @param tp The SentenceVisitor to be applied
     */
    public abstract void apply(SentenceVisitor tp);


    /** Return an Iterator over Sentences in the Sentencebank.
     */
    public abstract Iterator iterator();


    /** Return the whole Sentencebank as a series of big bracketed lists.
     *  Calling this is a really bad idea if your Sentencebank is large.
     */
    public String toString() {
        final StringBuffer sb = new StringBuffer();
	apply(new SentenceVisitor() {
		  public void visitSentence(Sentence t) {
		      sb.append(t.toString());
		      sb.append("\n");
		  }
	      });
	return sb.toString();
    }


    private final class CounterSentenceVisitor implements SentenceVisitor {
	int i; // = 0;

	public void visitSentence(Sentence t) {
	    i++;
	}

	public int total() {
	    return i;
	}
    }

    /** Returns the size of the Sentencebank.
     */
    public int size() {
	CounterSentenceVisitor counter = new CounterSentenceVisitor();
	apply(counter);
	return counter.total();
    }


    /** This operation isn't supported for a Sentencebank.
     *  Tell them immediately.
     *  @param o The object that would be removed
     */
    public boolean remove(Object o) {
	throw new UnsupportedOperationException("Sentencebank is read-only");
    }

}
