package edu.stanford.nlp.trees;

import java.util.*;

/** A <code>SimpleTree</code> is a minimal concrete implementation of an
 *  unlabeled, unscored <code>Tree</code>.  It has a tree structure, but no
 *  annotation at nodes.
 *
 *  @author Christopher Manning
 */
public class SimpleTree extends Tree {

    /** Daughters of the parse tree.
     */
    private Tree[] daughterTrees;

    /** Create an empty parse tree.
     */
    public SimpleTree() {
    }

    /** Create parse tree with given root and null daughters.
     *  @param label root label of new tree to construct.  For a SimpleTree
     *          this parameter is ignored.
     */
    public SimpleTree(Label label) {
	this();
    }

    /** Create parse tree with given root and array of daughter trees.
     *  @param label root label of tree to construct.  For a SimpleTree
     *         this parameter is ignored
     *  @param daughterTreesList list of daughter trees to construct.
     */
    public SimpleTree(Label label, List daughterTreesList) {
	setChildren(daughterTreesList);
    }


    /** Returns an array of children for the current node, or null
     *  if it is a leaf.
     */
    public Tree[] children() {
	return daughterTrees;
    }


    public void setChildren(Tree[] children) {
	daughterTrees = children;
    }



    // extra class guarantees correct lazy loading (Bloch p.194)
    private static class TreeFactoryHolder {
	static final TreeFactory tf = new SimpleTreeFactory();
    }


    /** Return a <code>TreeFactory</code> that produces trees of the
     *  <code>SimpleTree</code> type.
     *  The factory returned is always the same one (a singleton).
     *  @return a factory to produce simple (unlabelled) trees
     */
    public TreeFactory treeFactory() {
	return TreeFactoryHolder.tf;
    }


    /** Return a <code>TreeFactory</code> that produces trees of the
     *  <code>SimpleTree</code> type.
     *  The factory returned is always the same one (a singleton).
     *  @return a factory to produce simple (unlabelled) trees
     */
    public static TreeFactory factory() {
	return TreeFactoryHolder.tf;
    }

}
