package edu.stanford.nlp.trees;

import java.util.*;
import java.io.*;

import edu.stanford.nlp.util.Filter;
import edu.stanford.nlp.util.Filters;
import edu.stanford.nlp.util.Scored;
import edu.stanford.nlp.util.Pair;

/** The abstract class <code>Tree</code> is used to collect
 *  all of the tree types, and acts as a generic composite type.  This
 *  is the standard implementation of inheritance-based polymorphism.
 *  Trees, like standard containers, are not parametrized in any way by
 *  the type of object in the <code>data</code> field.
 *  All <code>Tree</code> objects support accessors for their children
 *  (a <code>Tree[]</code>), their data (a <code>Label</code>), and their
 *  score (a <code>double</code>).  However, different concrete
 *  implementations may or may not include the latter two, in which case
 *  a default value is returned.  The class Tree defines no data fields.
 *  The three abstract methods that must be implemented are:
 *  <code>children()</code>, <code>setChildren(Tree[])</code>, and
 *  <code>treeFactory()</code>.
 *  There is now support for finding the parent of a tree.  This may be
 *  done by search from a tree root, or via a directly stored parent.
 *  The <code>Tree</code> class now implements the <code>Collection</code>
 *  interface: in terms of this, each <i>node</i> of the tree is an
 *  element of the collection; hence one can explore the tree by using
 *  the methods of this interface.  A <code>Tree</code> is regarded as a
 *  read-only <code>Collection</code> (even though the <code>Tree</code>
 *  class has various methods that modify trees).  Moreover, the
 *  implementation is <i>not</i> thread-safe: no attempt is made to detect
 *  and report concurrent modifications.
 *
 *  @author Christopher Manning
 *  @author Dan Klein
 */
public abstract class Tree extends AbstractCollection
					implements Labeled, Scored {

    /** Says whether a node is a leaf.  Can be
     *  used on an arbitrary <code>Tree</code>.  Being a leaf is defined as
     *  having no children.  This is the preferred alternative
     *  to running meta checks on types, as it works independent of Tree
     *  implementation.  
     *  @return true if this object is a leaf
     */
    public boolean isLeaf() {
	Tree[] kids = children();
	return (kids == null) || (kids.length == 0);
    }


    /** Return whether this node is a preterminal or not.  A preterminal is
     *  defined to be a node with one child which is itself a leaf.
     *  @return true if the node is a preterminal; false otherwise
     */
    public boolean isPreTerminal() {
	Tree[] kids = children();
	return (kids != null) && (kids.length == 1) && (kids[0].isLeaf());
    }


    /** Return whether all the children of this node are preterminals or not.
     *  A preterminal is
     *  defined to be a node with one child which is itself a leaf.
     *  Considered false if the node has no children
     *  @return true if the node is a prepreterminal; false otherwise
     */
    public boolean isPrePreTerminal() {
	Tree[] kids = children();
	if (kids == null || kids.length == 0) {
	    return false;
	} else {
	    boolean allPreTerminals = true;
	    for (int j = 0; j < kids.length; j++) {
		if ( ! kids[j].isPreTerminal()) {
		    return false;
		}
	    }
	    return true;
	}
    }


    /** Return whether this node is a phrasal node or not.  A phrasal node 
     *  is defined to be a node which is not a leaf or a preterminal.
     *  Worded positively, this means that it must have two or more children,
     *  or one child that is not a leaf.
     *  @return <code>true</code> if the node is phrasal; 
     *        <code>false</code> otherwise
     */
    public boolean isPhrasal() {
	Tree[] kids = children();
	return ! (kids == null || kids.length == 0 || 
		  (kids.length == 1 && kids[0].isLeaf()));
    }


    /** Returns an array of children for the current node.  If there are
     *  no children (if the node is a leaf), this may return <code>null</code>
     *  or an array of length 0.  Other code in this class will work
     *  correctly for both these cases.  A caller may assume that either
     *  <code>isLeaf()</code> returns true, or this node has a nonzero number
     *  of children.  So the following idiom is frequently useful:
     *  <pre>
     *    if ( ! tr.isLeaf()) {
     *        Tree[] kids = children();
     *        for (int i = 0; i < kids.length; i++) {
     *            doSomethingTo(kids[i]);
     *        }
     *    }
     *  </pre>
     *  @return The children of the node
     *  @see #getChildrenAsList()
     */
    public abstract Tree[] children();


    /** Returns an array of children for the current node.  If there are no
     *  children, then a (non-null) <code>List</code> of size 0 will be
     *  returned.
     *  @return The children of the node
     */
    public List getChildrenAsList() {
	Tree[] kids = children();
	int length;
	if (kids == null) {
	    length = 0;
	} else {
	    length = kids.length;
	}
	ArrayList kidsList = new ArrayList(length);
	for (int i = 0; i < length; i++) {
	    kidsList.add(kids[i]);
	}
	return kidsList;
    }


    /** Set the children of this node to be the children given in
     *  the array.  The children array may be either an array of
     *  length 0 or <code>null</code> if there are no children,
     *  and other methods in this class work on either representation.
     *  @param children The array of children, each a <code>Tree</code>
     *  @see #setChildren(List)
     */
    public abstract void setChildren(Tree[] children);


    /** Set the children of this tree node to the given list.  This method
     *  is implemented in the <code>Tree</code> class by converting the
     *  <code>List</code> into a tree array and calling the array-based
     *  method.  Subclasses which use a <code>List</code>-based representation
     *  of tree children should override this method.
     *  This implementation allows the case that the <code>List</code> is
     *  <code>null</code>: it yields a node with no children.
     *  @param childTreesList list of trees to become children
     *  @see #setChildren(Tree[])
     */
    public void setChildren(List childTreesList) {
	if (childTreesList != null) {
	    int leng = childTreesList.size();
	    Tree[] childTrees = new Tree[leng];
	    childTreesList.toArray(childTrees);
	    setChildren(childTrees);
	} else {
	    setChildren((Tree[]) null);
	}
    }


    /** Returns the label associated with the current node, or null
     *  if there is no label.  The default implementation always
     *  returns <code>null</code>.
     *  @return The label of the node
     */
    public Label label() {
	return null;
    }


    /** Sets the label associated with the current node, if there is one.
     *  @param label The label
     */
    public void setLabel(Label label) {
	// a noop
    }


    /** Returns the score associated with the current node, or NaN
     *  if there is no score.  The default implementation returns NaN.
     *  @return The score
     */
    public double score() {
	return Double.NaN;
    }


    /** Sets the score associated with the current node, if there is one.
     *  @param score The score
     */
    public void setScore(double score) {
	// a no-op
    }


    /** Returns the first child of a tree, or <code>null</code> if none.
     *  @return The first child
     */
    public Tree firstChild() {
	Tree[] kids = children();
	if (kids == null || kids.length == 0) {
	    return null;
	} else {
	    return kids[0];
	}
    }


    /** Returns the last child of a tree, or <code>null</code> if none.
     *  @return The last child
     */
    public Tree lastChild() {
	Tree[] kids = children();
	if (kids == null || kids.length == 0) {
	    return null;
	} else {
	    return kids[kids.length - 1];
	}
    }


    /** Returns the Constituents generated by the parse tree.
     *  @return a Set of the constituents as constituents of
     *  type <code>SimpleConstituent</code>
     */
    public Set constituents() {
	return constituents(new SimpleConstituentFactory());
    }


    /** Returns the Constituents generated by the parse tree.
     *  The Constituents of a sentence include the preterminal categories
     *  but not the leaves.
     *  @param cf ConstituentFactory used to build the Constituent objects
     *  @return a Set of the constituents as SimpleConstituent type
     *        (in the current implementation, a <code>HashSet</code>
     */
    public Set constituents(ConstituentFactory cf) {
	Set constituentsSet = new HashSet();
	constituents(constituentsSet, 0, cf);
	return constituentsSet;
    }


    /** Adds the constituents derived from <code>this</code> tree to
     *  the ordered <code>Constituent</code> <code>Set</code>, beginning
     *  numbering from the second argument and returning the number of
     *  the right edge.  The reason for the return of the right frontier
     *  is in order to produce bracketings recursively by threading through
     *  the daughters of a given tree.
     *  @param constituentSet set of constituents to add results of bracketing
     *     this tree to
     *  @param left left position to begin labeling the bracketings with
     *  @param cf ConstituentFactory used to build the Constituent objects
     *  @return Index of right frontier of Constituent
     */
    private int constituents(Set constituentsSet, int left,
			     ConstituentFactory cf) {
	if (isLeaf()) {
	    // lexical leaves do not add any Constituents
	    // but increment position
	    // System.err.println("In bracketing trees leaf is " + label());
	    return (left+1);
	} else {
	    int position = left;

	    // System.err.println("In bracketing trees left is " + left);
	    // System.err.println("  label is " + label() +
	    //		       "; num daughters: " + children().length);
	    // enumerate through daughter trees
	    Tree[] kids = children();
	    for (int i = 0; i < kids.length; i++) {
		// compute bracketings for daughter tree
		// update position to end of daughter tree
		position = kids[i].constituents(constituentsSet, position, cf);
		// System.err.println("  position went to " + position);
	    }
	    // need to wait until result position is known in order to
	    // calculate span of whole tree
	    constituentsSet.add(cf.newConstituent(left, position, label(),
						  score()));
	    // System.err.println("  added " + label());
	    return position;
	}
    }


    /** Most instances of <code>Tree</code> will take a lot more than
     *  than the default <code>StringBuffer</code> size of 16 to print
     *  as an indented list of the whole tree, so we enlarge the default.
     */
    private static final int initialPrintStringBufferSize = 500;


    /** Appends the printed form of a parse tree (as a bracketed String)
     *  to a <code>StringBuffer</code>.
     *  The implementation of this may be more efficient than for
     *  <code>toString()</code> on complex trees.
     *  @param sb The StringBuffer  to append to
     *  @return StringBuffer returns the <code>StringBuffer</code>
     */
    public StringBuffer toStringBuffer(StringBuffer sb) {
	sb.append("(");
	if (label() != null) {
	    sb.append(label());
	}
	Tree[] kids = children();
	if (kids != null) {
	    for (int i = 0; i < kids.length; i++) {
		sb.append(" ");
		kids[i].toStringBuffer(sb);
	    }
	}
	return sb.append(")");
    }


    /** Converts parse tree to string in Penn Treebank format.
     *  Get efficiency by chaining a single <code>StringBuffer</code>
     *  through it all.
     *  @return the tree as a bracketed list on one line
     */
    public String toString() {
	return toStringBuffer(new
		StringBuffer(Tree.initialPrintStringBufferSize)).toString();
    }


    private static final int indentIncr = 2;


    private String makeIndentString(int indent) {
	StringBuffer sb = new StringBuffer(indent);
        for (int i = 0; i < indentIncr; i++) {
            sb.append(" ");
        }
	return sb.toString();
    }


    /** Indented list printing of a tree.  The tree is printed in an
     *  indented list notation, with nodel labels followed by node scores.
     */
    public void indentedListPrint() {
        indentedListPrint(new PrintWriter(System.out, true));
    }


    /** Indented list printing of a tree.  The tree is printed in an
     *  indented list notation, with nodel labels followed by node scores.
     *  @param pw The PrintWriter to print the tree to
     */
    public void indentedListPrint(PrintWriter pw) {
        indentedListPrint("", makeIndentString(indentIncr), pw);
    }


    /** Indented list printing of a tree.  The tree is printed in an
     *  indented list notation, with nodel labels followed by node scores.
     *  String parameters are used rather than integer levels for efficiency.
     *  @param indent The base <code>String</code> (normally just spaces)
     *     to print before each line of tree
     *  @param pad The additional <code>String</code> (normally just more
     *     spaces) to add when going to a deeper level of <code>Tree</code>.
     *  @param pw The PrintWriter to print the tree to
     */
    public void indentedListPrint(String indent, String pad, PrintWriter pw) {
	StringBuffer sb = new StringBuffer(indent);
	Label label = label();
	if (label != null) {
	    sb.append(label.toString());
	}
	sb.append("  ");
	sb.append(score());
	pw.println(sb.toString());
	Tree[] children = children();
	if (children != null) {
	    String newIndent = indent + pad;
	    for (int i = 0; i < children.length; i++) {
		children[i].indentedListPrint(newIndent, pad, pw);
	    }
	}
    }


    private static void displayChildren(Tree[] trChildren, int indent,
					boolean parentLabelNull,
					PrintWriter pw) {
	boolean firstSibling = true;
	boolean leftSibIsPreTerm = true;  // counts as true at beginning
	for (int i = 0; i < trChildren.length; i++) {
            Tree currentTree = trChildren[i];
	    currentTree.display(indent, parentLabelNull, firstSibling,
				leftSibIsPreTerm, false, pw);
	    leftSibIsPreTerm = trChildren[i].isPreTerminal();
	    // CC is a special case
	    if (trChildren[i].label() != null &&
		trChildren[i].label().toString().startsWith("CC")) {
		leftSibIsPreTerm = false;
	    }
	    firstSibling = false;
        }
    }


    /** Display a node, implementing Penn Treebank style layout
     */
    private void display(int indent, boolean parentLabelNull,
			 boolean firstSibling,
			 boolean leftSiblingPreTerminal,
			 boolean topLevel,
			 PrintWriter pw) {
	// the condition for staying on the same line in Penn Treebank
	boolean suppressIndent = (parentLabelNull ||
				  (firstSibling && isPreTerminal()) ||
				  (leftSiblingPreTerminal && isPreTerminal() &&
				   (label() == null ||
				    ! label().toString().startsWith("CC"))));
	if (suppressIndent) {
	    pw.print(" ");
	    pw.flush();
	} else {
	    if (! topLevel) {
		pw.println();
	    }
	    for (int i = 0; i < indent; i++) {
		pw.print("  ");
		pw.flush();
	    }
	}
        if (isLeaf() || isPreTerminal()) {
	    pw.print(toString());
	    pw.flush();
            return;
        }
        pw.print("(");
        if (label() != null && label().toString() != null) {
            pw.print(label().toString());
	    pw.flush();
	}
	displayChildren(children(), indent + 1,
			label() == null || label().toString() == null, pw);
        pw.print(")");
	pw.flush();
    }


    /** Print the tree as done in Penn Treebank merged files.
     *  The formatting should be exactly the same, but we don't print the
     *  trailing whitespace found in Penn Treebank trees.
     *  The basic deviation from a bracketed indented tree is to in general
     *  collapse the printing of adjacent preterminals onto one line of
     *  tags and words.  Additional complexities are that conjunctions
     *  (tag CC) are not collapsed in this way, and that the unlabeled
     *  outer brackets are collapsed onto the same line as the next
     *  bracket down.
     *  @param pw The tree is printed to this <code>PrintWriter</code>
     */
    public void pennPrint(PrintWriter pw) {
        display(0, false, false, false, true, pw);
	pw.println();
    }


    /** Print the tree as done in Penn Treebank merged files.
     *  The formatting should be exactly the same, but we don't print the
     *  trailing whitespace found in Penn Treebank trees.
     *  The basic deviation from a bracketed indented tree is to in general
     *  collapse the printing of adjacent preterminals onto one line of
     *  tags and words.  Additional complexities are that conjunctions
     *  (tag CC) are not collapsed in this way, and that the unlabeled
     *  outer brackets are collapsed onto the same line as the next
     *  bracket down.
     *  @param ps The tree is printed to this <code>PrintStream</code>
     */
    public void pennPrint(PrintStream ps) {
      pennPrint(new PrintWriter(new OutputStreamWriter(ps), true));
    }


    /** Print the tree as done in Penn Treebank merged files.
     *  The formatting should be exactly the same, but we don't print the
     *  trailing whitespace found in Penn Treebank trees.
     *  The tree is printed to <code>System.out</code>. The basic deviation
     *  from a bracketed indented tree is to in general
     *  collapse the printing of adjacent preterminals onto one line of
     *  tags and words.  Additional complexities are that conjunctions
     *  (tag CC) are not collapsed in this way, and that the unlabeled
     *  outer brackets are collapsed onto the same line as the next
     *  bracket down.
     */
    public void pennPrint() {
	pennPrint(System.out);
    }


    /** Finds the depth of the tree.  The depth is defined as the length
     *  of the longest path from this node to a leaf node.  Leaf nodes
     *  have depth zero.
     *  @return the depth
     */
    public int depth() {
	if (isLeaf()) {
	    return 0;
	} else {
	    int maxDepth = 0;
	    Tree[] kids = children();
	    for (int i = 0; i < kids.length; i++) {
		int curDepth = kids[i].depth();
		if (curDepth > maxDepth) {
		    maxDepth = curDepth;
		}
	    }
	    return maxDepth+1;
	}
    }


    /** Returns the tree leaf that is the head of the tree.
     *  @param hf The headfinding algorithm to use
     *  @return The head tree leaf if any, else <code>null</code>
     */
    public Tree headTerminal(HeadFinder hf) {
	if (isLeaf()) {
	    return this;
	} else {
	    Tree head = hf.determineHead(this);
	    if (head != null) {
		return head.headTerminal(hf);
	    } else {
		System.err.println("Head is null: " + this);
		return null;
	    }
	}
    }


    /** Returns the preterminal tree that is the head of the tree.
     *  @param hf The headfinding algorithm to use
     *  @return The head preterminal tree, if any, else <code>null</code>
     */
    public Tree headPreTerminal(HeadFinder hf) {
	if (isPreTerminal()) {
	    return this;
	} else {
	    Tree head = hf.determineHead(this);
	    if (head != null) {
		return head.headPreTerminal(hf);
	    } else {
		System.err.println("Head is null: " + this);
		return null;
	    }
	}
    }


    /** Finds the heads of the tree.  This code assumes that the label
     *  does store and return sensible values for the category, word, and tag.
     *  It will be a no-op otherwise.  The tree is modified.  The routine
     *  assumes the Tree has word leaves and tag preterminals, and copies
     *  their category to word and tag respectively, if they have a null
     *  value.
     *  @param hf The headfinding algorithm to use
     */
    public void percolateHeads(HeadFinder hf) {
	Label cwt = label();
	if (isLeaf()) {
	    if (cwt instanceof HasWord) {
		HasWord w = (HasWord) cwt;
		if (w.word() == null) {
		    w.setWord(cwt.value());
		}
	    }
	} else {
	    Tree[] kids = children();
	    if (kids != null) {
		for (int i = 0; i < kids.length; i++) {
		    kids[i].percolateHeads(hf);
		}
	    }
	    Tree head = hf.determineHead(this);
	    if (head != null) {
		Label headCwt = head.label();
		String headTag = null;
		if (headCwt instanceof HasTag) {
		    headTag = ((HasTag) headCwt).tag();
		}
		if (headTag == null && head.isLeaf()) {
		    // below us is a leaf
		    headTag = cwt.value();
		}
		String headWord = null;
		if (headCwt instanceof HasWord) {
		    headWord = ((HasWord) headCwt).word();
		}
		if (headWord == null && head.isLeaf()) {
		    // below us is a leaf
		    // this might be useful despite case for leaf above in
		    // case the leaf label type doesn't support word()
		    headWord = headCwt.value();
		}
		if (cwt instanceof HasWord) {
		    ((HasWord) cwt).setWord(headWord);
		}
		if (cwt instanceof HasTag) {
		    ((HasTag) cwt).setTag(headTag);
		}
	    } else {
		System.err.println("Head is null: " + this);
	    }
	}
    }


    /** Return a set of Word-Word dependencies, represented as 
     *  edu.stanford.nlp.util.Pair objects
     *  (where the first() is the head) for the Tree.  This will only give 
     *  useful results if the internal tree node labels support HasWord and
     *  head percolation has already been done (see percolateHeads()).
     *  @return Set of dependencies (each a Pair)
     */
    public Set dependencies() {
	return dependencies(Filters.acceptFilter());
    }

    /** Return a set of Word-Word dependencies, represented as Pair objects
     *  (where the first() is the head) for the Tree.  This will only give 
     *  useful results if the internal tree node labels support HasWord and
     *  head percolation has already been done (see percolateHeads()).
     *  @param f Dependencies are excluded for which the Pair is not
     *       accepted by the Filter
     *  @return Set of dependencies (each a Pair)
     */
    public Set dependencies(Filter f) {
	return dependencies(f, null);
    }


    /** Return a set of Word-Word dependencies, represented as Pair objects
     *  (where the first() is the head) for the Tree. 
     *  @param hf The HeadFinder to use to identify the head of constituents
     *  @return Set of dependencies (each a <code>Pair</code>)
     */
    public Set dependencies(HeadFinder hf) {
	return dependencies(Filters.acceptFilter(), hf);
    }

    /** Return a set of Word-Word dependencies, represented as Pair objects
     *  (where the first() is the head) for the Tree. 
     *  @param f Dependencies are excluded for which the Pair is not
     *       accepted by the Filter
     *  @param hf The HeadFinder to use to identify the head of constituents.  
     *       If this is <code>null</code>, then nodes are assumed to already
     *       be marked with their heads.
     *  @return Set of dependencies (each a <code>Pair</code>)
     */
    public Set dependencies(Filter f, HeadFinder hf) {
	Set deps = new HashSet();
	for (Iterator nodeI = subTrees().iterator(); nodeI.hasNext(); ) {
	    Tree node = (Tree) nodeI.next();
	    if (node.isLeaf() || node.children().length < 2) {
		continue;
	    }
	    // every child with a different head (or repated) is an argument
	    Label l = node.label();
	    Word w = null;
	    if (hf != null) {
		Tree hwt = node.headTerminal(hf);
		if (hwt != null) {
		    w = new Word(hwt.label());
		}
	    } else {
		if (l instanceof HasWord) {
		    w = new Word(((HasWord) l).word());
		}
	    }
	    Tree[] kids = node.children();
	    boolean seenHead = false;
	    for (int cNum = 0; cNum < kids.length; cNum++) {
		Tree child = kids[cNum];
		Label dl = child.label();
		Word dw = null;
		if (hf != null) {
		    Tree dwt = child.headTerminal(hf);
		    if (dwt != null) {
			dw = new Word(dwt.label());
		    }
		} else {
		    if (dl instanceof HasWord) {
			dw = new Word(((HasWord) dl).word());
		    }
		}
		// System.out.println("XX Doing pair: " + l + ", " + dl +
		//		   " gives " + w + ", " +dw);
		if (w != null && w.word() != null && dw != null && 
		    w.word().equals(dw.word()) && ! seenHead) {
		    seenHead = true;
		} else {
		    Pair p = new Pair(w, dw);
		    if (f.accept(p)) {
			deps.add(p);
		    }
		}
	    }
	}
	return deps;
    }


    /** Return a Set of TaggedWord-TaggedWord dependencies, represented as 
     *  edu.stanford.nlp.util.Pair objects
     *  (where the first() is the head) for the Tree.  This will only give 
     *  useful results if the internal tree node labels support HasWord and
     *  HasTag, and head percolation has already been done (see
     *  percolateHeads()).
     *  @return Set of dependencies (each a Pair)
     */
    public Set taggedDependencies() {
	return taggedDependencies(Filters.acceptFilter());
    }

    /** Return a set of TaggedWord-TaggedWord dependencies, represented as
     *  Pair objects
     *  (where the first() is the head) for the Tree.  This will only give 
     *  useful results if the internal tree node labels support HasWord and
     *  head percolation has already been done (see percolateHeads()).
     *  <p><i>Implementation note:</i> It would be nice to generalize this with
     *  dependencies() to use a LabelFactory, but it seems impossible with the
     *  current setup. 
     *  @param f Dependencies are excluded for which the Pair is not
     *       accepted by the Filter
     *  @return Set of dependencies (each a Pair)
     */
    public Set taggedDependencies(Filter f) {
	Set deps = new HashSet();
	for (Iterator nodeI = subTrees().iterator(); nodeI.hasNext(); ) {
	    Tree node = (Tree) nodeI.next();
	    if (node.isLeaf() || node.children().length < 2) {
		continue;
	    }
	    // every child with a different head (or repated) is an argument
	    Label l = node.label();
	    TaggedWord tw = null;
	    if (l instanceof HasWord && l instanceof HasTag) {
		tw = new TaggedWord(((HasWord) l).word(), 
				    ((HasTag) l).tag());
	    }
	    Tree[] kids = node.children();
	    boolean seenHead = false;
	    for (int cNum = 0; cNum < kids.length; cNum++) {
		Tree child = kids[cNum];
		Label dl = child.label();
		TaggedWord dtw = null;
		if (dl instanceof HasWord && dl instanceof HasTag) {
		    dtw = new TaggedWord(((HasWord) dl).word(), 
				    ((HasTag) dl).tag());
		}
		if (tw != null && tw.word() != null && dtw != null && 
		    tw.word().equals(dtw.word()) && 
		    tw.tag().equals(dtw.tag()) && ! seenHead) {
		    seenHead = true;
		} else {
		    Pair p = new Pair(tw, dtw);
		    if (f.accept(p)) {
			deps.add(p);
		    }
		}
	    }
	}
	return deps;
    }


    /** Gets the yield of the tree.  The <code>Label</code> of all leaf nodes
     *  is returned
     *  as a list ordered by the natural left to right order of the
     *  leaves.  Null values, if any, are inserted into the list like any
     *  other value.
     *
     *  @return a <code>List</code> of the data in the tree's leaves.
     */
    public Sentence yield() {
	return (Sentence) yield(new Sentence());
    }

    /** Gets the yield of the tree.  The <code>Label</code> of all leaf nodes
     *  is returned
     *  as a list ordered by the natural left to right order of the
     *  leaves.  Null values, if any, are inserted into the list like any
     *  other value.  This has been rewritten to thread, so only one List
     *  is used.
     *  @param y The list in which the yield of the tree will be placed.
     *      Normally, this will be empty when the routine is called, but
     *      if not, the new yield is added to the end of the list.
     *  @return a <code>List</code> of the data in the tree's leaves.
     */
    public List yield(List y) {
	if (isLeaf()) {
	    y.add(label());
	} else {
	    Tree[] kids = children();
	    for (int i = 0; i < kids.length; i++) {
		kids[i].yield(y);
	    }
	}
	return y;
    }


    /** Gets the tagged yield of the tree.
     *  The <code>Label</code> of all leaf nodes is returned
     *  as a list ordered by the natural left to right order of the
     *  leaves.  Null values, if any, are inserted into the list like any
     *  other value.
     *
     *  @return a <code>List</code> of the data in the tree's leaves.
     */
    public Sentence taggedYield() {
	return (Sentence) taggedYield(new Sentence());
    }


    /** Gets the tagged yield of the tree -- that is, get the preterminals
     *  as well as the terminals.  The <code>Label</code> of all leaf nodes
     *  is returned
     *  as a list ordered by the natural left to right order of the
     *  leaves.  Null values, if any, are inserted into the list like any
     *  other value.  This has been rewritten to thread, so only one List
     *  is used.
     *  @param ty The list in which the tagged yield of the tree will be
     *      placed. Normally, this will be empty when the routine is called,
     *      but if not, the new yield is added to the end of the list.
     *  @return a <code>List</code> of the data in the tree's leaves.
     */
    public List taggedYield(List ty) {
	Tree[] kids = children();
	// this inlines the content of isPreTerminal()
	if (kids != null) {
	    if (kids.length == 1 && kids[0].isLeaf()) {
		ty.add(new TaggedWord(kids[0].label(), label()));
	    } else {
		for (int i = 0; i < kids.length; i++) {
		    kids[i].taggedYield(ty);
		}
	    }
	}
	return ty;
    }


    /** Gets the preterminal yield (i.e., tags) of the tree.  All data in
     *  preleaf nodes is returned as a list ordered by the natural left to
     *  right order of the tree.  Null values, if any, are inserted into the
     *  list like any other value.  Pre-leaves are nodes of height 1.
     *  @return a <code>List</code> of the data in the tree's pre-leaves.
     */
    public List preTerminalYield() {
	return preTerminalYield(new ArrayList());
    }


    /** Gets the preterminal yield (i.e., tags) of the tree.  All data in
     *  preleaf nodes is returned as a list ordered by the natural left to
     *  right order of the tree.  Null values, if any, are inserted into the
     *  list like any other value.  Pre-leaves are nodes of height 1.
     *  @param y The list in which the preterminals of the tree will be
     *      placed. Normally, this will be empty when the routine is called,
     *      but if not, the new yield is added to the end of the list.
     *  @return a <code>List</code> of the data in the tree's pre-leaves.
     */
    public List preTerminalYield(List y) {
	if (isPreTerminal()) {
	    y.add(label());
	} else {
	    Tree[] kids = children();
	    for (int i = 0; i < kids.length; i++) {
		kids[i].preTerminalYield(y);
	    }
	}
	return y;
    }


    /** Get the set of all node and leaf <code>Label</code>s,
     *  null or otherwise, contained in the tree.
     *
     *  @return the <code>Collection</code> (actually, Set) of all values
     *       in the tree.
     */
    public Collection labels() {
	Set n = new HashSet();
	n.add(label());
	if (! isLeaf()) {
	    Tree[] kids = children();
	    for (int i = 0; i < kids.length; i++) {
		n.addAll(kids[i].labels());
	    }
	}
	return n;
    }


    public void setLabels(Collection c) {
        throw new UnsupportedOperationException("Can't set Tree labels");
    }


    /** Return a flattened version of a tree.  In many circumstances, this
     *  will just return the tree, but if the tree is something like a 
     *  binarized version of a dependency grammar tree, then it will be
     *  flattened back to a dependency grammar tree representation.  Formally, 
     *  a node will be removed from the tree when: it is not a terminal or
     *  preterminal, and its <code>label()</code is <code>equal()</code> to
     *  the <code>label()</code> of its parent, and all its children will
     *  then be promoted to become children of the parent (in the same
     *  position in the sequence of daughters.
     *  @return A flattened version of this tree.
     */
    public Tree flatten() {
	return flatten(treeFactory());
    }

    /** Return a flattened version of a tree.  In many circumstances, this
     *  will just return the tree, but if the tree is something like a 
     *  binarized version of a dependency grammar tree, then it will be
     *  flattened back to a dependency grammar tree representation.  Formally, 
     *  a node will be removed from the tree when: it is not a terminal or
     *  preterminal, and its <code>label()</code is <code>equal()</code> to
     *  the <code>label()</code> of its parent, and all its children will
     *  then be promoted to become children of the parent (in the same
     *  position in the sequence of daughters. <p>
     *  Note: In the current implementation, the tree structure is mainly
     *  duplicated, but the links between preterminals and terminals aren't.
     *  @param tf TreeFactory used to create tree structure for flattened tree
     *  @return A flattened version of this tree.
     */
    public Tree flatten(TreeFactory tf) {
	if (isLeaf() || isPreTerminal()) {
	    return this;
	}
	Tree[] kids = children();
	List newChildren = new ArrayList(kids.length);
	for (int c = 0; c < kids.length; c++) {
	    Tree child = kids[c];
	    if (child.isLeaf() || child.isPreTerminal()) {
		newChildren.add(child);
	    } else {
		Tree newChild = child.flatten(tf);
		if (label().equals(newChild.label())) {
		    newChildren.addAll(newChild.getChildrenAsList());
		} else {
		    newChildren.add(newChild);
		}
	    }
	}
	return tf.newTreeNode(label(), newChildren);
    }


    /** Get the set of all subtrees inside the tree by returning a tree
     *  rooted at each node.  These are <i>not</i> copies, but all share
     *  structure.  The tree is regarded as a subtree of itself.
     *
     *  @return the <code>Set</code> of all subtrees in the tree.
     */
    public Set subTrees() {
        return subTrees(new HashSet());
    }


    /** Add the set of all subtrees inside a tree (including the tree itself) 
     *  to the given <code>Set</code>.
     *  @param n A set of nodes to which the subtrees will be added
     *  @return The set parameter with the subtrees added
     */
    public Set subTrees(Set n) {
        n.add(this);
        if ( ! isLeaf()) {
            Tree[] kids = children();
            for (int i = 0; i < kids.length; i++) {
                kids[i].subTrees(n);
            }
        }
        return n;
    }


    /** Create a deep copy of the tree.  The entire structure is
     *  recursively copied, but label data themselves are not cloned.
     *  The copy is built using a <code>TreeFactory</code> that will
     *  produce a <code>Tree</code> like the input one.
     *  @return a deep structural copy of the tree.
     */
    public Tree deepCopy() {
	return deepCopy(treeFactory());
    }


    /** Create a deep copy of the tree.  The entire structure is
     *  recursively copied, but label data themselves are not cloned.
     *  By specifying an appropriate <code>TreeFactory</code>, this
     *  method can be used to change the type of a <code>Tree</code>.
     *  @param tf The <code>TreeFactory</code> to be used for creating
     *         the returned <code>Tree</code>
     *  @return a deep structural copy of the tree.
     */
    public Tree deepCopy(TreeFactory tf) {
	Tree t;
	if (isLeaf()) {
	    t = tf.newLeaf(label());
	} else {
	    Tree[] kids = children();
	    List newKids;
	    if (kids == null) {
		newKids = null;
	    } else {
		newKids = new ArrayList(kids.length);
		for (int i = 0; i < kids.length; i++) {
		    newKids.add(kids[i].deepCopy(tf));
		}
	    }
	    t = tf.newTreeNode(label(), newKids);
	}
	return t;
    }


    /** Create a transformed Tree.  The tree is traversed in a depth-first,
     *  left-to-right order, and the <code>TreeTransformer</code> is called
     *  on each node.  It returns some <code>Tree</code>.  The transformed
     *  tree has a new tree structure (i.e., a "deep copy" is done), but it
     *  will usually share its labels with the original tree.
     *  @param transformer The function that transforms tree nodes or subtrees
     *  @return a transformation of this <code>Tree</code>
     */
    public Tree transform(final TreeTransformer transformer) {
	return transform(transformer, treeFactory());
    }


    /** Create a transformed Tree.  The tree is traversed in a depth-first,
     *  left-to-right order, and the <code>TreeTransformer</code> is called
     *  on each node.  It returns some <code>Tree</code>.  The transformed
     *  tree has a new tree structure (i.e., a "deep copy" is done), but it
     *  will usually share its labels with the original tree.
     *  @param transformer The function that transforms tree nodes or subtrees
     *  @param tf The <code>TreeFactory</code> which will be used for creating
     *         new nodes for the returned <code>Tree</code>
     *  @return a transformation of this <code>Tree</code>
     */
    public Tree transform(final TreeTransformer transformer,
			 final TreeFactory tf) {
	Tree t;
	if (isLeaf()) {
	    t = tf.newLeaf(label());
	} else {
	    Tree[] kids = children();
	    List newKids;
	    if (kids == null) {
		newKids = null;
	    } else {
		newKids = new ArrayList(kids.length);
		for (int i = 0; i < kids.length; i++) {
		    newKids.add(kids[i].transform(transformer, tf));
		}
	    }
	    t = tf.newTreeNode(label(), newKids);
	}
	return transformer.transformTree(t, tf);
    }


    /** Creates a (partial) deep copy of the tree, where all nodes that the
     *  filter does not accept are spliced out.  If the result is not a tree
     *  (that is, it's a forest), an empty root node is generated.
     *  @param nodeFilter a Filter method which returns true to mean delete
     *          this node
     *  @return a filtered copy of the tree
     */
    public Tree spliceOut(final Filter nodeFilter) {
	return spliceOut(nodeFilter, treeFactory());
    }


    /** Creates a (partial) deep copy of the tree, where all nodes that the
     *  filter does not accept are spliced out.  That is, the particular
     *  modes for which the <code>Filter</code> returns <code>false</code>
     *  are removed from the <code>Tree</code>, but those nodes' children
     *  are kept (assuming they pass the <code>Filter</code>, and they are
     *  added in the appropriate left-to-right ordering as new children of
     *  the parent node.  If the root node is deleted, so that the result
     *  would not be a tree (that is, it's a forest), an empty root node is
     *  generated.  If nothing is accepted, <code>null</code> is returned.
     *  @param nodeFilter a Filter method which returns true to mean delete
     *          this node
     *  @param tf A <code>TreeFactory</code> for making new trees. Used if
     *          the root node is deleted.
     *  @return a filtered copy of the tree.
     */
    public Tree spliceOut(final Filter nodeFilter, final TreeFactory tf) {
	List l = spliceOutHelper(nodeFilter, tf);
	if (l.isEmpty()) {
	    return null;
	} else if (l.size() == 1) {
	    return (Tree) l.get(0);
	}
	// for a forest, make a new root
	return tf.newTreeNode((Label) null, l);
    }


    private List spliceOutHelper(Filter nodeFilter, TreeFactory tf) {
	// recurse over all children first
	Tree[] kids = children();
	List l = new ArrayList();
	if (kids != null) {
	    for (int i = 0; i < kids.length; i++) {
		l.addAll(kids[i].spliceOutHelper(nodeFilter, tf));
	    }
	}
	// check if this node is being spliced out
	if (nodeFilter.accept(this)) {
	    // no, so add our children and return
            Tree t = null;
            if (l.size() != 0)
                t = tf.newTreeNode(label(), l);
            else
                t = tf.newLeaf(label());
            l = new ArrayList(1);
            l.add(t);
            return l;
	} else {
	    // we're out, so return our children
	    return l;
	}
    }


    /** Creates a deep copy of the tree, where all nodes that the filter
     *  does not accept and all children of such nodes are pruned.  If all '
     *  of a node's children are pruned, that node is cut as well.
     *  A <code>Filter</code> can assume
     *  that it will not be called with a <code>null</code> argument.
     *  @param filter the filter to be apply
     *  @return a filtered copy of the tree.
     */
    public Tree prune(final Filter filter) {
	return prune(filter, treeFactory());
    }


    /** Creates a deep copy of the tree, where all nodes that the filter
     *  does not accept and all children of such nodes are pruned.  If all '
     *  of a node's children are pruned, that node is cut as well.
     *  A <code>Filter</code> can assume
     *  that it will not be called with a <code>null</code> argument.
     *  @param filter the filter to be apply
     *  @param tf the TreeFactory to be used to make new Tree nodes if needed
     *  @return a filtered copy of the tree.
     */
    public Tree prune(Filter filter, TreeFactory tf) {
	// is the current node to be pruned?
	if (! filter.accept(this))
	    return null;
	// if not, recurse over all children
	List l = new ArrayList();
	Tree[] kids = children();
	if (kids != null) {
	    for (int i = 0; i < kids.length; i++) {
		Tree prunedChild = kids[i].prune(filter, tf);
		if (prunedChild != null)
		    l.add(prunedChild);
	    }
	    // and check if this node has lost all its children
	    if (l.isEmpty() && ! (kids == null)) {
		return null;
	    }
	    // if we're still ok, copy the node
	}
	if (isLeaf()) {
	    return tf.newLeaf(label());
	} else {
	    return tf.newTreeNode(label(), l);
	}
    }


    /** Return a <code>TreeFactory</code> that produces trees of the
     *  appropriate type.
     *  @return A factory to produce Trees
     */
    public abstract TreeFactory treeFactory();


    /** Return the parent of the tree node.  This routine may return
     *  <code>null</code> meaning simply that the implementation doesn't
     *  know how to determine the parent node, rather than there is no
     *  such node.
     *  @return The parent <code>Tree</code> node or <code>null</code>
     *  @see Tree#parent(Tree)
     */
    public Tree parent() {
        return null;
    }


    /** Return the parent of the tree node.  This routine will traverse
     *  a tree (depth first) from the given <code>root</code>, and will
     *  correctly find the parent, regardless of whether the concrete
     *  class stores parents.  It will only return <code>null</code> if this
     *  node is the <code>root</code> node, or if this node is not
     *  contained within the tree rooted at <code>root</code>.
     *  @param root The root node of the whole Tree
     *  @return the parent <code>Tree</code> node if any;
     *            else <code>null</code>
     */
    public Tree parent(Tree root) {
	if (root.isLeaf()) {
	    return null;
	}
	Tree[] kids = root.children();
	return parentHelper(root, kids, this);
    }


    private static Tree parentHelper(Tree parent,  Tree[] kids, Tree node) {
	for (int i = 0; i < kids.length; i++) {
	    if (kids[i] == node) {
		return parent;
	    }
	    Tree ret = node.parent(kids[i]);
	    if (ret != null) {
		return ret;
	    }
	}
        return null;
    }


    /** Returns the number of nodes the tree contains.  This method
     *  implements the <code>size()</code> function required by the
     *  <code>Collections</code> interface.  The size of the tree is the
     *  number of nodes it contains (of all types, including the leaf nodes
     *  and the root).
     *  @return The size of the tree
     *  @see #depth()
     */
    public int size() {
	int size = 1;
	if ( ! isLeaf()) {
	    Tree[] kids = children();
	    for (int i = 0; i < kids.length; i++) {
		size += kids[i].size();
	    }
	}
	return size;
    }


    private class TreeIterator implements Iterator {

        private List treeStack;

        private TreeIterator () {
            treeStack = new ArrayList();
            treeStack.add(Tree.this);
	}

	public boolean hasNext() {
	    return ( ! treeStack.isEmpty());
	}

	public Object next() {
	    int lastIndex = treeStack.size() - 1;
	    Tree tr = (Tree) treeStack.remove(lastIndex);
	    if ( ! tr.isLeaf()) {
		Tree[] kids = tr.children();
		// so that we can efficiently use one List, we reverse them
		for (int i = kids.length - 1; i >= 0; i--) {
		    treeStack.add(kids[i]);
		}
	    }
	    return tr;
	}

	/** Not supported */
	public void remove() {
	    throw new UnsupportedOperationException();
	}

    }


    /** Returns an iterator over the nodes of the tree.  This method
     *  implements the <code>iterator()</code> method required by the
     *  <code>Collections</code> interface.  It does a preorder
     *  (children after node) traversal of the tree.  (A possible
     *  extension to the class at some point would be to allow different
     *  traversal orderings via variant iterators.)
     *  @return An interator over the nodes of the tree
     */
    public Iterator iterator() {
	return new TreeIterator();
    }


    /** This gives you a tree from a String representation (as a 
     *  bracketed Tree, of the kind produced by <code>toString()</code>,
     *  <code>pennPrint()</code>, or as in the Penn Treebank).
     *  It's not the most efficient thing to do for heavy duty usage.
     *  The Tree returned is created by a 
     *  StringLabeledScoredTreeReaderFactory.
     *  @param str The tree as a bracketed list in a String.
     *  @throws IOException If Tree format is not valid
     *  @return The Tree
     */
    public static Tree valueOf(String str) throws IOException {
	return valueOf(str, new StringLabeledScoredTreeReaderFactory());
    }

    /** This gives you a tree from a String representation (as a 
     *  bracketed Tree, of the kind produced by <code>toString()</code>,
     *  <code>pennPrint()</code>, or as in the Penn Treebank.
     *  It's not the most efficient thing to do for heavy duty usage.
     *  @param str The tree as a bracketed list in a String.
     *  @throws IOException If Tree format is not valid
     *  @return The Tree
     */
    public static Tree valueOf(String str, TreeReaderFactory trf)
	    throws IOException {
	return trf.newTreeReader(new StringReader(str)).readTree();
    }

}
