package edu.stanford.nlp.trees;

/**
 * This is a utility class which vends tree transformers to translate
 * trees from one factory type to trees of another.  For example,
 * StringLabel trees need to be made into Category < Tag < Word trees before
 * they can be head-percolated.  Enter
 * StringLabelTreeToCategoryWordTagTreeAppliable.
 *
 * @author <a href="mailto:klein@cs.stanford.edu">Dan Klein</a>
 * @version 1.0
 * @since 1.0
 */

import edu.stanford.nlp.process.*;
import java.util.*;
import java.io.*;

public class TreeJugglers {

  private static class StringLabelTreeToCategoryWordTagTreeAppliable implements Appliable {

    protected TreeFactory tf = new LabeledScoredTreeFactory();

    public Tree helper(Tree t) {
      if (t == null)
	return null;
      if (t.isLeaf())
	return tf.newLeaf(new CategoryWordTag(t.label().value()));
      if (t.children().length == 1 && t.children()[0].isLeaf()) {
	return tf.newTreeNode(new CategoryWordTag(t.label().value()),
			      Collections.singletonList(helper(t.children()[0])));
      }
      List children = new ArrayList(t.children().length);
      for (int k=0; k<t.children().length; k++) {
	children.add(helper(t.children()[k]));
      }
      return tf.newTreeNode(new CategoryWordTag(t.label().value()),
			    children);
    }

    public Object apply(Object o) {
      try {
	return helper((Tree)o);
      } catch (ClassCastException e) {
	throw new IllegalArgumentException();
      }
    }
  };

  /** Return an Appliable that maps from StringLabel labeled trees to
   *  LabeledScoredTrees with a CategoryWordTag label.
   *  @return The Appliable object
   */
  public static Appliable getStringLabelTreeToCategoryWordTagTreeAppliable() {
    return new StringLabelTreeToCategoryWordTagTreeAppliable();
  }

  /** This just test the functionality of the included Jugglers. */
  public static void main(String[] args) {
    TreeFactory tf = new LabeledScoredTreeFactory();
    Tree stringyTree = null;
    try {
      stringyTree = (new PennTreeReader(new StringReader("(S (VP (VBZ Try) (NP (DT this))) (. .))"), new LabeledScoredTreeFactory(new StringLabelFactory()))).readTree();
    } catch (IOException e) {
    }
    System.out.println(stringyTree);
    Appliable a = getStringLabelTreeToCategoryWordTagTreeAppliable();
    Tree adaptyTree = (Tree)a.apply(stringyTree);
    System.out.println(adaptyTree);
    adaptyTree.percolateHeads(new CollinsHeadFinder());
    System.out.println(adaptyTree);
  }
}
