package edu.stanford.nlp.trees;

import java.io.*;
import java.util.*;

import edu.stanford.nlp.io.ExtensionFileFilter;

/** A <code>Treebank</code> object provides access to a corpus of examples with
 *  given tree structures.
 *  This class now implements the Collection interface. However, it may offer
 *  less than the full power of the Collection interface: some Treebanks are
 *  read only, and so may throw the UnsupportedOperationException.
 *  @author Christopher Manning
 */
public abstract class Treebank extends AbstractCollection {

    /** Stores the <code>TreeReaderFactory</code> that will be used to
     *  create a <code>TreeReader</code> to process a file of trees.
     */
    private TreeReaderFactory trf;

    /** Create a new Treebank (using a SimpleTreeReaderFactory).
    */
    public Treebank() {
        this(new SimpleTreeReaderFactory());
    }


    /** Create a new Treebank.
     *  @param trf the factory class to be called to create a new
     *      <code>TreeReader</code>
     */
    public Treebank(TreeReaderFactory trf) {
        this.trf = trf;
    }


    /** Create a new Treebank.
     *  @param initialCapacity The initial size of the underlying Collection,
     *  (if a Collection-based storage mechanism is being provided)
     */
    public Treebank(int initialCapacity) {
        this(initialCapacity, new SimpleTreeReaderFactory());
    }


    /** Create a new Treebank.
     *  @param initialCapacity The initial size of the underlying Collection,
     *  (if a Collection-based storage mechanism is being provided)
     *  @param trf the factory class to be called to create a new
     *      <code>TreeReader</code>
     */
    public Treebank(int initialCapacity, TreeReaderFactory trf) {
        this.trf = trf;
    }


    /** Get the <code>TreeReaderFactory</code> for a <code>Treebank</code> --
     *  this method is provided in order to make the
     *  <code>TreeReaderFactory</code> available to subclasses.
     *  @return The TreeReaderFactory
     */
    protected TreeReaderFactory treeReaderFactory() {
        return trf;
    }


    /** Empty a <code>Treebank</code>.
     */
    public abstract void clear();


    /** Load a sequence of trees from given directory and its subdirectories.
     *  Trees should reside in files with the suffix "mrg".
     *  Or: load a single file with the given pathName (including extension)
     *  @param pathName file or directory name
     */
    public void loadPath(String pathName) {
        loadPath(new File(pathName));
    }


    /** Load a sequence of trees from given directory and its subdirectories.
     *  Trees should reside in files with the suffix "mrg".
     *  Or: load a single file with the given path (including extension)
     *  @param path File specification
     */
    public void loadPath(File path) {
        loadPath(path, "mrg", true);
    }


    /** Load trees from given directory.
     *  @param pathName File or directory name
     *  @param suffix Extension of files to load: If <code>pathName</code>
     *    is a directory, then, if this is
     *    non-<code>null</code>, all and only files ending in "." followed
     *    by this extension will be loaded; if it is <code>null</code>,
     *    all files in directories will be loaded.  If <code>pathName</code>
     *    is not a directory, this parameter is ignored.
     *  @param Whether to ecursively descend into subdirectories as well
     */
    public void loadPath(String pathName, String suffix, boolean recursively) {
        loadPath(new File(pathName),
                 new ExtensionFileFilter(suffix, recursively));
    }


    /** Load trees from given directory.
     * @param path  file or directory to load from
     * @param suffix suffix of files to load
     * @param recursively descend into subdirectories as well
     */
    public void loadPath(File path, String suffix, boolean recursively) {
        loadPath(path, new ExtensionFileFilter(suffix, recursively));
    }


    /** Load a sequence of trees from given directory and its subdirectories
     *  which match the file filter.
     *  Or: load a single file with the given pathName (including extension)
     *  @param pathName file or directory name
     *  @param filt A filter used to determine which files match
     */
    public void loadPath(String pathName, FileFilter filt) {
        loadPath(new File(pathName), filt);
    }


    /** Load trees from given path specification.
     * @param path  file or directory to load from
     * @param filt a FilenameFilter of files to load
     */
    public abstract void loadPath(File path, FileFilter filt);


    /** Return an Iterator over Trees in the Treebank.
     *  @return The Iterator
     */
    public abstract Iterator iterator();


    /** Apply a TreeProcessor to each tree in the Treebank.
     *  For all current implementations of Treebank, this is the fastest
     *  way to traverse all the trees in the Treebank.
     *  @param tp The TreeProcessor to be applied
     */
    public abstract void apply(TreeProcessor tp);



    /** Return the whole treebank as a series of big bracketed lists.
     *  Calling this is a really bad idea if your treebank is large.
     */
    public String toString() {
        final StringBuffer sb = new StringBuffer();
        apply(new TreeProcessor() {
                  public void processTree(Tree t) {
                      sb.append(t.toString());
                      sb.append("\n");
                  }
              });
        return sb.toString();
    }


    private final static class CounterTreeProcessor implements TreeProcessor {
        int i; // = 0;

        public void processTree(Tree t) {
            i++;
        }

        public int total() {
            return i;
        }
    }


    /** Returns the size of the Treebank.
     *  @return size How many trees are in the treebank
     */
    public int size() {
        CounterTreeProcessor counter = new CounterTreeProcessor();
        apply(counter);
        return counter.total();
    }


    /** This operation isn't supported for a Treebank.  Tell them immediately.
     */
    public boolean remove(Object o) {
        throw new UnsupportedOperationException("Treebank is read-only");
    }

}
