package edu.stanford.nlp.trees;

/** A WordTag corresponds to a tagged (e.g., for part of speech) word
 * and is implemented with String-valued word and tag.  It implements
 * the Label interface; the <code>value()<code> method for that
 * interface corresponds to the word of the WordTag.
 *
 * The equality relation for WordTag is defined as identity of both
 * word and tag.  Note that this is different from
 * <code>TaggedWord</code>, for which equality derives from
 * <code>AbstractLabel</code> and requires only identity of value.
 * TaggedWord should be considered deprecated and the use of WordTag
 * is preferable.
 *
 * @author Roger Levy
 */

public class WordTag implements Label,HasWord,HasTag {

  private String word;
  private String tag;
  private static String divider = "/";

  public WordTag(String word) {
    this.word = word;
    setTag(null);
  }

  public WordTag(Label word) {
    this(word.value());
  }

  /** Create a new <code>WordTag</code>.
   *  @param word This word is passed to the supertype constructor
   *  @param tag  The <code>value()</code> of this label is set as the 
   *        tag of this Label
   */
  public WordTag(String word, String tag) {
    this(word);
    setTag(tag);
  }

  /** Create a new <code>WordTag</code> from a Label.  The value of
   *  the Label corresponds to the word of the WordTag.
   *  @param word This word is passed to the supertype constructor
   *  @param tag  The <code>value()</code> of this label is set as the 
   *        tag of this Label
   */
  public WordTag(Label word, Label tag) {
    this(word);
    setTag(tag.value());
  }


  /** Return a String representation of just the "main" value of this label.
   *  @return the "value" of the label
   */
  public String value() {
    return word;
  }

  public String word() {
    return value();
  }

  /** Set the value for the label (if one is stored).
   *  @param value - the value for the label
   */
  public void setValue(String value) {
    word = value;
  }

  public String tag() {
    return tag;
  }

  public void setWord(String word) {
    setValue(word);
  }
  
  public void setTag(String tag) {
    this.tag = tag;
  }


  /** Return a String representation of the label.  For a multipart label,
   *  this will return all parts.  The <code>toString()</code> method
   *  causes a label to spill its guts.  It should always return an
   *  empty string rather than <code>null</code> if there is no value.
   *  @return a text representation of the full label contents
   */
  public String toString() {
    return word() + divider + tag;
  }



  /** Sets a TaggedWord from decoding 
   *  the <code>String</code> passed in.  The String is divided according
   *  to the divider character (usually, "/").  We assume that we can 
   *  always just  
   *  divide on the rightmost divider character, rather than trying to
   *  parse up escape sequences.  If the divider character isn't found
   *  in the word, then the whole string becomes the word, and the tag
   *  is <code>null</code>.
   *  @param taggedWord The word that will go into the <code>Word</code>
   */
  public void setFromString(String taggedWord) {
    int where = taggedWord.lastIndexOf(divider);
    if (where >= 0) {
      setWord(taggedWord.substring(0, where));
      setTag(taggedWord.substring(where + 1));
    } else {
      setWord(taggedWord);
      setTag(null);
    }
  }

  /** Equality is satisfied only if the compared object is a WordTag
   * and has String-equal word and tag fields. */
  public boolean equals(Object obj) {
    return (obj instanceof WordTag) &&
      (word().equals(((WordTag) obj).word())) && (tag().equals(((WordTag) obj).tag()));
  }

  public int hashCode() {
    return toString().hashCode();
  }

  /** Implementation of comparison for <code>Label</code>.
   *  By default this orders by value's string lexicographic ordering.
   *  @param obj object to compare to
   *  @return result (positive if <code>this</code> is greater than
   *       <code>obj</code>, 0 if equal, negative otherwise)
   */
  public int compareTo(Object obj) {
    throw new UnsupportedOperationException();
  }
    

  // extra class guarantees correct lazy loading (Bloch p.194)
  private static class LabelFactoryHolder {

    private static final LabelFactory lf = new WordTagFactory();

  }

  /** Return a factory for this kind of label
   *  (i.e., <code>TaggedWord</code>).
   *  The factory returned is always the same one (a singleton).
   *  @return The label factory
   */
  public LabelFactory labelFactory() {
    return LabelFactoryHolder.lf;
  }


  /** Return a factory for this kind of label.
   *  @return The label factory
   */
  public static LabelFactory factory() {
    return LabelFactoryHolder.lf;
  }


  /** Set the String (usually just a single character) used to mark
   *  the separation between the word and the tag.  This is used when
   *  printing out a <code>TaggedWord</code> via the <code>toString()</code>
   *  method, and when reading in a <code>TaggedWord</code> via the
   *  <code>setFromString()</code> method.
   *  Unless altered using this method, the default is "/".
   *  @param divider The String between words and tags
   */
  public static void setDivider(String divider) {
    WordTag.divider = divider;
  }
 
}
