package edu.stanford.nlp.util;

import java.util.*;
import java.io.Serializable;

/** 
 * Specialized Map for storing numeric counts for objects. Works like a normal
 * Map but with extra methods for easily getting/setting/incrementing counts
 * for objects and computing various functions with the counts. Any attempt to
 * put a non-Double value into this Counter will result in an 
 * IllegalArgumentException being thrown. Note however that the Map constructor
 * and <tt>putAll</tt> method can be used to copy another Counter's contents 
 * over (see also <tt>addAll(Counter)</tt>). This class also provides access
 * to Comparators that can be used to sort the keys or entries of this Counter
 * by the counts, in either ascending or descending order.
 * @author Dan Klein
 * @author Joseph Smarr (jsmarr@stanford.edu)
 */
public class Counter extends HashMap implements Serializable 
{
    /** Default comparator for breaking ties in argmin and argmax. */
    private static final Comparator naturalComparator=new NaturalComparator();
    private static final long serialVersionUID = 3;
    
    private double total; // current total count of all entries
    
    /** Constructs a new (empty) Counter. */
    public Counter() { super(); }
    
    /**
     * Constructs a new Counter with the contents of the given Map.
     * The values in <tt>m</tt> must all be Doubles or an 
     * IllegalArgumentException will be thrown when they are added.
     */
    public Counter(Map m) { this(); putAll(m); }
    
    // --- Standard access/modification methods --- //
    
    /** 
     * Returns the current total count for all objects in this Counter. The
     * total is maintained as counts are adjusted so it can be returned quickly
     * without having to sum all the counts on demand.
     */
    public double totalCount() { return(total); }
      
    /** 
     * Returns the total count for all objects in this Counter that pass the 
     * given Filter. Passing in a filter that always returns true is equivalent 
     * to calling {@link #total()}, though the latter is faster since the total
     * is maintained internally as counts are adjusted.
     */
    public double totalCount(Filter filter)
    {
        double total=0.0;
        for(Iterator iter=seenSet().iterator();iter.hasNext();)
        {
            Object key=iter.next();
            if(filter.accept(key)) total+=getCount(key);
        }
        return(total);
    }
    
    /** Returns the mean of all the counts (totalCount/size). */
    public double averageCount() { return(totalCount()/size()); }   
    
    /**
     * Returns the current count for the given key, which is 0 if it hasn't
     * been
     * seen before. This is a convenient version of <code>get</code> that casts
     * and extracts the primitive value.
     */
    public double getCount(Object key)
    {
        Number count=(Number)get(key);
        if(count==null) return(0); // haven't seen this object before -> 0 count
        return(count.doubleValue());
    }
    
    /**
     * Returns the current count for the given key as a fraction of the total
     * count in the counter. This is equivalent to assuming that all the counts
     * sum to one, but it doesn't actually change the raw counts.
     */
    public double getNormalizedCount(Object key)
    {
        return(getCount(key)/total());
    }
    
    /**
     * Sets the current count for the given key. This will wipe out any existing
     * count for that key. 
     * <p>
     * To add to a count instead of replacing it, use 
     * {@link #incrementCount(Object,double)}.
     */
    public void setCount(Object key,double count)
    {
        put(key,new Double(count));
    }
    
    /**
     * Sets the current count for each of the given keys. This will wipe out
     * any existing counts for these keys.
     * <p>
     * To add to the counts of a collection of objects instead of replacing them,
     * use {@link #incrementCounts(Collection,double)}.
     */
    public void setCounts(Collection keys,double count)
    {
        for(Iterator iter=keys.iterator();iter.hasNext();)
            setCount(iter.next(),count);
    }
    
    /**
     * Adds the given count to the current count for the given key. If the key
     * hasn't been seen before, it is assumed to have count 0, and thus this 
     * method will set its count to the given amount. Negative increments are
     * equivalent to calling <tt>decrementCount</tt>.
     * <p>
     * To more conviently increment the count by 1.0, use 
     * {@link #incrementCount(Object)}.
     * <p>
     * To set a count to a specifc value instead of incrementing it, use 
     * {@link #setCount(Object,double)}.
     */
    public void incrementCount(Object key,double count)
    {
        put(key,new Double(count+getCount(key)));
    }
    
    /**
     * Adds 1.0 to the count for the given key. If the key hasn't been seen 
     * before, it is assumed to have count 0, and thus this method will set
     * its count to 1.0. 
     * <p>
     * To increment the count by a value other than 1.0, use 
     * {@link #incrementCount(Object,double)}.
     * <p>
     * To set a count to a specifc value instead of incrementing it, use 
     * {@link #setCount(Object,double)}.
     */
    public void incrementCount(Object key) { incrementCount(key,1.0); }   
    
    /**
     * Adds the given count to the current counts for each of the given keys.
     * If any of the keys haven't been seen before, they are assumed to have
     * count 0, and thus this method will set their counts to the given
     * amount. Negative increments are equivalent to calling <tt>decrementCounts</tt>.
     * <p>
     * To more conviniently increment the counts of a collection of objects by
     * 1.0, use {@link #incrementCounts(Collection)}.
     * <p>
     * To set the counts of a collection of objects to a specific value instead
     * of incrementing them, use {@link #setCounts(Collection,double)}.
     */
    public void incrementCounts(Collection keys,double count)
    {
        for(Iterator iter=keys.iterator();iter.hasNext();)
            incrementCount(iter.next(),count);
    }
    
    /**
     * Adds 1.0 to the counts for each of the given keys. If any of the keys 
     * haven't been seen before, they are assumed to have count 0, and thus 
     * this method will set their counts to 1.0.
     * <p>
     * To increment the counts of a collection of object by a value other 
     * than 1.0, use {@link #incrementCounts(Collection,double)}.
     * <p>
     * To set the counts of a collection of objects  to a specifc value instead
     * of incrementing them, use  {@link #setCounts(Collection,double)}.
     */
    public void incrementCounts(Collection keys) { incrementCounts(keys,1.0); }    
    
    /**
     * Subtracts the given count from the current count for the given key. 
     * If the key hasn't been seen before, it is assumed to have count 0, and 
     * thus this  method will set its count to the negative of the given amount. 
     * Negative increments are equivalent to calling <tt>incrementCount</tt>.
     * <p>
     * To more conviently decrement the count by 1.0, use 
     * {@link #decrementCount(Object)}.
     * <p>
     * To set a count to a specifc value instead of decrementing it, use 
     * {@link #setCount(Object,double)}.
     */
    public void decrementCount(Object key,double count)
    {
        incrementCount(key,-count);
    }
    
    /**
     * Subtracts 1.0 from the count for the given key. If the key hasn't been 
     * seen  before, it is assumed to have count 0, and thus this method will 
     * set its count to -1.0. 
     * <p>
     * To decrement the count by a value other than 1.0, use 
     * {@link #decrementCount(Object,double)}.
     * <p>
     * To set a count to a specifc value instead of decrementing it, use 
     * {@link #setCount(Object,double)}.
     */
    public void decrementCount(Object key) { decrementCount(key,1.0); }   
            
    /**
     * Subtracts the given count from the current counts for each of the given keys.
     * If any of the keys haven't been seen before, they are assumed to have
     * count 0, and thus this method will set their counts to the negative of the given
     * amount. Negative increments are equivalent to calling <tt>incrementCounts</tt>.
     * <p>
     * To more conviniently decrement the counts of a collection of objects by
     * 1.0, use {@link #decrementCounts(Collection)}.
     * <p>
     * To set the counts of a collection of objects to a specific value instead
     * of decrementing them, use {@link #setCounts(Collection,double)}.
     */
    public void decrementCounts(Collection keys,double count)
    {
        incrementCounts(keys,-count);
    }
    
    /**
     * Subtracts 1.0 from the counts of each of the given keys. If any of the keys 
     * haven't been seen before, they are assumed to have count 0, and thus 
     * this method will set their counts to -1.0.
     * <p>
     * To decrement the counts of a collection of object by a value other 
     * than 1.0, use {@link #decrementCounts(Collection,double)}.
     * <p>
     * To set the counts of a collection of objects  to a specifc value instead
     * of decrementing them, use  {@link #setCounts(Collection,double)}.
     */
    public void decrementCounts(Collection keys) { decrementCounts(keys,1.0); }

    /**
     * Adds the counts in the given Counter to the counts in this Counter.
     * <p>
     * To copy the values from another Counter rather than adding them, use
     * {@link #putAll(Map)} or {@link #Counter(Map)}.
     */
    public void addAll(Counter counter)
    {
        for(Iterator iter=counter.keySet().iterator();iter.hasNext();)
        {
            Object key=iter.next();
            incrementCount(key,counter.getCount(key));
        }
    }
    
    /**
     * Subtracts the counts in the given Counter from the counts in this Counter.
     * <p>
     * To copy the values from another Counter rather than subtracting them, use
     * {@link #putAll(Map)} or {@link #Counter(Map)}.
     */
    public void subtractAll(Counter counter)
    {
        for(Iterator iter=counter.keySet().iterator();iter.hasNext();)
        {
            Object key=iter.next();
            decrementCount(key,counter.getCount(key));
        }
    }
    
    // --- Collection methods (to constrain values and keep total) --- //
    
    /**
     * Adds a count for the given key if value is a Number. Throws an
     * IllegalArgumentException otherwise. All of HashMap's various put
     * functions route through this method, so it's the single choke point
     * to ensure that only Object-->Double entries are contained in this map.
     * If the value is a non-Double Number, its doubleValue is extracted and
     * inserted in a new Double, so the values should always be of type Double,
     * but this way adding an Integer doesn't break it.
     * For normal use, use <tt>incrementCount</tt> instead.
     */
    public Object put(Object key,Object value) throws IllegalArgumentException
    {   
        if(value!=null && value instanceof Number)
        {     
            double count=((Number)value).doubleValue();
            total+=count; // add new count
            
            Double oldCount=(Double)super.put(key,new Double(count)); 
            if(oldCount!=null) total-=oldCount.doubleValue(); // subtract old count (it's being replaced)
            
            return(oldCount); // this is what put() normally returns
        }
        else throw(new IllegalArgumentException("value must be a Number"));
    }

    /** 
     * Removes the given key from this Counter. Its count will now be 0 and it
     * will no longer be considered previously seen.
     */
    public Object remove(Object key)
    {
        total-=getCount(key); // subtract removed count from total (may be 0)
        return(super.remove(key));
    }
    
    /** Removes all the given keys from this Counter. */
    public void removeAll(Collection c)
    {
        for(Iterator iter=c.iterator();iter.hasNext();)
            remove(iter.next());
    }
    
    /**  Removes all counts from this Counter. */
    public void clear()
    {
        super.clear();
        total=0;
    }
       
    // --- Extra calculation methods --- //
        
    /** 
     * Divides all the counts by the total so they collectively sum to 1.0. 
     * This effectively turns this Counter into a probability distribution.
     * The behavior will be weird if there are any negative counts.
     * Note that calling {@link #getNormalizedCount} returns the same thing
     * as {@link #getCount} after calling <tt>normalize</tt> but it doesn't
     * change the raw counts.
     */
    public void normalize()
    {
        double total=total(); // cache value since it will change as we normalize
        for(Iterator iter=keySet().iterator();iter.hasNext();)
        {
            Object key=iter.next();
            setCount(key,getCount(key)/total);
        }
    }
    
    /**
     * Removes all keys whose count is 0. After incrementing and decrementing
     * counts or adding and subtracting Counters, there may be keys left whose
     * count is 0, though normally this is undesirable. This method cleans up
     * the map. 
     * <p>
     * Maybe in the future we should try to do this more on-the-fly, though it's
     * not clear whether a distinction should be made between "never seen" (i.e.
     * null count) and "seen with 0 count". Certainly there's no distinction in
     * getCount() but there is in containsKey().
     */
    public void removeZeroCounts()
    {
        for(Iterator iter=keySet().iterator();iter.hasNext();)
            if(getCount(iter.next())==0) 
                iter.remove();            
    }
    
    /** Finds and returns the largest count in this Counter. */
    public double max()
    {
        double max=Double.NEGATIVE_INFINITY;
        for(Iterator iter=keySet().iterator();iter.hasNext();)
            max=Math.max(max,getCount(iter.next()));
        return(max);
    }
    
    /** Finds and returns the smallest count in this Counter. */
    public double min()
    {
        double min=Double.POSITIVE_INFINITY;
        for(Iterator iter=keySet().iterator();iter.hasNext();)
            min=Math.min(min,getCount(iter.next()));
        return(min);
    }
    
    /** 
     * Finds and returns the key in this Counter with the largest count.
     * Ties are broken by comparing the objects using the given tie breaking
     * Comparator, favoring Objects that are sorted to the front. This is useful
     * if the keys are numeric and there is a bias to prefer smaller or larger
     * values, and can be useful in other circumstances where random tie-breaking
     * is not desirable. Returns null if this Counter is empty.
     */
    public Object argmax(Comparator tieBreaker)
    {
        double max=Double.NEGATIVE_INFINITY;
        Object argmax=null;
        for(Iterator iter=keySet().iterator();iter.hasNext();)
        {
            Object key=iter.next();
            double count=getCount(key);
            if(argmax==null || count>max || (count==max && tieBreaker.compare(key,argmax)<0))
            {
                max=count;
                argmax=key;
            }
        }
        return(argmax);
    }
    
    
    /** 
     * Finds and returns the key in this Counter with the largest count.
     * Ties are broken according to the natural ordering of the objects. 
     * This will prefer smaller numeric keys and lexicographically earlier
     * String keys. To use a different tie-breaking Comparator, use
     * {@link #argmax(Comparator)}. Returns null if this Counter is empty.
     */
    public Object argmax() { return(argmax(naturalComparator)); }
    
    /** 
     * Finds and returns the key in this Counter with the smallest count.
     * Ties are broken by comparing the objects using the given tie breaking
     * Comparator, favoring Objects that are sorted to the front. This is useful
     * if the keys are numeric and there is a bias to prefer smaller or larger
     * values, and can be useful in other circumstances where random tie-breaking
     * is not desirable. Returns null if this Counter is empty.
     */
    public Object argmin(Comparator tieBreaker)
    {
        double min=Double.POSITIVE_INFINITY;
        Object argmin=null;
        for(Iterator iter=keySet().iterator();iter.hasNext();)
        {
            Object key=iter.next();
            double count=getCount(key);
            if(argmin==null || count<min || (count==min && tieBreaker.compare(key,argmin)<0))
            {
                min=count;
                argmin=key;
            }
        }
        return(argmin);
    }
    
    /** 
     * Finds and returns the key in this Counter with the smallest count.
     * Ties are broken according to the natural ordering of the objects. 
     * This will prefer smaller numeric keys and lexicographically earlier
     * String keys. To use a different tie-breaking Comparator, use
     * {@link #argmin(Comparator)}. Returns null if this Counter is empty.
     */
    public Object argmin() { return(argmin(naturalComparator)); }
    
    /**
     * Returns the set of keys whose counts are at or above the given threshold.
     * This set may have 0 elements but will not be null.
     */
    public Set keysAbove(double countThreshold)
    {
        Set keys=new HashSet();
        for(Iterator iter=keySet().iterator();iter.hasNext();)
        {
            Object key=iter.next();
            if(getCount(key)>=countThreshold)
                keys.add(key);
        }
        return(keys);
    }
    
    /**
     * Returns the set of keys whose counts are at or below the given threshold.
     * This set may have 0 elements but will not be null.
     */
    public Set keysBelow(double countThreshold)
    {
        Set keys=new HashSet();
        for(Iterator iter=keySet().iterator();iter.hasNext();)
        {
            Object key=iter.next();
            if(getCount(key)<=countThreshold)
                keys.add(key);
        }
        return(keys);
    }
    
    /**
     * Returns the set of keys that have exactly the given count.
     * This set may have 0 elements but will not be null.
     */
    public Set keysAt(double count)
    {
        Set keys=new HashSet();
        for(Iterator iter=keySet().iterator();iter.hasNext();)
        {
            Object key=iter.next();
            if(getCount(key)==count)
                keys.add(key);
        }
        return(keys);
    }   
    
    /** 
     * Calculates the entropy of this counter (in bits). This method internally
     * uses normalized counts (so they sum to one), but the value returned is 
     * meaningless if some of the counts are negative.
     * @return The entropy of this counter (in bits)
     */
    public double entropy() 
    {   
    	double entropy=0.0;
        for(Iterator iter=keySet().iterator();iter.hasNext();) 
        {
    	    double count=getCount(iter.next());
    	    if(count==0) continue; // 0.0 doesn't add entropy but may cause -Inf
            count/=total(); // use normalized count
            entropy-=count*(Math.log(count)/Math.log(2.0));
	    }
        return(entropy);
    }

    /** 
     * Calculates the KL divergence between this counter and another 
     * counter.  That is, it calculates KL(this||other). This method internally
     * uses normalized counts (so they sum to one), but the value returned is
     * meaningless if some of the counts are negative.
     * @param other The other <code>Counter</code>
     * @return The KL divergence between the distributions
     */
    public double klDivergence(Counter other) 
    {
        double kl = 0.0;
        double tot = total();
        double tot2 = other.total();
        // System.out.println("tot is " + tot + " tot2 is " + tot2);
        for (Iterator it = seenSet().iterator(); it.hasNext(); ) {
            Object key = it.next();
            double num = countOf(key);
            if (num == 0) continue;
            num /= tot;
            double num2 = other.countOf(key);
            num2 /= tot2;
            // System.out.println("num is " + num + " num2 is " + num2);
            kl += num * (Math.log(num/num2)/Math.log(2.0));
        }
        return kl;
    }
    
    /**
     * Calculates the information radius (aka the Jensen-Shannon divergence) 
     * between
     * this Counter and another counter.  This measure is defined as:
     * <blockquote> iRad(p,q) = D(p||(p+q)/2)+D(q,(p+q)/2) </blockquote>
     * where p is this Counter, q is the other counter, and D(p||q) is the 
     * KL divergence bewteen p and q.  Note that iRad(p,q) = iRad(q,p).
     * @param other The other <code>Counter</code>
     * @return The information radius between the distributions
     */
    public double informationRadius(Counter other)
    {
        Counter avg=average(other); // (p+q)/2
        return(klDivergence(avg)+other.klDivergence(avg));
    }
    
    /**
     * Returns a new Counter with counts averaged from this Counter and the 
     * given Counter. The average Counter will contain the union of keys in
     * both
     * source Counters, and each count will be the average of the two source 
     * counts for that key, where as usual a missing count in one Counter
     * is treated as count 0.
     */
    public Counter average(Counter other)
    {
        Counter average=new Counter();
        Set allKeys=new HashSet(keySet());
        allKeys.addAll(other.keySet());
        for(Iterator iter=allKeys.iterator();iter.hasNext();)
        {
            Object key=iter.next();
            average.setCount(key,(getCount(key)+other.getCount(key))*0.5);
        }
        return(average);
    }

    /**
     * Returns a comparator suitable for sorting this Counter's keys or entries
     * by their respective counts. If <tt>ascending</tt> is true, lower counts
     * will be returned first, otherwise higher counts will be returned first.
     * <p>
     * Sample usage:
     * <pre>
     * Counter c = new Counter();
     * // add to the counter...
     * List biggestKeys = Collections.sort(new ArrayList(c.keySet()), c.comparator(false));
     * List smallestEntries = Collections.sort(new ArrayList(c.entrySet()), c.comparator(true))
     * </pre>
     */
    public Comparator comparator(boolean ascending)
    {
        return(new EntryValueComparator(this,ascending));
    }
    
    /** 
     * Comparator that sorts objects by (increasing) count. Shortcut for calling
     * {@link #comparator(boolean) comparator(true)}.
     */
    public Comparator comparator() { return(comparator(true)); } 
    
    /** Comparator that uses natural ordering. Returns 0 if o1 is not Comparable. */
    private static class NaturalComparator implements Comparator
    {
        public int compare(Object o1,Object o2)
        {
            if(o1 instanceof Comparable)
                return(((Comparable)o1).compareTo(o2));
            return(0); // soft-fail
        }
    }    
    
    /** For internal debugging purposes only. */
    public static void main(String[] args)
    {
        Counter c=new Counter();
        c.setCount("p",0);
        c.setCount("q",2);
        System.out.println(c+" -> "+c.total());
        c.incrementCount("p");
        System.out.println(c+" -> "+c.total());
        c.put("p",new Integer(3));
        System.out.println(c+" -> "+c.total());
        System.out.println(c.entropy());
        System.out.println(c.klDivergence(c));
        System.out.println(c.min());
        System.out.println(c.argmin());
    }
    
    // --- OLD DEPRECATED METHODS --- //

    /** 
     * Returns whether count of object is non-zero.  This normally corresponds to 
     * the object having been previously seen. 
     * @deprecated use the standard containsKey function of Map.
     */
    public boolean hasSeen(Object o) { return containsKey(o); }
    
    /**
     * Returns the set of objects with non-zero counts. This normally corresponds
     * to the objects that have been previously seen.
     * @deprecated use the standard keySet() function of Map.
     */
    public Set seenSet() { return keySet(); }
    
    /**
     * Returne the current count for the given object, or 0 if it hasn't been
     * seen before.
     * @deprecated use getCount instead.
     */
    public double countOf(Object o) { return(getCount(o)); }
    
    /**
     * Returns the total count.
     * @deprecated use totalCount instead.
     */
    public double total() { return(totalCount()); }
    
    /**
     * Adds the given count to the given object.
     * @deprecated use incrementCount instead.
     */
    public void add(Object o,double count) { incrementCount(o,count); }
    
    /**
     * Adds all the counts from the given Counter.
     * @deprecated use addAll instead.
     */
    public void addCounter(Counter c) { addAll(c); }
    
    /**
     * Adds 1.0 to the count of the given Object.
     * @deprecated use incrementCount instead.
     */
    public void increment(Object o) { incrementCount(o); }
}
