package edu.stanford.nlp.util;

import java.io.*;

/** This class now combines two sets of debugging calls:
 *  Chris': Centralize the debugging print-outs, and make them more
 *  efficient when turned off at runtime.  Always use
 *  	if (Dbg.WHATEVER) Dbg.print(str);
 *  Then string isn't constructed if debugging is turned off
 *  <p>
 *  Sep's:  Debugging class.  Variables and Methods involved in debugging code.
 *  The variable <code>verbosity</code> gives a level of "loudness".  The 
 *  higher the value of <code>verbosity</code>, the more debugging statements
 *  will be printed.
 *  <p><code>verbosity</code>
 *  <br> 0 -- silent
 *  <br> 1 -- prints necessary output, but no debugging statements
 *  <br> 10 -- very loud . . . prints all debugging values
 *
 *  @author Christopher Manning
 *  @author Sep Kamvar
 */
public class Dbg {

    public static final boolean ERROR = true;
    public static final boolean WARNING = false;
    public static final boolean TIMING = false;
    public static final boolean VERBOSE = false;
    public static final boolean PARSE = false;
    public static final boolean PROGRESS = false;
    public static final boolean ONE = false;
    public static final boolean TWO = false;
    public static final boolean THREE = false;
    public static final boolean FOUR = false;
    public static final boolean FIVE = false;
    public static final boolean SIX = false;

    private static final boolean DEBUGTOFILE = false;
    private static PrintWriter debugFile;

    private final static String debugFilename = "debug.log";

    {
        // initialization block
	try {
	    if (DEBUGTOFILE) {
	        debugFile = new PrintWriter(new
			     BufferedWriter(new FileWriter(debugFilename)));
	    }
	} catch (IOException ioe) {
	    ioe.printStackTrace(System.err);
        }
    }


    /** Print the debug message to a file or stderr
     *  @param s The string to print
     */
    public static void print(String s) {
        if (DEBUGTOFILE && debugFile != null) {
	    debugFile.println(s);
	    // flush it to get as much as possible if it dies.
	    debugFile.flush();
	} else {
            System.err.println(s);
        }
    }


    public static void linePrint(int line, String s) {
        print(line + ": " + s);
    }


    public static void close() {
        if (debugFile != null) {
	    debugFile.close();
	}
	debugFile = null;
    }


    // the stuff below here is from Sep's dbg.java .  
    // It is conceptually unrelated to the stuff above, but what the hell
    // let's just combine them into one class.
    // Christopher Manning.  2002/06/27

    public static int verbosity=7;  //levels of debugging 
    //0 -- silent
    //1 -- current debugging needs
    //10 -- very loud . . . prints all debugging values

    /** Sets the verbosity to v.
     */
    public static void setVerbosity(int v) {
        verbosity=v;
    }

    
    /** prints string s only if the verbosity exceeds v 
     */
    public static void println(int v, String s) {
        if (verbosity>=v) {
            System.err.println(s);
        }
    }


    /** Prints to a file.  If the file already exists, appends if 
     *  <code>append=true</code>, and overwrites if <code>append=false</code>
     */
    public static void printToFile(String filename, String message,
				   boolean append) {
	FileWriter fw = null;
        PrintWriter pw = null;
        try {
	    fw=new FileWriter(filename, append);
            pw = new PrintWriter(fw);
            pw.print(message);
	    pw.close();
	    fw.close();
	} catch(Exception e) {
	    System.out.println("Exception: " + e + " in printToFile " + filename + " " +
			       message); 
	}
    }


    /** Prints to a file.  If the file does not exist, rewrites the file;
     *	does not append.
     */
    public static void printToFile(String filename, String message) {
	printToFile(filename, message, false);
    }

}
