package edu.stanford.nlp.util;

import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.net.*;

import edu.stanford.nlp.trees.Word;

/**
 *  Stringtils is a class for random String things. 
 *
 *  @author Dan Klein
 *  @author Christopher Manning
 *  @author Tim Grow (grow@stanford.edu)
 *  @version 2003/02/03
 */
public class StringUtils {

    /**
     * Don't let anyone instantiate this class.
     */
    private StringUtils() {}


    /** Say whether this regular expression can be found inside 
     *  this String.  This method provides one of the two "missing"
     *  convenience methods for regular expressions in the String class
     *  in JDK1.4.  This is the one you'll want to use all the time if
     *  you're used to Perl.  What were they smoking?
     *  @param str String to search for match in
     *  @param regex String to compile as the regular expression
     *  @return Whether the regex can be found in str
     */
    public static boolean find(String str, String regex) {
	return Pattern.compile(regex).matcher(str).find();
    }


    /** Say whether this regular expression can be found at the beginning of
     *  this String.  This method provides one of the two "missing"
     *  convenience methods for regular expressions in the String class
     *  in JDK1.4. 
     *  @param str String to search for match at start of
     *  @param regex String to compile as the regular expression
     *  @return Whether the regex can be found at the start of str
     */
    public static boolean lookingAt(String str, String regex) {
	return Pattern.compile(regex).matcher(str).lookingAt();
    }


    /** Say whether this regular expression matches
     *  this String.  This method is the same as the String.matches() method,
     *  and is included just to give a call that is parallel to the other
     *  static regex methods in this class.
     *  @param str String to search for match at start of
     *  @param regex String to compile as the regular expression
     *  @return Whether the regex matches the whole of this str
     */
    public static boolean matches(String str, String regex) {
	return Pattern.compile(regex).matcher(str).matches();
    }

    /** Returns all the text in the given File. */
    public static String slurpFile(File file) throws IOException {
        String lineSeparator = System.getProperty("line.separator");
	
	BufferedReader br = new BufferedReader(new FileReader(file));
	String temp;
	StringBuffer buff = new StringBuffer(16000);  // make biggish
	while ((temp = br.readLine()) != null) {
	    buff.append(temp);
	    buff.append(lineSeparator);
	}
	br.close();
	return buff.toString();
    }
  
    /** Returns all the text at the given URL. */
    public static String slurpURL(URL u) throws IOException {
        String lineSeparator = System.getProperty("line.separator");
	URLConnection uc = u.openConnection();
	InputStream is = uc.getInputStream();
	BufferedReader br = new BufferedReader(new InputStreamReader(is));
	String temp;
	StringBuffer buff = new StringBuffer(16000);  // make biggish
	while ((temp = br.readLine()) != null) {
	    buff.append(temp);
	    buff.append(lineSeparator);
	}
	br.close();
	return buff.toString();
    }
  
  /**
   * Joins each elem in the List with the given glue. For example, given a list
   * of Integers, you can create a comma-separated list by calling 
   * <tt>join(numbers, ", ")</tt>.
   */
  public static String join(List l, String glue) {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < l.size(); i++) {
      if (i > 0)
	sb.append(glue);
      sb.append(l.get(i).toString());
    }
    return sb.toString();
  }
  
  /**
   * Joins each elem in the array with the given glue. For example, given a list
   * of ints, you can create a comma-separated list by calling 
   * <tt>join(numbers, ", ")</tt>.
   */
  public static String join(Object[] elements,String glue)
  {
      return(join(Arrays.asList(elements),glue));
  }

  /** Joins elems with a space. */
  public static String join(List l) {
    return join(l, " ");
  }
  
  /** Joins elems with a space. */
  public static String join(Object[] elements) {
      return(join(elements," "));
  }


  /** Splits on whitespace (\\s+). */
  public static List split(String s) {
    return(split(s,"\\s+"));
  }
  
    /** Splits the given string using the given regex as delimiters.  
     *  This method is the same as the String.split() method (except it throws
     *  the results in a List),
     *  and is included just to give a call that is parallel to the other
     *  static regex methods in this class.
     *  @param str String to split up
     *  @param regex String to compile as the regular expression
     *  @return List of Strings resulting from splitting on the regex
     */  
  public static List split(String str, String regex) {
      return(Arrays.asList(str.split(regex)));
  }


  /** Return a String of length a minimum of totalChars characters by
   *  padding the input String str with spaces.  If str is already longer
   *  than totalChars, it is returned unchanged. 
   */
  public static String pad(String str, int totalChars) {
    int slen = str.length();
    if (slen > totalChars) {
      return str;
    } else {
      StringBuffer sb = new StringBuffer(str);
      for (int i = 0; i < totalChars-slen; i++) {
	sb.append(" ");
      }
      return sb.toString();
    }
  }

  /** Pads the toString value of the given Object. */
  public static String pad(Object obj, int totalChars) {
    return pad(obj.toString(), totalChars);
  }


  /** Pad or trim so as to produce a string of exactly a certain length.
   *  @param inStr The String to be padded or truncated
   *  @param num   The desired length
   */
  public static String exactlyN(String inStr, int num) {
    int leng = inStr.length();
    if (leng < num) {
      StringBuffer sb = new StringBuffer(inStr);
      for (int i = 0; i < num - leng; i++) {
        sb.append(" ");
      }
      return sb.toString();
    } else if (leng > num) {
      return inStr.substring(0, num);
    } else {
      return inStr;
    }
  }

  /** Pad or trim the toString value of the given Object. */
  public static String exactlyN(Object obj, int totalChars) {
    return exactlyN(obj.toString(), totalChars);
  }


  /** 
   * Pads the given String to the left with spaces to ensure that it's 
   * at least totalChars long.
   */
  public static String leftPad(String str, int totalChars) {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < totalChars-str.length(); i++) {
      sb.append(" ");
    }
    sb.append(str);
    return sb.toString();
  }


  public static String leftPad(Object obj, int totalChars) {
    return leftPad(obj.toString(), totalChars);
  }
  
  public static String leftPad(int i, int totalChars) {
      return leftPad(new Integer(i), totalChars);
  }

  public static String leftPad(double d, int totalChars) {
      return leftPad(new Double(d), totalChars);
  }
  
  /** Returns s if it's at most maxWidth chars, otherwise chops right side to fit. */
  public static String trim(String s, int maxWidth) {
      if(s.length()<=maxWidth) return(s);
      return(s.substring(0,maxWidth));
  }
  
  public static String trim(Object obj, int maxWidth) {
      return trim(obj.toString(), maxWidth);
  }

  /** 
   * Returns a "clean" version of the given filename in which spaces have
   * been converted to dashes and all non-alphaneumeric chars are underscores.
   */
  public static String fileNameClean(String s) {
    char[] chars = s.toCharArray();
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < chars.length; i++) {
      char c = chars[i];
      if ((c >= 'A' && c <= 'Z') ||
	  (c >= 'a' && c <= 'z') ||
	  (c >= '0' && c <= '9') ||
	  (c == '_'))
	  sb.append(c);
      else {
	if (c == ' ' ||
	    c == '-')
	  sb.append('_');
	else 
	  sb.append("x"+(int)c+"x");
      }
    }
    return sb.toString();
  }

  /**
   * Returns a presentable version of the given PTB-tokenized text.
   * PTB tokenization splits up punctuation and does various other things
   * that makes simply joining the tokens with spaces look bad. So join
   * the tokens with space and run it through this method to produce nice 
   * looking text. It's not perfect, but it works pretty well.
   */
  public static String ptb2Text(String ptbText)
  {
    StringBuffer sb=new StringBuffer(ptbText.length()); // probably an overestimate
    PTB2TextLexer lexer=new PTB2TextLexer(new StringReader(ptbText));
    String token;
    try { while((token=lexer.next())!=null) sb.append(token); }
    catch(IOException e) { e.printStackTrace(); }
    return(sb.toString());
  }

  /**
   * Returns a presentable version of the given PTB-tokenized words.
   * Pass in a List of Words or Strings, or a Document and this method will
   * join the words with spaces and call {@link #ptb2Text(String) } on the
   * output. This method will check if the elements in the list are subtypes
   * of Word, and if so, it will take the word() values to prevent additional
   * text from creeping in (e.g. POS tags). Otherwise the toString value will
   * be used.
   */
  public static String ptb2Text(List ptbWords)
  {
    for(int i=0;i<ptbWords.size();i++)
      if(ptbWords.get(i) instanceof Word)
        ptbWords.set(i,((Word)ptbWords.get(i)).word());

    return(ptb2Text(join(ptbWords)));
  }

}
