// StanfordLexicalizedParser -- a probabilistic lexicalized NL CFG parser
// Copyright (c) 2002, 2003 Leland Stanford Junior University
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// For information contact:
//    Christopher Manning
//    Dept of Computer Science, Gates 4A
//    Stanford CA 94305-9040
//    USA
//    manning@cs.stanford.edu

package edu.stanford.nlp.parser.lexparser;

import java.io.*;
import java.util.*;
import java.text.NumberFormat;

import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.util.*;
import edu.stanford.nlp.math.SloppyMath;
import edu.stanford.nlp.io.NumberRangeFileFilter;
import edu.stanford.nlp.io.EncodingPrintWriter;

/** @author Dan Klein (92% of the code)
 *  @author Christopher Manning (random improvements)
 *  @author Roger Levy (internationalization)
 *  @author Teg Grenager (grammars, tokenization)
 */

class Test {
  /* THESE OPTIONS AFFECT ONLY TEST TIME */

  /** If false, then failure of the PCFG factor to parse a sentence
   * will trigger parse recovery mode, in which more flexible tagging
   * is allowed. False by default. Use option -noRecoveryTagging to set
   * to true.
   */
  public static boolean noRecoveryTagging = false;
      
  /** If true, the n^4 "speed-up" is not used. */
  public static boolean useN5 = false;

  /** The maximum sentence length (including punctuation, etc.) to parse. */
  public static int maxLength = 40;

  /** Parse trees in test treebank in order of increasing length. */
  public static boolean increasingLength = false;

  /** Tag the sentences first, then parse given those (coarse) tags. */
  public static boolean preTag = false;
  /** Parse using only tags given from correct answer or the POS tagger */
  public static boolean forceTags = preTag;

  /** "double count" Tags in PCFG and Dep parser.  Good. Leave true. */
  public static boolean dcTags = true;

  /** Write EvalB-readable output files. */
  public static boolean evalb = false;

  /** Print a lot of extra output as you parse. */
  public static boolean verbose = false; // Don't change this; set with -v

  public static final boolean exhaustiveTest = false;
  public static final boolean pcfgThreshold = false;
  public static final double pcfgThresholdValue = -2.0;

  /** If true, remove any node which improves the PCFG score. */
  public static boolean nodePrune = false;

  /** Weighting on dependency log probs */
  public static final double depWeight = 1.0;
  public static boolean prunePunc = false;


  /** Determines format of output trees: choose among penn, oneline  */
  public static String outputTreeFormat = "penn";

  /** determines whether output trees have lexicalized categories displayed */
  public static boolean lexicalizeOutput = false;
  public static boolean printDepsSet = false;


  public static TreePrint treePrint() {
    TreePrint tp = new TreePrint(outputTreeFormat);
    tp.setLexicalize(lexicalizeOutput);
    tp.setPrintDepsSet(printDepsSet);
    return tp;
  }

  public static void display() {
    String str = "Test parameters maxLength="+maxLength+" dcTags="+dcTags+" nPrune="+nodePrune+" preTag="+preTag;
    System.out.println(str);
  }

} // end class Test


class Options implements Serializable {
  /* THESE OPTIONS MUST BE THE SAME AT TRAIN AND TEST TIME */

  private Options() {}  // not instantiable

  private static Options instance = new Options();
  public static Options get() { return instance; }
  public static void set(Options pt) { instance = pt; };

  /** Use suffix and capitalization information for unknowns. 
   *  0 means a single unknown token.  1 uses suffix, and capitalization.
   *  2 uses a variant (richer) form of signature.  Good.
   *  Use this one.  Using the richer signatures in versions 3 or 4 seems
   *  to have very marginal or no positive value.
   *  3 uses a richer form of signature that mimics the NER word type 
   *  patterns.  4 is avariant of 2.  5 is another that _may_ beat 2.
   */
  public int useUnknownWordSignatures = 0;

  /** Smarter smoothing for rare words. */
  public boolean smartMutation = false;


  /** Do a PCFG parse of the sentence.  If both variables are on,
   *  also do a combined parse of the sentence.
   */
  public boolean doPCFG = true;

  /** Do a dependency parse of the sentence. */
  public boolean doDep = true;

  /** Words more common than this are tagged with MLE P(t|w). Default 100. The
   *  smoothing is sufficiently slight that changing this has little effect. 
   */
  public int SMOOTH_IN_UNKNOWNS_THRESHOLD = 100;

  public TreebankLanguagePack langpack;

  /** if true, any child can be the head (seems rather bad!) */
  public boolean freeDependencies = false;

  /** Whether dependency grammar considers left/right direction. Good. */
  public boolean directional = true;
  public boolean genStop = true;

  /** Use distance bins in the dependency calculations */
  public boolean distance = true;
  /** Use coarser distance (4 bins) in dependency calculations */
  public boolean coarseDistance = false;

  /* Setting this option to true means that the parser generates ALL tags in
   * the lexicon for every word (some may still be scored at -Inf, but lots
   * of tags that it's never been seen with will have positive probability).
   * If false, it only generates seen tags, and if the word was of low
   * enough frequency in the training data, other tags deemed as open class.
   */
  public boolean flexiTag = false;

  public void display() {
    String str = "Options parameters dist="+distance+
	" coarseDist="+coarseDistance+
	" useUnknownWordSignatures="+useUnknownWordSignatures;
    System.out.println(str);
  }

  private static final long serialVersionUID = 1L;

} // end class Options


class Train {
  /* THESE OPTIONS AFFECT ONLY TRAIN TIME */

  /** if true, leave all PTB (functional tag) annotations (bad) */
  public static boolean leaveItAll = false;
  public static boolean cheatPCFG = false;

  public static boolean markovFactor = false;
  public static int markovOrder = 1;
  public static boolean hSelSplit = false; // good with true;
  public static int HSEL_CUT = 10;

  /** A POS tag has to have been attributed to more than this number of word
   *  types before it is regarded as an open-class tag.  Unknown words will
   *  only possibly be tagged as open-class tags.
   */
  public static int openClassTypesThreshold = 50;


  /** If true, declare early -- leave this on except maybe with markov on. */
  public static boolean outsideFactor() { return !markovFactor; }

  /** This variable controls doing parent annotation of phrasal nodes.  Good.
   */
  public static boolean PA = true;
  /** This variable controls doing 2 levels of parent annotation.  Bad.
   */
  public static boolean gPA = false;
  /** Only split the "common high KL divergence" categories.... Good. */
  public static boolean selectiveSplit = false; //true;

  /** Selective Sister annotation */
  public static boolean sisterAnnotate = false;

  public static Set sisterSplitters;

  /** Mark all unary nodes specially.  Good for just PCFG. Bad for factored. (1 better than 2 in combos) */
  public static boolean markUnary = false;//true;
  public static boolean markUnary2 = false;
  public static boolean markUnaryTags = false;


  /** Mark all pre-preterminals (also does splitBaseNP: don't need both) */
  public static boolean splitPrePreT = false;


  /** Parent annotation on tags.  Good (for PCFG?) */
  public static boolean tagPA = false;//true;

  /** Do parent annotation on tags selectively.  Neutral, but less splits. */
  public static boolean tagSelectiveSplit = false;

  /** Verbal distance -- mark whether symbol dominates a verb (V*, MD).
   *  Very good. */
  public static boolean dominatesV = false;//true;
  /** Verbal distance -- mark whether symbol dominates a preposition (IN) */
  public static boolean dominatesI = false;
  /** Verbal distance -- mark whether symbol dominates a conjunction (CC) */
  public static boolean dominatesC = false;

  /** Right edge has a phrasal node.  Bad? */
  public static boolean rightPhrasal = false;

  /** Right edge is right-recursive (X << X) Bad. (NP only is good) */
  public static boolean rightRec = false;//true;

  /** Left edge is right-recursive (X << X)  Bad. */
  public static boolean leftRec = false;

  /** X over X is marked (subsumes baseNP marking) Bad. */
  public static boolean xOverX = false;

  /** Promote/delete punctuation like Collins.  Bad (!) */
  public static boolean collinsPunc = false;

  /** Set the splitter strings -- mainly used by language-specific
   * parser parameterizations */
  public static Set splitters;

  /** Just for debugging: check that your tree transforms work right */
  public static boolean printTreeTransformations = false;


  /** CHANGE ANYTHING BELOW HERE AT YOUR OWN RISK */

  public static boolean smoothing = false;
  public static boolean smoothedBound = false;

  /*  public static boolean factorOut = false;
  public static boolean rightBonus = false;
  public static boolean brokenDep = false;*/

  public static double ruleDiscount = 0.0;

  //public static boolean outsideFilter = false;



  public static void display() {
    String testStr = "Train parameters smooth="+smoothing+" PA="+PA+" GPA="+gPA+" selSplit="+selectiveSplit+" mUnary="+markUnary+" mUnary2="+markUnary2+" mUnaryTags="+markUnaryTags+" sPPT="+splitPrePreT+" tagPA="+tagPA+" tagSelSplit="+tagSelectiveSplit+" dominatesV="+dominatesV+" dominatesI="+dominatesI+" dominatesC="+dominatesC+" rightPhrasal="+rightPhrasal+" rightRec="+rightRec+" leftRec="+leftRec+" xOverX="+xOverX+" collinPunc="+collinsPunc+" markov="+markovFactor+" mOrd="+markovOrder+" hSelSplit="+hSelSplit+" ("+HSEL_CUT+")";

    System.out.println(testStr);
    //System.out.println("Splitting categories: " + splitters);
  }

} // end class Train



interface Extractor {
  public Object extract(Object object);
}


////  XXXX.  This class seems to never be used.  Delete it?
////  Change treeToEvents in Lexicon so it uses one tree like this though?
abstract class AbstractTreeEventExtractor implements Extractor {
  protected void extractTree(Tree tree, List events) {}
  public Object extract(Object object) {
    Collection trees = (Collection)object;
    List events = new ArrayList();
    for (Iterator i=trees.iterator(); i.hasNext();) {
      Tree tree = (Tree)i.next();
      extractTree(tree,events);
    }
    return events;
  }
}


abstract class AbstractTreeExtractor implements Extractor {

  protected void tallyLeaf(Tree lt) {}
  protected void tallyPreTerminal(Tree lt) {}
  protected void tallyInternalNode(Tree lt) {}
  protected void tallyRoot(Tree lt) {}
  protected Object formResult() { return null; }

  protected void tallyLocalTree(Tree lt) {
    if (lt.isLeaf()) {
      tallyLeaf(lt);
    } else if (lt.isPreTerminal()) {
      tallyPreTerminal(lt);
    } else {
      tallyInternalNode(lt);
    }
  }

  protected void tallyTree(Tree t) {
    tallyRoot(t);
    for (Iterator tI = t.subTrees().iterator(); tI.hasNext();) {
      Tree localTree = (Tree)tI.next();
      tallyLocalTree(localTree);
    }
  }

  protected void tallyTreeList(List treeList) {
    for (int i=0; i<treeList.size(); i++) {
      Tree tree = (Tree)treeList.get(i);
      tallyTree(tree);
    }
  }

  public void processTree(Tree tree) {
    tallyTree(tree);
  }

  public Object extract() {
    return formResult();
  }

  public Object extract(Object treeList) {
    tallyTreeList((List)treeList);
    return formResult();
  }
}


// LEXICON STUFF

class Lexicon implements Serializable {

  public static final String UNKNOWN_WORD = "UNK";
  public static final String BOUNDARY = ".$.";
  public static final String BOUNDARY_TAG = "*";

  // An array of Lists of rules, indexed by word
  protected transient List[] rulesWithWord;

  protected transient Set rules = new HashSet();
  protected transient Set tags = new HashSet();
  protected transient Set words = new HashSet();
  protected transient Set sigs = new HashSet();

  protected Counter seenCounter = new Counter();
  protected Counter unSeenCounter = new Counter();

  protected static final int nullWord = -1;
  protected static final short nullTag = -1;

  /** We cache the last signature looked up, because it asks for the same one
   *  many times when an unknown word is encountered!  (Note that under the
   *  current scheme, one unknown word, if seen sentence-initially and
   *  non-initially, will be parsed with two different signatures....)
   */
  private transient int lastSignatureIndex = -1;
  private transient int lastSentencePosition = -1;
  private transient int lastWordToSignaturize = -1;

  double[] smooth = {1.0,1.0};

  transient double[][] m_TT = null;
  transient double[] m_T = null;
    
  private transient Numberer tagNumberer = null;
  private transient Numberer wordNumberer = null;

  public Numberer tagNumberer() {
    if (tagNumberer == null)
      tagNumberer = Numberer.getGlobalNumberer("tags");
    return tagNumberer;
  }

  public Numberer wordNumberer() {
    if (wordNumberer == null)
      wordNumberer = Numberer.getGlobalNumberer("words");
    return wordNumberer;
  }

  public Iterator ruleIterator() {
    return rules.iterator();
  }

  public boolean isKnown(int word) {
    if (rulesWithWord == null)
      initRulesWithWord();
    return (word < rulesWithWord.length &&
	    rulesWithWord[word] != null &&
	    rulesWithWord[word].size() > 0);
  }

  public Iterator ruleIteratorByWord(int word, int loc) {
    if (rulesWithWord == null)
      initRulesWithWord();
    if (isKnown(word)) {
      Iterator i = rulesWithWord[word].iterator();
      if (i.hasNext())
        return i;
    }
    List newTaggings = new ArrayList(40);
    for (Iterator i=rulesWithWord[Numberer.number("words",Lexicon.UNKNOWN_WORD)].iterator(); i.hasNext();) {
      IntTaggedWord iTW = (IntTaggedWord)i.next();
      newTaggings.add(new IntTaggedWord(word,iTW.tag));
    }
    //System.out.println("UNKNOWN WORD ("+Numberer.number("words",Lexicon.UNKNOWN_WORD)+") has taggings: "+newTaggings);
    return newTaggings.iterator();
  }

  protected void initRulesWithWord() {
    //int numWords = words.size()+sigs.size()+1;
    int numWords = wordNumberer().total();
    rulesWithWord = new List[numWords];
    for (int w = 0; w < numWords; w++) {
      rulesWithWord[w] = new ArrayList(2);   // most have 1 or 2 items in them
    }
    //for (Iterator ruleI = rules.iterator(); ruleI.hasNext();) {
    Set tags = new HashSet();
    for (Iterator ruleI = seenCounter.keySet().iterator(); ruleI.hasNext();) {
      IntTaggedWord iTW = (IntTaggedWord)ruleI.next();
      if (iTW.tag() == nullTag)
	continue;
      if (iTW.word() == nullWord) {
	tags.add(iTW);
	continue;
      }
      //IntTaggedWord iW = new IntTaggedWord(iTW.word, nullTag);
      //if (seenCounter.countOf(iTW) > 2 || seenCounter.countOf(iW) < 1000) {
      rulesWithWord[iTW.word].add(iTW);
      //}
    }
    // tags for unknown words
    //System.out.println("Considering UNKNOWN WORD");
    for (Iterator tagI = tags.iterator(); tagI.hasNext(); ) {
      IntTaggedWord iT = (IntTaggedWord) tagI.next();
      double types = unSeenCounter.countOf(iT);
      if (types > Train.openClassTypesThreshold) { 
	// Number of types before it's treated as open class
        IntTaggedWord iTW = 
	  new IntTaggedWord(Numberer.number("words",Lexicon.UNKNOWN_WORD),
			    iT.tag);
	//System.out.println("Approved UNKNOWN WORD ("+Numberer.number("words",Lexicon.UNKNOWN_WORD)+") tag "+iTW+" (tag "+iTW.tag()+") length is "+rulesWithWord[iTW.word]);
        rulesWithWord[iTW.word].add(iTW);
      }
    }
  }

  private List treeToEvents(Tree tree) {
    List taggedWords = tree.taggedYield();
    List itwList = new ArrayList();
    for (int i=0; i<taggedWords.size(); i++) {
      TaggedWord tw = (TaggedWord)taggedWords.get(i);
      IntTaggedWord iTW = new IntTaggedWord(Numberer.number("words",tw.word()),
                                            Numberer.number("tags",tw.tag()));
      itwList.add(iTW);
    }
    return itwList;
  }

  private List treeToEvents(Collection trees) {
    List iTW = new ArrayList();
    for(Iterator i =trees.iterator(); i.hasNext() ;) {
      iTW.addAll(treeToEvents((Tree)i.next()));
    }
    return iTW;
  }

    /**
     * Trains this lexicon on the Collection of trees.
     */
  public void train(Collection trees) {
    // scan data
    int tNum = 0;
    for (Iterator tI = trees.iterator(); tI.hasNext();) {
      Tree tree = (Tree)tI.next();
      tNum++;
      List taggedWords = treeToEvents(tree);
      for (int w=0; w<taggedWords.size(); w++) {
        IntTaggedWord iTW = (IntTaggedWord)taggedWords.get(w);
        seenCounter.add(iTW,1.0);
        IntTaggedWord iT = new IntTaggedWord(nullWord, iTW.tag);
        seenCounter.add(iT,1.0);
        IntTaggedWord iW = new IntTaggedWord(iTW.word, nullTag);
        seenCounter.add(iW,1.0);
        IntTaggedWord i = new IntTaggedWord(nullWord, nullTag);
        seenCounter.add(i,1.0);
        rules.add(iTW);
        tags.add(iT);
        words.add(iW);
        if (tNum > trees.size()/2) {
	    // start doing this once we're halfway through the trees
          if (seenCounter.countOf(iW) < 2) {
            // it's an entirely unknown word
            int s = getSignatureIndex(iTW.word, w);
            Numberer.number("words",Lexicon.UNKNOWN_WORD); //cdm: outside loop?
            IntTaggedWord iTS = new IntTaggedWord(s, iTW.tag);
            IntTaggedWord iS = new IntTaggedWord(s, nullTag);
            unSeenCounter.add(iTS, 1.0);
            unSeenCounter.add(iT, 1.0);
            unSeenCounter.add(iS, 1.0);
            unSeenCounter.add(i, 1.0);
            //rules.add(iTS);
            sigs.add(iS);
          } else {
            if (seenCounter.countOf(iTW) < 2) {
              // it's a new tag for a known word
              // do nothing for now
            }
          }
        }
      }
    }
    // index the possible tags for each word
    //numWords = wordNumberer.total();
    //unknownWordIndex = Numberer.number("words",Lexicon.UNKNOWN_WORD);
    //initRulesWithWord();
  }

    /**
     * Adds the tagging with count to the data structures in this Lexicon.
     */
    private void addTagging(boolean seen, IntTaggedWord itw, double count) {
	if (seen) {
	    seenCounter.add(itw, count);
	    if (itw.tag() == nullTag)
		words.add(itw);
	    else if (itw.word() == nullWord)
		tags.add(itw);
	    else
		rules.add(itw);
	} else {
	    unSeenCounter.add(itw, count);
	    if (itw.tag() == nullTag)
		sigs.add(itw);
	}
    }

    /**
     * Returns the index of the signature of the word numbered wordIndex,
     * where the signature is the String representation of unknown word 
     * features.  Caches the last signature index returned.
     */
  private int getSignatureIndex(int wordIndex, int sentencePosition) {
    if (wordIndex == lastWordToSignaturize &&
        sentencePosition == lastSentencePosition) {
      // System.err.println("Signature: cache mapped " + wordIndex + " to " + lastSignatureIndex);
      return lastSignatureIndex;
    } else {
      String uwSig = getSignature((String)Numberer.object("words", wordIndex),
                               sentencePosition);
      int sig = Numberer.number("words", uwSig);
      lastSignatureIndex = sig;
      lastSentencePosition = sentencePosition;
      lastWordToSignaturize = wordIndex;
      // System.err.println("Signature: getSignature mapped " + wordIndex + " to " + lastSignatureIndex);
      return sig;
    }
  }


  /** This routine returns a String that is the "signature" of the class of a
   *  word.
   *  For, example, it might represent whether it is a number of ends in -s.
   *  The strings returned by convention match the pattern UNK-.* , which
   *  is just assumed to not match any real word.
   *  @param word The word to make a signature for
   *  @param loc Its position in the sentence (mainly so sentence-initial
   *     capitalized words can be treated differently)
   *  @return A String that is its signature (equivalence class)
   */
  private String getSignature(String word, int loc) {
    int unknownLevel = Options.get().useUnknownWordSignatures;
    StringBuffer sb = new StringBuffer("UNK");
    switch (unknownLevel) {

    case 5: {
      // { -ALLC aps, -INIT cap, -CAP capitalized, -LC starts lower, NONL } +
      // { -NUM, -DIG, -NODIG } +
      // { -DASH, 0 } +
      // { -PER, 0 } +
      // { last char if lowercase, 0}
      sb.append("-");
      boolean hasDigit = false;
      boolean hasNonDigit = false;
      boolean hasLetter = false;
      boolean hasNonLetterDigit = false;
      boolean hasLower = false;
      for (int i = 0; i < word.length(); i++) {
        char ch = word.charAt(i);
        if (Character.isDigit(ch)) {
          hasDigit = true;
        } else {
          hasNonDigit = true;
          if (Character.isLetter(ch)) {
            hasLetter = true;
            if (Character.isLowerCase(ch) || Character.isTitleCase(ch)) {
              hasLower = true;
            }
          } else {
            hasNonLetterDigit = true;
          }
        }
      }
      char ch0 = word.charAt(0);
      if (Character.isUpperCase(ch0) || Character.isTitleCase(ch0)) {
        if ( ! hasLower) {
          sb.append("ALLC");
        } else if (loc == 0) {
          sb.append("INITC");
        } else {
          sb.append("CAP");
        }
      } else if (Character.isLowerCase(ch0)) {
        sb.append("LC");
      } else {
        sb.append("NONL");
      }
      if (hasDigit) {
        if ( ! hasNonDigit) {
          sb.append("-NUM");
        } else {
          sb.append("-DIG");
        }
      } else {
        sb.append("-NODIG");
      }
      if (word.indexOf('-') >= 0) {
        sb.append("-DASH");
      }
      if (word.indexOf('.') >= 0) {
        sb.append("-PER");
      }
      if (word.length() >= 3) {
        // don't do for very short words; but if you go beyond 3, then you
        // miss out on ones like 80s
        char ch = word.charAt(word.length() - 1);
        if (Character.isLowerCase(ch)) {
          sb.append(ch);
        }
      }
      break; }

    case 4: {
      boolean hasDigit = false;
      boolean hasNonDigit = false;
      boolean hasLetter = false;
      boolean hasNonLetter = false;
      boolean hasLower = false;
      boolean hasDash = false;
      boolean hasPeriod = false;
      boolean hasComma = false;
      for (int i = 0; i < word.length(); i++) {
        char ch = word.charAt(i);
        if (Character.isDigit(ch)) {
          hasDigit = true;
        } else {
          hasNonDigit = true;
          if (Character.isLetter(ch)) {
            hasLetter = true;
            if (Character.isLowerCase(ch) || Character.isTitleCase(ch)) {
              hasLower = true;
            }
          } else {
            hasNonLetter = true;
            if (ch == '-') {
              hasDash = true;
            } else if (ch == '.') {
              hasPeriod = true;
            } else if (ch == ',') {
              hasComma = true;
            }
          }
        }
      }
      // 6 way on letters
      if (Character.isUpperCase(word.charAt(0)) ||
          Character.isTitleCase(word.charAt(0))) {
        if ( ! hasLower) {
          sb.append("-AC");
        } else if (loc == 0) {
          sb.append("-SC");
        } else {
          sb.append("-C");
        }
      } else if (hasLower) {
        sb.append("-L");
      } else if (hasLetter) {
        sb.append("-U");
      } else {
        // no letter
        sb.append("-S");
      }
      // 3 way on number
      if (hasDigit && ! hasNonDigit) {
        sb.append("-N");
      } else if (hasDigit) {
        sb.append("-n");
      }
      // binary on period, dash, comma
      if (hasDash) {
        sb.append("-H");
      }
      if (hasPeriod) {
        sb.append("-P");
      }
      if (hasComma) {
        sb.append("-C");
      }
      if (word.length() > 3) {
        // don't do for very short words: "yes" isn't an "-es" word
        // try doing to lower for further densening and skipping digits
        char ch = word.charAt(word.length() - 1);
        if (Character.isLetter(ch)) {
          sb.append("-");
          sb.append(Character.toLowerCase(ch));
        }
      }
      break; }

    case 3: {
      // This basically works right, except note that 'S' is applied to all
      // capitalized letters in first word of sentence, not just first....
      sb.append("-");
      char lastClass = '-';  // i.e., nothing
      char newClass;
      int num = 0;
      for (int i = 0; i < word.length(); i++) {
        char ch = word.charAt(i);
        if (Character.isUpperCase(ch) || Character.isTitleCase(ch)) {
          if (loc == 0) {
            newClass = 'S';
          } else {
            newClass = 'L';
          }
        } else if (Character.isLetter(ch)) {
          newClass = 'l';
        } else if (Character.isDigit(ch)) {
          newClass = 'd';
        } else if (ch == '-') {
          newClass = 'h';
        } else if (ch == '.') {
          newClass = 'p';
        } else {
          newClass = 's';
        }
        if (newClass != lastClass) {
          lastClass = newClass;
          sb.append(lastClass);
          num = 1;
        } else {
          if (num < 2) {
            sb.append('+');
          }
          num++;
        }
      }
      if (word.length() > 3) {
        // don't do for very short words: "yes" isn't an "-es" word
        // try doing to lower for further densening and skipping digits
        char ch = Character.toLowerCase(word.charAt(word.length() - 1));
        sb.append('-');
        sb.append(ch);
      }
      break; }

    case 2: {
      boolean hasDigit = false;
      boolean hasNonDigit = false;
      boolean hasLetter = false;
      boolean hasNonLetter = false;
      boolean hasLower = false;
      for (int i = 0; i < word.length(); i++) {
        char ch = word.charAt(i);
        if (Character.isDigit(ch)) {
          hasDigit = true;
        } else {
          hasNonDigit = true;
          if (Character.isLetter(ch)) {
            hasLetter = true;
            if (Character.isLowerCase(ch) || Character.isTitleCase(ch)) {
              hasLower = true;
            }
          } else {
            hasNonLetter = true;
          }
        }
      }
      if (Character.isUpperCase(word.charAt(0)) ||
          Character.isTitleCase(word.charAt(0))) {
        if ( ! hasLower) {
          sb.append("-ALLC");
        } else if (loc == 0) {
          sb.append("-INIT");
        } else {
          sb.append("-UC");
        }
      } else if (hasLower) {   // if (Character.isLowerCase(word.charAt(0))) {
        sb.append("-LC");
      }
      // no suffix = no (lowercase) letters
      if (word.indexOf('-') >= 0) {
        sb.append("-DASH");
      }
      if (hasDigit) {
        if ( ! hasNonDigit) {
          sb.append("-NUM");
        } else {
          sb.append("-DIG");
        }
      } else if (word.length() > 3) {
        // don't do for very short words: "yes" isn't an "-es" word
        // try doing to lower for further densening and skipping digits
        char ch = word.charAt(word.length() - 1);
        sb.append(Character.toLowerCase(ch));
      }
      // no suffix = short non-number
      break; }

    default:
      sb.append("-");
      sb.append(word.substring(Math.max(word.length()-2,0),word.length()));
      sb.append("-");
      if (Character.isLowerCase(word.charAt(0))) {
	sb.append("LOWER");
      } else {
	if (Character.isUpperCase(word.charAt(0))) {
	  if (loc == 0)
	    sb.append("INIT");
	  else
	    sb.append("UPPER");
	} else
	  sb.append("OTHER");
      }
    } // end switch (unknownLevel)
    // System.err.println("Summarized " + word + " to " + sb.toString());
    return sb.toString();
  } // end getSignature()


  void buildPT_T() {
    Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
    int numTags = tagNumberer.total();
    m_TT = new double[numTags][numTags];
    m_T = new double[numTags];
    double[] tmp = new double[numTags];
    for (Iterator i=words.iterator(); i.hasNext();) {
      IntTaggedWord word = (IntTaggedWord)i.next();
      IntTaggedWord iTW = new IntTaggedWord(word.word, nullTag);
      double tot = 0.0;
      for (int t=0; t<numTags; t++) {
        iTW.tag = (short)t;
        tmp[t] = seenCounter.countOf(iTW);
        tot += tmp[t];
      }
      if (tot < 10)
        continue;
      for (int t=0; t<numTags; t++) {
        for (int t2=0; t2<numTags; t2++) {
          if (tmp[t2] > 0.0) {
            double c = tmp[t]/tot;
            m_T[t] += c;
            m_TT[t2][t] += c;
          }
        }
      }
    }
  }

  public double score(Object event, int loc) {
    IntTaggedWord iTW = (IntTaggedWord)event;
    int word = iTW.word;
    short tag = iTW.tag;

    iTW.tag = nullTag;
    double c_W = seenCounter.countOf(iTW);
    //double x_W = xferCounter.countOf(iTW);
    iTW.tag = tag;

    double pb_W_T = 0.0;

    boolean seen = (c_W > 0.0);
    if (seen) {

      // known word model for P(T|W)

      double c_TW = seenCounter.countOf(iTW);
      //double x_TW = xferCounter.countOf(iTW);
      iTW.word = nullWord;
      double c_T = seenCounter.countOf(iTW);
      double c_Tunseen = unSeenCounter.countOf(iTW);
      iTW.tag = nullTag;
      double total = seenCounter.countOf(iTW);
      double totalUnseen = unSeenCounter.countOf(iTW);
      iTW.tag = tag;
      iTW.word = word;

      //c_TW = Math.sqrt(c_TW);
      //c_TW += 0.5;

      double p_T_U = c_Tunseen/totalUnseen;
      double pb_T_W = 0.0;

      // System.err.println("c_W is " + c_W + " THRESH is " + Options.get().SMOOTH_IN_UNKNOWNS_THRESHOLD + " mle = " + (c_TW/c_W));
      if (c_W > Options.get().SMOOTH_IN_UNKNOWNS_THRESHOLD) {
        pb_T_W = c_TW/c_W;
      } else {
        if (Options.get().smartMutation) {
          Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
          int numTags = tagNumberer.total();
          if (m_TT == null || numTags != m_T.length)
            buildPT_T();
          p_T_U *= 0.1;
          //System.out.println("Checking "+iTW);
          for (int t=0; t<numTags; t++) {
            IntTaggedWord iTW2 = new IntTaggedWord(word, t);
            double p_T_W2 = seenCounter.countOf(iTW2)/c_W;
            if (p_T_W2 > 0) {
              //System.out.println(" Observation of "+tagNumberer.object(t)+" ("+seenCounter.countOf(iTW2)+") mutated to "+tagNumberer.object(iTW.tag)+" at rate "+(m_TT[tag][t]/m_T[t]));
              p_T_U += p_T_W2*m_TT[tag][t]/m_T[t]*0.9;
            }
          }
        }
        pb_T_W = (c_TW+smooth[1]*p_T_U)/(c_W+smooth[1]);
      }

      //double pb_T_W = (c_TW+smooth[1]*x_TW)/(c_W+smooth[1]*x_W);

      // 1000 / 2

      double p_T = (c_T/total);
      double p_W = (c_W/total);
      pb_W_T = Math.log(pb_T_W * p_W / p_T);

    } else {

      // unknown word model for P(T|S)
      int sig = getSignatureIndex(iTW.word, loc);

      iTW.word = sig;
      double c_TS = unSeenCounter.countOf(iTW);
      iTW.tag = nullTag;
      double c_S = unSeenCounter.countOf(iTW);
      iTW.word = nullWord;
      double c_U = unSeenCounter.countOf(iTW);
      double total = seenCounter.countOf(iTW);
      iTW.tag = tag;
      double c_T = unSeenCounter.countOf(iTW);
      double c_Tseen = seenCounter.countOf(iTW);
      iTW.word = word;

      double p_T_U = c_T/c_U;
      if (Options.get().useUnknownWordSignatures == 0) {
        c_TS = 0;
        c_S = 0;
      }
      double pb_T_S = (c_TS+smooth[0]*p_T_U)/(c_S+smooth[0]);

      double p_T = (c_Tseen/total);
      double p_W = 1.0/total;
      pb_W_T = Math.log(pb_T_S * p_W / p_T);
      //System.out.println(event+" as "+pb_W_T);

    }


    if (pb_W_T > -100.0)
      return pb_W_T;
    return Double.NEGATIVE_INFINITY;
  } // end score()

  protected double scoreAll(List trees) {
    double totalScore = 0.0;
    for (Iterator tI = trees.iterator(); tI.hasNext();) {
      Tree tree = (Tree)tI.next();
      List taggedWords = treeToEvents(tree);
      for (int i=0; i<taggedWords.size(); i++) {
        IntTaggedWord iTW = (IntTaggedWord)taggedWords.get(i);
        double score = score(iTW,i);
        if (score > Double.NEGATIVE_INFINITY)
          totalScore += score;
      }
    }
    return totalScore;
  }

  public void tune(List trees) {
    double bestScore = Double.NEGATIVE_INFINITY;
    double[] bestSmooth = {0.0,0.0};
    for (smooth[0]=1; smooth[0]<=1; smooth[0] *= 2.0) {//64
      for (smooth[1]=0.2; smooth[1]<=0.2; smooth[1] *= 2.0) {//3
        //for (smooth[0]=0.5; smooth[0]<=64; smooth[0] *= 2.0) {//64
        //for (smooth[1]=0.1; smooth[1]<=12.8; smooth[1] *= 2.0) {//3
        double score = 0.0;
        //score = scoreAll(trees);
        if (Test.verbose)
          System.out.println("Tuning lexicon: s0 "+smooth[0]+" s1 "+smooth[1]+" is "+score+" "+trees.size()+" trees.");
        if (score > bestScore) {
          System.arraycopy(smooth, 0, bestSmooth, 0, smooth.length);
          bestScore = score;
        }
      }
    }
    System.arraycopy(bestSmooth, 0, smooth, 0, bestSmooth.length);
    if (Options.get().smartMutation) {
      smooth[0] = 8.0;
      //smooth[1] = 1.6;
      //smooth[0] = 0.5;
      smooth[1] = 0.1;
    }
    if (Test.verbose)
      System.out.println("Tuning selected s0 "+smooth[0]+" s1 "+smooth[1]+" at "+bestScore);
  }

  public void printLexStats() {
    System.out.println("Rules size: " + rules.size());
    System.out.println("Tags size: " + tags.size());
    System.out.println("Words size: " + words.size());
    System.out.println("Sigs size: " + sigs.size());
    System.out.println("rulesWithWord length: " + rulesWithWord.length);
    for (int i = 0; i < 10; i++) {
      int j = (rulesWithWord.length / 10) * i;
      System.out.println("  rulesWithWord[" + j +"] size: " +
                         rulesWithWord[j].size());
    }
  }

  /** Evaluates how many words (= terminals) in a collection of trees
   *  are covered by the lexicon. First arg is the collection of
   *  trees; second through fourth args get the results.
   */
  public double evaluateCoverage(Collection trees, Set missingWords, 
				 Set missingTags, Set missingTW) {

    List iTW = treeToEvents(trees);
    double total = 0;
    double unseen = 0;

    for(Iterator i=iTW.iterator(); i.hasNext() ;) {
      total++;
      IntTaggedWord itw = (IntTaggedWord)i.next();
      if(! words.contains(new IntTaggedWord(itw.word(),nullTag))) {
        missingWords.add(Numberer.object("word",itw.word()));
      }
      if(! tags.contains(new IntTaggedWord(nullWord,itw.tag()))) {
        missingTags.add(Numberer.object("tag",itw.tag()));
      }
      if(! rules.contains(itw)) {
        unseen++;
        missingTW.add(itw);
      }
    }
    return (double) unseen / (double) total;
  }


  public String showTags() {
    return tags.toString();
  }

  private void readObject(ObjectInputStream stream) 
    throws IOException, ClassNotFoundException {
    stream.defaultReadObject();
    //initRulesWithWord();
  }

    /**
     * Populates data in this Object from the character stream 
     * given by the Reader r.
     */
    public void readData(Reader r) throws IOException {
	final String SEEN = "SEEN";
	BufferedReader in = new BufferedReader(r);
	String line, seenString, itwString, countString;
	int lineNum = 1;
	StringTokenizer tok;
	boolean seen;
	// all lines have one tagging with raw count per line
	while ((line = in.readLine()) != null) {
	    try {
		tok = new StringTokenizer(line); // break on whitespace only
		seenString = tok.nextToken();
		if (seenString.equals(SEEN))
		    seen = true;
		else
		    seen = false;
		itwString = tok.nextToken();
		countString = tok.nextToken();
		addTagging(seen, new IntTaggedWord(itwString), Double.parseDouble(countString));
	    } catch (Exception e) {
		throw new IOException("Error on line " + lineNum);
	    }	    
	    lineNum++;   
	}
	//initRulesWithWord();
    }

    /**
     * Writes out data from this Object to the Writer w.
     */
    public void writeData(Writer w) throws IOException {
	PrintWriter out = new PrintWriter(w);
	IntTaggedWord itw;
	String tagString, wordString;
	// all lines have one rule per line
	for (Iterator sI = seenCounter.keySet().iterator(); sI.hasNext();) {
	    itw = (IntTaggedWord) sI.next();
	    out.println("SEEN " + itw + " " + seenCounter.getCount(itw));
	}
	for (Iterator uI = unSeenCounter.keySet().iterator(); uI.hasNext();) {
	    itw = (IntTaggedWord) uI.next();
	    out.println("UNSEEN " + itw + " " + unSeenCounter.getCount(itw));
	}
	out.flush();
    }

  private static final long serialVersionUID = 5873L;

} // end class Lexicon




class FA {
  protected boolean[] inStatePrev;
  protected boolean[] inStateNext;
  protected boolean[] loopState;
  protected int acceptingState;
  protected int initialState;
  protected int numStates;
  protected int numSymbols;
  protected int[][] transition; // state x tag

  public void init() {
    Arrays.fill(inStatePrev,false);
    Arrays.fill(inStateNext,false);
    inStatePrev[initialState] = true;
  }
  public void input(int symbol) {
    for (int prevState=0; prevState<numStates; prevState++) {
      if (inStatePrev[prevState])
        inStateNext[transition[prevState][symbol]] = true;
    }
  }
  public void advance() {
    boolean[] temp = inStatePrev;
    inStatePrev = inStateNext;
    inStateNext = temp;
    Arrays.fill(inStateNext,false);
    for (int state=0; state<numStates; state++) {
      if (inStatePrev[state] && loopState[state])
        inStateNext[state] = true;
    }
  }
  public boolean isAccepting() {
    return inStatePrev[acceptingState];
  }
  public void setTransition(int state, int symbol, int result) {
    transition[state][symbol] = result;
  }
  public void setLoopState(int state, boolean loops) {
    loopState[state] = loops;
  }
  public FA(int numStates, int numSymbols) {
    this.numStates = numStates;
    this.numSymbols = numSymbols;
    acceptingState = numStates-1;
    inStatePrev = new boolean[numStates];
    inStateNext = new boolean[numStates];
    loopState = new boolean[numStates];
    transition = new int[numStates][numSymbols];
  }

} // end class FA


class OutsideRuleFilter {

  private Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  private Numberer stateNumberer = Numberer.getGlobalNumberer("states");
  private int numTags;
  private int numFAs;

  protected FA[] leftFA;
  protected FA[] rightFA;

  protected List reverse(List list) {
    List reverse = new ArrayList(list.size());
    for (int i=list.size()-1; i>=0; i--) {
      reverse.add(list.get(i));
    }
    return reverse;
  }
  protected FA buildFA(List tags) {
    FA fa = new FA(tags.size()+1,numTags);
    fa.setLoopState(0,true);
    for (int state=1; state<=tags.size(); state++) {
      Object tagO = tags.get(state-1);
      if (tagO == null) {
        fa.setLoopState(state,true);
        for (int symbol=0; symbol<numTags; symbol++)
          fa.setTransition(state-1,symbol,state);
      } else {
        int tag = tagNumberer.number(tagO);
        fa.setTransition(state-1,tag,state);
      }
    }
    return fa;
  }
  protected void registerRule(List leftTags, List rightTags, int state) {
    leftFA[state] = buildFA(leftTags);
    rightFA[state] = buildFA(reverse(rightTags));
  }

  public void init() {
    for (int rule=0; rule<numFAs; rule++) {
      leftFA[rule].init();
      rightFA[rule].init();
    }
  }

  public void advanceRight(boolean[] tags) {
    for (int tag=0; tag<numTags; tag++) {
      if (!tags[tag])
        continue;
      for (int rule=0; rule<numFAs; rule++)
        leftFA[rule].input(tag);
    }
    for (int rule=0; rule<numFAs; rule++)
      leftFA[rule].advance();
  }

  public void leftAccepting(boolean[] result) {
    for (int rule=0; rule<numFAs; rule++)
      result[rule] = leftFA[rule].isAccepting();
  }

  public void advanceLeft(boolean[] tags) {
    for (int tag=0; tag<numTags; tag++) {
      if (!tags[tag])
        continue;
      for (int rule=0; rule<numFAs; rule++)
        rightFA[rule].input(tag);
    }
    for (int rule=0; rule<numFAs; rule++)
      rightFA[rule].advance();
  }

  public void rightAccepting(boolean[] result) {
    for (int rule=0; rule<numFAs; rule++)
      result[rule] = rightFA[rule].isAccepting();
  }

  private void allocate(int numFAs) {
    this.numFAs = numFAs;
    leftFA = new FA[numFAs];
    rightFA = new FA[numFAs];
  }

  public OutsideRuleFilter(BinaryGrammar bg) {
    int numStates = stateNumberer.total();
    numTags = tagNumberer.total();
    allocate(numStates);
    for (int state=0; state<numStates; state++) {
      String stateStr = (String)stateNumberer.object(state);
      List left = new ArrayList();
      List right = new ArrayList();
      if (!bg.isSynthetic(state)) {
        registerRule(left,right,state);
        continue;
      }
      boolean foundSemi = false;
      boolean foundDots = false;
      List array = left;
      StringBuffer sb = new StringBuffer();
      for (int c=0; c<stateStr.length(); c++) {
        if (stateStr.charAt(c)==':') {
          foundSemi = true;
          continue;
        }
        if (! foundSemi)
          continue;
        if (stateStr.charAt(c)==' ') {
          if (sb.length() > 0) {
            String str = sb.toString();
            if (! tagNumberer.hasSeen(str))
              str = null;
            array.add(str);
            sb = new StringBuffer();
          }
          continue;
        }
        if (! foundDots && stateStr.charAt(c)=='.') {
          c+=3;
          foundDots = true;
          array = right;
          continue;
        }
        sb.append(stateStr.charAt(c));
      }
      registerRule(left, right, state);
    }
  }

} // end class OutsideRuleFilter


// Binarization, Debinarization

class TreeHeadPair {
  public Tree tree; // = null;
  public TaggedWord head; // = null;
}


class BinarizerFactory {

  // Binarizes such that head-argument structure is respected
  // -- Looks only at the value of input tree nodes
  // -- Produces LSTrees with CWT labels.
  static class HeadRespectingBinarizer implements TreeTransformer {

    private HeadFinder hf;
    private TreeFactory tf;
    private TreebankLanguagePack tlp = Options.get().langpack;
    private boolean insideFactor; // true: DT JJ NN -> DT "JJ NN", false: DT "DT"
    private boolean doSubcategorization; // do the subcategorizing phase

    private static TreebankLangParserParams tlpParams;
    public static void setTreebankLang(TreebankLangParserParams tlpp) {
      tlpParams = tlpp;
    }

    boolean countStates = false;
    Counter stateCounter = new Counter();

    private String join(List treeList) {
      StringBuffer sb = new StringBuffer();
      for (Iterator i=treeList.iterator(); i.hasNext();) {
        Tree t = (Tree)i.next();
        sb.append(t.label().value());
        if (i.hasNext())
          sb.append(" ");
      }
      return sb.toString();
    }

    private String local(Tree t) {
      StringBuffer sb = new StringBuffer();
      for (int c=0; c<t.children().length; c++) {
        sb.append("<");
        sb.append(t.children()[c].label().value());
        sb.append("> ");
      }
      return sb.toString();
    }

    Tree binarizeTree(Tree t, int headNum, TaggedWord head) {
      //System.out.println("Working on: "+headNum+" -- "+t.label());
      if (Train.markovFactor) {
        String topCat = t.label().value();
        t.setLabel(new CategoryWordTag(topCat, head.word(), head.tag()));
        Tree t2 = null;
        if (insideFactor)
          t2 = markovInsideBinarizeLocalTree(t, head, headNum, topCat, false);
        else
          t2 = markovOutsideBinarizeLocalTree(t, head, headNum, topCat, new LinkedList(), false);
        if (false) {
          System.out.println("Made: "+local(t)+"\ninto: "+local(t2));
          if (!t2.children()[0].isLeaf()) {
            System.out.println("then0: "+local(t2.children()[0]));
          }
          if (t2.children().length > 1 && !t2.children()[1].isLeaf()) {
            System.out.println("then1: "+local(t2.children()[1]));
          }
          System.out.println("Tree: "+t2);
        }
        return t2;
      }
      if (insideFactor)
        return insideBinarizeLocalTree(t, headNum, head, 0, 0);
      return outsideBinarizeLocalTree(t, t.label().value(), t.label().value(),
                                      headNum, head, 0, "", 0, "");
    }

    private Tree markovOutsideBinarizeLocalTree(Tree t, TaggedWord head, int headLoc, String topCat, LinkedList ll, boolean doneLeft) {
      int markovOrder = Train.markovOrder;
      String word = head.word();
      String tag = head.tag();
      List newChildren = new ArrayList(2);
      // call with t, headNum, head, topCat, false
      if (headLoc == 0) {
        if (! doneLeft) {
          // insert a unary to separate the sides
          if (tlp.isStartSymbol(topCat))
            return markovOutsideBinarizeLocalTree(t, head, headLoc, topCat, new LinkedList(), true);
          String headStr = t.children()[headLoc].label().value();
          String subLabelStr = "@"+topCat+": "+headStr+" ]";
          Label subLabel = new CategoryWordTag(subLabelStr, word, tag);
          Tree subTree = tf.newTreeNode(subLabel, t.getChildrenAsList());
          newChildren.add(markovOutsideBinarizeLocalTree(subTree, head, headLoc, topCat, new LinkedList(), true));
        return tf.newTreeNode(t.label(),newChildren);

        }
        int len = t.children().length;
        // len = 1
        if (len == 1) {
          return tf.newTreeNode(t.label(),
                                Collections.singletonList(t.children()[0]));
        }
        ll.addFirst(t.children()[len-1]);
        if (ll.size() > markovOrder)
          ll.removeLast();
        // generate a right
        String headStr = t.children()[headLoc].label().value();
        String rightStr = (len > markovOrder-1 ? "... " : "")+join(ll);
        String subLabelStr = "@"+topCat+": "+headStr+" "+rightStr;
        Label subLabel = new CategoryWordTag(subLabelStr, word, tag);
        Tree subTree = tf.newTreeNode(subLabel, t.getChildrenAsList().subList(0,len-1));
        newChildren.add(markovOutsideBinarizeLocalTree(subTree, head, headLoc, topCat, ll, true));
        newChildren.add(t.children()[len-1]);
        return tf.newTreeNode(t.label(),newChildren);
      }
      if (headLoc > 0) {
        ll.addLast(t.children()[0]);
        if (ll.size() > markovOrder)
          ll.removeFirst();
        // generate a left
        String headStr = t.children()[headLoc].label().value();
        String leftStr = join(ll)+(headLoc > markovOrder-1 ? " ..." : "");
        String subLabelStr = "@"+topCat+": "+leftStr+" "+headStr+" ]";
        Label subLabel = new CategoryWordTag(subLabelStr, word, tag);
        Tree subTree = tf.newTreeNode(subLabel, t.getChildrenAsList().subList(1,t.children().length));
        newChildren.add(t.children()[0]);
        newChildren.add(markovOutsideBinarizeLocalTree(subTree, head, headLoc-1, topCat, ll, false));
        return tf.newTreeNode(t.label(),newChildren);
      }
      return t;
    }

    private Tree markovInsideBinarizeLocalTree(Tree t, TaggedWord head, int headLoc, String topCat, boolean doneLeft) {
      int markovOrder = Train.markovOrder;
      String word = head.word();
      String tag = head.tag();
      List newChildren = new ArrayList(2);
      // call with t, headNum, head, topCat, false
      if (headLoc == 0) {
        int len = t.children().length;
        // len = 1
        if (len == 1) {
          return tf.newTreeNode(t.label(),
                                Collections.singletonList(t.children()[0]));
        }
        // generate a right
        String headStr = t.children()[headLoc].label().value();
        int end = len-1;
        int start = Math.max(len-markovOrder-1, 1);
        List subList = t.getChildrenAsList().subList(start, end);
        String rightStr = (start > 1 ? "... " : "")+join(subList);
        String subLabelStr = "@"+topCat+": "+headStr+" "+rightStr;
        if (Train.hSelSplit) {
          if (countStates) {
            stateCounter.add(subLabelStr, 1.0);
          } else {
            double stateCount = stateCounter.countOf(subLabelStr);
            if (stateCount < Train.HSEL_CUT) {
              start = Math.max(len-markovOrder, 1);
              subList = t.getChildrenAsList().subList(start, end);
              rightStr = (start > 1 ? "... " : "")+join(subList);
              //System.out.print("Dumped "+subLabelStr+" ("+stateCount+")");
              subLabelStr = "@"+topCat+": "+headStr+" "+rightStr;
              //System.out.println("for "+subLabelStr);
            }
          }
        }
        Label subLabel = new CategoryWordTag(subLabelStr, word, tag);
        Tree subTree = tf.newTreeNode(subLabel, t.getChildrenAsList().subList(0,len-1));
        newChildren.add(markovInsideBinarizeLocalTree(subTree, head, headLoc, topCat, true));
        newChildren.add(t.children()[len-1]);
        return tf.newTreeNode(t.label(),newChildren);
      }
      if (headLoc > 0) {
        // generate a left
        String headStr = t.children()[headLoc].label().value();
        int start = 1;
        int end = Math.min(markovOrder+1, headLoc);
        List subList = t.getChildrenAsList().subList(start, end);
        String leftStr = join(subList)+(end < headLoc ? " ..." : "");
        String subLabelStr = "@"+topCat+": "+leftStr+" "+headStr+" ]";
        if (Train.hSelSplit) {
          if (countStates) {
            stateCounter.add(subLabelStr, 1.0);
          } else {
            double stateCount = stateCounter.countOf(subLabelStr);
            if (stateCount < Train.HSEL_CUT) {
              end = Math.min(markovOrder, headLoc);
              subList = t.getChildrenAsList().subList(start, end);
              leftStr = join(subList)+(end < headLoc ? " ..." : "");
              subLabelStr = "@"+topCat+": "+leftStr+" "+headStr+" ]";
            }
          }
        }
        Label subLabel = new CategoryWordTag(subLabelStr, word, tag);
        Tree subTree = tf.newTreeNode(subLabel, t.getChildrenAsList().subList(1,t.children().length));
        newChildren.add(t.children()[0]);
        newChildren.add(markovInsideBinarizeLocalTree(subTree, head, headLoc-1, topCat, false));
        return tf.newTreeNode(t.label(),newChildren);
      }
      return t;
    }

    private Tree insideBinarizeLocalTree(Tree t, int headNum, TaggedWord head, int leftProcessed, int rightProcessed) {
      String word = head.word();
      String tag = head.tag();
      List newChildren = new ArrayList(2);
      // check done
      if (t.children().length <= leftProcessed+rightProcessed+2) {
        Tree leftChild = t.children()[leftProcessed];
        newChildren.add(leftChild);
        if (t.children().length == leftProcessed+rightProcessed+1) {
          // unary ... so top level
          String finalCat = t.label().value();
          return tf.newTreeNode(new CategoryWordTag(finalCat,word,tag),newChildren);
        }
        // binary
        Tree rightChild = t.children()[leftProcessed+1];
        newChildren.add(rightChild);
        String labelStr = t.label().value();
        if (leftProcessed != 0 || rightProcessed != 0)
          labelStr = ("@ "+leftChild.label().value()+
                      " "+rightChild.label().value());
        return tf.newTreeNode(new CategoryWordTag(labelStr,word,tag),newChildren);
      }
      if (headNum > leftProcessed) {
        // eat left word
        Tree leftChild = t.children()[leftProcessed];
        Tree rightChild = insideBinarizeLocalTree(t, headNum, head,
                                                  leftProcessed+1, rightProcessed);
        newChildren.add(leftChild);
        newChildren.add(rightChild);
        String labelStr = ("@ "+leftChild.label().value()+
                           " "+rightChild.label().value().substring(2));
        if (leftProcessed == 0 && rightProcessed == 0)
          labelStr = t.label().value();
        return tf.newTreeNode(new CategoryWordTag(labelStr,word,tag),newChildren);
      } else {
        // eat right word
        Tree leftChild = insideBinarizeLocalTree(t, headNum, head,
                                                  leftProcessed, rightProcessed+1);
        Tree rightChild = t.children()[t.children().length-rightProcessed-1];
        newChildren.add(leftChild);
        newChildren.add(rightChild);
        String labelStr = ("@ "+leftChild.label().value().substring(2)+
                           " "+rightChild.label().value());
        if (leftProcessed == 0 && rightProcessed == 0)
          labelStr = t.label().value();
        return tf.newTreeNode(new CategoryWordTag(labelStr,word,tag),newChildren);
      }
    }

    private Tree outsideBinarizeLocalTree(Tree t, String labelStr, String finalCat, int headNum, TaggedWord head, int leftProcessed, String leftStr, int rightProcessed, String rightStr) {
      List newChildren = new ArrayList(2);
      Label label = new CategoryWordTag(labelStr,head.word(),head.tag());
      // check if there are <=2 children already
      if (t.children().length - leftProcessed - rightProcessed <= 2) {
        // done, return
        newChildren.add(t.children()[leftProcessed]);
        if (t.children().length - leftProcessed - rightProcessed == 2)
          newChildren.add(t.children()[leftProcessed+1]);
        return tf.newTreeNode(label,newChildren);
      }
      if (headNum > leftProcessed) {
        // eat a left word
        Tree leftChild = t.children()[leftProcessed];
        String childLeftStr = leftStr+" "+leftChild.label().value();
        String childLabelStr = "@"+finalCat+" :"+childLeftStr+" ..."+rightStr;
        Tree rightChild = outsideBinarizeLocalTree(t, childLabelStr, finalCat,
                                              headNum, head,
                                              leftProcessed+1, childLeftStr,
                                              rightProcessed, rightStr);
        newChildren.add(leftChild);
        newChildren.add(rightChild);
        return tf.newTreeNode(label,newChildren);
      } else {
        // eat a right word
        Tree rightChild = t.children()[t.children().length-rightProcessed-1];
        String childRightStr = " "+rightChild.label().value()+rightStr;
        String childLabelStr = "@"+finalCat+" :"+leftStr+" ..."+childRightStr;
        Tree leftChild = outsideBinarizeLocalTree(t, childLabelStr, finalCat,
                                              headNum, head,
                                              leftProcessed, leftStr,
                                              rightProcessed+1, childRightStr);
        newChildren.add(leftChild);
        newChildren.add(rightChild);
        return tf.newTreeNode(label,newChildren);
      }
    }

    /* Do the category splitting of the tree passed in.
     * The root is the root of the tree, must contain t or will throw
     * NullPointerException.
     */
    private TreeHeadPair transformTreeHelper(Tree t, Tree root) {
      Tree parent;
      Tree grandParent;
      String parentStr;
      String grandParentStr;
      if (root == null || t.equals(root)) {
        parent = null;
        parentStr = "";
      } else {
        parent = t.parent(root);
        parentStr = parent.label().value();
      }
      if (parent == null || parent.equals(root)) {
        grandParent = null;
        grandParentStr = "";
      } else {
        grandParent = parent.parent(root);
        grandParentStr = grandParent.label().value();
      }
      String cat = t.label().value();
      String baseParentStr = tlp.basicCategory(parentStr);
      String baseGrandParentStr = tlp.basicCategory(grandParentStr);
      TreeHeadPair result = new TreeHeadPair();

      //System.out.println(t.label().value() + " " + parentStr + " " + grandParentStr);

      // handle null
      if (t == null)
        return result;
      // handle words
      if (t.isLeaf()) {
        Label label = new StringLabel(cat);//new CategoryWordTag(cat,cat,"");
        result.tree = tf.newLeaf(label);
        result.head = new TaggedWord(cat, "");
        return result;
      }
      // handle tags
      if (t.isPreTerminal()) {
        TreeHeadPair childResult = transformTreeHelper(t.children()[0],null);
        String word = childResult.head.word();

        if (Train.tagPA) {
          String test = cat+"^"+baseParentStr;
          if ( ! Train.tagSelectiveSplit || Train.splitters.contains(test)) {
            cat = test;
          }
        }
        if (Train.markUnaryTags && parent.children().length == 1) {
          cat = cat+"^U";
        }
        List newChildren = new ArrayList(1);
        newChildren.add(childResult.tree);
        Label label = new CategoryWordTag(cat,word,cat);
        result.tree = tf.newTreeNode(label,newChildren);
        result.head = new TaggedWord(word,cat);

        // language-specific transforms
        return tlpParams.transformTree(t, root, result);
      }
      // handle categories
      Tree headChild = hf.determineHead(t);
      if (headChild == null)
	System.err.println("Warning -- null head found for tree\n" + t.toString());
      int headNum = -1;
      List newChildren = new ArrayList(t.children().length);
      Tree[] kids = t.children();
      for (int childNum=0; childNum<kids.length; childNum++) {
        Tree child = kids[childNum];
        TreeHeadPair childResult = transformTreeHelper(child, root);
        if (child == headChild) {
          result.head = childResult.head;
          headNum = childNum;
        }
        newChildren.add(childResult.tree);
      }
      String word = result.head.word();
      // tag is always really "baseTag"
      String tag = tlp.basicCategory(result.head.tag());
      String baseCat = tlp.basicCategory(cat);

      /* Sister annotation. Potential problem: if multiple sisters are
       * strong indicators for a single category's expansions.  This
       * happens concretely in the Chinese Treebank when NP (object)
       * has left sisters VV and AS.  Could lead to too much
       * sparseness.  The ideal solution would be to give the
       * splitting list an ordering, and take only the highest (~most
       * informative/reliable) sister annotation.
       */
      if (Train.sisterAnnotate && ! Train.smoothing && baseParentStr.length() > 0) {
        List leftSis = listBasicCategories(SisterAnnotationStats.leftSisterLabels(t,parent));
        List rightSis = listBasicCategories(SisterAnnotationStats.rightSisterLabels(t,parent));

        List leftAnn = new ArrayList();
        List rightAnn = new ArrayList();

        for(Iterator i = leftSis.iterator(); i.hasNext(); ){
          String s = (String) i.next();
          //s = baseCat+"=l="+tlp.basicCategory(s);
          leftAnn.add(baseCat+"=l="+tlp.basicCategory(s));
          //System.out.println("left-annotated test string " + s);
        }
        for(Iterator i = rightSis.iterator(); i.hasNext(); ){
          String s = (String) i.next();
          //s = baseCat+"=r="+tlp.basicCategory(s);
          rightAnn.add(baseCat+"=r="+tlp.basicCategory(s));
        }
        for(Iterator j = rightAnn.iterator(); j.hasNext(); ){
          //System.out.println("new rightsis " + (String)j.next()); //debugging
        }
        for(Iterator i = Train.sisterSplitters.iterator(); i.hasNext();){
          String annCat = (String) i.next();
          //System.out.println("annotated test string " + annCat);
          if(leftAnn.contains(annCat) || rightAnn.contains(annCat)) {
            cat = cat + annCat.replaceAll("^"+baseCat,"");
            break;
          }
        }
      }

      if (Train.PA && ! Train.smoothing && baseParentStr.length() > 0) {
        String cat2 = baseCat+"^"+baseParentStr;
        if ( ! Train.selectiveSplit || Train.splitters.contains(cat2)) {
          cat = cat+"^"+baseParentStr;
        }
      }
      if (Train.gPA && ! Train.smoothing && grandParentStr.length() > 0) {
        if (Train.selectiveSplit && (cat.indexOf("^") >= 0)) {
          String cat2 = baseCat+"^"+baseParentStr+"~"+baseGrandParentStr;
          if (Train.splitters.contains(cat2)) {
            cat = cat+"~"+baseGrandParentStr;
          }
        } else {
          cat = cat+"~"+baseGrandParentStr;
        }
      }
      if (Train.markUnary && kids.length == 1 && kids[0].depth() >= 2) {
        cat = cat+"-U";
      }
      if (Train.markUnary2 && parent != null &&
          parent.children().length == 1 && t.depth() >= 2) {
        cat = cat+"-u";
      }
      if (Train.dominatesV && hasV(t.preTerminalYield())) {
        cat = cat+"-v";
      }
      if (Train.dominatesI && hasI(t.preTerminalYield())) {
        cat = cat+"-i";
      }
      if (Train.dominatesC && hasC(t.preTerminalYield())) {
        cat = cat+"-c";
      }
      if (Train.rightPhrasal && rightPhrasal(t)) {
        cat = cat+"-RX";
      }
      if (Train.xOverX && xOverX(t, baseCat)) {
        cat = cat+"-X";
      }
      if (Train.rightRec && rightRec(t, baseCat)) {
        cat = cat+"-R";
      }
      if (Train.leftRec && leftRec(t, baseCat)) {
        cat = cat+"-L";
      }
      if (Train.splitPrePreT && t.isPrePreTerminal()) {
        cat = cat+"-PPT";
      }
      Label label = new CategoryWordTag(cat, word, tag);
      result.tree = tf.newTreeNode(label, newChildren);

      // System.out.println("internal transformTree: " +
      //                         baseCat + "^" + baseParentStr + "~" +
      //                         baseGrandParentStr + " to " +
      //                         label.value());
      return tlpParams.transformTree(t, root, result);
    }

    /* root is the root of the tree, must contain t or will throw
     * NullPointerException
     */
    private TreeHeadPair transformTreeBinarizer(Tree t) {
      String cat = t.label().value();
      TreeHeadPair result = new TreeHeadPair();

      // handle null
      if (t == null)
        return result;
      // handle words
      if (t.isLeaf()) {
        Label label = new StringLabel(cat);//new CategoryWordTag(cat,cat,"");
        result.tree = tf.newLeaf(label);
        result.head = new TaggedWord(cat, "");
        return result;
      }
      // handle tags
      if (t.isPreTerminal()) {
        TreeHeadPair childResult = transformTreeBinarizer(t.children()[0]);
        String word = childResult.head.word();

        List newChildren = new ArrayList(1);
        newChildren.add(childResult.tree);
        Label label = new CategoryWordTag(cat,word,cat);
        result.tree = tf.newTreeNode(label,newChildren);
        result.head = new TaggedWord(word,cat);
        return result;
      }
      // handle categories
      Tree headChild = hf.determineHead(t);
      if (headChild == null) {
        System.err.println("### No head found for:");
        t.pennPrint();
      }
      int headNum = -1;
      List newChildren = new ArrayList(t.children().length);
      Tree[] kids = t.children();
      for (int childNum=0; childNum<kids.length; childNum++) {
        Tree child = kids[childNum];
        TreeHeadPair childResult = transformTreeBinarizer(child);
        if (child == headChild) {
          result.head = childResult.head;
          headNum = childNum;
        }
        newChildren.add(childResult.tree);
      }
      String word = result.head.word();
      // tag is always really "baseTag"
      String tag = tlp.basicCategory(result.head.tag());

      Label label = new CategoryWordTag(cat, word, tag);
      result.tree = tf.newTreeNode(label, newChildren);
      result.tree = binarizeTree(result.tree, headNum, result.head);
      return result;
    }

    public List listBasicCategories(List l) {
      List l1 = new ArrayList();
      for(Iterator i = l.iterator();i.hasNext();) {
        l1.add(tlp.basicCategory((String)i.next()));
      }
      return l1;
    }


    private boolean hasV(List tags) {
      for (int i=0; i<tags.size(); i++) {
        if (tags.get(i).toString().startsWith("V") ||
            tags.get(i).toString().startsWith("MD"))
          return true;
      }
      return false;
    }

    private boolean hasI(List tags) {
      for (int i=0; i<tags.size(); i++) {
        if (tags.get(i).toString().startsWith("I"))
          return true;
      }
      return false;
    }

    private boolean hasC(List tags) {
      for (int i=0; i<tags.size(); i++) {
        if (tags.get(i).toString().startsWith("CC"))
          return true;
      }
      return false;
    }

    private boolean rightPhrasal(Tree t) {
      while (! t.isLeaf()) {
        t = t.lastChild();
        String str = t.label().value();
        if (str.startsWith("NP") ||
            str.startsWith("PP") ||
            str.startsWith("VP") ||
            str.startsWith("S") ||
            str.startsWith("Q") ||
            str.startsWith("A"))
          return true;
      }
      return false;
    }

    private boolean rightRec(Tree t, String baseCat) {
      if (//! baseCat.equals("S") &&
          ! baseCat.equals("NP"))
        return false;
      while (! t.isLeaf()) {
        t = t.lastChild();
        String str = t.label().value();
        if (str.startsWith(baseCat))
          return true;
      }
      return false;
    }

    private boolean leftRec(Tree t, String baseCat) {
      while (! t.isLeaf()) {
        t = t.firstChild();
        String str = t.label().value();
        if (str.startsWith(baseCat))
          return true;
      }
      return false;
    }

    private boolean xOverX(Tree t, String baseCat) {
      for (Iterator i=t.subTrees().iterator(); i.hasNext();) {
        Tree s = (Tree)i.next();
        if (s == t)
          continue;
        if (s.label().value().startsWith(baseCat))
          return true;
      }
      return false;
    }

    /** Changes the ROOT label, and adds a Lexicon.BOUNDARY daughter to it. 
     *  This is needed for the dependency parser.
     */
    private void addRoot(Tree t) {
      t.setLabel(new CategoryWordTag(tlp.startSymbol(),
                                     Lexicon.BOUNDARY, Lexicon.BOUNDARY_TAG));
      List preTermChildList = new ArrayList();
      Tree boundaryTerm = tf.newLeaf(new StringLabel(Lexicon.BOUNDARY));//CategoryWordTag(Lexicon.BOUNDARY,Lexicon.BOUNDARY,""));
      preTermChildList.add(boundaryTerm);
      Tree boundaryPreTerm = tf.newTreeNode(new CategoryWordTag(Lexicon.BOUNDARY_TAG,Lexicon.BOUNDARY,Lexicon.BOUNDARY_TAG),preTermChildList);
      List childList = t.getChildrenAsList();
      childList.add(boundaryPreTerm);
      t.setChildren(childList);
    }

    public Tree transformTree(Tree t, TreeFactory dummy) {
      if (Train.printTreeTransformations) {
	System.out.println("### Transforming:");
	t.pennPrint();
      }
      Tree trTree = t;
      if (doSubcategorization) {
	trTree = transformTreeHelper(t, t).tree;
	if (Train.printTreeTransformations) {
	  System.out.println("### to:");
	  trTree.pennPrint();
	}
      }
      Tree binarizedTree = transformTreeBinarizer(trTree).tree;
      addRoot(binarizedTree);
      if (Train.printTreeTransformations) {
	System.out.println("### binarized as:");
	binarizedTree.pennPrint();
      }
      return binarizedTree;
    }

    public HeadRespectingBinarizer (HeadFinder hf, 
		    boolean insideFactor, boolean doSubcategorization) {
      this.hf = hf;
      this.tf = new LabeledScoredTreeFactory(new CategoryWordTagFactory());
      this.insideFactor = insideFactor;
      this.doSubcategorization = doSubcategorization;
    }

  } // end static class HeadRespectingBinarizer


  private static class Debinarizer implements TreeTransformer {

    private TreeFactory tf;
    private TreebankLanguagePack tlp = Options.get().langpack;

    /** Remove the last daughter of a ROOT node.  This is a boundary symbol.
     *  This is needed because the dependency parser uses such symbols.
     */
    private void removeRoot(Tree t) {
      if ( ! tlp.isStartSymbol(t.label().value()))
        return;
      Tree[] origChildren = t.children();
      Tree[] children = new Tree[origChildren.length - 1];
      System.arraycopy(origChildren, 0, children, 0, origChildren.length - 1);
      t.setChildren(children);
    }

    protected Tree transformTreeHelper(Tree t) {
      if (t.isLeaf())
        return tf.newLeaf(new StringLabel(t.label().value()));
      List newChildren = new ArrayList(20);
      Label label = new StringLabel(t.label().value());
      for (int childNum=0; childNum<t.children().length; childNum++) {
        Tree child = t.children()[childNum];
        Tree newChild = transformTreeHelper(child);
        if ((! newChild.isLeaf()) && newChild.label().value().indexOf('@') >= 0)
          newChildren.addAll(newChild.getChildrenAsList());
        else
          newChildren.add(newChild);
      }
      return tf.newTreeNode(label,newChildren);
    }

    public Tree transformTree(Tree t, TreeFactory dummy) {
      Tree result = transformTreeHelper(t);
      removeRoot(result);
      return result;
    }

    public Debinarizer() {
      tf = new LabeledScoredTreeFactory(new StringLabelFactory());
    }
  }

  public static TreeTransformer getHeadRespectingBinarizer(HeadFinder hf,
						       boolean insideFactor) {
    return new HeadRespectingBinarizer(hf, insideFactor, true);
  }

  public static TreeTransformer getHeadRespectingBinarizer(HeadFinder hf,
			boolean insideFactor, boolean doSubcategorization) {
    return new HeadRespectingBinarizer(hf, insideFactor, doSubcategorization);
  }

  public static TreeTransformer getDebinarizer() {
    return new Debinarizer();
  }

} // end class BinarizerFactory


class Rule implements Serializable {

  public int parent = -1;

  public double score = Double.NaN;

  public double score() { return score; }

  public boolean isUnary() { return false; }

  private static final long serialVersionUID = 1L;

}


class UnaryRule extends Rule implements Serializable {

  public int child = -1;

    public UnaryRule() {}

    public UnaryRule(int parent, int child, double score) {
	this.parent = parent;
	this.child = child;
	this.score = score;
    }

  public boolean isUnary() { return true; }

  public int hashCode() { return (parent << 16)^child;}

  public boolean equals(Object o) {
    if (this == o)
      return true;
    if (o instanceof UnaryRule) {
      UnaryRule ur = (UnaryRule)o;
      if (parent == ur.parent && child == ur.child)
        return true;
    }
    return false;
  }


  public String toString() {
    return Numberer.getGlobalNumberer("states").object(parent) + " -> " + Numberer.getGlobalNumberer("states").object(child) + " " + score;
  }

  private static final long serialVersionUID = 1L;

}


class BinaryRule extends Rule implements Serializable {

  public int leftChild = -1;
  public int rightChild = -1;

    public BinaryRule() {}

    public BinaryRule(int parent, int leftChild, int rightChild, double score) {
	this.parent = parent;
	this.leftChild = leftChild;
	this.rightChild = rightChild;
	this.score = score;
    }

  public int hashCode() { return (parent << 16)^(leftChild << 8)^rightChild; }

  public boolean equals(Object o) {
    if (this == o)
      return true;
    if (o instanceof BinaryRule) {
      BinaryRule br = (BinaryRule)o;
      if (parent == br.parent &&
          leftChild == br.leftChild &&
          rightChild == br.rightChild)
        return true;
    }
    return false;
  }

  public String toString() {
    return Numberer.getGlobalNumberer("states").object(parent) + " -> " + Numberer.getGlobalNumberer("states").object(leftChild) + " " + Numberer.getGlobalNumberer("states").object(rightChild) + " " + score;
  }

  private static final long serialVersionUID = 1L;

}


class BinaryGrammar implements Serializable {
  private String stateSpace;

  private int numStates;
  private transient List[] rulesWithParent = null;
  private transient List[] rulesWithLC = null;
  private transient List[] rulesWithRC = null;
  private List allRules = null;
  private transient BinaryRule[][] splitRulesWithLC = null;
  private transient BinaryRule[][] splitRulesWithRC = null;
  private transient BinaryRule[][] splitRulesWithParent = null;
  private transient Map ruleMap = null;

  public int numRules() { return allRules.size(); }

  public String stateSpace() { return stateSpace; }

  // for super speed! (maybe)
  private Numberer stateNumberer;
  private boolean[] synthetic = null;

  public boolean isSynthetic(int state) {
    if (synthetic == null) {
      synthetic = new boolean[numStates];
      for (int s=0; s<numStates; s++) {
        try {
	  //System.out.println(((String)stateNumberer.object(s))); // debugging
          synthetic[s] = (((String)stateNumberer.object(s)).charAt(0) == '@');
        } catch (NullPointerException e) {
          synthetic[s] = true;
        }
      }
    }
    return synthetic[state];
  }

  private BinaryRule[] toBRArray(List list) {
    BinaryRule[] array = new BinaryRule[list.size()];
    for (int i=0; i<array.length; i++)
      array[i] = (BinaryRule)list.get(i);
    return array;
  }

  public void splitRules() {
    splitRulesWithLC = new BinaryRule[numStates][];
    splitRulesWithRC = new BinaryRule[numStates][];
    splitRulesWithParent = new BinaryRule[numStates][];
    // rules accesed by their "synthetic" child or left child if none
    for (int state=0; state<numStates; state++) {
      // check synthetic
      if (isSynthetic(state)) {
        splitRulesWithLC[state] = toBRArray(rulesWithLC[state]);
        splitRulesWithRC[state] = toBRArray(rulesWithRC[state]);
      } else {
        List ruleList = new ArrayList();
        for (int rNum=0; rNum<rulesWithLC[state].size(); rNum++) {
          BinaryRule br = (BinaryRule)rulesWithLC[state].get(rNum);
          if (!isSynthetic(br.rightChild))
            ruleList.add(br);
        }
        splitRulesWithLC[state] = toBRArray(ruleList);
        ruleList = new ArrayList();
        for (int rNum=0; rNum<rulesWithRC[state].size(); rNum++) {
          BinaryRule br = (BinaryRule)rulesWithRC[state].get(rNum);
          if (!isSynthetic(br.leftChild))
            ruleList.add(br);
        }
        splitRulesWithRC[state] = toBRArray(ruleList);
      }
      // parent accessor
      splitRulesWithParent[state] = toBRArray(rulesWithParent[state]);
    }
  }

  public BinaryRule[] splitRulesWithLC(int state) {
    return splitRulesWithLC[state];
  }

  public BinaryRule[] splitRulesWithRC(int state) {
    return splitRulesWithRC[state];
  }

  public BinaryRule[] splitRulesWithParent(int state) {
    return splitRulesWithParent[state];
  }

  // the sensible version

  public double scoreRule(BinaryRule br) {
    BinaryRule rule = (BinaryRule)ruleMap.get(br);
    return (rule != null ? rule.score : Double.NEGATIVE_INFINITY);
  }

  public void addRule(BinaryRule br) {
    rulesWithParent[br.parent].add(br);
    rulesWithLC[br.leftChild].add(br);
    rulesWithRC[br.rightChild].add(br);
    allRules.add(br);
    ruleMap.put(br,br);
  }

  public Iterator ruleIterator() {
    return allRules.iterator();
  }

  public Iterator ruleIteratorByParent(int state) {
    return rulesWithParent[state].iterator();
  }

  public Iterator ruleIteratorByRightChild(int state) {
    return rulesWithRC[state].iterator();
  }

  public Iterator ruleIteratorByLeftChild(int state) {
    return rulesWithLC[state].iterator();
  }

  public List ruleListByRightChild(int state) {
    return rulesWithRC[state];
  }

  public List ruleListByLeftChild(int state) {
    return rulesWithLC[state];
  }

  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject();
    init();
    for (int i=0; i<allRules.size(); i++) {
      BinaryRule br = (BinaryRule)allRules.get(i);
      rulesWithParent[br.parent].add(br);
      rulesWithLC[br.leftChild].add(br);
      rulesWithRC[br.rightChild].add(br);
      ruleMap.put(br,br);
    }
    splitRules();
  }

  void init() {
    stateNumberer = Numberer.getGlobalNumberer(stateSpace);
    ruleMap = new HashMap();
    rulesWithParent = new List[numStates];
    rulesWithLC = new List[numStates];
    rulesWithRC = new List[numStates];
    for (int s=0; s<numStates; s++) {
      rulesWithParent[s] = new ArrayList();
      rulesWithLC[s] = new ArrayList();
      rulesWithRC[s] = new ArrayList();
    }
  }

  public BinaryGrammar(int numStates) {
    this(numStates, "states");
  }

  public BinaryGrammar(int numStates, String stateSpace) {
    this.stateSpace = stateSpace;
    this.numStates = numStates;
    allRules = new ArrayList();
    init();
  }

    /**
     * Populates data in this Object from the character stream 
     * given by the Reader r.
     */
    public void readData(Reader r) throws IOException {
	BufferedReader in = new BufferedReader(r);
	String line, parentString, leftChildString, rightChildString, scoreString;
	int lineNum = 1;
	StringTokenizer tok;
	Numberer n = Numberer.getGlobalNumberer("states");
	// all lines have one rule per line
	while ((line = in.readLine()) != null) {
	    try {
		tok = new StringTokenizer(line);
		parentString = tok.nextToken();
		tok.nextToken();
		leftChildString = tok.nextToken();
		rightChildString = tok.nextToken();
		scoreString = tok.nextToken();
		addRule(new BinaryRule(n.number(parentString), n.number(leftChildString), n.number(rightChildString), 
				      Double.parseDouble(scoreString)));
	    } catch (Exception e) {
		throw new IOException("Error on line " + lineNum);
	    }	    
	    lineNum++;   
	}
	splitRules();
    }

    /**
     * Writes out data from this Object to the Writer w.
     */
    public void writeData(Writer w) throws IOException {
	PrintWriter out = new PrintWriter(w);
	// all lines have one rule per line
	for (Iterator rI = ruleIterator(); rI.hasNext();) {
	    out.println(((UnaryRule) rI.next()).toString());
	}
	out.flush();
    }

  private static final long serialVersionUID = 1L;

} // end class BinaryGrammar


class UnaryGrammar implements Serializable {

  private int numStates = -1;
  private String stateSpace;

  private transient List[] rulesWithParent = null;
  private transient List[] rulesWithChild = null;

  private transient List[] closedRulesWithParent = null;
  private transient List[] closedRulesWithChild = null;

  private transient UnaryRule[][] closedRulesWithP = null;
  private transient UnaryRule[][] closedRulesWithC = null;

  protected Map coreRules = null;
  private transient Map bestRules = null;
  private transient Map backTrace = null;

  public int numClosedRules() { return bestRules.keySet().size(); }
  public Iterator closedRuleIterator() { return bestRules.keySet().iterator(); }
  public int numRules() { return coreRules.keySet().size(); }
  public Iterator ruleIterator() { return coreRules.keySet().iterator(); }

  public void purgeRules() {
    Map bR = new HashMap();
    for (Iterator i=bestRules.keySet().iterator(); i.hasNext();) {
      UnaryRule ur = (UnaryRule)i.next();
      if (ur.parent != ur.child)
        bR.put(ur,ur);
    }
    bestRules = bR;
  }

  public List getBestPath(int parent, int child) {
    List path = new ArrayList();
    UnaryRule tempR = new UnaryRule();
    tempR.parent = parent;
    tempR.child = child;
    //System.out.println("Building path...");
    int loc = parent;
    while (loc != child) {
      path.add(new Integer(loc));
      //System.out.println("Path is "+path);
      tempR.parent = loc;
      Integer nextInt = (Integer)backTrace.get(tempR);
      if (nextInt == null)
        loc = child;
      else
        loc = nextInt.intValue();
      //System.out.println(Numberer.getGlobalNumberer("states").object(parent)+"->"+Numberer.getGlobalNumberer("states").object(child)+" went via "+Numberer.getGlobalNumberer("states").object(loc));
      if (path.size() > 10)
        System.exit(0);
    }
    path.add(new Integer(child));
    return path;
  }

  private void closeRules(UnaryRule ur) {
    for (int i=0; i<closedRulesWithChild[ur.parent].size(); i++) {
      UnaryRule pr = (UnaryRule)closedRulesWithChild[ur.parent].get(i);
      for (int j=0; j<closedRulesWithParent[ur.child].size(); j++) {
        UnaryRule cr = (UnaryRule)closedRulesWithParent[ur.child].get(j);
        UnaryRule resultR = new UnaryRule();
        resultR.parent = pr.parent;
        resultR.child = cr.child;
        resultR.score = pr.score+cr.score+ur.score;
        if (relaxRule(resultR)) {
          if (resultR.parent != ur.parent)
            backTrace.put(resultR, new Integer(ur.parent));
          else
            backTrace.put(resultR, new Integer(ur.child));
        }
      }
    }
  }

  private boolean relaxRule(UnaryRule ur) {
    UnaryRule bestR = (UnaryRule)bestRules.get(ur);
    if (bestR == null) {
      bestRules.put(ur,ur);
      closedRulesWithParent[ur.parent].add(ur);
      closedRulesWithChild[ur.child].add(ur);
      return true;
    } else {
      if (bestR.score < ur.score) {
        bestR.score = ur.score;
        return true;
      }
      return false;
    }
  }

  public double scoreRule(UnaryRule ur) {
    UnaryRule bestR = (UnaryRule)bestRules.get(ur);
    return (bestR != null ? bestR.score : Double.NEGATIVE_INFINITY);
  }

  public void addRule(UnaryRule ur) {
    // add rules' closure
    closeRules(ur);
    coreRules.put(ur,ur);
    rulesWithParent[ur.parent].add(ur);
    rulesWithChild[ur.child].add(ur);
  }
  //public Iterator closedRuleIterator() {
  //  return bestRules.keySet().iterator();
  //}

  void makeCRArrays() {
    int numStates = closedRulesWithParent.length;
    closedRulesWithP = new UnaryRule[numStates][];
    closedRulesWithC = new UnaryRule[numStates][];
    for (int i=0; i<numStates; i++) {
      closedRulesWithP[i] = (UnaryRule[])closedRulesWithParent[i].toArray(new UnaryRule[0]);
      closedRulesWithC[i] = (UnaryRule[])closedRulesWithChild[i].toArray(new UnaryRule[0]);
    }
  }

  public UnaryRule[] closedRulesByParent(int state) {
    if (closedRulesWithP == null) {
      makeCRArrays();
    }
    return closedRulesWithP[state];
  }

  public UnaryRule[] closedRulesByChild(int state) {
    if (closedRulesWithC == null) {
      makeCRArrays();
    }
    return closedRulesWithC[state];
  }

  public Iterator closedRuleIteratorByParent(int state) {
    return closedRulesWithParent[state].iterator();
  }
  public Iterator closedRuleIteratorByChild(int state) {
    return closedRulesWithChild[state].iterator();
  }
  public Iterator ruleIteratorByParent(int state) {
    return rulesWithParent[state].iterator();
  }
  public Iterator ruleIteratorByChild(int state) {
    return rulesWithChild[state].iterator();
  }

  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject();
    Set allRules = new HashSet(coreRules.keySet());
    init();
    for (Iterator i=allRules.iterator(); i.hasNext();) {
      UnaryRule ur = (UnaryRule)i.next();
      addRule(ur);
    }
    purgeRules();
  }


  void init() {
    coreRules = new HashMap();
    rulesWithParent = new List[numStates];
    rulesWithChild = new List[numStates];
    closedRulesWithParent = new List[numStates];
    closedRulesWithChild = new List[numStates];
    bestRules = new HashMap();
    backTrace = new HashMap();
    for (int s=0; s<numStates; s++) {
      rulesWithParent[s] = new ArrayList();
      rulesWithChild[s] = new ArrayList();
      closedRulesWithParent[s] = new ArrayList();
      closedRulesWithChild[s] = new ArrayList();
      UnaryRule selfR = new UnaryRule();
      selfR.parent = s;
      selfR.child = s;
      selfR.score = 0;
      relaxRule(selfR);
    }
  }

  public UnaryGrammar(int numStates) {
    this(numStates, "states");
  }

  public UnaryGrammar(int numStates, String stateSpace) {
      this.numStates = numStates;
      this.stateSpace = stateSpace;
      init();
  }

    /**
     * Populates data in this Object from the character stream 
     * given by the Reader r.
     */
    public void readData(Reader r) throws IOException {
	BufferedReader in = new BufferedReader(r);
	String line, parentString, childString, scoreString;
	int lineNum = 1;
	StringTokenizer tok;
	Numberer n = Numberer.getGlobalNumberer(stateSpace);
	// all lines have one rule per line
	while ((line = in.readLine()) != null) {
	    try {
		tok = new StringTokenizer(line);
		parentString = tok.nextToken();
		tok.nextToken();
		childString = tok.nextToken();
		scoreString = tok.nextToken();
		addRule(new UnaryRule(n.number(parentString), n.number(childString), 
				      Double.parseDouble(scoreString)));
	    } catch (Exception e) {
		throw new IOException("Error on line " + lineNum);
	    }	    
	    lineNum++;   
	}
	purgeRules();
    }

    /**
     * Writes out data from this Object to the Writer w.
     */
    public void writeData(Writer w) throws IOException {
	PrintWriter out = new PrintWriter(w);
	// all lines have one rule per line
	for (Iterator rI = ruleIterator(); rI.hasNext();) {
	    out.println(((UnaryRule) rI.next()).toString());
	}
	out.flush();
    }

  private static final long serialVersionUID = 1L;

} // end class UnaryGrammar



class BinaryGrammarExtractor extends AbstractTreeExtractor {
  private Numberer stateNumberer = Numberer.getGlobalNumberer("states");
  private Counter ruleCounter = new Counter();
  private Counter symbolCounter = new Counter();
  private Set binaryRules = new HashSet();
  private Set unaryRules = new HashSet();

  protected void tallyRule(Rule r) {
    symbolCounter.add(stateNumberer.object(r.parent),1);
    ruleCounter.add(r,1);
    if (r.isUnary())
      unaryRules.add(r);
    else
      binaryRules.add(r);
  }

  protected void tallyInternalNode(Tree lt) {
    Rule r = null;
    if (lt.children().length == 1) {
      UnaryRule ur = new UnaryRule();
      ur.parent = stateNumberer.number(lt.label().value());
      ur.child = stateNumberer.number(lt.children()[0].label().value());
      r = ur;
    } else {
      BinaryRule br = new BinaryRule();
      br.parent = stateNumberer.number(lt.label().value());
      br.leftChild = stateNumberer.number(lt.children()[0].label().value());
      br.rightChild = stateNumberer.number(lt.children()[1].label().value());
      r = br;
    }
    //System.err.println("Tallying rule: "+r);
    tallyRule(r);
  }

  protected Object formResult() {
    stateNumberer.number(Lexicon.BOUNDARY_TAG);
    BinaryGrammar bg = new BinaryGrammar(stateNumberer.total());
    UnaryGrammar ug = new UnaryGrammar(stateNumberer.total());
    // add unaries
    for (Iterator uI = unaryRules.iterator(); uI.hasNext();) {
      UnaryRule ur = (UnaryRule)uI.next();
      ur.score = Math.log((ruleCounter.countOf(ur))/symbolCounter.countOf(stateNumberer.object(ur.parent)));
      ug.addRule(ur);
    }
    // add binaries
    for (Iterator bI = binaryRules.iterator(); bI.hasNext();) {
      BinaryRule br = (BinaryRule)bI.next();
      br.score = Math.log((ruleCounter.countOf(br)-Train.ruleDiscount)/symbolCounter.countOf(stateNumberer.object(br.parent)));
      //System.out.println("Scoring "+br+" as "+br.score+" from "+ruleCounter.countOf(br)+"/"+symbolCounter.countOf(stateNumberer.object(br.parent)));
      bg.addRule(br);
    }
    return new Pair(ug,bg);
  }


} // end class BinaryGrammarExtractor


class SmoothedBinaryGrammarExtractor extends AbstractTreeExtractor {

  protected int HISTORY_DEPTH() { return (Train.gPA ? 3 :
                                          (Train.PA ? 2 : 1)); }

  protected Numberer stateNumberer = Numberer.getGlobalNumberer("states");
  protected Numberer tagNumberer = Numberer.getGlobalNumberer("tags");

  protected Map ruleToLabel = new HashMap();
  protected Counter rulePairs = new Counter();
  protected Counter labelPairs = new Counter();

    public void tallyTree(Tree t) {
      LinkedList ll = new LinkedList();
      //ll.add("ROOT");
      //ll.add("ROOT");
      tallyTree(t, ll);
    }

    protected void tallyTree(Tree t, LinkedList parents) {
      // traverse tree, building parent list
      String str = t.label().value();
      boolean strIsPassive = (str.indexOf('@') == -1);
      if (strIsPassive) {
        parents.addFirst(str);
      }
      if (! t.isLeaf()) {
        if (! t.children()[0].isLeaf()) {
          tallyInternalNode(t, parents);
          for (int c=0; c<t.children().length; c++) {
            Tree child = t.children()[c];
            tallyTree(child, parents);
          }
        } else {
          tagNumberer.number(t.label().value());
        }
      }
      if (strIsPassive) {
        parents.removeFirst();
      }
    }

  protected Rule ltToRule(Tree lt) {
    if (lt.children().length == 1) {
      UnaryRule ur = new UnaryRule();
      ur.parent = stateNumberer.number(lt.label().value());
      ur.child = stateNumberer.number(lt.children()[0].label().value());
      return ur;
    } else {
      BinaryRule br = new BinaryRule();
      br.parent = stateNumberer.number(lt.label().value());
      br.leftChild = stateNumberer.number(lt.children()[0].label().value());
      br.rightChild = stateNumberer.number(lt.children()[1].label().value());
      return br;
    }
  }

  protected boolean isSynthetic(int state) {
    return ((String)stateNumberer.object(state)).indexOf('@') > -1;
  }

  protected boolean isTag(int state) {
    return tagNumberer.hasSeen(stateNumberer.object(state));
  }

  protected Rule specifyRule(Rule rule, List history, int childDepth) {
    Rule r = null;
    String topHistoryStr = historyToString(history.subList(1,history.size()));
    String bottomHistoryStr = historyToString(history.subList(0,childDepth));
    if (rule instanceof UnaryRule) {
      UnaryRule ur = new UnaryRule();
      UnaryRule urule = (UnaryRule)rule;
      ur.parent = stateNumberer.number(stateNumberer.object(urule.parent)+topHistoryStr);
      if (isSynthetic(urule.child)) {
        ur.child = stateNumberer.number(stateNumberer.object(urule.child)+topHistoryStr);
      } else if (isTag(urule.child)) {
        ur.child = urule.child;
      } else {
        ur.child = stateNumberer.number(stateNumberer.object(urule.child)+bottomHistoryStr);
      }
      r = ur;
    } else {
      BinaryRule br = new BinaryRule();
      BinaryRule brule = (BinaryRule)rule;
      br.parent = stateNumberer.number(stateNumberer.object(brule.parent)+topHistoryStr);
      if (isSynthetic(brule.leftChild)) {
        br.leftChild = stateNumberer.number(stateNumberer.object(brule.leftChild)+topHistoryStr);
      } else if (isTag(brule.leftChild)) {
        br.leftChild = brule.leftChild;
      }else {
        br.leftChild = stateNumberer.number(stateNumberer.object(brule.leftChild)+bottomHistoryStr);
      }
      if (isSynthetic(brule.rightChild)) {
        br.rightChild = stateNumberer.number(stateNumberer.object(brule.rightChild)+topHistoryStr);
      } else if (isTag(brule.rightChild)) {
        br.rightChild = brule.rightChild;
      } else {
        br.rightChild = stateNumberer.number(stateNumberer.object(brule.rightChild)+bottomHistoryStr);
      }
      r = br;
    }
    return r;
  }

  protected void tallyInternalNode(Tree lt, List parents) {
    StringBuffer historySB = new StringBuffer();
    // form base rule
    String label = lt.label().value();
    Rule baseR = ltToRule(lt);
    ruleToLabel.put(baseR,label);
    // act on each history depth
    for (int depth = 0; depth <= Math.min(HISTORY_DEPTH(),parents.size()); depth++) {
      List history = new ArrayList(parents.subList(0,depth));
      // tally each history level / rewrite pair
      rulePairs.add(new Pair(baseR,history),1);
      labelPairs.add(new Pair(label,history),1);
    }
  }

  protected Map historyToString = new HashMap();
  protected String historyToString(List history) {
    String str = (String)historyToString.get(history);
    if (str == null) {
      StringBuffer sb = new StringBuffer();
      for (int i=0; i<history.size(); i++) {
        sb.append('^');
        sb.append(history.get(i));
      }
      str = sb.toString();
      historyToString.put(history, str);
    }
    return str;
  }

  protected Object formResult() {
    Set brs = new HashSet();
    Set urs = new HashSet();
    // scan each rule / history pair
    int ruleCount = 0;
    for (Iterator pairI = rulePairs.seenSet().iterator(); pairI.hasNext();) {
      if (ruleCount % 100 == 0)
        System.err.println("Rules multiplied: "+ruleCount);
      ruleCount++;
      Pair rulePair = (Pair)pairI.next();
      Rule baseRule = (Rule)rulePair.first;
      String baseLabel = (String)ruleToLabel.get(baseRule);
      List history = (List)rulePair.second;
      double totalProb = 0;
      for (int depth = 1; depth <= HISTORY_DEPTH() && depth <= history.size(); depth++) {
        List subHistory = history.subList(0,depth);
        double c_label = labelPairs.countOf(new Pair(baseLabel,subHistory));
        double c_rule = rulePairs.countOf(new Pair(baseRule,subHistory));
        //System.out.println("Multiplying out "+baseRule+" with history "+subHistory);
        //System.out.println("Count of "+baseLabel+" with "+subHistory+" is "+c_label);
        //System.out.println("Count of "+baseRule+" with "+subHistory+" is "+c_rule );

        double prob = (1.0/HISTORY_DEPTH())*(c_rule)/(c_label);
        totalProb += prob;
        for (int childDepth=0; childDepth<=Math.min(HISTORY_DEPTH()-1,depth); childDepth++) {
          Rule rule = specifyRule(baseRule, subHistory, childDepth);
          rule.score = Math.log(totalProb);
          //System.out.println("Created  "+rule+" with score "+rule.score);
          if (rule instanceof UnaryRule)
            urs.add(rule);
          else
            brs.add(rule);
        }
      }
    }
    System.out.println("Total states: "+stateNumberer.total());
    BinaryGrammar bg = new BinaryGrammar(stateNumberer.total());
    UnaryGrammar ug = new UnaryGrammar(stateNumberer.total());
    for (Iterator brI = brs.iterator(); brI.hasNext();) {
      BinaryRule br = (BinaryRule)brI.next();
      bg.addRule(br);
    }
    for (Iterator urI = urs.iterator(); urI.hasNext();) {
      UnaryRule ur = (UnaryRule)urI.next();
      ug.addRule(ur);
    }
    return new Pair(ug,bg);
  }
}

class XBarGrammarProjection implements GrammarProjection {
  UnaryGrammar sourceUG;
  BinaryGrammar sourceBG;
  Numberer sourceNumberer;

  UnaryGrammar targetUG;
  BinaryGrammar targetBG;
  Numberer targetNumberer;

  int[] projection;

  public int project(int state) { return projection[state]; }
  public UnaryGrammar sourceUG() { return sourceUG; }
  public BinaryGrammar sourceBG() { return sourceBG; }
  public UnaryGrammar targetUG() { return targetUG; }
  public BinaryGrammar targetBG() { return targetBG; }


  protected String projectString(String str) {
    if (str.indexOf('@') == -1) {
      if (str.indexOf('^') == -1)
        return str;
        return str.substring(0,str.indexOf('^'));
    }
    StringBuffer sb = new StringBuffer();
    sb.append(str.substring(0,str.indexOf(' ')));
    if (str.indexOf('^') > -1) {
      //sb.append(str.substring(str.indexOf('^'),str.length()));
    }
    int num = -2;
    for (int i=0; i<str.length(); i++) {
      if (str.charAt(i) == ' ')
        num++;
    }
    sb.append(" w "+num);
    return sb.toString();
  }

  protected void scanStates(Numberer source, Numberer target) {
    for (int i=0; i<source.total(); i++) {
      String stateStr = (String)source.object(i);
      String projStr = projectString(stateStr);
      projection[i] = target.number(projStr);
    }
  }

  protected BinaryRule projectBinaryRule(BinaryRule br) {
    BinaryRule br2 = new BinaryRule();
    br2.parent = projection[br.parent];
    br2.leftChild = projection[br.leftChild];
    br2.rightChild = projection[br.rightChild];
    br2.score = br.score;
    return br2;
  }

  protected UnaryRule projectUnaryRule(UnaryRule ur) {
    UnaryRule ur2 = new UnaryRule();
    ur2.parent = projection[ur.parent];
    ur2.child = projection[ur.child];
    ur2.score = ur.score;
    return ur2;
  }

  public XBarGrammarProjection(BinaryGrammar bg, UnaryGrammar ug) {
    Map binaryRules = new HashMap();
    Map unaryRules = new HashMap();
    sourceUG = ug;
    sourceBG = bg;
    sourceNumberer = Numberer.getGlobalNumberer(bg.stateSpace());
    targetNumberer = Numberer.getGlobalNumberer(bg.stateSpace()+"-xbar");
    projection = new int[sourceNumberer.total()];
    scanStates(sourceNumberer, targetNumberer);
    targetBG = new BinaryGrammar(targetNumberer.total(), bg.stateSpace()+"-xbar");
    targetUG = new UnaryGrammar(targetNumberer.total());
    for (Iterator brI = bg.ruleIterator(); brI.hasNext();) {
      Rule rule = projectBinaryRule((BinaryRule)brI.next());
      Rule old = (BinaryRule)binaryRules.get(rule);
      if (old == null || rule.score > old.score)
        binaryRules.put(rule, rule);
    }
    for (Iterator brI = binaryRules.keySet().iterator(); brI.hasNext();) {
      BinaryRule br = (BinaryRule)binaryRules.get(brI.next());
      targetBG.addRule(br);
      //System.out.println("BR: "+targetNumberer.object(br.parent)+" -> "+targetNumberer.object(br.leftChild)+" "+targetNumberer.object(br.rightChild)+" %% "+br.score);
    }
    targetBG.splitRules();
    for (int parent=0; parent<sourceNumberer.total(); parent++) {
      for (Iterator urI = ug.ruleIteratorByParent(parent); urI.hasNext();) {
        UnaryRule sourceRule = (UnaryRule)urI.next();
        Rule rule = projectUnaryRule(sourceRule);
        Rule old = (UnaryRule)unaryRules.get(rule);
        if (old == null || rule.score > old.score)
          unaryRules.put(rule, rule);
        /*
          if (((UnaryRule)rule).child == targetNumberer.number("PRP") &&
            ((String)sourceNumberer.object(rule.parent)).charAt(0) == 'N') {
          System.out.println("Source UR: "+sourceRule+" %% "+sourceRule.score);
          System.out.println("Score of "+rule+"is now: "+((UnaryRule)unaryRules.get(rule)).score);
        }
        */
      }
    }
    for (Iterator urI = unaryRules.keySet().iterator(); urI.hasNext();) {
      UnaryRule ur = (UnaryRule)unaryRules.get(urI.next());
      targetUG.addRule(ur);
      //System.out.println("UR: "+targetNumberer.object(ur.parent)+" -> "+targetNumberer.object(ur.child)+" %% "+ur.score);
    }
    targetUG.purgeRules();
    System.out.println("Projected "+sourceNumberer.total()+" states to "+targetNumberer.total()+" states.");
  }
}


interface GrammarProjection {
    int project(int state);
    UnaryGrammar sourceUG();
    BinaryGrammar sourceBG();
    UnaryGrammar targetUG();
    BinaryGrammar targetBG();
}

class NullGrammarProjection implements GrammarProjection {
    UnaryGrammar ug;
    BinaryGrammar bg;
    public int project(int state) { return state; }
    public UnaryGrammar sourceUG() { return ug; }
    public BinaryGrammar sourceBG() { return bg; }
    public UnaryGrammar targetUG() { return ug; }
    public BinaryGrammar targetBG() { return bg; }
    NullGrammarProjection(BinaryGrammar bg, UnaryGrammar ug) {
        this.ug = ug;
        this.bg = bg;
    }
}

class ProjectionScorer implements Scorer {
  protected GrammarProjection gp;
  protected Scorer scorer;

  protected Edge tempEdge = new Edge();
  protected Edge project(Edge edge) {
    tempEdge.start = edge.start;
    tempEdge.end = edge.end;
    tempEdge.state = gp.project(edge.state);
    tempEdge.head = edge.head;
    tempEdge.tag = edge.tag;
    return tempEdge;
  }

  protected Hook tempHook = new Hook();
  protected Hook project(Hook hook) {
    tempHook.start = hook.start;
    tempHook.end = hook.end;
    tempHook.state = gp.project(hook.state);
    tempHook.head = hook.head;
    tempHook.tag = hook.tag;
    tempHook.subState = gp.project(hook.subState);
    return tempHook;
  }

  public double oScore(Edge edge) { return scorer.oScore(project(edge)); }
  public double iScore(Edge edge) { return scorer.iScore(project(edge)); }
  public boolean oPossible(Hook hook) { return scorer.oPossible(project(hook)); }
  public boolean iPossible(Hook hook) { return scorer.iPossible(project(hook)); }

  public boolean parse(List words) {
    return scorer.parse(words);
  }

  public ProjectionScorer(Scorer scorer, GrammarProjection gp) {
    this.scorer = scorer;
    this.gp = gp;
  }
} // end ProjectionScorer



class DependencyGrammar implements Serializable {

  private static Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  private static Numberer wordNumberer = Numberer.getGlobalNumberer("words");

  protected int numWords;
  protected int numTags;
  transient private Counter argCounter;
  transient private Counter stopCounter;
  private Counter coreDependencies = new Counter();
  //private Counter rheadCounter;
  //private Counter lheadCounter;
  //private Counter argCounter;
  //private Counter headArgCounter;
  //private Counter headDistCounter;
  //private double total = 0.0;

  public double[] smooth = {16.0,16.0,4.0};
  public double[] interp = {0.6,0.6};
  public double distanceDecay = 0.0;

  public String toString() {
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setMaximumFractionDigits(2);
    StringBuffer sb = new StringBuffer(2000);
    String cl = getClass().getName();
    sb.append(cl.substring(cl.lastIndexOf(".")+1) +
              "[tags="+numTags+",words="+numWords+"; head -> arg\n");
    for (Iterator dI = coreDependencies.seenSet().iterator(); dI.hasNext();) {
      Dependency d = (Dependency)dI.next();
      double count = coreDependencies.countOf(d);
      sb.append(d + " count " + nf.format(count));
      if (dI.hasNext()) {
        sb.append(",");
      }
      sb.append("\n");
    }
    sb.append("]");
    return sb.toString();
  }

  public static boolean pruneTW(IntTaggedWord argTW) {
    String[] punctTags = Options.get().langpack.punctuationTags();
    for (int i = 0; i < punctTags.length; i++) {
      if (argTW.tag == tagNumberer.number(punctTags[i])) {
        return true;
      }
    }
    return false;
  }

  public static boolean rootTW(IntTaggedWord rTW) {
    return rTW.tag == tagNumberer.number(Lexicon.BOUNDARY_TAG);
  }

  static class EndHead {
    public int end;
    public int head;
  }

  protected static transient EndHead tempEndHead = new EndHead();

  protected static EndHead treeToDependencyHelper(Tree tree, List depList,
                                                  int loc) {
//       try {
// 	PrintWriter pw = new PrintWriter(new OutputStreamWriter(System.out,"GB18030"),true);
// 	tree.pennPrint(pw);
//       }
//       catch (UnsupportedEncodingException e) {}


    if (tree.isLeaf() || tree.isPreTerminal()) {
      tempEndHead.head = loc;
      tempEndHead.end = loc+1;
      return tempEndHead;
    }
    if (tree.children().length == 1)
      return treeToDependencyHelper(tree.children()[0], depList, loc);
    int start = loc;
    tempEndHead = treeToDependencyHelper(tree.children()[0], depList, loc);
    int lHead = tempEndHead.head;
    int split = tempEndHead.end;
    tempEndHead = treeToDependencyHelper(tree.children()[1], depList, tempEndHead.end);
    int end = tempEndHead.end;
    int rHead = tempEndHead.head;
    String hTag = ((HasTag)tree.label()).tag();
    String lTag = ((HasTag)tree.children()[0].label()).tag();
    String rTag = ((HasTag)tree.children()[1].label()).tag();
    String hWord = ((HasWord)tree.label()).word();
    String lWord = ((HasWord)tree.children()[0].label()).word();
    String rWord = ((HasWord)tree.children()[1].label()).word();
    boolean leftHeaded = hWord.equals(lWord);
    String aTag = (leftHeaded ? rTag : lTag);
    String aWord = (leftHeaded ? rWord : lWord);
    int hT = tagNumberer.number(hTag);
    int aT = tagNumberer.number(aTag);
    int hW = (wordNumberer.hasSeen(hWord) ?
              wordNumberer.number(hWord) :
              wordNumberer.number(Lexicon.UNKNOWN_WORD));
    int aW = (wordNumberer.hasSeen(aWord) ?
              wordNumberer.number(aWord) :
              wordNumberer.number(Lexicon.UNKNOWN_WORD));
    int head = (leftHeaded ? lHead : rHead);
    int arg = (leftHeaded ? rHead : lHead);
    Dependency dependency = new Dependency(hW,hT,aW,aT, leftHeaded, (leftHeaded ? split-head-1: head-split));
    depList.add(dependency);
    Dependency stopL = new Dependency(aW,aT,-2,-2, false, (leftHeaded ? arg-split : arg-loc));
    depList.add(stopL);
    Dependency stopR = new Dependency(aW,aT,-2,-2, true, (leftHeaded ? end-arg-1 : split-arg-1));
    depList.add(stopR);
    //System.out.println("Adding: "+dependency+" at "+tree.label());
    tempEndHead.head = head;
    return tempEndHead;
  }

  public static List treeToDependencyList(Tree tree) {
    List depList = new ArrayList();
    treeToDependencyHelper(tree, depList, 0);
    if (false) {
      System.out.println("----------------------------");
      tree.pennPrint();
      System.out.println(depList);
    }
    return depList;
  }

  protected double scoreAll(Collection deps) {
    double totalScore = 0.0;
    for (Iterator i=deps.iterator(); i.hasNext();) {
      Dependency d = (Dependency)i.next();
      //if (d.head.word == Numberer.number("words","via") ||
      //          d.arg.word == Numberer.number("words","via"))
      //System.out.println(d+" at "+score(d));
      double score = score(d);
      if (score > Double.NEGATIVE_INFINITY)
        totalScore += score;
    }
    return totalScore;
  }

  public void tune(List deps) {
    double bestScore = Double.NEGATIVE_INFINITY;
    double[] bestSmooth = {0.0,0.0,0.0};
    double[] bestInterp = {0.0,0.0};
    double bestDecay = 0.0;
    double maxDecay = (Options.get().distance ? -0.01 : 0.0);
    for (smooth[0]=16; smooth[0]<32; smooth[0] *= 2.0) {
      for (smooth[1]=16; smooth[1]<32; smooth[1] *= 2.0) {
        for (smooth[2]=4; smooth[2]<=(Options.get().distance ? 5 : 5); smooth[2] *= 2.0) {
          for (interp[0]=0.6; interp[0]<0.7; interp[0] += 0.1) {
            for (distanceDecay=maxDecay; distanceDecay == maxDecay; distanceDecay-=0.01) {
              //for (interp[1]=0.1; interp[1]<0.9; interp[1] += 0.1) {
              double score = scoreAll(deps);
              if (Test.verbose)
                System.out.println("Trying s0 "+smooth[0]+" s1 "+smooth[1]+" i0 "+interp[0]+" s2 "+smooth[2]+" score "+score);
              if (score > bestScore) {
                System.arraycopy(smooth, 0, bestSmooth, 0, smooth.length);
                System.arraycopy(interp, 0, bestInterp, 0, interp.length);
                bestDecay = distanceDecay;
                bestScore = score;
              }
              //}
            }
          }
        }
      }
    }
    System.arraycopy(bestSmooth, 0, smooth, 0, bestSmooth.length);
    System.arraycopy(bestInterp, 0, interp, 0, bestInterp.length);
    distanceDecay = bestDecay;
    if (Test.verbose)
      System.out.println("Tuning selected s0 "+smooth[0]+" s1 "+smooth[1]+" s2 "+smooth[2]+" i0 "+interp[0]);
  }

  private static final IntTaggedWord stopTW = new IntTaggedWord(-2,-2);
  private static final IntTaggedWord wildTW = new IntTaggedWord(-1,-1);

  public void addRule(Dependency dependency, double count) {
    if (! Options.get().directional)
      dependency.leftHeaded = false;
    coreDependencies.add(dependency, count);
    /*new Dependency(dependency.head.word,
                                        dependency.head.tag,
                                        dependency.arg.word,
                                        dependency.arg.tag,
                                        dependency.leftHeaded,
                                        dependency.distance), count);
    */
    expandDependency(dependency, count);
  }


  private transient Dependency tempDependency = null;
  private transient Map expandDependencyMap = new HashMap();
  private Dependency intern(IntTaggedWord headTW, IntTaggedWord argTW, boolean leftHeaded, int dist) {
    Map map = expandDependencyMap;
    if (tempDependency == null)
      tempDependency = new Dependency();
    tempDependency.head = headTW;
    tempDependency.arg = argTW;
    tempDependency.leftHeaded = leftHeaded;
    tempDependency.distance = (short) dist;
    Dependency returnDependency = tempDependency;
    if (map != null) {
      returnDependency = (Dependency) map.get(tempDependency);
      if (returnDependency == null)
	map.put(tempDependency, tempDependency);
      returnDependency = tempDependency;
    }
    if (returnDependency == tempDependency)
      tempDependency = null;
    return returnDependency;
  }
    
  private transient List tagITWList = null;//new ArrayList();

  private void expandDependency(Dependency dependency, double count) {
    //if (Test.prunePunc && pruneTW(dependency.arg))
    //  return;
    if (dependency.head == null)
      return;
    if (dependency.arg == null)
      return;
    if (tagITWList == null)
      tagITWList = new ArrayList();
    while (dependency.head.tag+2 >= tagITWList.size() || dependency.arg.tag+2 >= tagITWList.size())
      tagITWList.add(null);
    IntTaggedWord headT = (IntTaggedWord) tagITWList.get(dependency.head.tag+2);
    if (headT == null) {
      headT = new IntTaggedWord(-1,dependency.head.tag);
      tagITWList.set(dependency.head.tag+2, headT);
    }
    IntTaggedWord argT = (IntTaggedWord) tagITWList.get(dependency.arg.tag+2);
    if (argT == null) {
      argT = new IntTaggedWord(-1,dependency.arg.tag);
      tagITWList.set(dependency.arg.tag+2, argT);
    }
    IntTaggedWord head = dependency.head;
    IntTaggedWord arg = dependency.arg;
    boolean leftHeaded = dependency.leftHeaded;
    short valBinDist = valenceBin(dependency.distance);
    short distBinDist = distanceBin(dependency.distance);

    boolean isStop = (arg.word == -2);
    if (! isStop) {
      argCounter.add(intern(head,arg,leftHeaded,valBinDist),count);
      argCounter.add(intern(headT,arg,leftHeaded,valBinDist),count);
      argCounter.add(intern(head,argT,leftHeaded,valBinDist),count);
      argCounter.add(intern(headT,argT,leftHeaded,valBinDist),count);

      argCounter.add(intern(head,wildTW,leftHeaded,valBinDist),count);
      argCounter.add(intern(headT,wildTW,leftHeaded,valBinDist),count);

      argCounter.add(intern(wildTW,arg,false,-1),count);
      argCounter.add(intern(wildTW,argT,false,-1),count);
    }
    stopCounter.add(intern(head,arg,leftHeaded,distBinDist),count);
    stopCounter.add(intern(head,wildTW,leftHeaded,distBinDist),count);
    stopCounter.add(intern(headT,arg,leftHeaded,distBinDist),count);
    stopCounter.add(intern(headT,wildTW,leftHeaded,distBinDist),count);
  }

  protected static short valenceBin(int distance) {
    if (! Options.get().distance)
      return 0;
    if (distance < 0)
      return -1;
    if (distance == 0)
      return 0;
    return 1;
  }

  public static int NUMDISTBIN() { return Options.get().coarseDistance ? 4: 5; }

  protected static short distanceBin(int distance) {
    if (! Options.get().distance)
      return 0;
    else if (Options.get().coarseDistance) {
      if (distance <= 0)
        return 0;
      if (distance <= 2)
        return 1;
      if (distance <= 5)
        return 2;
      else
        return 3;
    } else {
      if (distance <= 0)
        return 0;
      if (distance <= 1)
        return 1;
      if (distance <= 5)
        return 2;
      if (distance <= 10)
        return 3;
      else
        return 4;
    }
  }

  /*
  protected Dependency tempDependency = new Dependency(-2,-2,-2,-2,false,0);
  public double oldSscore(Dependency dependency) {
    if (! Test.directional)
      dependency.leftHeaded = false;
    // score the link, plus the stops for the arg
    double score = 0.0;
    score += rawScore(dependency);
    tempDependency.head.word = dependency.arg.word;
    tempDependency.head.tag = dependency.arg.tag;
    if (Test.directional && Test.genStop && ! Test.brokenDep) {
      tempDependency.leftHeaded = true;
      score += rawScore(tempDependency);
    }
    if (Test.genStop) {
      tempDependency.leftHeaded = false;;
      score += rawScore(tempDependency);
    }
    return score;
    }*/

  public double score(Dependency dependency) {
    boolean verbose = false;
    if (! Options.get().directional)
      dependency.leftHeaded = false;
    if (verbose)
      System.out.println("Generating "+dependency);
    short distance = dependency.distance;
    boolean leftHeaded = dependency.leftHeaded;
    IntTaggedWord aTW = dependency.arg;
    IntTaggedWord hTW = dependency.head;
    int hW = dependency.head.word;
    int aW = dependency.arg.word;

    dependency.distance = valenceBin(distance);

    double c_aTW_hTWd = argCounter.countOf(dependency);
    dependency.arg.word = -1;
    double c_aT_hTWd = argCounter.countOf(dependency);
    dependency.arg.word = aW;
    dependency.arg = wildTW;
    double c_hTWd = argCounter.countOf(dependency);
    dependency.arg = aTW;

    dependency.head.word = -1;
    double c_aTW_hTd = argCounter.countOf(dependency);
    dependency.arg.word = -1;
    double c_aT_hTd = argCounter.countOf(dependency);
    dependency.arg.word = aW;
    dependency.arg = wildTW;
    double c_hTd = argCounter.countOf(dependency);
    dependency.arg = aTW;
    dependency.head.word = hW;

    dependency.distance = distanceBin(distance);
    dependency.arg = stopTW;
    double c_stop_hTWds = stopCounter.countOf(dependency);
    dependency.head.word = -1;
    double c_stop_hTds = stopCounter.countOf(dependency);
    dependency.head.word = hW;
    dependency.arg = wildTW;
    double c_hTWds = stopCounter.countOf(dependency);
    dependency.head.word = -1;
    double c_hTds = stopCounter.countOf(dependency);
    dependency.head.word = hW;
    dependency.arg = aTW;

    dependency.head = wildTW;
    dependency.leftHeaded = false;
    dependency.distance = -1;
    double c_aTW = argCounter.countOf(dependency);
    dependency.arg.word = -1;
    double c_aT = argCounter.countOf(dependency);
    dependency.arg.word = aW;
    dependency.leftHeaded = leftHeaded;
    dependency.head = hTW;
    dependency.distance = distance;

    // decide whether to generate something

    double p_stop_hTWds = (c_hTWds>0.0 ? c_stop_hTWds/c_hTWds : 0.0);
    double p_stop_hTds = (c_hTds>0.0 ? c_stop_hTds/c_hTds : 1.0);

    double pb_stop_hTWds = (c_stop_hTWds+smooth[2]*p_stop_hTds)/(c_hTWds+smooth[2]);

    if (verbose) {
      System.out.println("  c_stop_hTWds: "+c_stop_hTWds+ "; c_hTWds: "+c_hTWds+
                         "; c_stop_hTds: "+c_stop_hTds+"; c_hTds: "+c_hTds);
      System.out.println("  Generate STOP prob: "+pb_stop_hTWds);
    }

    if (aW == -2) {
      // did we generate stop?
      if (rootTW(hTW))
        return 0.0;
      return Test.depWeight*Math.log(pb_stop_hTWds);
    }

    double pb_go_hTWds = 1.0-pb_stop_hTWds;

    if (rootTW(hTW)) {
      pb_go_hTWds = 1.0;
    }

    // generate the argument

    // do the magic
    double p_aTW_hTWd = (c_hTWd>0.0 ? c_aTW_hTWd/c_hTWd : 0.0);
    double p_aTW_hTd = (c_hTd>0.0 ? c_aTW_hTd/c_hTd : 0.0);
    double p_aT_hTWd = (c_hTWd>0.0 ? c_aT_hTWd/c_hTWd : 0.0);
    double p_aT_hTd = (c_hTd>0.0 ? c_aT_hTd/c_hTd : 0.0);
    double p_aTW_aT = (c_aTW>0.0 ? c_aTW/c_aT : 1.0);
    //double p_hd_hTW = (c_hTW>0.0 ? c_hTWd/c_hTW : 0.0);
    //double p_hd_hT = (c_hT>0.0 ? c_hTd/c_hT : 0.0);

    double pb_aTW_hTWd = (c_aTW_hTWd+smooth[0]*p_aTW_hTd)/(c_hTWd+smooth[0]);
    double pb_aT_hTWd = (c_aT_hTWd+smooth[1]*p_aT_hTd)/(c_hTWd+smooth[1]);

    //double nScore = Math.log(p_aT_hTd);

    double score = (Math.log(interp[0]*pb_aTW_hTWd+
                             (1.0-interp[0])*p_aTW_aT*pb_aT_hTWd)+
                    Math.log(pb_go_hTWds));

    if (verbose) {
      NumberFormat nf = NumberFormat.getNumberInstance();
      nf.setMaximumFractionDigits(2);
      System.out.println("  c_aTW_hTWd: "+c_aTW_hTWd+ "; c_aT_hTWd: "+c_aT_hTWd+"; c_hTWd: "+c_hTWd);
      System.out.println("  c_aTW_hTd: "+c_aTW_hTd+"; c_aT_hTd: "+c_aT_hTd+"; c_hTd: "+c_hTd);
      System.out.println("  Generated with pb_go_hTWds: "+nf.format(Math.log(pb_go_hTWds))+" pb_aTW_hTWd: "+nf.format(Math.log(pb_aTW_hTWd))+" p_aTW_aT: "+nf.format(Math.log(p_aTW_aT))+" pb_aT_hTWd: "+nf.format(Math.log(pb_aT_hTWd)));
      System.out.println("  NoDist log score: "+score);
    }

    if (Test.prunePunc && pruneTW(aTW))
      return 0.0;

    if (Double.isNaN(score))
      score = Double.NEGATIVE_INFINITY;

    //if (Test.rightBonus && ! dependency.leftHeaded)
    //  score -= 0.2;

    if (score < -100)
      score = Double.NEGATIVE_INFINITY;

    return Test.depWeight*score;
  }

  private Dependency d = new Dependency(-2,-2,-2,-2,false,0);

  public double score(int headWord, int headTag, int argWord, int argTag, boolean leftHeaded, int dist) {
    d.head.word = headWord;
    d.head.tag = (short)headTag;
    d.arg.word = argWord;
    d.arg.tag = (short)argTag;
    d.leftHeaded = leftHeaded;
    d.distance = (short)dist;
    return score(d);
  }

  public DependencyGrammar(int numWords, int numTags) {
    this.numWords = numWords;
    this.numTags = numTags;
    //headArgCounter = new Counter();
    //headDistCounter = new Counter();
    //headCounter = new Counter();
    //lheadCounter = new Counter();
    //rheadCounter = new Counter();
    argCounter = new Counter();
    stopCounter = new Counter();
  }

  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject();
    argCounter = new Counter();
    stopCounter = new Counter();
    for (Iterator dI = coreDependencies.seenSet().iterator(); dI.hasNext();) {
      Dependency d = (Dependency)dI.next();
      double count = coreDependencies.countOf(d);
      expandDependency(d,count);
    }
    expandDependencyMap = null;
  }

  public Counter getCoreDependencies() {
    return coreDependencies;
  }

    /**
     * Populates data in this Object from the character stream 
     * given by the Reader r.
     */
    public void readData(Reader r) throws IOException {
	final String LEFT = "left";
	BufferedReader in = new BufferedReader(r);
	String line, headString, argString, dirString, distanceString, countString;
	int lineNum = 1;
	StringTokenizer tok;
	Dependency d;
	boolean leftHeaded;
	// all lines have one rule per line
	while ((line = in.readLine()) != null) {
	    try {
		tok = new StringTokenizer(line);
		headString = tok.nextToken();
		tok.nextToken();
		argString = tok.nextToken();
		dirString = tok.nextToken();
		if (dirString.equals(LEFT))
		    leftHeaded = true;
		else
		    leftHeaded = false;
		distanceString = tok.nextToken();
		countString = tok.nextToken();
		d = new Dependency(new IntTaggedWord(headString), new IntTaggedWord(argString), 
				   leftHeaded, Integer.parseInt(distanceString));
		addRule(d, Double.parseDouble(countString));
	    } catch (Exception e) {
		throw new IOException("Error on line " + lineNum);
	    }	    
	    lineNum++;   
	}
    }

    /**
     * Writes out data from this Object to the Writer w.
     */
    public void writeData(Writer w) throws IOException {
	PrintWriter out = new PrintWriter(w);
	// all lines have one rule per line
	for (Iterator dI = coreDependencies.seenSet().iterator(); dI.hasNext();) {
	    Dependency d = (Dependency) dI.next();
	    double count = coreDependencies.countOf(d);
	    out.println(d + " " + count);
	}
	out.flush();
    }

  private static final long serialVersionUID = 1L;

} // end class DependencyGrammar


class IntTaggedWord implements Serializable, Comparable {

    Numberer wordNumberer = Numberer.getGlobalNumberer("words");
    Numberer tagNumberer = Numberer.getGlobalNumberer("tags");

  public int word;
  public short tag;
    public static final String ANY = "*";
    public static final String STOP = "STOP";

  public int tag() { return tag; }
  public int word() { return word; }

  public int hashCode() { return word^(tag << 16); }

  public boolean equals(Object o) {
    if (this == o) {
      return true;
    } else if (o instanceof IntTaggedWord) {
      IntTaggedWord i = (IntTaggedWord)o;
      return (word==i.word && tag==i.tag);
    } else {
      return false;
    }
  }

  public int compareTo(Object obj) {
    IntTaggedWord that = (IntTaggedWord) obj;
    if (tag != that.tag) {
      return tag - that.tag;
    } else {
      return word - that.word;
    }
  }

  public String toString() {
    String wordStr;
    String tagStr;
    if (word >= 0) {
      wordStr = (String)wordNumberer.object(word);
    } else if (word == -1) {
	wordStr = ANY;
    } else {
      wordStr = STOP;
    }
    if (tag >= 0) {
      tagStr = (String)tagNumberer.object(tag);
    } else if (tag == -1) {
      tagStr = ANY;
    } else {
      tagStr = STOP;
    }
    return wordStr+"|"+tagStr;
  }

  public IntTaggedWord(int word, int tag) {
    this.word = word;
    this.tag = (short)tag;
  }

  public TaggedWord toTaggedWord() {
    String wordStr;
    String tagStr;
    if (word >= 0) {
      wordStr = (String)wordNumberer.object(word);
    } else if (word == -1) {
      wordStr = ANY;
    } else {
      wordStr = STOP;
    }
    if (tag >= 0) {
      tagStr = (String)tagNumberer.object(tag);
    } else if (tag == -1) {
      tagStr = ANY;
    } else {
      tagStr = STOP;
    }
    return new TaggedWord(wordStr, tagStr);
  }

    /**
     * Creates an IntTaggedWord given by the String representation
     * of the form <word>|<tag>
     */
    public IntTaggedWord(String s) throws Exception {
	String wordString, tagString;
	int word, tag;
	int n = s.indexOf('|');
	wordString = s.substring(0,n);
	tagString = s.substring(n+1, s.length());
	if (wordString.equals(ANY)) {
	    word = -1;
	} else {
	    word = wordNumberer.number(wordString);
	}
	if (tagString.equals(ANY)) {
	    tag = -1;
	} else {
	    tag = tagNumberer.number(tagString);
	}
	this.word = word;
	this.tag = (short)tag;
    }

  private static final long serialVersionUID = 1L;

} // end class IntTaggedWord


class Dependency implements Serializable {

  public static final String LEFT = "left";
  public static final String RIGHT = "right";

  public IntTaggedWord head;
  public IntTaggedWord arg;
  public boolean leftHeaded;
  public short distance;

  public int hashCode() {
    return head.hashCode()^(arg.hashCode() << 8)^((leftHeaded ? 1 : 0) << 15)^
      (distance << 16);
  }

  public boolean equals(Object o) {
    if (this == o)
      return true;
    if (o instanceof Dependency) {
      Dependency d = (Dependency)o;
      return (head.equals(d.head) &&
              arg.equals(d.arg) &&
              distance == d.distance &&
              leftHeaded == d.leftHeaded);
    } else {
      return false;
    }
  }

  public String toString() {
    return head + " -> " + arg + " " + (leftHeaded ? LEFT : RIGHT) +
      " " + distance;
  }

  public Dependency() {
    // head = null;
    // arg = null;
    // leftHeaded = false;
  }

  public Dependency(IntTaggedWord head, IntTaggedWord arg, boolean leftHeaded, int distance) {
    this.head = head;
    this.arg = arg;
    this.distance = (short)distance;
    this.leftHeaded = leftHeaded;
  }

  public Dependency(int headWord, int headTag, int argWord, int argTag, boolean leftHeaded, int distance) {
    this.head = new IntTaggedWord(headWord,headTag);
    this.arg = new IntTaggedWord(argWord,argTag);
    this.distance = (short)distance;
    this.leftHeaded = leftHeaded;
  }

  private static final long serialVersionUID = 1L;

} // end class Dependency


class DependencyGrammarExtractor extends AbstractTreeExtractor {
  private Numberer wordNumberer = Numberer.getGlobalNumberer("words");
  private Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  private Counter dependencyCounter = new Counter();
  private Set dependencies = new HashSet();

  protected void tallyRoot(Tree lt) {
    List deps = DependencyGrammar.treeToDependencyList(lt);
    for (Iterator i=deps.iterator(); i.hasNext();) {
      Dependency dependency = (Dependency)i.next();
      dependencyCounter.add(dependency,1.0);
      dependencies.add(dependency);
      /*
      int aW = dependency.arg.word;
      int aT = dependency.arg.tag;
      if (Test.genStop) {
        Dependency dependencyEndL = new Dependency(aW,aT,-2,-2,false,0);
        dependencyCounter.add(dependencyEndL,1.0);
        dependencies.add(dependencyEndL);
      }
      if (Test.genStop && Test.directional && ! Test.brokenDep) {
        Dependency dependencyEndR = new Dependency(aW,aT,-2,-2,true,0);
        dependencyCounter.add(dependencyEndR,1.0);
        dependencies.add(dependencyEndR);
      }
      */
    }
  }

  /*
  protected void tallyInternalNodeOld(Tree lt) {
    if (lt.children().length != 2)
      return;
    String hTag = ((HasTag)lt.label()).tag();
    String lTag = ((HasTag)lt.children()[0].label()).tag();
    String rTag = ((HasTag)lt.children()[1].label()).tag();
    String hWord = ((HasWord)lt.label()).word();
    String lWord = ((HasWord)lt.children()[0].label()).word();
    String rWord = ((HasWord)lt.children()[1].label()).word();
    String aTag = (hTag.equals(lTag) ? rTag : lTag);
    String aWord = (hWord.equals(lWord) ? rWord : lWord);
    int hT = tagNumberer.number(hTag);
    int aT = tagNumberer.number(aTag);
    int hW = wordNumberer.number(hWord);
    int aW = wordNumberer.number(aWord);
    boolean leftHeaded = (hWord.equals(lWord) ? true : false);
    Dependency dependency = new Dependency(hW,hT,aW,aT, leftHeaded,0);
    dependencyCounter.add(dependency,1.0);
    dependencies.add(dependency);
    if (Test.genStop) {
      Dependency dependencyEndL = new Dependency(aW,aT,-2,-2,false,0);
      dependencyCounter.add(dependencyEndL,1.0);
      dependencies.add(dependencyEndL);
    }
    if (Test.genStop && Test.directional && ! Test.brokenDep) {
      Dependency dependencyEndR = new Dependency(aW,aT,-2,-2,true,0);
      dependencyCounter.add(dependencyEndR,1.0);
      dependencies.add(dependencyEndR);
    }
  }
  */

  protected Object formResult() {
    wordNumberer.number(Lexicon.UNKNOWN_WORD);
    int numWords = wordNumberer.total();
    int numTags = tagNumberer.total();
    DependencyGrammar dg = new DependencyGrammar(numWords, numTags);
    for (Iterator dependencyI=dependencies.iterator(); dependencyI.hasNext();) {
      Dependency dependency = (Dependency)dependencyI.next();
      dg.addRule(dependency,dependencyCounter.countOf(dependency));
    }
    return dg;
  }

} // end class DependencyGrammarExtractor


class ExhaustiveDependencyParser implements Scorer {

  private Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  private Numberer wordNumberer = Numberer.getGlobalNumberer("words");
  //private Numberer stateNumberer = Numberer.getGlobalNumberer("states");

  private TreeFactory tf;

  private DependencyGrammar dg;
  private Lexicon lex;

  private List sentence;
  private int[] words;

  private float[][][] iScoreH; // headPos, headTag, cornerPos
  private float[][][] oScoreH; // headPos, headTag, cornerPos

  private int[][] rawDistance;
  int[][] binDistance;       // reused in other class, so can't be private
  float[][][][][] headScore;
  float[][][] headStop; // headPos, headTag, split
  private boolean[][][] oPossibleByL;
  private boolean[][][] oPossibleByR;
  private boolean[][][] iPossibleByL;
  private boolean[][][] iPossibleByR;

  float oScore(int start, int end, int head, int tag) {
    return oScoreH[head][tag][start]+oScoreH[head][tag][end];
  }

  float iScore(int start, int end, int head, int tag) {
    return iScoreH[head][tag][start]+iScoreH[head][tag][end];
  }

  public double oScore(Edge edge) {
    return oScore(edge.start,edge.end,edge.head,edge.tag);
  }

  public double iScore(Edge edge) {
    return iScore(edge.start,edge.end,edge.head,edge.tag);
  }

  public boolean oPossible(Hook hook) {
    return (hook.isPreHook() ?
            oPossibleByR[hook.end][hook.head][hook.tag] :
            oPossibleByL[hook.start][hook.head][hook.tag]);
  }

  public boolean iPossible(Hook hook) {
    return (hook.isPreHook() ?
            iPossibleByR[hook.start][hook.head][hook.tag] :
            iPossibleByL[hook.end][hook.head][hook.tag]);
  }

  public boolean parse(List sentence) {
    if (Test.verbose) {
      Timing.tick("Starting dependency parse.");
    }
    this.sentence = sentence;
    int length = sentence.size();
    if (length > Test.maxLength+1)
      return false;
    if (Test.verbose) {
      System.out.print("Initializing...");
    }
    // map to words
    words = new int[length];
    int numTags = tagNumberer.total();
    //System.out.println("\nNumTags: "+numTags);
    //System.out.println(tagNumberer);
    boolean[][] hasTag = new boolean[length][numTags];
    for (int i=0; i<length; i++) {
      //if (wordNumberer.hasSeen(sentence.get(i).toString()))
      Object o = sentence.get(i);
      if (o instanceof HasWord)
        o = ((HasWord)o).word();
      words[i] = wordNumberer.number(o.toString());
        //else
        //words[i] = wordNumberer.number(Lexicon.UNKNOWN_WORD);
    }
    for (int head=0; head<length; head++) {
      for (int tag=0; tag<numTags; tag++) {
	Arrays.fill(iScoreH[head][tag], Float.NEGATIVE_INFINITY);
	Arrays.fill(oScoreH[head][tag], Float.NEGATIVE_INFINITY);
      }
    }
    for (int head=0; head<length; head++) {
      for (int loc=0; loc<=length; loc++) {
        rawDistance[head][loc] = (head >= loc ? head-loc : loc-head-1);
        binDistance[head][loc] = dg.distanceBin(rawDistance[head][loc]);
      }
    }
    // do tags
    if (! Options.get().flexiTag) {
      for (int start = 0; start+1 <= length; start++) {
        int word = words[start];
        for (Iterator taggingI = lex.ruleIteratorByWord(word, start);
             taggingI.hasNext(); ) {
          IntTaggedWord tagging = (IntTaggedWord)taggingI.next();
          int tag = tagging.tag;
          float score = (float) lex.score(tagging,start);
          iScoreH[start][tag][start] = (Test.dcTags ? (float)Test.depWeight*score : 0.0f);
          iScoreH[start][tag][start+1] = 0.0f;
          //System.out.println(tagging+" got score "+score);
        }
      }
    } else {
      for (int start = 0; start+1 <= length; start++) {
        int word = words[start];
        for (int tag=0; tag<numTags; tag++) {
	  iScoreH[start][tag][start] = (float)(Test.dcTags ? (float)Test.depWeight*lex.score(new IntTaggedWord(word, tag), start) : 0.0f);
          iScoreH[start][tag][start+1] = 0.0f;
        }
      }
    }
    for (int hWord=0; hWord<length; hWord++) {
      for (int hTag=0; hTag<numTags; hTag++) {
	hasTag[hWord][hTag] = (iScore(hWord,hWord+1,hWord,hTag) > Float.NEGATIVE_INFINITY);
	Arrays.fill(headStop[hWord][hTag], Float.NEGATIVE_INFINITY);
	for (int aWord=0; aWord<length; aWord++) {
	  for (int dist=0; dist < DependencyGrammar.NUMDISTBIN(); dist++) {
	    Arrays.fill(headScore[dist][hWord][hTag][aWord], Float.NEGATIVE_INFINITY);
	  }
	}
      }
    }
    // score all pairs
    //int hit = 0;
    for (int hWord=0; hWord<length; hWord++) {
      for (int hTag=0; hTag<numTags; hTag++) {
        //Arrays.fill(headStopL[hWord][hTag], Float.NEGATIVE_INFINITY);
        //Arrays.fill(headStopR[hWord][hTag], Float.NEGATIVE_INFINITY);
        //Arrays.fill(headStop[hWord][hTag], Float.NEGATIVE_INFINITY);
        if (! hasTag[hWord][hTag])
	  continue;
	for (int split=0; split<=length; split++) {
	  if (split <= hWord) {
	    headStop[hWord][hTag][split] = (float) dg.score(words[hWord],
							    hTag,-2,-2,false,hWord-split);
	    //System.out.println("headstopl " + hWord +" " + hTag + " " + split + " " + headStopL[hWord][hTag][split]); // debugging
	  } else {
	    headStop[hWord][hTag][split] = (float) dg.score(words[hWord],
							    hTag,-2,-2,true,split-hWord-1);
	    //System.out.println("headstopr " + hWord +" " + hTag + " " + split + " " + headStopR[hWord][hTag][split]); // debugging
	  }
	  //hit++;
	}
	//Timing.tick("hWord: "+hWord+" hTag: "+hTag+" piddle count: "+hit);
	for (int aWord=0; aWord<length; aWord++) {
	  for (int aTag=0; aTag<numTags; aTag++) {
	    if (! hasTag[aWord][aTag])
	      continue;
	    for (int split=0; split<=length; split++) {
	      boolean leftHeaded = hWord < aWord;
	      if ((leftHeaded && (split <= hWord || split > aWord)) ||
		  ((! leftHeaded) && (split <= aWord || split > hWord)))
		continue;
	      int headDistance = rawDistance[hWord][split];
	      int binDist = binDistance[hWord][split];
	      headScore[binDist][hWord][hTag][aWord][aTag] =
		(float)dg.score(words[hWord],hTag,words[aWord],aTag,leftHeaded,headDistance);
	      //hit++;
	      while (split+1 <= length && binDistance[hWord][split+1] == binDist)
		split++;
	      //System.out.println("headscore " + headScore[binDistance[hWord][split]][hWord][hTag][aWord][aTag]);
	      if (false) {
		System.out.println("Dependency score for " +
				   wordNumberer.object(words[hWord]) + " -> " +
				   wordNumberer.object(words[aWord]) + " = " +
				   headScore[binDistance[hWord][split]][hWord][hTag][aWord][aTag]);
	      }
	    }
	  }
	}
      }
    }
    if (Test.verbose) {
      Timing.tick("done.");
      // displayHeadScores();
      System.out.print("Starting insides...");
    }
    // do larger spans
    for (int diff = 2; diff <= length; diff++) {
      for (int start = 0; start+diff <= length; start++) {
        int end = start+diff;
	// left extension
	int endHead = end-1;
	for (int endTag = 0; endTag < numTags; endTag++) {
	  if (! hasTag[endHead][endTag])
	    continue;
	  for (int argHead = start; argHead < end-1; argHead++) {
	    for (int argTag = 0; argTag < numTags; argTag++) {
	      if (! hasTag[argHead][argTag])
		continue;
	      float bestScore = iScoreH[endHead][endTag][start];
	      for (int split = argHead+1; split < end; split++) {
		float score = (iScoreH[endHead][endTag][split]+
				iScore(start,split,argHead,argTag)+
				headScore[binDistance[endHead][split]][endHead][endTag][argHead][argTag]+
				headStop[argHead][argTag][start]+
				headStop[argHead][argTag][split]);
		if (score > bestScore)
		  bestScore = score;
	      }
	      iScoreH[endHead][endTag][start] = bestScore;
	    }
	  }
	}
	// right extension
	int startHead = start;
	for (int startTag = 0; startTag < numTags; startTag++) {
	  if (! hasTag[startHead][startTag])
	    continue;
	  for (int argHead = start+1; argHead < end; argHead++) {
	    for (int argTag = 0; argTag < numTags; argTag++) {
	      if (! hasTag[argHead][argTag])
		continue;
	      float bestScore = iScoreH[startHead][startTag][end];
	      for (int split = start+1; split <= argHead ; split++) {
		float score = (iScoreH[startHead][startTag][split]+
				iScore(split,end,argHead,argTag)+
				headScore[binDistance[startHead][split]][startHead][startTag][argHead][argTag]+
				headStop[argHead][argTag][end]+
				headStop[argHead][argTag][split]);
		if (score > bestScore)
		  bestScore = score;
	      }
	      iScoreH[startHead][startTag][end] = bestScore;
	    }
	  }
	}
      }
    }
    int goalTag = tagNumberer.number(Lexicon.BOUNDARY_TAG);
    if (Test.verbose) {
      Timing.tick("done.");
      System.out.println("Dep  "+length+" iScore "+iScore(0,length,length-1,goalTag));
      System.out.print("Starting outsides...");
    }
    oScoreH[length-1][goalTag][0] = 0.0f;
    oScoreH[length-1][goalTag][length] = 0.0f;
    for (int diff = length; diff > 1; diff--) {
      for (int start = 0; start+diff <= length; start++) {
        int end = start+diff;
	// left half
	int endHead = end-1;
	for (int endTag = 0; endTag < numTags; endTag++) {
	  if (! hasTag[endHead][endTag])
	    continue;
	  for (int argHead = start; argHead < endHead; argHead++) {
	    for (int argTag = 0; argTag < numTags; argTag++) {
	      if (! hasTag[argHead][argTag])
		continue;
	      for (int split = argHead; split <= endHead; split++) {
		float subScore = (oScoreH[endHead][endTag][start]+
				  headScore[binDistance[endHead][split]][endHead][endTag][argHead][argTag]+
				  headStop[argHead][argTag][start]+
				  headStop[argHead][argTag][split]);
		float scoreRight = (subScore+
				    iScoreH[argHead][argTag][start]+
				    iScoreH[argHead][argTag][split]);
		float scoreMid = (subScore+
				  iScoreH[argHead][argTag][start]+
				  iScoreH[endHead][endTag][split]);
		float scoreLeft = (subScore+
				   iScoreH[argHead][argTag][split]+
				   iScoreH[endHead][endTag][split]);
		if (scoreRight > oScoreH[endHead][endTag][split])
		  oScoreH[endHead][endTag][split] = scoreRight;
		if (scoreMid > oScoreH[argHead][argTag][split])
		  oScoreH[argHead][argTag][split] = scoreMid;
		if (scoreLeft > oScoreH[argHead][argTag][start])
		  oScoreH[argHead][argTag][start] = scoreLeft;
	      }
	    }
	  }
	}
	// right half
	int startHead = start;
	for (int startTag = 0; startTag < numTags; startTag++) {
	  if (! hasTag[startHead][startTag])
	    continue;
	  for (int argHead = startHead+1; argHead < end; argHead++) {
	    for (int argTag = 0; argTag < numTags; argTag++) {
	      if (! hasTag[argHead][argTag])
		continue;
	      for (int split = startHead+1; split <= argHead ; split++) {
		float subScore = (oScoreH[startHead][startTag][end]+
				  headScore[binDistance[startHead][split]][startHead][startTag][argHead][argTag]+
				  headStop[argHead][argTag][split]+
				  headStop[argHead][argTag][end]);
		float scoreLeft = (subScore+
				   iScoreH[argHead][argTag][split]+
				   iScoreH[argHead][argTag][end]);
		float scoreMid = (subScore+
				  iScoreH[startHead][startTag][split]+
				  iScoreH[argHead][argTag][end]);
		float scoreRight = (subScore+
				    iScoreH[startHead][startTag][split]+
				    iScoreH[argHead][argTag][split]);
		if (scoreLeft > oScoreH[startHead][startTag][split])
		  oScoreH[startHead][startTag][split] = scoreLeft;
		if (scoreMid > oScoreH[argHead][argTag][split])
		  oScoreH[argHead][argTag][split] = scoreMid;
		if (scoreRight > oScoreH[argHead][argTag][end])
		  oScoreH[argHead][argTag][end] = scoreRight;
	      }
	    }
	  }
	}
      }
    }
    if (Test.verbose) {
      Timing.tick("done.");
      System.out.print("Starting half-filters...");
    }
    for (int loc=0; loc<=length; loc++) {
      for (int head=0; head<length; head++) {
        Arrays.fill(iPossibleByL[loc][head],false);
        Arrays.fill(iPossibleByR[loc][head],false);
        Arrays.fill(oPossibleByL[loc][head],false);
        Arrays.fill(oPossibleByR[loc][head],false);
      }
    }
    for (int head=0; head<length; head++) {
      for (int tag=0; tag<numTags; tag++) {
        if (! hasTag[head][tag])
	  continue;
	for (int start=0; start<=head; start++) {
          for (int end=head+1; end<=length; end++) {
            if (iScore(start,end,head,tag) > Float.NEGATIVE_INFINITY &&
                oScore(start,end,head,tag) > Float.NEGATIVE_INFINITY) {
              iPossibleByR[end][head][tag] = true;
              iPossibleByL[start][head][tag] = true;
              oPossibleByR[end][head][tag] = true;
              oPossibleByL[start][head][tag] = true;
            }
          }
        }
      }
    }
    if (Test.verbose) {
      Timing.tick("done");
    }
    return hasParse();
  }

  public boolean hasParse() {
    int length = sentence.size();
    int goalTag = tagNumberer.number(Lexicon.BOUNDARY_TAG);
    return (iScore(0,length,length-1,goalTag) > Float.NEGATIVE_INFINITY);
  }

  /** This displays a headScore matrix, which will be valid after parsing
   *  a sentence.  Unclear yet whether this is valid/useful [cdm].
   */
  public void displayHeadScores() {
    int numTags = tagNumberer.total();
    System.out.println("---- headScore matrix (head x dep, best tags) ----");
    System.out.print(StringUtils.exactlyN("", 6));
    for (int i = 0; i < words.length; i++) {
      System.out.print(" " + StringUtils.exactlyN(wordNumberer.object(words[i]).toString(),
                                      2));
    }
    System.out.println();
    for (int hWord = 0; hWord < words.length; hWord++) {
      System.out.print(StringUtils.exactlyN(wordNumberer.object(words[hWord]).toString(),
                                6));
      int bigBD = -1, bigHTag = -1, bigATag = -1;
      for (int aWord = 0; aWord < words.length; aWord++) {
        // we basically just max of all the variables, but for distance > 0, we
        // include a factor for generating something at distance 0, or else
        // the result is too whacked out to be useful
        float biggest = Float.NEGATIVE_INFINITY;
        for (int bd = 0; bd < DependencyGrammar.NUMDISTBIN(); bd++) {
          for (int hTag = 0; hTag < numTags; hTag++) {
            /*
            float penalty = 0.0f;
            if (bd != 0) {
              penalty = (float) dg.score(words[hWord], hTag, -2, -2, aWord > hWord, 0);
              penalty = (float) Math.log(1.0 - Math.exp(penalty));
            }
            for (int aTag = 0; aTag < numTags; aTag++) {
              if (headScore[bd][hWord][hTag][aWord][aTag] + penalty > biggest) {
                biggest = headScore[bd][hWord][hTag][aWord][aTag] + penalty;
            */
            for (int aTag = 0; aTag < numTags; aTag++) {
              if (headScore[bd][hWord][hTag][aWord][aTag] > biggest) {
                biggest = headScore[bd][hWord][hTag][aWord][aTag];
                bigBD = bd;
                bigHTag = hTag;
                bigATag = aTag;
              }
            }
          }
        }
        if (Float.isInfinite(biggest)) {
          System.out.print(" " + StringUtils.exactlyN("in", 2));
        } else {
          int score = Math.round(Math.abs(headScore[bigBD][hWord][bigHTag][aWord][bigATag]));
          System.out.print(" " + StringUtils.exactlyN(Integer.toString(score), 2));
        }
      }
      System.out.println();
    }
  }

  private static final double TOL = 1e-5;
  private static boolean matches(double x, double y) {
    return (Math.abs(x-y)/(Math.abs(x)+Math.abs(y)+1e-5) < TOL);
  }
  private Tree extractBestParse(int start, int end, int hWord, int hTag) {
    final boolean verbose = true;
    //System.out.println("Span "+start+" to "+end+" word "+wordNumberer.object(words[hWord])+"/"+hWord+" tag "+tagNumberer.object(hTag)+"/"+hTag+" score "+iScore[start][end][hWord][hTag]);
    String headWordStr = (String)wordNumberer.object(words[hWord]);
    String headTagStr = (String)tagNumberer.object(hTag);
    Label headLabel = new CategoryWordTag(headWordStr,headWordStr,headTagStr);
    List children = new ArrayList();
    // deal with span 1
    if (end-start == 1) {
      Tree leaf = tf.newLeaf(new Word(headWordStr));
      children.add(leaf);
      return tf.newTreeNode(headLabel, children);
    }
    // find backtrace
    double bestScore = iScore(start,end,hWord,hTag);
    for (int split = start+1; split < end; split++) {
      int binD = binDistance[hWord][split];
      if (hWord < split) {
        for (int aWord = split; aWord < end; aWord++) {
          for (int aTag = 0; aTag < tagNumberer.total(); aTag++) {
            if (matches(iScore(start,split,hWord,hTag)+
                        iScore(split,end,aWord,aTag)+
                        headScore[binD][hWord][hTag][aWord][aTag]+
                        headStop[aWord][aTag][split]+
                        headStop[aWord][aTag][end],
                        bestScore)) {
              if (verbose) {
                //String argWordStr=(String)wordNumberer.object(words[aWord]);
                //String argTagStr = (String)tagNumberer.object(aTag);
                //System.out.println(headWordStr+"|"+headTagStr+" -> "+argWordStr+"|"+argTagStr+" "+((int)(10.0*pairScore[hWord][hTag][aWord][aTag]))/10.0);
              }
              // build it
              children.add(extractBestParse(start,split,hWord,hTag));
              children.add(extractBestParse(split,end,aWord,aTag));
              return tf.newTreeNode(headLabel,children);
            }
          }
        }
      } else {
        for (int aWord = start; aWord < split; aWord++) {
          for (int aTag = 0; aTag < tagNumberer.total(); aTag++) {
            if (matches(iScore(start,split,aWord,aTag)+
                        iScore(split,end,hWord,hTag)+
                        headScore[binD][hWord][hTag][aWord][aTag]+
                        headStop[aWord][aTag][start]+
                        headStop[aWord][aTag][split],
                        bestScore)) {
              if (verbose) {
                //String argWordStr=(String)wordNumberer.object(words[aWord]);
                //String argTagStr = (String)tagNumberer.object(aTag);
                //System.out.println(headWordStr+"|"+headTagStr+" -> "+argWordStr+"|"+argTagStr+" "+((int)(10.0*pairScore[hWord][hTag][aWord][aTag]))/10.0);
              }
              children.add(extractBestParse(start,split,aWord,aTag));
              children.add(extractBestParse(split,end,hWord,hTag));
              return tf.newTreeNode(headLabel,children);
              // build it
            }
          }
        }
      }
    }
    System.err.println("Problem in ExhaustiveDependencyParser::extractBestParse");
    return null;
  }

  private Tree flatten(Tree tree) {
    if (tree.isLeaf() || tree.isPreTerminal())
      return tree;
    List newChildren = new ArrayList();
    Tree[] children = tree.children();
    for (int c=0; c<children.length; c++) {
      Tree child = children[c];
      Tree newChild = flatten(child);
      if (!newChild.isPreTerminal() &&
          newChild.label().toString().equals(tree.label().toString()))
        newChildren.addAll(newChild.getChildrenAsList());
      else
        newChildren.add(newChild);
    }
    return tf.newTreeNode(tree.label(),newChildren);
  }

  public Tree getBestParse() {
    return flatten(extractBestParse(0,words.length,words.length-1,tagNumberer.number(Lexicon.BOUNDARY_TAG)));
  }

  public ExhaustiveDependencyParser(DependencyGrammar dg, Lexicon lex) {
    this.dg = dg;
    this.lex = lex;
    tf = new LabeledScoredTreeFactory(new StringLabelFactory());
    int length = Test.maxLength+1;
    int tagNum = tagNumberer.total();

    iScoreH = new float[length+1][tagNum][length+1];
    oScoreH = new float[length+1][tagNum][length+1];
    iPossibleByL = new boolean[length+1][length+1][tagNum];
    iPossibleByR = new boolean[length+1][length+1][tagNum];
    oPossibleByL = new boolean[length+1][length+1][tagNum];
    oPossibleByR = new boolean[length+1][length+1][tagNum];
    headScore = new float[DependencyGrammar.NUMDISTBIN()][length][tagNum][length][tagNum];
    headStop = new float[length+1][tagNum][length+1];
    rawDistance = new int[length+1][length+1];
    binDistance = new int[length+1][length+1];

  }
}

// A*

interface Scorer {
  public double oScore(Edge edge);
  public double iScore(Edge edge);
  public boolean oPossible(Hook hook);
  public boolean iPossible(Hook hook);
  public boolean parse(List words);
}

class TwinScorer implements Scorer {
  private Scorer scorer1;
  private Scorer scorer2;
  public double oScore(Edge edge) {
    return scorer1.oScore(edge)+scorer2.oScore(edge);
  }
  public double iScore(Edge edge) {
    return scorer1.iScore(edge)+scorer2.iScore(edge);
  }
  public boolean oPossible(Hook hook) {
    return scorer1.oPossible(hook) && scorer2.oPossible(hook);
  }
  public boolean iPossible(Hook hook) {
    return scorer1.iPossible(hook) && scorer2.iPossible(hook);
  }
  public boolean parse(List words) {
    boolean b1 = scorer1.parse(words);
    boolean b2 = scorer2.parse(words);
    return (b1 && b2);
  }
  public TwinScorer(Scorer scorer1, Scorer scorer2) {
    this.scorer1 = scorer1;
    this.scorer2 = scorer2;
  }
}

class MaxScorer implements Scorer {
  private Scorer scorer1;
  private Scorer scorer2;
  public double oScore(Edge edge) {
    return SloppyMath.min(scorer1.oScore(edge),scorer2.oScore(edge));
  }
  public double iScore(Edge edge) {
    return SloppyMath.min(scorer1.iScore(edge),scorer2.iScore(edge));
  }
  public boolean oPossible(Hook hook) {
    return scorer1.oPossible(hook) && scorer2.oPossible(hook);
  }
  public boolean iPossible(Hook hook) {
    return scorer1.iPossible(hook) && scorer2.iPossible(hook);
  }
  public boolean parse(List words) {
    boolean b1 = scorer1.parse(words);
    boolean b2 = scorer2.parse(words);
    return (b1 && b2);
  }
  public MaxScorer(Scorer scorer1, Scorer scorer2) {
    this.scorer1 = scorer1;
    this.scorer2 = scorer2;
  }
}

// A* Guts

class ExhaustivePCFGParser implements Scorer {

  // public static long insideTime = 0;  // for profiling
  // public static long outsideTime = 0;

  private static final String goalStr =
    Options.get().langpack.startSymbol();
  private String stateSpace;
  private Numberer stateNumberer;
  private Numberer wordNumberer = Numberer.getGlobalNumberer("words");
  private Numberer tagNumberer = Numberer.getGlobalNumberer("tags");

  private TreeFactory tf;

  private BinaryGrammar bg;
  private UnaryGrammar ug;
  private Lexicon lex;

  private OutsideRuleFilter orf;

  private float[][][] iScore;  // start idx, end idx, state
  private float[][][] oScore;  // start idx, end idx, state
  private float bestScore;

  private boolean[][] oFilteredStart;
  private boolean[][] oFilteredEnd;

  private boolean[][] iPossibleByL;
  private boolean[][] iPossibleByR;
  private boolean[][] oPossibleByL;
  private boolean[][] oPossibleByR;

  private int[] words;
  private int length;
  private boolean[][] tags;

  private int numStates;

  public double oScore(Edge edge) {
    if (Test.pcfgThreshold) {
      double oS = oScore[edge.start][edge.end][edge.state];
      double iS = iScore[edge.start][edge.end][edge.state];
      if (iS+oS-bestScore < Test.pcfgThresholdValue)
        return Double.NEGATIVE_INFINITY;
      else
        return oS;
    } else {
      return oScore[edge.start][edge.end][edge.state];
    }
  }

  public double iScore(Edge edge) {
    return iScore[edge.start][edge.end][edge.state];
  }

  public boolean oPossible(Hook hook) {
    return (hook.isPreHook() ?
            oPossibleByR[hook.end][hook.state] :
            oPossibleByL[hook.start][hook.state]);
  }

  public boolean iPossible(Hook hook) {
    return (hook.isPreHook() ?
            iPossibleByR[hook.start][hook.subState] :
            iPossibleByL[hook.end][hook.subState]);
  }


  public boolean oPossibleL(int state, int start) {
    return oPossibleByL[start][state];
  }
  public boolean oPossibleR(int state, int end) {
    return oPossibleByR[end][state];
  }
  public boolean iPossibleL(int state, int start) {
    return iPossibleByL[start][state];
  }
  public boolean iPossibleR(int state, int end) {
    return iPossibleByR[end][state];
  }

  private void buildOFilter() {
    orf.init();
    for (int start = 0; start<length; start++) {
      orf.leftAccepting(oFilteredStart[start]);
      orf.advanceRight(tags[start]);
    }
    for (int end=length; end>0; end--) {
      orf.rightAccepting(oFilteredEnd[end]);
      orf.advanceLeft(tags[end-1]);
    }
  }


  public double validateBinarizedTree(Tree tree, int start) {
    if (tree.isLeaf())
        return 0.0;
    if (tree.isPreTerminal()) {
      int tag = Numberer.number("tags",tree.label().value());
      int word = Numberer.number("words",tree.children()[0].label().value());
      IntTaggedWord iTW = new IntTaggedWord(word,tag);
      double score = lex.score(iTW,start);
      double bound = iScore[start][start+1][Numberer.number(stateSpace,tree.label().value())];
      if (score > bound) {
        System.out.println("Invalid tagging:");
        System.out.println("  Tag: "+tree.label().value());
        System.out.println("  Word: "+tree.children()[0].label().value());
        System.out.println("  Score: "+score);
        System.out.println("  Bound: "+bound);
      }
      return score;
    }
    int parent = Numberer.number(stateSpace,tree.label().value());
    int firstChild = Numberer.number(stateSpace,tree.children()[0].label().value());
    if (tree.children().length == 1) {
      UnaryRule ur = new UnaryRule();
      ur.parent = parent;
      ur.child = firstChild;
      double score = SloppyMath.max(ug.scoreRule(ur),-10000)+validateBinarizedTree(tree.children()[0],start);
      double bound = iScore[start][start+tree.yield().size()][parent];
      if (score > bound) {
        System.out.println("Invalid unary:");
        System.out.println("  Parent: "+tree.label().value());
        System.out.println("  Child: "+tree.children()[0].label().value());
        System.out.println("  Start: "+start);
        System.out.println("  End: "+(start+tree.yield().size()));
        System.out.println("  Score: "+score);
        System.out.println("  Bound: "+bound);
      }
      return score;
    }
    BinaryRule br = new BinaryRule();
    br.parent = parent;
    br.leftChild = firstChild;
    br.rightChild = Numberer.number(stateSpace,tree.children()[1].label().value());
    double score = SloppyMath.max(bg.scoreRule(br),-10000)+
      validateBinarizedTree(tree.children()[0],start)+
      validateBinarizedTree(tree.children()[1],start+tree.children()[0].yield().size());
    double bound = iScore[start][start+tree.yield().size()][parent];
    if (score > bound) {
      System.out.println("Invalid binary:");
      System.out.println("  Parent: "+tree.label().value());
      System.out.println("  LChild: "+tree.children()[0].label().value());
      System.out.println("  RChild: "+tree.children()[1].label().value());
      System.out.println("  Start: "+start);
      System.out.println("  End: "+(start+tree.yield().size()));
      System.out.println("  Score: "+score);
      System.out.println("  Bound: "+bound);
    }
    return score;
  }

  //
  public double scoreBinarizedTree(Tree tree, boolean leftmost) {
    if (tree.isLeaf())
        return 0.0;
    if (tree.isPreTerminal()) {
      int tag = Numberer.number("tags",tree.label().value());
      int word = Numberer.number("words",tree.children()[0].label().value());
      IntTaggedWord iTW = new IntTaggedWord(word,tag);
      // if (lex.score(iTW,(leftmost ? 0 : 1)) == Double.NEGATIVE_INFINITY) {
      //   System.out.println("NO SCORE FOR: "+iTW);
      // }
      return lex.score(iTW,(leftmost ? 0 : 1));
    }
    int parent = Numberer.number(stateSpace,tree.label().value());
    int firstChild = Numberer.number(stateSpace,tree.children()[0].label().value());
    if (tree.children().length == 1) {
      UnaryRule ur = new UnaryRule();
      ur.parent = parent;
      ur.child = firstChild;
      //+ DEBUG
      // if (ug.scoreRule(ur) < -10000) {
      //        System.out.println("Grammar doesn't have rule: " + ur);
      // }
      return SloppyMath.max(ug.scoreRule(ur),-10000)+scoreBinarizedTree(tree.children()[0],leftmost);
    }
    BinaryRule br = new BinaryRule();
    br.parent = parent;
    br.leftChild = firstChild;
    br.rightChild = Numberer.number(stateSpace,tree.children()[1].label().value());
    //+ DEBUG
    // if (bg.scoreRule(br) < -10000) {
    //  System.out.println("Grammar doesn't have rule: " + br);
    // }
    return SloppyMath.max(bg.scoreRule(br),-10000)+
      scoreBinarizedTree(tree.children()[0],leftmost)+
      scoreBinarizedTree(tree.children()[1],false);
  }


  private static final boolean spillGuts = false;
  private static final boolean dumpTagging = false;
  private static long time = System.currentTimeMillis();
  private static void tick(String str) {
    long time2 = System.currentTimeMillis();
    long diff = time2-time;
    time = time2;
    System.err.print("done.  "+diff+"\n"+str);
  }

  private boolean floodTags = false;
  private List sentence = null;

  private int[][] narrowRExtent = null;
  private int[][] wideRExtent = null;
  private int[][] narrowLExtent = null;
  private int[][] wideLExtent = null;

  private boolean[] isTag = null;

  public boolean parse(List sentence) { // ExhaustivePCFGParser

      //System.out.println("is it a taggedword?" + (sentence.get(0) instanceof TaggedWord)); //debugging
    if (sentence != this.sentence) {
      this.sentence = sentence;
      floodTags = false;
    }
    if (Test.verbose) {
      Timing.tick("Starting pcfg parse.");
    }
    if (spillGuts) tick("Starting PCFG parse...");
    length = sentence.size();
    if (length > Test.maxLength+1)
      return false;
    int goal = stateNumberer.number(goalStr);
    if (Test.verbose) {
      System.out.println(numStates+" states, "+goal+" is the goal state.");
      System.out.println(bg.numRules()+" binary rules, "+ug.numRules()+" unary rules.");
      //System.out.println(new ArrayList(ug.coreRules.keySet()));
      System.out.print("Initializing...");
    }
    // map to words
    words = new int[length];
    int[] inputTags = new int[length];
    int unk = 0;
    StringBuffer unkWords = new StringBuffer("[");
    for (int i=0; i<length; i++) {
      Object o = sentence.get(i);
      if (o instanceof HasWord)
        o = ((HasWord)o).word();
      if (Test.verbose && ! lex.isKnown(wordNumberer.number(o.toString()))) {
        unk++;
        unkWords.append(" ");
        unkWords.append(o);
      }
      words[i] = wordNumberer.number(o.toString());
        //else
        //      words[i] = wordNumberer.number(Lexicon.UNKNOWN_WORD);

      /*
      if (o instanceof HasTag) { // collect the tags of the input sentence
	String tag = ((HasTag)o).tag();
	if(tagNumberer.hasSeen(tag))
	  inputTags[i] = tagNumberer.number(tag);
      }
      */
    }
    if (isTag == null) {
      isTag = new boolean[numStates];
      for (int state=0; state<numStates; state++) {
	isTag[state] = Numberer.getGlobalNumberer("tags").hasSeen(Numberer.getGlobalNumberer(stateSpace).object(state));
      }
    }
    // CKY + unaries
    if (spillGuts) tick("Wiping arrays...");
    for (int start=0; start<length; start++) {
      for (int end=start+1; end<=length; end++) {
        Arrays.fill(iScore[start][end], Float.NEGATIVE_INFINITY);
        if (Options.get().doDep)
          Arrays.fill(oScore[start][end], Float.NEGATIVE_INFINITY);
      }
    }
    for (int loc=0; loc<=length; loc++) {
      Arrays.fill(narrowRExtent[loc],length+1);
      Arrays.fill(narrowLExtent[loc],-1);
      Arrays.fill(wideRExtent[loc],-1);
      Arrays.fill(wideLExtent[loc],length+1);
    }
    // int puncTag = stateNumberer.number(".");
    // boolean lastIsPunc = false;
    if (Test.verbose) {
      Timing.tick("done.");
      unkWords.append(" ]");
      System.out.println("Unknown words: "+unk + " " + unkWords);
      System.out.print("Starting filters...");
    }
    // do tags
    if (spillGuts) tick("Tagging...");
    for (int start = 0; start+1 <= length; start++) {
      int word = words[start];
      int end = start+1;
      Arrays.fill(tags[start],false);
      // setting the below line to false generates ALL tags in lexicon for
      // every word (though some may still be scored at -Inf).  Beware! XXXX
      if ( ! Options.get().flexiTag) {
        for (Iterator taggingI = lex.ruleIteratorByWord(word,start);
             taggingI.hasNext(); ) {
          IntTaggedWord tagging = (IntTaggedWord)taggingI.next();
          int state = Numberer.translate("tags",stateSpace,tagging.tag);
          try {
            iScore[start][end][state] = (float)lex.score(tagging,start);
	    if (iScore[start][end][state] > Float.NEGATIVE_INFINITY) {
	      narrowRExtent[start][state] = end;
	      narrowLExtent[end][state] = start;
	      wideRExtent[start][state] = end;
	      wideLExtent[end][state] = start;
	    }
          } catch (Exception e) {
            System.out.println("State: "+state+" tags "+Numberer.getGlobalNumberer("tags").object(tagging.tag));
          }
          int tag = tagging.tag;
          tags[start][tag] = true;
          if (dumpTagging)
            System.out.println("Word "+start+" tagging "+tagging+" score "+iScore[start][start+1][state]+" -- "+state+" = "+Numberer.object("states",state));
          //if (start == length-2 && tagging.parent == puncTag)
          //  lastIsPunc = true;
        }
      }  
      else {
        for (int state=0; state<numStates; state++) {
          float iS = iScore[start][end][state];
          if (iS == Float.NEGATIVE_INFINITY && isTag[state]) {
            iScore[start][end][state] = (float)lex.score(new IntTaggedWord(word, Numberer.translate(stateSpace,"tags",state)), start);
	    if (iScore[start][end][state] > Float.NEGATIVE_INFINITY) {
	      narrowRExtent[start][state] = end;
	      narrowLExtent[end][state] = start;
	      wideRExtent[start][state] = end;
	      wideLExtent[end][state] = start;
	    }
            if (dumpTagging)
              System.out.println("Word "+start+" tagging "+(new IntTaggedWord(word, Numberer.translate(stateSpace,"tags",state)))+" score "+iScore[start][start+1][state]+" -- "+state+" = "+Numberer.object("states",state));
          }
        }
      }
      // null out any tags that violate prespecified basic tags
      if (! floodTags && sentence.get(start) instanceof HasTag) {
	//System.out.println("Input sentence is tagged; permitting only these taggings in parse");
	String tagStr = ((HasTag)sentence.get(start)).tag();
	TreebankLanguagePack tlp = Options.get().langpack;
	Numberer stateNumberer = Numberer.getGlobalNumberer(stateSpace);
	for (int state=0; state<numStates; state++) {
	  if (! isTag[state])
	    continue;
	  String stateStr = (String)stateNumberer.object(state);
	  if (! tlp.basicCategory(stateStr).equals(tagStr)) {
	    iScore[start][end][state] -= 1000.0f;
	    //iScore[start][end][state] = Float.NEGATIVE_INFINITY;
	    //narrowRExtent[start][state] = length+1;
	    //narrowLExtent[end][state] = -1;
	    //wideRExtent[start][state] = -1;
	    //wideLExtent[end][state] = length+1;
	  } else {
	    //System.out.println("At "+start+" allowed "+stateStr+" for "+tagStr);
	  }
	}
      }
      // if parse failed because of tag coverage
      if (floodTags && (! Test.noRecoveryTagging)) { // sends to recovery parse with more flexible tagging
        for (int state=0; state<numStates; state++) {
          float iS = iScore[start][end][state];
          if (iS == Float.NEGATIVE_INFINITY && isTag[state]) {
            iScore[start][end][state] = -1000.0f;
	    narrowRExtent[start][state] = end;
	    narrowLExtent[end][state] = start;
	    wideRExtent[start][state] = end;
	    wideLExtent[end][state] = start;
          }
        }
      }
      if (spillGuts) tick("Terminal Unary...");
      for (int state=0; state<numStates; state++) {
        float iS = iScore[start][end][state];
        if (iS == Float.NEGATIVE_INFINITY)
          continue;
        UnaryRule[] unaries = ug.closedRulesByChild(state);
        for (int r=0; r<unaries.length; r++) {
          UnaryRule ur = unaries[r];
          int parent = ur.parent;
          float pS = (float) ur.score;
          float tot = iS+pS;
          if (tot > iScore[start][end][parent]) {
            iScore[start][end][parent] = tot;
	    narrowRExtent[start][parent] = end;
	    narrowLExtent[end][parent] = start;
	    wideRExtent[start][parent] = end;
	    wideLExtent[end][parent] = start;
          }
        }
      }
      if (spillGuts) tick("Next word...");
    }
    //if (Test.outsideFilter)
    // buildOFilter();
    if (Test.verbose) {
      Timing.tick("done.");
      System.err.print("Starting insides...");
    }
    for (int diff = 2; diff <= length; diff++) {
      for (int start = 0; start+diff <= length; start++) {
        if (spillGuts) tick("Binaries for "+diff+"...");
        int end = start+diff;
        if (end == length && start != 0)
          continue;
        // do binary rules
        for (int s=0; s<numStates; s++) {
	  int narrowR = narrowRExtent[start][s];
	  boolean iPossibleL = (narrowR < end);
	  if (! iPossibleL)
	    continue;
	  BinaryRule[] leftRules = bg.splitRulesWithLC(s);
	  for (int i=0; i<leftRules.length; i++) {
	    BinaryRule r = leftRules[i];
	    int narrowL = narrowLExtent[end][r.rightChild];
	    boolean iPossibleR = (narrowL >= narrowR);
	    if (! iPossibleR)
	      continue;
	    int min1 = narrowR;
	    int min2 = wideLExtent[end][r.rightChild];
	    int min = (min1 > min2 ? min1 : min2);
	    if (min > narrowL)
	      continue;
	    int max1 = wideRExtent[start][s];
	    int max2 = narrowL;
	    int max = (max1 < max2 ? max1 : max2);
	    if (min > max)
	        continue;
	    float pS = (float) r.score;
	    int state = r.parent;
	    float oldIScore = iScore[start][end][state];
	    float curIScore = oldIScore;
	    //System.out.println("Min "+min+" max "+max+" start "+start+" end "+end);
	    for (int split = min; split <= max; split++) {
	      float lS = iScore[start][split][s];
	      if (lS == Float.NEGATIVE_INFINITY)
		continue;
	      float rS = iScore[split][end][r.rightChild];
	      if (rS == Float.NEGATIVE_INFINITY)
		continue;
	      float tot = pS+lS+rS;
	      if (tot > curIScore)
		curIScore = tot;
	    }
	    if (curIScore > oldIScore) {
	      iScore[start][end][state] = curIScore;
	      if (oldIScore == Float.NEGATIVE_INFINITY) {
		if (start > narrowLExtent[end][state]) {
		  narrowLExtent[end][state] = start;
		  wideLExtent[end][state] = start;
		} else {
		  if (start < wideLExtent[end][state])
		    wideLExtent[end][state] = start;
		}
		if (end < narrowRExtent[start][state]) {
		  narrowRExtent[start][state] = end;
		  wideRExtent[start][state] = end;
		} else {
		  if (end > wideRExtent[start][state])
		    wideRExtent[start][state] = end;
		}
	      }
	    }
	  }
	}
	for (int s=0; s<numStates; s++) {
	  int narrowL = narrowLExtent[end][s];
	  boolean iPossibleR = (narrowL > start);
	  if (! iPossibleR)
	    continue;
	  BinaryRule[] rightRules = bg.splitRulesWithRC(s);
	  for (int i=0; i<rightRules.length; i++) {
	    BinaryRule r = rightRules[i];
	    int narrowR = narrowRExtent[start][r.leftChild];
	    boolean iPossibleL = (narrowR <= narrowL);
	    if (! iPossibleL)
	      continue;
	    int min1 = narrowR;
	    int min2 = wideLExtent[end][s];
	    int min = (min1 > min2 ? min1 : min2);
	    if (min > narrowL)
	      continue;
	    int max1 = wideRExtent[start][r.leftChild];
	    int max2 = narrowL;
	    int max = (max1 < max2 ? max1 : max2);
	    if (min > max)
	      continue;
	    float pS = (float) r.score;
	    int state = r.parent;
	    float oldIScore = iScore[start][end][state];
	    float curIScore = oldIScore;
	    //System.out.println("Start "+start+" end "+end+" min "+min+" max "+max);
	    for (int split = min; split <= max; split++) {
	      float lS = iScore[start][split][r.leftChild];
	      if (lS == Float.NEGATIVE_INFINITY)
		continue;
	      float rS = iScore[split][end][s];
	      if (rS == Float.NEGATIVE_INFINITY)
		continue;
	      float tot = pS+lS+rS;
	      if (tot > curIScore)
		curIScore = tot;
	    }
	    if (curIScore > oldIScore) {
	      iScore[start][end][state] = curIScore;
	      if (oldIScore == Float.NEGATIVE_INFINITY) {
		if (start > narrowLExtent[end][state]) {
		  narrowLExtent[end][state] = start;
		  wideLExtent[end][state] = start;
		} else {
		  if (start < wideLExtent[end][state])
		    wideLExtent[end][state] = start;
		}
		if (end < narrowRExtent[start][state]) {
		  narrowRExtent[start][state] = end;
		  wideRExtent[start][state] = end;
		} else {
		  if (end > wideRExtent[start][state])
		    wideRExtent[start][state] = end;
		}
	      }
	    }
	  }
	}
        if (spillGuts) tick("Unaries for "+diff+"...");
        // do unary rules -- one could promote this loop and put start inside
        for (int state=0; state<numStates; state++) {
          float iS = iScore[start][end][state];
          if (iS == Float.NEGATIVE_INFINITY)
            continue;
          UnaryRule[] unaries = ug.closedRulesByChild(state);
          for (int r=0; r<unaries.length; r++) {
            UnaryRule ur = unaries[r];
            int parent = ur.parent;
            float pS = (float) ur.score;
            float tot = iS+pS;
	    float cur = iScore[start][end][parent];
            if (tot > cur) {
              iScore[start][end][parent] = tot;
	      if (cur == Float.NEGATIVE_INFINITY) {
		if (start > narrowLExtent[end][parent]) {
		  narrowLExtent[end][parent] = start;
		  wideLExtent[end][parent] = start;
		} else {
		  if (start < wideLExtent[end][parent])
		    wideLExtent[end][parent] = start;
		}
		if (end < narrowRExtent[start][parent]) {
		  narrowRExtent[start][parent] = end;
		  wideRExtent[start][parent] = end;
		} else {
		  if (end > wideRExtent[start][parent])
		    wideRExtent[start][parent] = end;
		}
	      }
            }
          }
        } // for unary rules
      } // for start
    } // for diff

    if (Test.verbose) {
      // insideTime += Timing.tick("done.");
      Timing.tick("done.");
      System.out.println("PCFG " + length + " words (incl. stop) iScore " +
                         iScore[0][length][goal]);
    }
    bestScore = iScore[0][length][goal];
    boolean succeeded = hasParse();
    if (bestScore == Float.NEGATIVE_INFINITY && !floodTags) {
      floodTags = true; // sentence will try to reparse 
      System.err.println("Trying recovery parse...");
      //System.out.println("Trying recovery parse...");
      return parse(sentence);
    }
    if (! Options.get().doDep)
      return succeeded;
    if (Test.verbose) {
      System.err.print("Starting outsides...");
    }
    // outside scores
    oScore[0][length][goal] = 0.0f;
    for (int diff = length; diff >= 1; diff--) {
      for (int start = 0; start+diff <= length; start++) {
        int end = start+diff;
        // do unaries
	for (int s = 0; s < numStates; s++) {
          float oS = oScore[start][end][s];
          if (oS == Float.NEGATIVE_INFINITY)
            continue;
	  UnaryRule[] rules = ug.closedRulesByParent(s);
	  for (int r=0; r<rules.length; r++) {
	    UnaryRule ur = rules[r];
	    float pS = (float) ur.score;
	    float tot = oS+pS;
	    if (tot > oScore[start][end][ur.child] &&
		iScore[start][end][ur.child] > Float.NEGATIVE_INFINITY)
	      oScore[start][end][ur.child] = tot;
          }
        }
        // do binaries
        for (int s = 0; s < numStates; s++) {
	  int min1 = narrowRExtent[start][s];
	  if (end < min1)
	    continue;
	  BinaryRule[] rules = bg.splitRulesWithLC(s);
          for (int r=0; r<rules.length; r++) {
            BinaryRule br = rules[r];
	    float oS = oScore[start][end][br.parent];
	    if (oS == Float.NEGATIVE_INFINITY)
	      continue;
	    int max1 = narrowLExtent[end][br.rightChild];
	    if (max1 < min1)
	      continue;
	    int min = min1;
	    int max = max1;
	    if (max-min > 2) {
	      int min2 = wideLExtent[end][br.rightChild];
	      min = (min1 > min2 ? min1 : min2);
	      if (max1 < min)
		continue;
	      int max2 = wideRExtent[start][br.leftChild];
	      max = (max1 < max2 ? max1 : max2);
	      if (max < min)
		continue;
	    }
            float pS = (float) br.score;
            for (int split = min; split <= max; split++) {
              float lS = iScore[start][split][br.leftChild];
              if (lS == Float.NEGATIVE_INFINITY)
		continue;
              float rS = iScore[split][end][br.rightChild];
	      if (rS == Float.NEGATIVE_INFINITY)
                continue;
              float totL = pS+rS+oS;
              if (totL > oScore[start][split][br.leftChild]) {
                oScore[start][split][br.leftChild] = totL;
              }
              float totR = pS+lS+oS;
              if (totR > oScore[split][end][br.rightChild]) {
                oScore[split][end][br.rightChild] = totR;
              }
            }
          }
        }
        for (int s = 0; s < numStates; s++) {
	  int max1 = narrowLExtent[end][s];
	  if (max1 < start)
	    continue;
	  BinaryRule[] rules = bg.splitRulesWithRC(s);
          for (int r=0; r<rules.length; r++) {
            BinaryRule br = rules[r];
	    float oS = oScore[start][end][br.parent];
	    if (oS == Float.NEGATIVE_INFINITY)
	      continue;
	    int min1 = narrowRExtent[start][br.leftChild];
	    if (max1 < min1)
	      continue;
	    int min = min1;
	    int max = max1;
	    if (max-min > 2) {
	      int min2 = wideLExtent[end][br.rightChild];
	      min = (min1 > min2 ? min1 : min2);
	      if (max1 < min)
		continue;
	      int max2 = wideRExtent[start][br.leftChild];
	      max = (max1 < max2 ? max1 : max2);
	      if (max < min)
		continue;
	    }
            float pS = (float) br.score;
            for (int split = min; split <= max; split++) {
              float lS = iScore[start][split][br.leftChild];
              if (lS == Float.NEGATIVE_INFINITY)
		continue;
              float rS = iScore[split][end][br.rightChild];
	      if (rS == Float.NEGATIVE_INFINITY)
                continue;
              float totL = pS+rS+oS;
              if (totL > oScore[start][split][br.leftChild]) {
                oScore[start][split][br.leftChild] = totL;
              }
              float totR = pS+lS+oS;
              if (totR > oScore[split][end][br.rightChild]) {
                oScore[split][end][br.rightChild] = totR;
              }
            }
          }
        }
	/*
        for (int s = 0; s < numStates; s++) {
          float oS = oScore[start][end][s];
	  //if (iScore[start][end][s] == Float.NEGATIVE_INFINITY ||
	  //             oS == Float.NEGATIVE_INFINITY)
	  if (oS == Float.NEGATIVE_INFINITY)
            continue;
	  BinaryRule[] rules = bg.splitRulesWithParent(s);
          for (int r=0; r<rules.length; r++) {
            BinaryRule br = rules[r];
	    int min1 = narrowRExtent[start][br.leftChild];
	    if (end < min1)
	      continue;
	    int max1 = narrowLExtent[end][br.rightChild];
	    if (max1 < min1)
	      continue;
	    int min2 = wideLExtent[end][br.rightChild];
	    int min = (min1 > min2 ? min1 : min2);
	    if (max1 < min)
	      continue;
	    int max2 = wideRExtent[start][br.leftChild];
	    int max = (max1 < max2 ? max1 : max2);
	    if (max < min)
	      continue;
            float pS = (float) br.score;
            for (int split = min; split <= max; split++) {
              float lS = iScore[start][split][br.leftChild];
              if (lS == Float.NEGATIVE_INFINITY)
		continue;
              float rS = iScore[split][end][br.rightChild];
	      if (rS == Float.NEGATIVE_INFINITY)
                continue;
              float totL = pS+rS+oS;
              if (totL > oScore[start][split][br.leftChild]) {
                oScore[start][split][br.leftChild] = totL;
              }
              float totR = pS+lS+oS;
              if (totR > oScore[split][end][br.rightChild]) {
                oScore[split][end][br.rightChild] = totR;
              }
            }
          }
        }
	*/
      }
    }
    //System.out.println("State rate: "+((int)(1000*ohits/otries))/10.0);
    //System.out.println("Traversals: "+ohits);
    if (Test.verbose) {
      // outsideTime += Timing.tick("Done.");
      Timing.tick("Done.");
      System.out.print("Starting half-filters...");
    }
    for (int loc=0; loc<=length; loc++) {
      Arrays.fill(iPossibleByL[loc],false);
      Arrays.fill(iPossibleByR[loc],false);
      Arrays.fill(oPossibleByL[loc],false);
      Arrays.fill(oPossibleByR[loc],false);
    }
    for (int start=0; start<length; start++) {
      for (int end=start+1; end<=length; end++) {
        for (int state=0; state<numStates; state++) {
          if (iScore[start][end][state] > Float.NEGATIVE_INFINITY &&
              oScore[start][end][state] > Float.NEGATIVE_INFINITY) {
            iPossibleByL[start][state] = true;
            iPossibleByR[end][state] = true;
            oPossibleByL[start][state] = true;
            oPossibleByR[end][state] = true;
          }
        }
      }
    }
    if (Test.verbose) {
      Timing.tick("done.");
    }
    if (false) {
      long iNZ = 0;
      long oNZ = 0;
      long bNZ = 0;
      long tot = 0;
      for (int start=0; start<length; start++) {
        for (int end=start+1; end<=length; end++) {
          for (int s=0; s<numStates; s++) {
            tot++;
            if (iScore[start][end][s] != Double.NEGATIVE_INFINITY)
              iNZ++;
            if (oScore[start][end][s] != Double.NEGATIVE_INFINITY)
              oNZ++;
            if (oScore[start][end][s] != Double.NEGATIVE_INFINITY &&
                iScore[start][end][s] != Double.NEGATIVE_INFINITY)
              bNZ++;
          }
        }
      }
      System.out.println("Zero Saturation (i): "+((int)(1000*iNZ/tot))/10.0);
      System.out.println("Zero Saturation (o): "+((int)(1000*oNZ/tot))/10.0);
      System.out.println("Zero Saturation (b): "+((int)(1000*bNZ/tot))/10.0);
      Timing.tick("Done with zero scan.");
    }
    return succeeded;
  }

  public boolean hasParse() {
    int goal = stateNumberer.number(goalStr);
    return (iScore[0][length][goal] > Float.NEGATIVE_INFINITY);
  }


  private static final double TOL = 1e-5;
  private static boolean matches(double x, double y) {
    return (Math.abs(x-y)/(Math.abs(x)+Math.abs(y)+1e-5) < TOL);
  }

  protected Tree extractBestParse(int goal, int start, int end) {
    // note that the returned tree may be missing intermediate nodes in
    // a unary chain because it parses with a unary-closed grammar.

    // find source of inside score
    // no backtraces so we can speed up the parsing for its primary use
    double bestScore = iScore[start][end][goal];
    String goalStr = (String)stateNumberer.object(goal);
    //System.out.println("Searching for "+goalStr+" from "+start+" to "+end+" scored "+bestScore);
    // check tags
    if (end-start == 1 &&
        tagNumberer.hasSeen(goalStr)) {
      IntTaggedWord tagging = new IntTaggedWord(words[start],tagNumberer.number(goalStr));
      double tagScore = lex.score(tagging,start);
      if (tagScore > Double.NEGATIVE_INFINITY || floodTags) {
        // return a pre-terminal tree
        String wordStr = (String)wordNumberer.object(words[start]);
        Tree wordNode = tf.newLeaf(new StringLabel(wordStr));
        List childList = new ArrayList();
        childList.add(wordNode);
        Tree tagNode = tf.newTreeNode(new StringLabel(goalStr),childList);
        //System.out.println("Tag node: "+tagNode);
        return tagNode;
      }
    }
    // check binaries first
    for (int split=start+1; split<end; split++) {
      for (Iterator binaryI = bg.ruleIteratorByParent(goal);
           binaryI.hasNext(); ) {
        BinaryRule br = (BinaryRule)binaryI.next();
        double score = br.score+iScore[start][split][br.leftChild]+iScore[split][end][br.rightChild];
        if (matches(score,bestScore)) {
          // build binary split
          Tree leftChildTree = extractBestParse(br.leftChild,start,split);
          Tree rightChildTree = extractBestParse(br.rightChild,split,end);
          List children = new ArrayList();
          children.add(leftChildTree);
          children.add(rightChildTree);
          Tree result = tf.newTreeNode(new StringLabel(goalStr),children);
          //System.out.println("Binary node: "+result);
          return result;
        }
      }
    }
    // check unaries
    for (Iterator unaryI = ug.ruleIteratorByParent(goal); unaryI.hasNext(); ) {
      UnaryRule ur = (UnaryRule)unaryI.next();
      double score = ur.score+iScore[start][end][ur.child];
      if (ur.child != ur.parent && matches(score,bestScore)) {
        // build unary
        Tree childTree = extractBestParse(ur.child,start,end);
          List children = new ArrayList();
          children.add(childTree);
          Tree result = tf.newTreeNode(new StringLabel(goalStr),children);
          //System.out.println("Unary node: "+result);
          return result;
      }
    }
    System.err.println("Failing on: ["+start+", "+end+"] looking for "+goalStr);
    System.err.println("Warning: no parse found in ExhaustivePCFGParser::extractBestParse");
    return null;
  }

  protected void restoreUnaries(Tree t) {
    //System.out.println("In restoreUnaries...");
    for (Iterator nodeI = t.subTrees().iterator(); nodeI.hasNext();) {
      Tree node = (Tree)nodeI.next();
      //System.out.println("Doing node: "+node.label());
      if (node.isLeaf() || node.isPreTerminal() || node.children().length != 1) {
        //System.out.println("Skipping node: "+node.label());
        continue;
      }
      //System.out.println("Not skipping node: "+node.label());
      Tree parent = node;
      Tree child = node.children()[0];
      List path = ug.getBestPath(stateNumberer.number(parent.label().value().toString()),
                                 stateNumberer.number(child.label().value().toString()));
      //System.out.println("Got path: "+path);
      int pos = 1;
      while (pos < path.size()-1) {
        int interState = ((Integer)path.get(pos)).intValue();
        Tree intermediate = tf.newTreeNode(new StringLabel((String)stateNumberer.object(interState)),
                                           parent.getChildrenAsList());
        List children = new ArrayList();
        children.add(intermediate);
        parent.setChildren(children);
        pos++;
      }
      //System.out.println("Done with node: "+node.label());
    }
  }

  public Tree getBestParse() {//String goalStr, int start, int end) {
    int start = 0;
    int end = words.length;
    int goal = stateNumberer.number(goalStr);
    Tree internalTree = extractBestParse(goal, start, end);
    //System.out.println("Got internal best parse...");
    if (internalTree == null) {
      System.err.println("Warning: no parse found in ExhaustivePCFGParser::extractBestParse");
    } else {
      restoreUnaries(internalTree);
    }
    //System.out.println("Restored unaries...");
    return internalTree;
    //TreeTransformer debinarizer = BinarizerFactory.getDebinarizer();
    //return debinarizer.transformTree(internalTree, null);
  }

  public ExhaustivePCFGParser(BinaryGrammar bg, UnaryGrammar ug, Lexicon lex) {
    this.stateSpace = bg.stateSpace();
    stateNumberer = Numberer.getGlobalNumberer(stateSpace);
    this.bg = bg;
    this.ug = ug;
    this.lex = lex;
    tf = new LabeledScoredTreeFactory(new StringLabelFactory());
    //if (Test.outsideFilter)
    //  orf = new OutsideRuleFilter(bg);
    int length = Test.maxLength+1;
    numStates = stateNumberer.total();
    int numTags = tagNumberer.total();
    // allocate just the parts of iScore and oScore used (end > start, etc.)
    iScore = new float[length+1][length+1][];
    for (int start=0; start<=length; start++) {
      for (int end = start+1; end<=length; end++) {
        iScore[start][end] = new float[numStates];
      }
    }
    if (Options.get().doDep) {
      oScore = new float[length+1][length+1][];
      for (int start=0; start<=length; start++) {
        for (int end= start+1; end<=length; end++) {
          oScore[start][end] = new float[numStates];
        }
      }
    }
    iPossibleByL = new boolean[length+1][numStates];
    iPossibleByR = new boolean[length+1][numStates];
    narrowRExtent = new int[length+1][numStates];
    wideRExtent = new int[length+1][numStates];
    narrowLExtent = new int[length+1][numStates];
    wideLExtent = new int[length+1][numStates];
    if (Options.get().doDep) {
      oPossibleByL = new boolean[length+1][numStates];
      oPossibleByR = new boolean[length+1][numStates];

      oFilteredStart = new boolean[length+1][numStates];
      oFilteredEnd = new boolean[length+1][numStates];
    }
    tags = new boolean[length+1][numTags];
  }

} // end class ExhaustivePCFGParser



class Interner {
  private Map oToO = new HashMap();
  public Object intern(Object o) {
    Object i=oToO.get(o);
    if (i==null) {
      i = o;
      oToO.put(o,o);
    }
    return i;
  }
}


abstract class Item implements Scored {
  public int start;
  public int end;
  public int state;
  public int head;
  public int tag;
  public Edge backEdge;
  public double iScore = Double.NEGATIVE_INFINITY;
  public double oScore = Double.NEGATIVE_INFINITY;
  public double score() { return iScore+(Test.exhaustiveTest ? 0.0 : oScore); }
  public boolean isEdge() { return false; }
  public boolean isPreHook() { return false; }
  public boolean isPostHook() { return false; }
}


class Edge extends Item {
  public Hook backHook;
  public boolean isEdge() { return true; }
  public String toString() {
    return "Edge("+Numberer.getGlobalNumberer("states").object(state)+":"+
      start+"-"+end+","+head+"/"+
      Numberer.getGlobalNumberer("tags").object(tag)+")";
  }
  public int hashCode() {
    return (state<<1)^(head << 8)^(tag << 16)^(start << 4)^(end << 24);
  }
  public boolean equals(Object o) {
    if (this == o) {
      return true;
    } else if (o instanceof Edge) {
      Edge e = (Edge) o;
      if (state == e.state &&
          head == e.head &&
          tag == e.tag &&
          start == e.start &&
          end == e.end)
        return true;
      return false;
    } else {
      return false;
    }
  }
}


class Hook extends Item {
  public int subState;
  public boolean isPreHook() { return head < start; }
  public boolean isPostHook() { return head >= end; }
  public String toString() { return (isPreHook()?"Pre":"Post")+"Hook("+Numberer.getGlobalNumberer("states").object(state)+"|"+Numberer.getGlobalNumberer("states").object(subState)+":"+start+"-"+end+","+head+"/"+Numberer.getGlobalNumberer("tags").object(tag)+")"; }
  public int hashCode() { return 1+(state<<14)^(subState<<16)^(head<<22)^(tag<<27)^(start<<1)^(end<<7); }

  public boolean equals(Object o) {
    //System.out.println("\nCHECKING HOOKS: "+this+" vs "+o);
    if (this == o)
      return true;
    try {
      if (((Item)o).isEdge())
        return false;
      Hook e = (Hook)o;
      if (state == e.state &&
          subState == e.subState &&
          head == e.head &&
          tag == e.tag &&
          start == e.start &&
          end == e.end) {
        return true;
      }
      return false;
    } catch (ClassCastException e) {
      return false;
    }
  }
}


class HookChart {
  private Map registeredPreHooks = new HashMap();
  private Map registeredPostHooks = new HashMap();
  private Map registeredEdgesByLeftIndex = new HashMap();
  private Map registeredEdgesByRightIndex = new HashMap();

  private Map realEdgesByL = new HashMap();
  private Map realEdgesByR = new HashMap();
  private Set builtLIndexes = new HashSet();
  private Set builtRIndexes = new HashSet();

  private Interner interner = new Interner();

  private static class ChartIndex {
    public int state;
    public int head;
    public int tag;
    public int loc;
    public int hashCode() { return state^(head << 8)^(tag << 16)^(loc << 24); }
    public boolean equals(Object o) {
      if (this == o)
        return true;
      try {
      ChartIndex ci = (ChartIndex)o;
      if (state == ci.state &&
          head == ci.head &&
          tag == ci.tag &&
          loc == ci.loc)
        return true;
      return false;
      } catch (ClassCastException e) {
        return false;
      }
    }
  }

  private static class WeakChartIndex {
    public int state;
    public int loc;
    public int hashCode() { return state^(loc << 16); }
    public boolean equals(Object o) {
      if (this == o)
        return true;
      try {
      WeakChartIndex ci = (WeakChartIndex)o;
      if (state == ci.state &&
          loc == ci.loc)
        return true;
      return false;
      } catch (ClassCastException e) {
        return false;
      }
    }
  }

  private static final Collection empty = new ArrayList(0);
  private WeakChartIndex tempWeakIndex = new WeakChartIndex();

  public void registerEdgeIndexes(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.start;
    ChartIndex index = (ChartIndex)interner.intern(tempIndex);
    if (index == tempIndex)
      tempIndex = new ChartIndex();
    builtLIndexes.add(index);
    //System.out.println("Edge registered: "+edge);
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.end;
    index = (ChartIndex)interner.intern(tempIndex);
    if (index == tempIndex)
      tempIndex = new ChartIndex();
    builtRIndexes.add(index);
  }

  public void registerRealEdge(Edge edge) {
    tempWeakIndex.state = edge.state;
    tempWeakIndex.loc = edge.start;
    WeakChartIndex index = (WeakChartIndex)interner.intern(tempWeakIndex);
    if (index == tempWeakIndex)
      tempWeakIndex = new WeakChartIndex();
    insert(realEdgesByL,index,edge);
    tempWeakIndex.state = edge.state;
    tempWeakIndex.loc = edge.end;
    index = (WeakChartIndex)interner.intern(tempWeakIndex);
    if (index == tempWeakIndex)
      tempWeakIndex = new WeakChartIndex();
    insert(realEdgesByR,index,edge);
  }

  public boolean isBuiltL(int state, int start, int head, int tag) {
    tempIndex.state = state;
    tempIndex.head = head;
    tempIndex.tag = tag;
    tempIndex.loc = start;
    return builtLIndexes.contains(tempIndex);
  }

  public boolean isBuiltR(int state, int end, int head, int tag) {
    tempIndex.state = state;
    tempIndex.head = head;
    tempIndex.tag = tag;
    tempIndex.loc = end;
    return builtRIndexes.contains(tempIndex);
  }

  public Collection getRealEdgesWithL(int state, int start) {
    tempWeakIndex.state = state;
    tempWeakIndex.loc = start;
    Collection edges = (Collection)realEdgesByL.get(tempWeakIndex);
    if (edges == null)
      return empty;
    return edges;
  }

  public Collection getRealEdgesWithR(int state, int end) {
    tempWeakIndex.state = state;
    tempWeakIndex.loc = end;
    Collection edges = (Collection)realEdgesByR.get(tempWeakIndex);
    if (edges == null)
      return empty;
    return edges;
  }

  public Collection getPreHooks(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.end;
    Collection result = (Collection)registeredPreHooks.get(tempIndex);
    if (result == null)
      result = empty;
    //System.out.println("For "+edge+" returning "+result.size()+" pre hooks");
    return result;
  }

  public Collection getPostHooks(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.start;
    Collection result = (Collection)registeredPostHooks.get(tempIndex);
    if (result == null)
      result = empty;
    //System.out.println("For "+edge+" returning "+result.size()+" post hooks");
    return result;
  }

  public Collection getEdges(Hook hook) {
    tempIndex.state = hook.subState;
    tempIndex.head = hook.head;
    tempIndex.tag = hook.tag;
    Collection result;
    if (hook.isPreHook()) {
      tempIndex.loc = hook.start;
      result = (Collection)registeredEdgesByRightIndex.get(tempIndex);
    } else {
      tempIndex.loc = hook.end;
      result = (Collection)registeredEdgesByLeftIndex.get(tempIndex);
    }
    if (result == null)
      result = empty;
    //System.out.println("For "+hook+" returning "+result.size()+" edges");
    return result;
  }

  private void insert(Map map, Object index, Object item) {
    List list = (List)map.get(index);
    if (list == null) {
      list = new ArrayList();
      map.put(index,list);
    }
    list.add(item);
  }

  ChartIndex tempIndex = new ChartIndex();

  public void addEdge(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    // left index
    tempIndex.loc = edge.start;
    ChartIndex index = (ChartIndex)interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
      tempIndex.state = edge.state;
      tempIndex.head = edge.head;
      tempIndex.tag = edge.tag;
    }
    insert(registeredEdgesByLeftIndex,index,edge);
    tempIndex.loc = edge.end;
    index = (ChartIndex)interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
    }
    insert(registeredEdgesByRightIndex,index,edge);
  }

  public void addHook(Hook hook) {
    Map map;
    tempIndex.state = hook.subState;
    tempIndex.head = hook.head;
    tempIndex.tag = hook.tag;
    if (hook.isPreHook()) {
      tempIndex.loc = hook.start;
      map = registeredPreHooks;
    } else {
      tempIndex.loc = hook.end;
      map = registeredPostHooks;
    }
    ChartIndex index = (ChartIndex)interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
    }
    insert(map,index,hook);
  }
} // HookChart


class N5BiLexPCFGParser extends BiLexPCFGParser {
  protected void relaxTempHook() {
    relaxHook1++;
    if (verbose) System.out.println("Considering: "+tempHook+" iP: "+scorer.iPossible(tempHook)+" oP: "+scorer.oPossible(tempHook));
    if (! Test.exhaustiveTest)
      if (! scorer.oPossible(tempHook) || ! scorer.iPossible(tempHook))
        return;
    relaxHook2++;
    Hook resultHook = tempHook;
    //Hook resultHook = (Hook)interner.intern(tempHook);
    if (verbose) System.out.println("Formed "+resultHook+" "+(resultHook == tempHook ? "new" : "old")+" "+tempHook.iScore+" was "+resultHook.iScore);
    if (resultHook == tempHook) {
      relaxHook3++;
      tempHook = new Hook();
      processHook(resultHook);
      builtHooks++;
    }
  }

  N5BiLexPCFGParser(Scorer scorer, ExhaustivePCFGParser fscorer, ExhaustiveDependencyParser leach, BinaryGrammar bg, UnaryGrammar ug, DependencyGrammar dg, Lexicon lex) {
    super(scorer, fscorer, leach, bg, ug, dg, lex, new NullGrammarProjection(bg, ug));
  }

  N5BiLexPCFGParser(Scorer scorer, ExhaustivePCFGParser fscorer, ExhaustiveDependencyParser leach, BinaryGrammar bg, UnaryGrammar ug, DependencyGrammar dg, Lexicon lex, GrammarProjection proj) {
    super(scorer, fscorer, leach, bg, ug, dg, lex, proj);
  }
}


class BiLexPCFGParser {
  protected static final boolean verbose = false;
  protected int MAX_ITEMS = -1;

  protected HookChart chart;
  protected Heap agenda;
  protected int length;
  protected int[] words;
  protected Edge goal;
  protected Interner interner;
  protected Scorer scorer;
  protected ExhaustivePCFGParser fscorer;
  protected GrammarProjection projection;
  //pair dep scores

  protected BinaryGrammar bg;
  protected UnaryGrammar ug;
  protected DependencyGrammar dg;
  protected Lexicon lex;
  protected List[] taggedWordList;
  protected float[][][][][] depScore;
  protected float[][][] headStopL;
  protected float[][][] headStopR;
  protected int[][] binDistance;

  protected Numberer wordNumberer = Numberer.getGlobalNumberer("words");
  protected Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  protected Numberer stateNumberer = Numberer.getGlobalNumberer("states");

  protected TreeFactory tf = new LabeledScoredTreeFactory();

  // temp
  protected long relaxHook1 = 0;
  protected long relaxHook2 = 0;
  protected long relaxHook3 = 0;
  protected long relaxHook4 = 0;

  protected long builtHooks = 0;
  protected long builtEdges = 0;
  protected long extractedHooks = 0;
  protected long extractedEdges = 0;


  protected static final double TOL = 1e-10;

  protected boolean better(double x, double y) {
    if ((x-y)/(Math.abs(x)+Math.abs(y)+1e-100) > TOL)
      return true;
    return false;
  }

  protected Tree extractParse(Edge edge) {
    String head = (String)wordNumberer.object(words[edge.head]);
    String tag = (String)tagNumberer.object(edge.tag);
    String state = (String)stateNumberer.object(edge.state);
    Label label = new CategoryWordTag(state,head,tag);
    if (edge.backEdge == null && edge.backHook == null) {
      // leaf, but needs word terminal
      List childList = new ArrayList(1);
      childList.add(tf.newLeaf(new StringLabel(head)));
      return tf.newTreeNode(label,childList);
    }
    List children = new ArrayList();
    if (edge.backHook == null) {
      // unary
      children.add(extractParse(edge.backEdge));
      return tf.newTreeNode(label,children);
    }
    // binary
    if (edge.backHook.isPreHook()) {
      children.add(extractParse(edge.backEdge));
      children.add(extractParse(edge.backHook.backEdge));
    } else {
      children.add(extractParse(edge.backHook.backEdge));
      children.add(extractParse(edge.backEdge));
    }
    return tf.newTreeNode(label,children);
  }

  public Tree getBestParse() {
    return extractParse(goal);
  }

  protected Edge tempEdge = new Edge();

  protected void combine(Edge edge, Hook hook) {
    if (verbose) System.out.println("Combining: "+edge+" and "+hook);
    // make result edge
    if (hook.isPreHook()) {
      tempEdge.start = edge.start;
      tempEdge.end = hook.end;
    } else {
      tempEdge.start = hook.start;
      tempEdge.end = edge.end;
    }
    tempEdge.state = hook.state;
    tempEdge.head = hook.head;
    tempEdge.tag = hook.tag;
    tempEdge.iScore = hook.iScore+edge.iScore;
    tempEdge.backEdge = edge;
    tempEdge.backHook = hook;
    relaxTempEdge();
  }

  protected void relaxTempEdge() {
    // if (tempEdge.iScore > scorer.iScore(tempEdge)+1e-4) {
    //   System.out.println(tempEdge+" has i "+tempEdge.iScore+" iE "+scorer.iScore(tempEdge));
    // }
    Edge resultEdge = (Edge)interner.intern(tempEdge);
    if (verbose) System.out.println("Formed "+resultEdge+" "+(resultEdge == tempEdge ? "new" : "old")+" "+tempEdge.iScore+" was "+resultEdge.iScore+" better? "+better(tempEdge.iScore, resultEdge.iScore));
    if (resultEdge == tempEdge) {
      //System.out.println(resultEdge+" was new.");
      tempEdge = new Edge();
      discoverEdge(resultEdge);
    } else {
      if (better(tempEdge.iScore, resultEdge.iScore) && resultEdge.oScore > Double.NEGATIVE_INFINITY) {
        double back = resultEdge.iScore;
        Edge backE = resultEdge.backEdge;
        Hook backH = resultEdge.backHook;
        resultEdge.iScore = tempEdge.iScore;
        resultEdge.backEdge = tempEdge.backEdge;
        resultEdge.backHook = tempEdge.backHook;
        try {
          agenda.decreaseKey(resultEdge);
        } catch(NullPointerException e) {
          if (false) {
            System.out.println("");
            System.out.println("Old backEdge: "+backE+" i "+backE.iScore+" o "+backE.oScore+" s "+backE.score());
            System.out.println("Old backEdge: "+backE+" iE "+scorer.iScore(backE));
            System.out.println("Old backHook: "+backH+" i "+backH.iScore+" o "+backH.oScore+" s "+backH.score());
            System.out.println("New backEdge: "+tempEdge.backEdge+" i "+tempEdge.backEdge.iScore+" o "+tempEdge.backEdge.oScore+" s "+tempEdge.backEdge.score());
            System.out.println("New backEdge: "+tempEdge.backEdge+" iE "+scorer.iScore(tempEdge.backEdge));
            System.out.println("New backHook: "+tempEdge.backHook+" i "+tempEdge.backHook.iScore+" o "+tempEdge.backHook.oScore+" s "+tempEdge.backHook.score());
            System.out.println("ERROR: Formed "+resultEdge+" i "+tempEdge.iScore+" o "+resultEdge.oScore+" s "+resultEdge.score());
            System.out.println("ERROR: Formed "+resultEdge+" "+(resultEdge == tempEdge ? "new" : "old")+" "+tempEdge.iScore+" was "+back+" better? "+better(tempEdge.iScore, back));
          }
        }
      }
    }
  }

  protected void discoverEdge(Edge edge) {
    // create new edge
    edge.oScore = scorer.oScore(edge);
    agenda.add(edge);
    builtEdges++;
  }

  protected void discoverHook(Hook hook) {
    hook.oScore = buildOScore(hook);
    if (hook.oScore == Double.NEGATIVE_INFINITY)
      relaxHook4++;
    builtHooks++;
    agenda.add(hook);
  }

  protected Edge iTemp = new Edge();
  protected Edge oTemp = new Edge();

  protected double buildOScore(Hook hook) {
    double bestOScore = Double.NEGATIVE_INFINITY;
    iTemp.head = hook.head;
    iTemp.tag = hook.tag;
    iTemp.state = hook.subState;
    oTemp.head = hook.head;
    oTemp.tag = hook.tag;
    oTemp.state = hook.state;
    if (hook.isPreHook()) {
      iTemp.end = hook.start;
      oTemp.end = hook.end;
      for (int start = 0; start <= hook.head; start++) {
        iTemp.start = start;
        oTemp.start = start;
        double oScore = scorer.oScore(oTemp)+scorer.iScore(iTemp);
        //System.out.println("Score for "+hook+" is i "+iTemp+" ("+scorer.iScore(iTemp)+") o "+oTemp+" ("+scorer.oScore(oTemp)+")");
        bestOScore = SloppyMath.max(bestOScore,oScore);
      }
    } else {
      iTemp.start = hook.end;
      oTemp.start = hook.start;
      for (int end = hook.head+1; end <= length; end++) {
        iTemp.end = end;
        oTemp.end = end;
        double oScore = scorer.oScore(oTemp)+scorer.iScore(iTemp);
        bestOScore = SloppyMath.max(bestOScore,oScore);
      }
    }
    return bestOScore;
  }

  protected Hook tempHook = new Hook();

  protected void projectHooks(Edge edge) {
    // form hooks
    // POST HOOKS
    //for (Iterator rI = bg.ruleIteratorByLeftChild(edge.state);
    //      rI.hasNext(); ) {
    List ruleList = bg.ruleListByLeftChild(edge.state);
    for (int r=0; r<ruleList.size(); r++) {
      //BinaryRule br = (BinaryRule)rI.next();
      BinaryRule br = (BinaryRule)ruleList.get(r);
      if (! fscorer.oPossibleL(project(br.parent),edge.start) ||
          ! fscorer.iPossibleL(project(br.rightChild),edge.end))
        if (! Test.exhaustiveTest)
          continue;
      for (int head=edge.end; head<length; head++) {
        for (Iterator iTWI=taggedWordList[head].iterator(); iTWI.hasNext();) {
          IntTaggedWord iTW = (IntTaggedWord)iTWI.next();
          int tag = iTW.tag;
          int word = iTW.word;
          tempHook.start = edge.start;
          tempHook.end = edge.end;
          tempHook.head = head;
          tempHook.tag = tag;
          tempHook.state = br.parent;
          tempHook.subState = br.rightChild;
          if (! chart.isBuiltL(tempHook.subState,tempHook.end,tempHook.head,tempHook.tag))
            continue;
          tempHook.iScore = edge.iScore+br.score+depScore[binDistance[head][edge.end]][head][tag][edge.head][edge.tag]+
            headStopL[edge.head][edge.tag][edge.start]+
            headStopR[edge.head][edge.tag][edge.end];
          tempHook.backEdge = edge;
          relaxTempHook();
        }
      }
    }
    // PRE HOOKS
    //for (Iterator rI = bg.ruleIteratorByRightChild(edge.state);
    //     rI.hasNext(); ) {
    ruleList = bg.ruleListByRightChild(edge.state);
    for (int r=0; r<ruleList.size(); r++) {
      //BinaryRule br = (BinaryRule)rI.next();
      BinaryRule br = (BinaryRule)ruleList.get(r);
      if (! fscorer.oPossibleR(project(br.parent),edge.end) ||
          ! fscorer.iPossibleR(project(br.leftChild),edge.start))
        if (! Test.exhaustiveTest)
          continue;
      for (int head=0; head<edge.start; head++) {
        for (Iterator iTWI=taggedWordList[head].iterator(); iTWI.hasNext();) {
          IntTaggedWord iTW = (IntTaggedWord)iTWI.next();
          int tag = iTW.tag;
          int word = iTW.word;
          tempHook.start = edge.start;
          tempHook.end = edge.end;
          tempHook.head = head;
          tempHook.tag = tag;
          tempHook.state = br.parent;
          tempHook.subState = br.leftChild;
          if (! chart.isBuiltR(tempHook.subState,tempHook.start,tempHook.head,tempHook.tag))
            continue;
          tempHook.iScore = edge.iScore+br.score+depScore[binDistance[head][edge.start]][head][tag][edge.head][edge.tag]+
            headStopL[edge.head][edge.tag][edge.start]+
            headStopR[edge.head][edge.tag][edge.end];
          tempHook.backEdge = edge;
          relaxTempHook();
        }
      }
    }
  }

  protected void registerReal(Edge real) {
    chart.registerRealEdge(real);
  }

  protected void triggerHooks(Edge edge) {
    // we might have built a synth edge, enabling some old real edges to project hooks
    boolean newL = ! chart.isBuiltL(edge.state,edge.start,edge.head,edge.tag);
    boolean newR = ! chart.isBuiltR(edge.state,edge.end,edge.head,edge.tag);
    //if (newL)
    //  System.out.println("Triggering on L: "+edge);
    //if (newR)
    //  System.out.println("Triggering on R: "+edge);
    chart.registerEdgeIndexes(edge);
    if (newR) {
      // PRE HOOKS
      BinaryRule[] rules = bg.splitRulesWithLC(edge.state);
      for (int i=0; i<rules.length; i++) {
        BinaryRule br = rules[i];
        Collection realEdges = chart.getRealEdgesWithL(br.rightChild,edge.end);
        for (Iterator realI = realEdges.iterator(); realI.hasNext();) {
          Edge real = (Edge)realI.next();
          tempHook.start = real.start;
          tempHook.end = real.end;
          tempHook.state = br.parent;
          tempHook.subState = br.leftChild;
          tempHook.head = edge.head;
          tempHook.tag = edge.tag;
          tempHook.backEdge = real;
          tempHook.iScore = real.iScore+br.score+depScore[binDistance[edge.head][edge.end]][edge.head][edge.tag][real.head][real.tag]+
            headStopL[real.head][real.tag][real.start]+
            headStopR[real.head][real.tag][real.end];
          relaxTempHook();
        }
      }
    }
    if (newL) {
      // POST HOOKS
      BinaryRule[] rules = bg.splitRulesWithRC(edge.state);
      for (int i=0; i<rules.length; i++) {
        BinaryRule br = rules[i];
        Collection realEdges = chart.getRealEdgesWithR(br.leftChild,edge.start);
        for (Iterator realI = realEdges.iterator(); realI.hasNext();) {
          Edge real = (Edge)realI.next();
          tempHook.start = real.start;
          tempHook.end = real.end;
          tempHook.state = br.parent;
          tempHook.subState = br.rightChild;
          tempHook.head = edge.head;
          tempHook.tag = edge.tag;
          tempHook.backEdge = real;
          tempHook.iScore = real.iScore+br.score+depScore[binDistance[edge.head][edge.start]][edge.head][edge.tag][real.head][real.tag]+
            headStopL[real.head][real.tag][real.start]+
            headStopR[real.head][real.tag][real.end];
          relaxTempHook();
        }
      }
    }
  }

  protected void triggerAllHooks(Edge edge) {
    // we might have built a new edge, enabling some old edges to project hooks
    boolean newL = ! chart.isBuiltL(edge.state,edge.start,edge.head,edge.tag);
    boolean newR = ! chart.isBuiltR(edge.state,edge.end,edge.head,edge.tag);
    if (verbose) {
      if (newL)
        System.out.println("Triggering on L: "+edge);
      if (newR)
        System.out.println("Triggering on R: "+edge);
    }
    chart.registerEdgeIndexes(edge);
    if (newR) {
      // PRE HOOKS
      for (Iterator rI = bg.ruleIteratorByLeftChild(edge.state); rI.hasNext();) {
        BinaryRule br = (BinaryRule)rI.next();
        Collection edges = chart.getRealEdgesWithL(br.rightChild,edge.end);
        for (Iterator realI = edges.iterator(); realI.hasNext();) {
          Edge real = (Edge)realI.next();
          tempHook.start = real.start;
          tempHook.end = real.end;
          tempHook.state = br.parent;
          tempHook.subState = br.leftChild;
          tempHook.head = edge.head;
          tempHook.tag = edge.tag;
          tempHook.backEdge = real;
          tempHook.iScore = real.iScore+br.score+depScore[binDistance[edge.head][edge.end]][edge.head][edge.tag][real.head][real.tag]+
            headStopL[real.head][real.tag][real.start]+
            headStopR[real.head][real.tag][real.end];
          relaxTempHook();
        }
      }
    }
    if (newL) {
      // POST HOOKS
      for (Iterator rI = bg.ruleIteratorByRightChild(edge.state); rI.hasNext();) {
        BinaryRule br = (BinaryRule)rI.next();
        Collection edges = chart.getRealEdgesWithR(br.leftChild,edge.start);
        if (false && verbose) {
          System.out.println("Looking for: "+Numberer.getGlobalNumberer("states").object(br.leftChild)+" ending at "+edge.start);
          System.out.println("Found: "+edges);
        }
        for (Iterator realI = edges.iterator(); realI.hasNext();) {
          Edge real = (Edge)realI.next();
          tempHook.start = real.start;
          tempHook.end = real.end;
          tempHook.state = br.parent;
          tempHook.subState = br.rightChild;
          tempHook.head = edge.head;
          tempHook.tag = edge.tag;
          tempHook.backEdge = real;
          tempHook.iScore = real.iScore+br.score+depScore[binDistance[edge.head][edge.start]][edge.head][edge.tag][real.head][real.tag]+
            headStopL[real.head][real.tag][real.start]+
            headStopR[real.head][real.tag][real.end];
          relaxTempHook();
        }
      }
    }
  }


  protected void relaxTempHook() {
    relaxHook1++;
    if (verbose) System.out.println("Considering: "+tempHook+" iP: "+scorer.iPossible(tempHook)+" oP: "+scorer.oPossible(tempHook));
    if (! Test.exhaustiveTest)
      if (! scorer.oPossible(tempHook) || ! scorer.iPossible(tempHook))
        return;
    relaxHook2++;
    Hook resultHook = (Hook)interner.intern(tempHook);
    if (verbose) System.out.println("Formed "+resultHook+" "+(resultHook == tempHook ? "new" : "old")+" "+tempHook.iScore+" was "+resultHook.iScore);
    if (resultHook == tempHook) {
      relaxHook3++;
      tempHook = new Hook();
      discoverHook(resultHook);
    }
    if (better(tempHook.iScore, resultHook.iScore)) {
      resultHook.iScore = tempHook.iScore;
      resultHook.backEdge = tempHook.backEdge;
      try {
        agenda.decreaseKey(resultHook);
      } catch (NullPointerException e) {
      }
    }
  }

  protected void projectUnaries(Edge edge) {
    for (Iterator rI = ug.ruleIteratorByChild(edge.state); rI.hasNext(); ) {
      UnaryRule ur = (UnaryRule)rI.next();
      if (ur.child == ur.parent)
        continue;
      tempEdge.start = edge.start;
      tempEdge.end = edge.end;
      tempEdge.head = edge.head;
      tempEdge.tag = edge.tag;
      tempEdge.state = ur.parent;
      tempEdge.backEdge = edge;
      tempEdge.backHook = null;
      tempEdge.iScore = edge.iScore+ur.score;
      relaxTempEdge();
    }
  }

  protected void processEdge(Edge edge) {
    // add to chart
    if (verbose) System.out.println("Adding to chart: "+edge);
    chart.addEdge(edge);
    // fetch existing hooks that can combine with it
    Collection preHooks = chart.getPreHooks(edge);
    Collection postHooks = chart.getPostHooks(edge);
    // combine them
    for (Iterator preI=preHooks.iterator(); preI.hasNext();) {
      Hook hook = (Hook)preI.next();
      combine(edge,hook);
    }
    for (Iterator postI=postHooks.iterator(); postI.hasNext();) {
      Hook hook = (Hook)postI.next();
      combine(edge,hook);
    }
    // do projections
    //if (verbose) System.out.println("Projecting: "+edge);
    projectUnaries(edge);
    if (! bg.isSynthetic(edge.state) && ! Options.get().freeDependencies) {
      projectHooks(edge);
      registerReal(edge);
    }
    if (Options.get().freeDependencies) {
      projectHooks(edge);
      registerReal(edge);
      triggerAllHooks(edge);
    } else {
      triggerHooks(edge);
    }
  }

  protected void processHook(Hook hook) {
    // add to chart
    //if (verbose) System.out.println("Adding to chart: "+hook);
    chart.addHook(hook);
    Collection edges = chart.getEdges(hook);
    for (Iterator edgeI=edges.iterator(); edgeI.hasNext();) {
      Edge edge = (Edge)edgeI.next();
      combine(edge,hook);
    }
  }

  protected void processItem(Item item) {
    if (item.isEdge())
      processEdge((Edge)item);
    else
      processHook((Hook)item);
  }

  protected void discoverItem(Item item) {
    if (item.isEdge())
      discoverEdge((Edge)item);
    else
      discoverHook((Hook)item);
  }

  protected Item makeInitialItem(int pos, int tag, int state, double iScore) {
    Edge edge = new Edge();
    edge.start = pos;
    edge.end = pos+1;
    edge.state = state;
    edge.head = pos;
    edge.tag = tag;
    edge.iScore = iScore;
    return edge;
  }

  protected List makeInitialItems(List wordList) {
    List itemList = new ArrayList();
    int length = wordList.size();
    int numTags = tagNumberer.total();
    words = new int[length];
    taggedWordList = new List[length];
    int terminalCount = 0;
    for (int i=0; i<length; i++) {
      taggedWordList[i] = new ArrayList(numTags);
      String wordStr = "";
      String tagStr = "";
      Object wordObject = wordList.get(i);
      if (wordObject instanceof HasWord) {
        wordStr = ((HasWord)wordObject).word();
      } else {
        wordStr = wordObject.toString();
      }
      //if (wordObject instanceof HasTag) {
      //  tagStr = ((HasTag)wordObject).tag();
      //}
      if (! wordNumberer.hasSeen(wordStr))
        wordStr = Lexicon.UNKNOWN_WORD;
      int word = wordNumberer.number(wordStr);
      words[i] = word;
      for (Iterator tagI = lex.ruleIteratorByWord(word,i); tagI.hasNext(); ) {
        IntTaggedWord tagging = (IntTaggedWord)tagI.next();
        int tag = tagging.tag;
        String curTagStr = (String)tagNumberer.object(tag);
        //if (!tagStr.equals("") && !tagStr.equals(curTagStr))
        //  continue;
        int state = Numberer.translate("tags","states",tag);
        //itemList.add(makeInitialItem(i,tag,state,1.0*tagging.score));
        // THIS WILL CAUSE BUGS!!!  Don't use with another A* scorers
        tempEdge.state = state;
        tempEdge.head = i;
        tempEdge.start = i;
        tempEdge.end = i+1;
        tempEdge.tag = tag;
        itemList.add(makeInitialItem(i,tag,state,scorer.iScore(tempEdge)));
        terminalCount++;
        taggedWordList[i].add(new IntTaggedWord(word,tag));
      }
    }
    if (Test.verbose) {
      System.out.println("Terminals:"+terminalCount);
    }
    return itemList;
  }

  protected void scoreDependencies() {
    // just leach it off the dparser for now...
    /*
    Dependency dependency = new Dependency();
    for (int head = 0; head < words.length; head++) {
      for (int hTag = 0; hTag < tagNumberer.total(); hTag++) {
        for (int arg = 0; arg < words.length; arg++) {
          for (int aTag = 0; aTag < tagNumberer.total(); aTag++) {
            Arrays.fill(depScore[head][hTag][arg][aTag],Float.NEGATIVE_INFINITY);
          }
        }
      }
    }
    for (int head = 0; head < words.length; head++) {
      for (int arg = 0; arg < words.length; arg++) {
        if (head == arg)
          continue;
        for (Iterator headTWI=taggedWordList[head].iterator(); headTWI.hasNext();) {
          IntTaggedWord headTW = (IntTaggedWord)headTWI.next();
          for (Iterator argTWI=taggedWordList[arg].iterator(); argTWI.hasNext();) {
            IntTaggedWord argTW = (IntTaggedWord)argTWI.next();
            dependency.head = headTW;
            dependency.arg = argTW;
            dependency.leftHeaded = (head < arg);
            dependency.distance = Math.abs(head-arg);
            depScore[head][headTW.tag][arg][argTW.tag] =
              dg.score(dependency);
            if (false && depScore[head][headTW.tag][arg][argTW.tag] > -100)
              System.out.println(wordNumberer.object(headTW.word)+"/"+tagNumberer.object(headTW.tag)+" -> "+wordNumberer.object(argTW.word)+"/"+tagNumberer.object(argTW.tag)+" score "+depScore[head][headTW.tag][arg][argTW.tag]);
          }
        }
      }
    }
    */
  }

  protected void setGoal(int length) {
    goal = new Edge();
    goal.start = 0;
    goal.end = length;
    goal.state = stateNumberer.number(Options.get().langpack.startSymbol());
    goal.tag = tagNumberer.number(Lexicon.BOUNDARY_TAG);
    goal.head = length-1;
    //goal = (Edge)interner.intern(goal);
  }

  protected void initialize(List words) {
    length = words.size();
    interner = new Interner();
    agenda = new ArrayHeap(new ScoredComparator(ScoredComparator.DESCENDING));
    chart = new HookChart();
    setGoal(length);
    List initialItems = makeInitialItems(words);
    scoreDependencies();
    for (int i=0; i<initialItems.size(); i++) {
      Item item = (Item)initialItems.get(i);
      item = (Item)interner.intern(item);
      //if (verbose) System.out.println("Initial: "+item);
      discoverItem(item);
    }
  }

  public boolean parse(List words) {
    int spanFound = 0;
    long last = 0;
    int exHook = 0;
    relaxHook1 = 0;
    relaxHook2 = 0;
    relaxHook3 = 0;
    relaxHook4 = 0;
    builtHooks = 0;
    builtEdges = 0;
    extractedHooks = 0;
    extractedEdges = 0;
    if (Test.verbose) {
      Timing.tick("Starting combined parse.");
    }
    initialize(words);
    while (!agenda.isEmpty()) {
      Item item = (Item)agenda.extractMin();
      if (!item.isEdge()) {
        exHook++;
        extractedHooks++;
      } else {
        extractedEdges++;
      }
      if (relaxHook1 > last + 1000000) {
        last = relaxHook1;
        if (Test.verbose) {
          System.out.println("Proposed hooks:   "+relaxHook1);
          System.out.println("Unfiltered hooks: "+relaxHook2);
          System.out.println("Built hooks:      "+relaxHook3);
          System.out.println("Waste hooks:      "+relaxHook4);
          System.out.println("Extracted hooks:  "+exHook);
        }
      }
      if (item.end-item.start > spanFound) {
        spanFound = item.end-item.start;
        if (Test.verbose) {
          System.out.print(spanFound+" ");
        }
      }
      //if (item.end < 5) System.out.println("Extracted: "+item+" iScore "+item.iScore+" oScore "+item.oScore+" score "+item.score());
      if (item.equals(goal)) {
        if (Test.verbose) {
          System.out.println("Found goal!");
          System.out.println("Comb iScore "+item.iScore); // was goal.iScore
          Timing.tick("Done, parse found.");
          System.out.println("Built items:      "+(builtEdges+builtHooks));
          System.out.println("Built hooks:      "+builtHooks);
          System.out.println("Built edges:      "+builtEdges);
          System.out.println("Extracted items:  "+(extractedEdges+extractedHooks));
          System.out.println("Extracted hooks:  "+extractedHooks);
          System.out.println("Extracted edges:  "+extractedEdges);
          //postMortem();
        }
        goal = (Edge)item;
        interner = null;
        agenda = null;
        return true;
      }
      if (item.score() == Double.NEGATIVE_INFINITY) {
        System.err.println("Hit A*-blocked edges, failing.");
        if (Test.verbose) {
          Timing.tick("Hit A*-blocked edges, failing.");
	}
        return false;
      }
      if (MAX_ITEMS > 0 && (builtEdges+builtHooks) >= MAX_ITEMS) {
        System.err.println("Hit MAX_ITEMS, failing.");
        if (Test.verbose) {
          Timing.tick("Hit MAX_ITEMS, failing.");
	}
        return false;
      }
      if (verbose && item.score() != Double.NEGATIVE_INFINITY) System.out.println("Removing from agenda: "+item+" score i "+item.iScore+" + o "+item.oScore+" = "+item.score());
      if (verbose && item.score() != Double.NEGATIVE_INFINITY) System.out.println("Backtrace: "+(item.backEdge != null ? item.backEdge.toString() : "null")+" "+(item.isEdge() ? (((Edge)item).backHook != null ? ((Edge)item).backHook.toString() : "") : ""));
      processItem(item);
    }
    System.err.println("Done, parse NOT found.");
    if (Test.verbose) {
      Timing.tick("Done, parse NOT found.");
    }
    return false;
  }

  protected void postMortem() {
    int numHooks = 0;
    int numEdges = 0;
    int numUnmatchedHooks = 0;
    int total = agenda.size();
    int done = 0;
    while (! agenda.isEmpty()) {
      Item item = (Item)agenda.extractMin();
      done++;
      //if(done % (total/10) == 0)
      //        System.out.println("Scanning: "+100*done/total);
      if (item.isEdge()) {
        numEdges++;
      } else {
        numHooks++;
        Collection edges = chart.getEdges((Hook)item);
        if (edges.size() == 0)
          numUnmatchedHooks++;
      }
    }
    System.out.println("--- Agenda Post-Mortem ---");
    System.out.println("Edges:           "+numEdges);
    System.out.println("Hooks:           "+numHooks);
    System.out.println("Unmatched Hooks: "+numUnmatchedHooks);
  }

  protected int project(int state) {
    return projection.project(state);
  }

  public void setMaxItems(int maxItems) {
    MAX_ITEMS = maxItems;
  }

  BiLexPCFGParser(Scorer scorer, ExhaustivePCFGParser fscorer, ExhaustiveDependencyParser leach, BinaryGrammar bg, UnaryGrammar ug, DependencyGrammar dg, Lexicon lex) {
    this(scorer, fscorer, leach, bg, ug, dg, lex, new NullGrammarProjection(bg, ug));
  }

  BiLexPCFGParser(Scorer scorer, ExhaustivePCFGParser fscorer, ExhaustiveDependencyParser leach, BinaryGrammar bg, UnaryGrammar ug, DependencyGrammar dg, Lexicon lex, GrammarProjection projection) {
    this.fscorer = fscorer;
    this.projection = projection;
    this.scorer = scorer;
    this.bg = bg;
    this.ug = ug;
    this.dg = dg;
    this.lex = lex;
    int length = Test.maxLength+1;
    int numTags = tagNumberer.total();
    depScore = leach.headScore;//new double[length][numTags][length][numTags][5];
    headStopL = leach.headStop;
    headStopR = leach.headStop;
    binDistance = leach.binDistance; // THIS IS TERRIBLE, BUT SAVES MEMORY
  }
}


abstract class AbstractEval {

  protected String str="";

  private double precision = 0.0;
  private double recall = 0.0;
  private double f1 = 0.0;
  private double num = 0.0;
  private double exact = 0.0;

  private double precision2 = 0.0;
  private double recall2 = 0.0;
  private double pnum2 = 0.0;
  private double rnum2 = 0.0;

  // should be able to pass in a comparator!
  protected double precision(Set s1, Set s2) {
    double n = 0.0;
    double p = 0.0;
    for (Iterator i1 = s1.iterator(); i1.hasNext();) {
      Object o1 = i1.next();
      if (s2.contains(o1))
        p += 1.0;
      //if (s2.contains(o1))
      //        System.out.println("Found: "+o1);
      //else
      //        System.out.println("Failed to find: "+o1);
      n += 1.0;
    }
    //System.out.println("Matched " + p + " of " + n);
    return (n > 0.0 ? p/n : 0.0);
  }

  abstract Set makeObjects(Tree tree);

  public void evaluate(Tree guess, Tree gold) {
    evaluate(guess,gold,new PrintWriter(System.out,true));
  }

  public void evaluate(Tree guess, Tree gold, PrintWriter pw) {
    Set dep1 = makeObjects(guess);
    //List dep1s = new ArrayList(dep1);
    //Collections.sort(dep1s);
    //System.out.println("Guess set is " + dep1s);
    Set dep2 = makeObjects(gold);
    //List dep2s = new ArrayList(dep2);
    //Collections.sort(dep2s);
    //System.out.println("Gold set is " + dep2s);
    //System.out.println("Precision");
    double curPrecision = precision(dep1, dep2);
    //System.out.println("Recall");
    double curRecall = precision(dep2, dep1);
    double curF1 = (curPrecision > 0.0 && curRecall > 0.0 ?
                    2.0/(1.0/curPrecision+1.0/curRecall) : 0.0);
    precision += curPrecision;
    recall += curRecall;
    f1 += curF1;
    num += 1.0;

    precision2 += dep1.size()*curPrecision;
    pnum2 += dep1.size();

    recall2 += dep2.size()*curRecall;
    rnum2 += dep2.size();

    if (curF1 > 0.9999)
      exact += 1.0;

    pw.println(" P: "+((int)(curPrecision*10000))/100.0 +
                       " (sent ave "+((int)(precision*10000/num))/100.0+") (evalb "+((int)(precision2*10000/pnum2))/100.0+")");
    pw.println(" R: "+((int)(curRecall*10000))/100.0 +
                       " (sent ave "+((int)(recall*10000/num))/100.0+") (evalb "+((int)(recall2*10000/rnum2))/100.0+")");

    double cF1 = 2.0/(rnum2/(double)recall2+pnum2/(double)precision2);
    String emit = str + " F1: " + ((int)(curF1*10000))/100.0 + " (sent ave " +
        ((int)(10000*f1/num))/100.0+", evalb "+((int)(10000*cF1))/100.0+
        ")   Exact: "+((int)(10000*exact/num))/100.0+" N: "+(int)num;
    pw.println(emit);
    //System.err.println(emit);

    /*
      Sentence s = guess.yield();
      for (Iterator it2 = s.iterator(); it2.hasNext(); ) {
        Object obj = it2.next();
        if (curF1 < 0.7) {
          badwords.incrementCount(obj);
        } else {
          goodwords.incrementCount(obj);
        }
      }
    */
  }

  /*
  private Counter goodwords = new Counter();
  private Counter badwords = new Counter();

  public void printGoodBad() {
    System.out.println("Printing bad categories");
    for (Iterator it = badwords.keysAbove(5.0).iterator(); it.hasNext(); ) {
      Object key = it.next();
      System.out.println("In badwords 5 times: " + key);
      double numb = badwords.getCount(key);
      double numg = goodwords.getCount(key);
      if (numb / (numb + numg) > 0.1) {
        System.out.println("Bad word!  " + key + " (" +
                           (numb / (numb + numg)) + " bad)");
        // EncodingPrintWriter.out.println("Bad word!  " + key + " (" +
        //                 (numb / (numb + numg)) + " bad)",
        //                              "GB18030");
      }
    }
  }
  */

  public void display(boolean verbose) {
    display(verbose, new PrintWriter(System.out,true));
  }

  public void display(boolean verbose, PrintWriter pw) {
    double prec = precision2/pnum2;//(num > 0.0 ? precision/num : 0.0);
    double rec = recall2/rnum2;//(num > 0.0 ? recall/num : 0.0);
    double f = 2.0/(1.0/prec+1.0/rec);//(num > 0.0 ? f1/num : 0.0);
    //System.out.println(" Precision: "+((int)(10000.0*prec))/100.0);
    //System.out.println(" Recall:    "+((int)(10000.0*rec))/100.0);
    //System.out.println(" F1:        "+((int)(10000.0*f))/100.0);
    pw.println("LP: "+((int)(10000.0*prec))/100.0+
                       " LR: "+((int)(10000.0*rec))/100.0+" F1: "+((int)(10000.0*f))/100.0+
                       " Exact: "+((int)(10000.0*exact/num))/100.0);
    /*
    double prec = (num > 0.0 ? precision/num : 0.0);
    double rec = (num > 0.0 ? recall/num : 0.0);
    double f = (num > 0.0 ? f1/num : 0.0);
    System.out.println(" Precision: "+prec);
    System.out.println(" Recall:    "+rec);
    System.out.println(" F1:        "+f);
    */
  }
}

class CBEval extends LabeledConstituentEval {
  private double cb = 0.0;
  private double num = 0.0;
  private double zeroCB = 0.0;

  protected void checkCrossing(Set s1, Set s2) {
    double c = 0.0;
    for (Iterator i1 = s1.iterator(); i1.hasNext();) {
      Constituent constit = (Constituent) i1.next();
      if (constit.crosses(s2))
        c += 1.0;
    }
    if (c == 0)
      zeroCB += 1.0;
    cb += c;
    num += 1;
  }

  public void evaluate(Tree t1, Tree t2, PrintWriter pw) {
    Set b1 = makeObjects(t1);
    Set b2 = makeObjects(t2);
    checkCrossing(b1, b2);
  }

  public void display(boolean verbose, PrintWriter pw) {
    pw.println("AvgCB: "+((int)(100.0*cb/num))/100.0+" ZeroCB: "+((int)(10000.0*zeroCB/num))/100.0);
  }

  public CBEval(String str) {
    super(str);
  }

}


class DependencyEval extends AbstractEval {

  private static class StringDependency {
    public String head;
    public String arg;
    public int hashCode() { return head.hashCode()^arg.hashCode(); }
    public boolean equals(Object o) {
      if (this == o)
        return true;
      if (o instanceof StringDependency) {
        StringDependency d = (StringDependency) o;
        return (head.equals(d.head) && arg.equals(d.arg));
      } 
      return false;
    }
    public String toString() {
      return head+" -> "+arg;
    }
    public StringDependency(String head, String arg) {
      this.head = head;
      this.arg = arg;
    }
  }

  /** Build the set of dependencies for evaluation.  This set excludes
   *  all dependencies for which the argument is a punctuation tag.
   */
  Set makeObjects(Tree tree) {
    Set deps = new HashSet();
    Filter punctFilter = Options.get().langpack.punctuationWordAcceptFilter();
    for (Iterator nodeI = tree.subTrees().iterator(); nodeI.hasNext();) {
      // every child with a different head is an argument, and ones with the
      // same head after the first one found
      Tree node = (Tree)nodeI.next();
      if (node.isLeaf() || node.children().length < 2)
        continue;
      String head = ((HasWord)node.label()).word();
      boolean seenHead = false;
      for (int cNum = 0; cNum < node.children().length; cNum++) {
        Tree child = node.children()[cNum];
        String arg = ((HasWord)child.label()).word();
        if (head.equals(arg) && ! seenHead)
          seenHead = true;
        else if ( ! punctFilter.accept(arg)) {
          deps.add(new StringDependency(head, arg));
        }
      }
    }
    return deps;
  }

  public DependencyEval(String str) {
    this.str = str;
  }

} // end class DependencyEval


class LabeledConstituentEval extends AbstractEval {

  private ConstituentFactory cf = new LabeledScoredConstituentFactory();
  private TreeFactory tf = new LabeledScoredTreeFactory();
  private TreebankLanguagePack tlp = Options.get().langpack;

  protected Tree stripLeaves(Tree tree) {
    if (tree.isLeaf())
      return null;
    if (tree.isPreTerminal())
      return tf.newLeaf(tree.label());
    List children = new ArrayList(tree.children().length);
    for (int cNum=0; cNum<tree.children().length; cNum++) {
      children.add(stripLeaves(tree.children()[cNum]));
    }
    return tf.newTreeNode(tree.label(),children);
  }

  Set makeObjects(Tree tree) {
    // Strip off a root node. Allow it to have multiple daughters.
    // Assumes boundary symbol is already removed.
    Tree noLeafTree = stripLeaves(tree);
    Set set = new HashSet();
    if (tlp.isStartSymbol(noLeafTree.label().value())) {
      for (int i = 0; i < noLeafTree.children().length - 1; i++) {
	set.addAll(noLeafTree.children()[i].constituents(cf));
      }
    } else {
      set.addAll(noLeafTree.constituents(cf));
    }
    /*
    System.err.println("Constituents: "+set.size());
    for (Iterator i = set.iterator(); i.hasNext();) {
      Object o = i.next();
      System.err.println(" "+o);
    }
    */
    return set;
  }

  public void evaluate(Tree guess, Tree gold, PrintWriter pw) {
    // System.err.println("XXXX Guess constits: " + stripRoot(stripLeaves(guess)).constituents(cf));
    // System.err.println("XXXX Gold  constits: " + stripRoot(stripLeaves(gold)).constituents(cf));
    if (guess.yield().size() != gold.yield().size()) {
      pw.println("Warning: yield differs:");
      pw.println("Guess: " + guess.yield());
      pw.println("Gold: " + gold.yield());
    }
    super.evaluate(guess, gold,pw);
  }

  public LabeledConstituentEval(String str) {
    this.str = str;
  }

}// LabeledConstituentEval



class TaggingEval extends AbstractEval {
  Set makeObjects(Tree tree) {
    List twList = tree.taggedYield();
    Set set = new HashSet();
    for (int i=0; i<twList.size(); i++) {
      TaggedWord tw = (TaggedWord)twList.get(i);
      //IntTaggedWord iTW = new IntTaggedWord(Numberer.number("words",tw.word()), Numberer.number("tags",tw.tag()));
      Pair positionWT = new Pair(new Integer(i), new WordTag(tw.value(),tw.tag()));
      //WordTag positionWT = new WordTag(tw.value(),tw.tag());
      //System.out.println(iTW);
      //if (! tw.tag.equals("*"))
      set.add(positionWT);
    }
    return set;
  }
  public TaggingEval(String str) {
    this.str = str;
  }
}

class RuleErrorEval extends AbstractEval {

  protected boolean verbose = false;

  protected Counter over = new Counter();
  protected Counter under = new Counter();

  protected static String localize(Tree tree) {
    if (tree.isLeaf())
      return "";
    StringBuffer sb = new StringBuffer();
    sb.append(tree.label().toString());
    sb.append(" -> ");
    for (int i=0; i<tree.children().length; i++) {
      sb.append(tree.children()[i].label().toString());
    }
    return sb.toString();
  }

  Set makeObjects(Tree tree) {
    Set localTrees = new HashSet();
    for (Iterator i=tree.subTrees().iterator(); i.hasNext();) {
      Tree st = (Tree)i.next();
      localTrees.add(localize(st));
    }
    return localTrees;
  }

  public void evaluate(Tree t1, Tree t2,PrintWriter pw) {
    Set s1 = makeObjects(t1);
    Set s2 = makeObjects(t2);
    for (Iterator i1 = s1.iterator(); i1.hasNext();) {
      Object o1 = i1.next();
      if (! s2.contains(o1))
        over.add(o1, 1.0);
    }
    for (Iterator i2 = s2.iterator(); i2.hasNext();) {
      Object o2 = i2.next();
      if (! s1.contains(o2))
        under.add(o2, 1.0);
    }
  }

  protected void display(Counter c, int num) {
    display(c, num, new PrintWriter(System.out, true));
  }

  protected void display(Counter c, int num, PrintWriter pw) {
    List rules = new ArrayList(c.seenSet());
    Collections.sort(rules, c.comparator());
    for (int i=0; i < num; i++) {
      pw.println(rules.get(i)+" "+c.countOf(rules.get(i)));
    }
  }

  public void display(boolean verbose, PrintWriter pw) {
    this.verbose = verbose;
    pw.println("Most freqently underproposed rules:");
    display(under, (verbose ? 100 : 10));
    pw.println("Most freqently overproposed rules:");
    display(over, (verbose ? 100 : 10));
  }
}

class ParentAnnotatorFactory {
  private static class Annotator implements TreeTransformer {
    protected TreeFactory tf = new LabeledScoredTreeFactory();
    protected Tree helper(Tree tree, String parentStr) {
      if (tree.isLeaf())
        return tf.newLeaf(new Word(tree.label().value()));
      String labelStr = tree.label().value();
      List children = new ArrayList(tree.children().length);
      if (tree.isPreTerminal()) {
        children.add(helper(tree.children()[0],""));
        return tf.newTreeNode(new StringLabel(labelStr),children);
      }
      Label newLabel = (parentStr.length() > 0 ?
                        new StringLabel(labelStr+"^"+parentStr) :
                        new StringLabel(labelStr));
      for (int cNum=0; cNum<tree.children().length; cNum++) {
        Tree child = tree.children()[cNum];
        children.add(helper(child,labelStr));
      }
      return tf.newTreeNode(newLabel,children);
    }
    public Tree transformTree(Tree tree, TreeFactory dummy) {
      return helper(tree, "");
    }
  }

  private static class DeAnnotator implements TreeTransformer {
    private TreebankLanguagePack tlp = Options.get().langpack;
    protected TreeFactory tf = new LabeledScoredTreeFactory();
    protected Tree helper(Tree tree) {
      if (tree.isLeaf())
        return tf.newLeaf(new StringLabel(tree.label().value()));
      String labelStr = tree.label().value();
      List children = new ArrayList(tree.children().length);
      if (tree.isPreTerminal()) {
        children.add(helper(tree.children()[0]));
        return tf.newTreeNode(new StringLabel(labelStr),children);
      }
      Label newLabel = new StringLabel(tlp.basicCategory(labelStr));
      for (int cNum=0; cNum<tree.children().length; cNum++) {
        Tree child = tree.children()[cNum];
        children.add(helper(child));
      }
      return tf.newTreeNode(newLabel,children);
    }
    public Tree transformTree(Tree tree, TreeFactory dummy) {
      return helper(tree);
    }
  }

  public static TreeTransformer getAnnotator() {
    return new Annotator();
  }
  public static TreeTransformer getDeAnnotator() {
    return new DeAnnotator();
  }

}

class CollinsPuncTransformer implements TreeTransformer {

  private TreebankLanguagePack tlp = Options.get().langpack;

  boolean isPunc(Tree t) {
    if (t.isPreTerminal()) {
      String s = t.label().value();
      if (tlp.isEvalBIgnoredPunctuationTag(s)) {
        return true;
      }
    }
    return false;
  }

  LinkedList preTerms(Tree t) {
    LinkedList l = new LinkedList();
    preTermHelper(t, l);
    return l;
  }

  void preTermHelper(Tree t, List l) {
    if (t.isLeaf())
      return;
    if (t.isPreTerminal()) {
      l.add(t);
      return;
    }
    Tree[] children = t.children();
    for (int i=0; i<children.length; i++)
      preTermHelper(children[i], l);
  }

  Tree transformRoot(Tree tree, TreeFactory tf) {
    // XXXX fix this: use tlp and don't assume 1 daughter of ROOT!
    // leave the root intact
    // if (tlp.isStartSymbol(tlp.basicCategory(tree.label().value())))
    if (tree.label().toString().startsWith("ROOT"))
      return tf.newTreeNode(tree.label(),
            Collections.singletonList(transformNode(tree.children()[0], tf)));
    return transformNode(tree, tf);
  }

  Tree transformNode(Tree tree, TreeFactory tf) {
    if (tree.isLeaf())
      return tf.newLeaf(tree.label());
    if (tree.isPreTerminal())
      return tf.newTreeNode(tree.label(),
                            Collections.singletonList(tf.newLeaf(tree.children()[0].label())));
    List children = tree.getChildrenAsList();
    LinkedList newChildren = new LinkedList();
    // promote lower punctuation
    for (Iterator i=children.iterator(); i.hasNext();) {
      Tree child = (Tree)i.next();
      LinkedList preTerms = preTerms(child);
      while (! preTerms.isEmpty() && isPunc((Tree)preTerms.getFirst())) {
        newChildren.add(preTerms.getFirst());
        preTerms.removeFirst();
      }
      Tree newChild = transformNode(child, tf);
      LinkedList temp = new LinkedList();
      if (newChild.children().length > 0)
        newChildren.add(newChild);
      while (! preTerms.isEmpty() && isPunc((Tree)preTerms.getLast())) {
        temp.addFirst(preTerms.getLast());
        preTerms.removeLast();
      }
      newChildren.addAll(temp);
    }
    // remove local punctuation
    while (! newChildren.isEmpty() && isPunc((Tree)newChildren.getFirst()))
      newChildren.removeFirst();
    while (! newChildren.isEmpty() && isPunc((Tree)newChildren.getLast()))
      newChildren.removeLast();
    return tf.newTreeNode(tree.label(), newChildren);
  }

  public Tree transformTree(Tree tree, TreeFactory tf) {
    //System.out.println("PUNCTUATION TRANSFORM:");
    //tree.pennPrint();
    //System.out.println("BECOMES:");
    //transformRoot(tree, tf).pennPrint();
    return transformRoot(tree, tf);
  }

  public static Tree transformTree(Tree tree) {
    return (new CollinsPuncTransformer()).transformTree(tree, tree.treeFactory());
  }

}


class TreeCollinizer implements TreeTransformer {

  private TreebankLanguagePack tlp = Options.get().langpack;
  private final boolean deletePunct;

  public TreeCollinizer() {
    this(true);
  }

  public TreeCollinizer(boolean deletePunct) {
    this.deletePunct = deletePunct;
  }

  protected TreeFactory tf = new LabeledScoredTreeFactory();

  public Tree transformTree(Tree tree, TreeFactory dummy) {
    Label l = tree.label();
    String s = l.value();
    if (tree.isLeaf())
      return tf.newLeaf(s);   // makes it a StringLabel
    s = tlp.basicCategory(s);
    if (deletePunct) {
      // this is broken as it's not the right thing to do when there
      // is any tag ambiguity -- and there is for ' (POS/'').  Sentences
      // can then have more or less words.  It's also unnecessary for EVALB,
      // since it ignores punctuation anyway
      if (tree.isPreTerminal() &&
          Options.get().langpack.isEvalBIgnoredPunctuationTag(s)) {
        return null;
      }
    }
    /*
    if (EnglishTrain.collinsBaseNP && s.equals("NP")) {
      Tree[] kids = tree.children();
      if (kids.length == 1 && kids[0].label().value().startsWith("NP")) {
        return transformTree(kids[0], null);
      }
    }
    */
    // XXXX below isn't correct if root isn't unary.
    if (tlp.isStartSymbol(s)) {
      // NB: This deletes the boundary symbol, which is in the tree!
      return transformTree(tree.children()[0],null);
    }
    if (s.equals("PRT"))
      s = "ADVP";
    List children = new ArrayList();
    for (int cNum=0; cNum<tree.children().length; cNum++) {
      Tree child = tree.children()[cNum];
      Tree newChild = transformTree(child,null);
      if (newChild != null)
        children.add(newChild);
    }
    if (children.size() == 0)
      return null;
    return tf.newTreeNode(new StringLabel(s), children);
  }
}



/** Currently only used in LexicalizedParser */
class SubcategoryStripper implements TreeTransformer {

  protected TreeFactory tf = new LabeledScoredTreeFactory();
  private TreebankLanguagePack tlp = Options.get().langpack;

  public Tree transformTree(Tree tree, TreeFactory dummy) {
    Label l = tree.label();
    String s = l.value();
    if (tree.isLeaf())
      return tf.newLeaf(new StringLabel(s));
    s = tlp.basicCategory(s);
    // XXXX below isn't correct if root isn't unary.
    if (tlp.isStartSymbol(s))
      return transformTree(tree.children()[0],null);
    List children = new ArrayList();
    for (int cNum=0; cNum<tree.children().length; cNum++) {
      Tree child = tree.children()[cNum];
      Tree newChild = transformTree(child,null);
      if (newChild != null)
        children.add(newChild);
    }
    if (children.size() == 0)
      return null;
    Label newLabel = new CategoryWordTag(l);
    ((HasCategory)newLabel).setCategory(s);
    return tf.newTreeNode(newLabel, children);
  }
}


class CorrectDependencyExtractor extends AbstractTreeExtractor {
  protected Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  protected Numberer wordNumberer = Numberer.getGlobalNumberer("words");
  protected List dependencies = new ArrayList();

  protected void tallyRoot(Tree lt) {
    dependencies.addAll(DependencyGrammar.treeToDependencyList(lt));
  }

  /*
  protected void tallyInternalNodeOld(Tree lt) {
        if (lt.children().length != 2)
      return;
    String hTag = ((HasTag)lt.label()).tag();
    String lTag = ((HasTag)lt.children()[0].label()).tag();
    String rTag = ((HasTag)lt.children()[1].label()).tag();
    String hWord = ((HasWord)lt.label()).word();
    String lWord = ((HasWord)lt.children()[0].label()).word();
    String rWord = ((HasWord)lt.children()[1].label()).word();
    String aTag = (hTag.equals(lTag) ? rTag : lTag);
    String aWord = (hWord.equals(lWord) ? rWord : lWord);
    int hT = tagNumberer.number(hTag);
    int aT = tagNumberer.number(aTag);
    int hW = (wordNumberer.hasSeen(hWord) ?
              wordNumberer.number(hWord) :
              wordNumberer.number(Lexicon.UNKNOWN_WORD));
    int aW = (wordNumberer.hasSeen(aWord) ?
              wordNumberer.number(aWord) :
              wordNumberer.number(Lexicon.UNKNOWN_WORD));
    boolean leftHeaded = (hWord.equals(lWord) ? true : false);
    Dependency dependency = new Dependency(hW,hT,aW,aT,leftHeaded,-1);
    dependencies.add(dependency);
    // Dependency dependencyEndL = new Dependency(aW,aT,-1,-1,false);
    // dependencies.add(dependencyEndL);
    // Dependency dependencyEndR = new Dependency(aW,aT,-1,-1,true);
    // dependencies.add(dependencyEndR);
  }
  */
  public Object formResult() {
    return dependencies;
  }
}


class NodePruner {
  private ExhaustivePCFGParser parser;
  private TreeTransformer debinarizer;

  List prune(List treeList, Label label, int start, int end) {
    // get reference tree
    if (treeList.size() == 1)
      return treeList;
    Tree testTree = ((Tree)treeList.get(0)).treeFactory().newTreeNode(label, treeList);
    int goal = Numberer.getGlobalNumberer("states").number(label.value());
    Tree tempTree = parser.extractBestParse(goal, start, end);
    parser.restoreUnaries(tempTree);
    Tree pcfgTree = debinarizer.transformTree(tempTree, null);
    Set pcfgConstituents = pcfgTree.constituents(new LabeledScoredConstituentFactory());
    // delete child labels that are not in reference but do not cross reference
    List prunedChildren = new ArrayList();
    int childStart = 0;
    for (int c=0; c<testTree.children().length; c++) {
      Tree child = testTree.children()[c];
      boolean isExtra = true;
      int childEnd = childStart + child.yield().size();
      Constituent childConstituent = new LabeledScoredConstituent(childStart, childEnd, child.label(), 0);
      if (pcfgConstituents.contains(childConstituent)) {
        isExtra = false;
      }
      if (childConstituent.crosses(pcfgConstituents)) {
        isExtra = false;
      }
      if (child.isLeaf() || child.children()[0].isLeaf()) {
        isExtra = false;
      }
      if (pcfgTree.yield().size() != testTree.yield().size()) {
        isExtra = false;
      }
      if (! label.value().startsWith("NP^NP")) {
        isExtra = false;
      }
      if (isExtra) {
        System.err.println("Pruning: "+child.label()+" from "+(childStart+start)+" to "+(childEnd+start));
        System.err.println("Was: "+testTree+" vs "+pcfgTree);
        prunedChildren.addAll(child.getChildrenAsList());
      } else {
        prunedChildren.add(child);
      }
      childStart = childEnd;
    }
    return prunedChildren;
  }

  List helper(List treeList, int start) {
    List newTreeList = new ArrayList(treeList.size());
    for (Iterator i=treeList.iterator(); i.hasNext();) {
      Tree tree = (Tree)i.next();
      int end = start+tree.yield().size();
      newTreeList.add(prune(tree, start));
      start = end;
    }
    return newTreeList;
  }

  public Tree prune(Tree tree) {
    return prune(tree, 0);
  }

  Tree prune(Tree tree, int start) {
    if (tree.isLeaf() || tree.children()[0].isLeaf())
      return tree;
    // check each node's children for deletion
    List children = helper(tree.getChildrenAsList(), start);
    children = prune(children, tree.label(), start, start+tree.yield().size());
    return tree.treeFactory().newTreeNode(tree.label(), children);
  }

  public NodePruner(ExhaustivePCFGParser parser, TreeTransformer debinarizer) {
    this.parser = parser;
    this.debinarizer = debinarizer;
  }
}

class EvalB {

  private static Writer goldWriter, testWriter;

  static void initEVALBfiles() {
    try {
      goldWriter = new FileWriter("parses.gld", false);
      testWriter = new FileWriter("parses.tst", false);
    } catch (IOException e) {
      System.exit(0);
    }
  }

  static void closeEVALBfiles() {
    try {
      goldWriter.close();
      testWriter.close();
    } catch (IOException e) {
      System.exit(0);
    }
  }

  static void writeEVALBline(Tree gold, Tree test) {
    try {
      if (false) {
        if (gold.yield().length() <= 1) {
          goldWriter.write("(ROOT (X X))\n");
          testWriter.write("(ROOT (X X))\n");
          return;
        }
      }
      goldWriter.write(gold.toString()+"\n");
      testWriter.write(test.toString()+"\n");
      goldWriter.flush();
      testWriter.flush();
      System.err.println("Wrote EVALB lines.");
    } catch (IOException e) {
      System.err.println("Problem writing EVALB files!");
    }
  }

}

/* some documentat for Roger's convenience
 * {pcfg,dep,combo}{PE,DE,TE} are precision/dep/tagging evals for the models

 * parser is the PCFG parser
 * dparser is the dependency parser
 * bparser is the combining parser

 * during testing:
 * tree is the test tree
 * tree2 is the best PCFG parse (debinarized)
 * tree3 is the dependency parse
 * tree4 is the combo parse
 */

public class FactoredParser {

  public static void main(String[] args) throws IOException {

    TreebankLangParserParams tlpParams = new EnglishTreebankParserParams();
    // tlpParams may be changed to something else later, so don't use it till
    // after options are parsed.

    System.out.println("Currently "+new Date());
    System.out.print("Invoked with arguments:");
    for(int i = 0; i < args.length; i++) {
      System.out.print(" " + args[i]);
    }
    System.out.println();

    String path = "/u/nlp/stuff/corpora/Treebank3/parsed/mrg/wsj";
    int trainLow = 200, trainHigh = 2199, testLow = 2200, testHigh = 2219;
    String serializeFile = null;

    int i=0;
    while (i < args.length && args[i].startsWith("-")) {
      if (args[i].equalsIgnoreCase("-path") && (i+1 < args.length)) {
        path = args[i+1];
        i+=2;
      } else if (args[i].equalsIgnoreCase("-train") && (i+2 < args.length)) {
        trainLow = Integer.parseInt(args[i+1]);
        trainHigh = Integer.parseInt(args[i+2]);
        i+=3;
      } else if (args[i].equalsIgnoreCase("-test") && (i+2 < args.length)) {
        testLow = Integer.parseInt(args[i+1]);
        testHigh = Integer.parseInt(args[i+2]);
        i+=3;
      } else if (args[i].equalsIgnoreCase("-serialize") &&
                 (i+1 < args.length)) {
        serializeFile = args[i+1];
        i+=2;
      } else if (args[i].equalsIgnoreCase("-tLPP") && (i+1 < args.length)) {
        try {
          tlpParams =
            (TreebankLangParserParams) Class.forName(args[i+1]).newInstance();
        } catch (Exception e) {
          System.err.println("Couldn't instantiate " + args[i+1]);
        }
        i+=2;
      } else if (args[i].equals("-encoding")) {
	// sets encoding for TreebankLangParserParams
        tlpParams.setInputEncoding(args[i+1]);
        tlpParams.setOutputEncoding(args[i+1]);
        i+=2;
      } else {
        int j = setOptionFlag(args, i);
        if (j == i) {
          j = tlpParams.setOptionFlag(args, i);
        }
        if (j == i) {
          System.err.println("Unknown option ignored: " + args[i]);
          j++;
        }
        i = j;
      }
    }
    // System.out.println(tlpParams.getClass());

    Options.get().langpack = tlpParams.treebankLanguagePack();
    Train.splitters = new HashSet(Arrays.asList(tlpParams.splitters()));
    Train.sisterSplitters = new HashSet(Arrays.asList(tlpParams.sisterSplitters()));
    BinarizerFactory.HeadRespectingBinarizer.setTreebankLang(tlpParams);
    PrintWriter pw = tlpParams.pw();

    Test.display();
    Train.display();
    Options.get().display();
    tlpParams.display();

    // setup tree transforms
    Treebank trainTreebank = tlpParams.memoryTreebank();
    MemoryTreebank testTreebank = tlpParams.testMemoryTreebank();

    Timing.startTime();
    System.out.print("Reading trees...");
    testTreebank.loadPath(path, new NumberRangeFileFilter(testLow, testHigh,
                                                          true));
    if (Test.increasingLength) Collections.sort(testTreebank,
                                                new TreeLengthComparator());

    trainTreebank.loadPath(path, new NumberRangeFileFilter(trainLow, trainHigh,
                                                           true));
    Timing.tick("done.");
    System.out.print("Binarizing trees...");
    TreeTransformer binarizer = BinarizerFactory.getHeadRespectingBinarizer(tlpParams.headFinder(), !Train.outsideFactor());
    TreeTransformer debinarizer = BinarizerFactory.getDebinarizer();
    List binaryTrainTrees = new ArrayList();
    if (Train.hSelSplit) {
      ((BinarizerFactory.HeadRespectingBinarizer)binarizer).countStates = true;
      for (Iterator tI=trainTreebank.iterator(); tI.hasNext();) {
        Tree tree = (Tree)tI.next();
        if (Train.collinsPunc)
          tree = CollinsPuncTransformer.transformTree(tree);
        //tree.pennPrint(tlpParams.pw());
        tree = binarizer.transformTree(tree, null);
        //binaryTrainTrees.add(tree);
      }
      ((BinarizerFactory.HeadRespectingBinarizer)binarizer).countStates = false;
    }
    for (Iterator tI=trainTreebank.iterator(); tI.hasNext();) {
      Tree tree = (Tree)tI.next();
      if (Train.collinsPunc)
        tree = CollinsPuncTransformer.transformTree(tree);
      tree = binarizer.transformTree(tree, null);
      binaryTrainTrees.add(tree);
    }


    List binaryTestTrees = new ArrayList();
    for (Iterator tI=testTreebank.iterator(); tI.hasNext();) {
      Tree tree = (Tree)tI.next();
      if (Train.collinsPunc)
        tree = CollinsPuncTransformer.transformTree(tree);
      tree = binarizer.transformTree(tree, null);
      binaryTestTrees.add(tree);
    }
    Timing.tick("done.");
    BinaryGrammar bg = null;
    UnaryGrammar ug = null;
    DependencyGrammar dg = null;
    Lexicon lex = null;
    // extract grammars
    Extractor bgExtractor = new BinaryGrammarExtractor();
    //Extractor bgExtractor = new SmoothedBinaryGrammarExtractor();//new BinaryGrammarExtractor();
    // Extractor lexExtractor = new LexiconExtractor();
    Extractor dgExtractor = new DependencyGrammarExtractor();
    Extractor cdExtractor = new CorrectDependencyExtractor();
    if (Options.get().doPCFG) {
      System.out.print("Extracting PCFG...");
      Pair bgug = null;
      if (Train.cheatPCFG) {
        List allTrees = new ArrayList(binaryTrainTrees);
        allTrees.addAll(binaryTestTrees);
        bgug = (Pair)bgExtractor.extract(allTrees);
      } else {
        bgug = (Pair)bgExtractor.extract(binaryTrainTrees);
      }
      bg = (BinaryGrammar)bgug.second;
      bg.splitRules();
      ug = (UnaryGrammar)bgug.first;
      ug.purgeRules();
      Timing.tick("done.");
    }
    System.out.print("Extracting Lexicon...");
    lex =  tlpParams.lex();
    lex.train(binaryTrainTrees);
    Timing.tick("done.");

    System.out.print("Tuning Lexicon...");
    lex.tune(binaryTrainTrees);
    Timing.tick("done.");

    if (Options.get().doDep) {
      System.out.print("Extracting Dependencies...");
      dg = (DependencyGrammar)dgExtractor.extract(binaryTrainTrees);
      Timing.tick("done.");
      //System.out.print("Extracting Unknown Word Model...");
      //UnknownWordModel uwm = (UnknownWordModel)uwmExtractor.extract(binaryTrainTrees);
      //Timing.tick("done.");
      System.out.print("Extracting Tune Dependencies...");
      List tuneDeps = (List)cdExtractor.extract(binaryTestTrees);
      Timing.tick("done.");
      System.out.print("Tuning Dependency Model...");
      dg.tune(tuneDeps);
      //System.out.println("TUNE DEPS: "+tuneDeps);
      Timing.tick("done.");
    }

    BinaryGrammar boundBG = bg;
    UnaryGrammar boundUG = ug;

    GrammarProjection gp = new NullGrammarProjection(bg, ug);

    // serialization
    if (serializeFile != null) {
      System.out.print("Serializing parser...");
      LexicalizedParser.serializeParser(
              new LexicalizedParser.ParserData(lex, bg, ug, dg,
                             Numberer.getNumberers(), Options.get()),
              serializeFile);
      Timing.tick("done.");
    }

    // test: pcfg-parse and output

    ExhaustivePCFGParser parser = null;
    if (Options.get().doPCFG) parser = new ExhaustivePCFGParser(boundBG, boundUG, lex);


    ExhaustiveDependencyParser dparser = (Options.get().doDep ? new ExhaustiveDependencyParser(dg, lex) : null);

    Scorer scorer = (Options.get().doPCFG ? new TwinScorer(new ProjectionScorer(parser, gp),dparser) : null);
    //Scorer scorer = parser;
    BiLexPCFGParser bparser = null;
    if (Options.get().doPCFG && Options.get().doDep) {
      bparser = (Test.useN5) ?
        new N5BiLexPCFGParser(scorer, parser, dparser, bg, ug, dg, lex, gp):
        new BiLexPCFGParser(scorer, parser, dparser, bg, ug, dg, lex, gp);
    }

    LabeledConstituentEval pcfgPE = new LabeledConstituentEval("pcfg  PE");
    LabeledConstituentEval comboPE = new LabeledConstituentEval("combo PE");
    CBEval pcfgCB = new CBEval("pcfg  CB");

    TaggingEval pcfgTE = new TaggingEval("pcfg  TE");
    TaggingEval comboTE = new TaggingEval("combo TE");
    TaggingEval pcfgTEnoPunct = new TaggingEval("pcfg nopunct TE");
    TaggingEval comboTEnoPunct = new TaggingEval("combo nopunct TE");
    TaggingEval depTE = new TaggingEval("depnd TE");

    DependencyEval depDE = new DependencyEval("depnd DE");
    DependencyEval comboDE = new DependencyEval("combo DE");

    if (Test.evalb)
      EvalB.initEVALBfiles();

    // int[] countByLength = new int[Test.maxLength+1];

    // use a reflection ruse, so one can run this without needing the tagger
    //edu.stanford.nlp.process.SentenceTagger tagger = (Test.preTag ? new edu.stanford.nlp.process.SentenceTagger("/u/nlp/data/tagger.params/wsj0-21.holder") : null);
    SentenceProcessor tagger = null;
    if (Test.preTag) {
      try {
        Class[] argsClass = new Class[] { String.class };
        Object[] arguments =
          new Object[] { "/u/nlp/data/tagger.params/wsj0-21.holder" };
        tagger = (SentenceProcessor) Class.forName("edu.stanford.nlp.process.SentenceTagger").getConstructor(argsClass).newInstance(arguments);
      } catch (Exception e) {
        System.err.println(e);
        System.err.println("Warning: No pretagging of sentences will be done.");
      }
    }

    for (int tNum = 0, ttSize = testTreebank.size(); tNum < ttSize; tNum++) {
      Tree tree = (Tree) testTreebank.get(tNum);
      int testTreeLen = tree.yield().size();
      if (testTreeLen > Test.maxLength)
        continue;
      Tree binaryTree = (Tree)binaryTestTrees.get(tNum);
      // countByLength[testTreeLen]++;
      System.out.println("-------------------------------------");
      System.out.println("Number: " + (tNum+1));
      System.out.println("Length: " + testTreeLen);

      //tree.pennPrint(pw);
      // System.out.println("XXXX The binary tree is");
      // binaryTree.pennPrint(pw);
      //System.out.println("Here are the tags in the lexicon:");
      //System.out.println(lex.showTags());
      //System.out.println("Here's the tagnumberer:");
      //System.out.println(Numberer.getGlobalNumberer("tags").toString());

      long timeMil1 = System.currentTimeMillis();
      Timing.tick("Starting parse.");
      if (Options.get().doPCFG) {
	//System.err.println(Test.forceTags);
        if (Test.forceTags) {
          if (tagger != null) {
	    //System.out.println("Using a tagger to set tags");
            //System.out.println("Tagged sentence as: " + tagger.processSentence(cutLast(wordify(binaryTree.yield()))).toString(false));
            parser.parse(addLast(tagger.processSentence(cutLast(wordify(binaryTree.yield())))));
          } else {
	    //System.out.println("Forcing tags to match input.");
            parser.parse(cleanTags(binaryTree.taggedYield()));
          }
        } else {
	  // System.out.println("XXXX Parsing " + binaryTree.yield());
          parser.parse(binaryTree.yield());
        }
        //Timing.tick("Done with pcfg phase.");
      }
      if (Options.get().doDep) {
        dparser.parse(binaryTree.yield());
        //Timing.tick("Done with dependency phase.");
      }
      boolean bothPassed = false;
      if (Options.get().doPCFG && Options.get().doDep) {
        bothPassed = bparser.parse(binaryTree.yield());
        //Timing.tick("Done with combination phase.");
      }
      long timeMil2 = System.currentTimeMillis();
      long elapsed = timeMil2-timeMil1;
      System.err.println("Time: " + ((int) (elapsed/100))/10.00 + " sec.");
      //System.out.println("PCFG Best Parse:");
      Tree tree2b = null; 
      Tree tree2 = null;
      //System.out.println("Got full best parse...");
      if (Options.get().doPCFG) {
	tree2b = parser.getBestParse();
        tree2 = debinarizer.transformTree(tree2b,null);
      }
      //System.out.println("Debinarized parse...");
      //tree2.pennPrint();
      //System.out.println("DepG Best Parse:");
      Tree tree3 = null;
      Tree tree3db = null;
      if (Options.get().doDep) {
        tree3 = dparser.getBestParse();
        // was: but wrong Tree tree3db = debinarizer.transformTree(tree2,null);
        tree3db = debinarizer.transformTree(tree3,null);
        tree3.pennPrint(pw);
      }
      //tree.pennPrint();
      //((Tree)binaryTrainTrees.get(tNum)).pennPrint();
      //System.out.println("Combo Best Parse:");
      Tree tree4 = null;
      if (Options.get().doPCFG && Options.get().doDep) {
        try {
          tree4 = bparser.getBestParse();
          if (tree4 == null)
            tree4 = tree2b;
        } catch (NullPointerException e) {
          System.err.println("Blocked, using PCFG parse!");
          tree4 = tree2b;
        }
      }
      if (Options.get().doPCFG && !bothPassed) {
        tree4 = tree2b;
      }
      //tree4.pennPrint();
      if (Options.get().doDep) {
        depDE.evaluate(tree3,binaryTree,pw);
        depTE.evaluate(tree3db,tree,pw);
      }
      TreeTransformer tc = tlpParams.collinizer();
      TreeTransformer tcEvalb = tlpParams.collinizerEvalb();
      Tree tree4b = null;
      if (Options.get().doPCFG) {
	// System.out.println("XXXX Best PCFG was: ");
	// tree2.pennPrint();
	// System.out.println("XXXX Transformed best PCFG is: ");
	// tc.transformTree(tree2, null).pennPrint();
        //System.out.println("True Best Parse:");
        //tree.pennPrint();
        //tc.transformTree(tree,null).pennPrint();
        pcfgPE.evaluate(tc.transformTree(tree2, null),
                        tc.transformTree(tree, null),pw);
        pcfgCB.evaluate(tc.transformTree(tree2, null),
                        tc.transformTree(tree, null),pw);
        if (Options.get().doDep) {
          comboDE.evaluate((bothPassed ? tree4 : tree3),binaryTree,pw);
          tree4b = tree4;
          tree4 = debinarizer.transformTree(tree4,null);
          if (Test.nodePrune) {
            NodePruner np = new NodePruner(parser, debinarizer);
            tree4 = np.prune(tree4);
          }
          //tree4.pennPrint();
          comboPE.evaluate(tc.transformTree(tree4,null),
                           tc.transformTree(tree,null),pw);
        }
        //pcfgTE.evaluate(tree2, tree);
        pcfgTE.evaluate(tcEvalb.transformTree(tree2,null),
                               tcEvalb.transformTree(tree,null),pw);
        pcfgTEnoPunct.evaluate(tc.transformTree(tree2,null),
                               tc.transformTree(tree,null),pw);

        if (Options.get().doDep) {
          comboTE.evaluate(tcEvalb.transformTree(tree4,null),
                           tcEvalb.transformTree(tree,null),pw);
          comboTEnoPunct.evaluate(tc.transformTree(tree4,null),
                                  tc.transformTree(tree,null),pw);
        }
        System.out.println("PCFG only: "+parser.scoreBinarizedTree(tree2b,true));

        //tc.transformTree(tree2, null).pennPrint();
        tree2.pennPrint(pw);

        if (Options.get().doDep) {
          System.out.println("Combo: "+parser.scoreBinarizedTree(tree4b,true));
          // tc.transformTree(tree4, null).pennPrint(pw);
          tree4.pennPrint(pw);
        }
        System.out.println("Correct:"+parser.scoreBinarizedTree(binaryTree,true));
        /*
        if (parser.scoreBinarizedTree(tree2b,true) < parser.scoreBinarizedTree(binaryTree,true)) {
          System.out.println("SCORE INVERSION");
          parser.validateBinarizedTree(binaryTree,0);
        }
        */
        tree.pennPrint(pw);
      } // end if doPCFG

      if (Test.evalb) {
        if (Options.get().doPCFG && Options.get().doDep) {
          EvalB.writeEVALBline(tcEvalb.transformTree(tree, null),
                         tcEvalb.transformTree(tree4, null));
        } else if (Options.get().doPCFG) {
          EvalB.writeEVALBline(tcEvalb.transformTree(tree, null),
                         tcEvalb.transformTree(tree2, null));
        } else if (Options.get().doDep) {
          EvalB.writeEVALBline(tcEvalb.transformTree(tree, null),
                         tcEvalb.transformTree(tree3db, null));
        }
      }
    } // end for each tree in test treebank

    if (Test.evalb)
      EvalB.closeEVALBfiles();

    // Test.display();
    if (Options.get().doPCFG) {
      System.out.print("ParseEval for PCFG: ");
      pcfgPE.display(false,pw);
      System.out.println("Grammar size: "+Numberer.getGlobalNumberer("states").total());
      pcfgCB.display(false,pw);
      if (Options.get().doDep) {
        System.out.print("ParseEval for Combo: ");
        comboPE.display(false,pw);
      }
      System.out.print("TagEval for PCFG: ");
      pcfgTE.display(false,pw);
      System.out.print("TagEval for PCFG excluding punct.: ");
      pcfgTEnoPunct.display(false,pw);
      if (Options.get().doDep) {
        System.out.print("TagEval for Combo: ");
        comboTE.display(false,pw);
        System.out.print("TagEval for Combo excluding punct.: ");
        comboTEnoPunct.display(false,pw);
      }
    }
    if (Options.get().doDep) {
      System.out.print("TagEval for Dependency: ");
      depTE.display(false,pw);
      System.out.print("DepEval for Dependency: ");
      depDE.display(false,pw);
    }
    if (Options.get().doPCFG && Options.get().doDep) {
      System.out.print("DepEval for Combo: ");
      comboDE.display(false,pw);
    }
    // pcfgPE.printGoodBad();
  }


  static int setOptionFlag(String[] args, int i) {
    if (args[i].equalsIgnoreCase("-PCFG")) {
      Options.get().doDep = false;
      Options.get().doPCFG = true;
      i++;
    } else if (args[i].equalsIgnoreCase("-dep")) {
      Options.get().doDep = true;
      Options.get().doPCFG = false;
      i++;
    } else if (args[i].equalsIgnoreCase("-factored")) {
      Options.get().doDep = true;
      Options.get().doPCFG = true;
      i++;
    } else if (args[i].equalsIgnoreCase("-noRecoveryTagging")) {
      Test.noRecoveryTagging = true;
      i++;
    } else if (args[i].equalsIgnoreCase("-maxLength") && (i+1 < args.length)) {
      int length = Integer.parseInt(args[i+1]);
      Test.maxLength = length;
      i+=2;
    } else if (args[i].equalsIgnoreCase("-vMarkov") && (i+1 < args.length)) {
      int order = Integer.parseInt(args[i+1]);
      if (order <= 1) {
        Train.PA = false;
        Train.gPA = false;
      } else if (order == 2) {
        Train.PA = true;
        Train.gPA = false;
      } else if (order >= 3) {
        Train.PA = true;
        Train.gPA = true;
      }
      i+=2;
    } else if (args[i].equalsIgnoreCase("-vSelSplit")) {
      Train.selectiveSplit = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-hMarkov") && (i+1 < args.length)) {
      int order = Integer.parseInt(args[i+1]);
      Train.markovOrder = order;
      Train.markovFactor = true;
      i+=2;
    } else if (args[i].equalsIgnoreCase("-smoothTagsThresh") &&
	       (i+1 < args.length)) {
      int thresh = Integer.parseInt(args[i+1]);
      Options.get().SMOOTH_IN_UNKNOWNS_THRESHOLD = thresh;
      i+=2;
    } else if (args[i].equalsIgnoreCase("-hSelSplitThresh") &&
	       (i+1 < args.length)) {
      int thresh = Integer.parseInt(args[i+1]);
      Train.HSEL_CUT = thresh;
      i+=2;
    } else if (args[i].equalsIgnoreCase("-hSelSplit")) {
      Train.hSelSplit = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-tagPA")) {
      Train.tagPA = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-tagSelSplit")) {
      Train.tagSelectiveSplit = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-uwm") && (i+1 < args.length)) {
      int order = Integer.parseInt(args[i+1]);
      if (order < 0) {
	order = 0;
      } else if (order > 5) {
	order = 5;
      }
      Options.get().useUnknownWordSignatures = order;
      i+=2;
    } else if (args[i].equalsIgnoreCase("-openClassThreshold") &&
	       (i+1 < args.length)) {
      int thresh = Integer.parseInt(args[i+1]);
      Train.openClassTypesThreshold = thresh;
      i+=2;
    } else if (args[i].equalsIgnoreCase("-leaveItAll")) {
      Train.leaveItAll = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-unary")) {
      Train.markUnary = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-unary2")) {
      Train.markUnary2 = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-mutate")) {
      Options.get().smartMutation = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-distanceV")) {
      Train.dominatesV = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-rightRec")) {
      Train.rightRec = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-preTag")) {
      Test.preTag = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-forceTags")) {
      Test.forceTags = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-scTags")) {
      Test.dcTags = false;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-v")) {
      Test.verbose = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-outputTreeFormat")) {
      Test.outputTreeFormat = args[i+1];
      i+=2;
    } else if (args[i].equalsIgnoreCase("-lexicalizeOutput")) {
      Test.lexicalizeOutput = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-printDepsSet")) {
      Test.printDepsSet = true;
      i+=1;
    } else if (args[i].equalsIgnoreCase("-acl03pcfg")) {
      Options.get().doDep = false;
      Options.get().doPCFG = true;
      // Options.get().SMOOTH_IN_UNKNOWNS_THRESHOLD = 30;
      Train.markUnary = true;
      Train.PA = true;
      Train.gPA = false;
      Train.tagPA = true;
      Train.tagSelectiveSplit = false;
      Train.dominatesV = true;
      Train.rightRec = true;
      Train.selectiveSplit = true;
      Train.markovFactor = true;
      Train.markovOrder = 2;
      Train.hSelSplit = true;
      Options.get().useUnknownWordSignatures = 2;
      Options.get().flexiTag = true;
      Test.dcTags = true;
      // don't increment i so it gets language specific stuff as well
    } else if (args[i].equalsIgnoreCase("-ijcai03")) {
      Options.get().doDep = true;
      Options.get().doPCFG = true;
      Train.markUnary = false;
      Train.PA = true;
      Train.gPA = false;
      Train.tagPA = false;
      Train.tagSelectiveSplit = false;
      Train.dominatesV = false;
      Train.rightRec = false;
      Train.selectiveSplit = true;
      Train.markovFactor = true;
      Train.markovOrder = 2;
      Train.hSelSplit = true;
      Options.get().useUnknownWordSignatures = 2;
      Options.get().flexiTag = false;
      Test.dcTags = true;
      Test.nodePrune = true;
      // don't increment i so it gets language specific stuff as well
    } else if (args[i].equalsIgnoreCase("-printTT")) {
      Train.printTreeTransformations = true;
      i++;
    }
    return i;
  }


  private static List cleanTags(List twList) {
    List l = new ArrayList();
    for (int i=0; i<twList.size(); i++) {
      TaggedWord tw = (TaggedWord)twList.get(i);
      TaggedWord tw2 = new TaggedWord(tw.word(),
                    Options.get().langpack.basicCategory(tw.tag()));
      l.add(tw2);
    }
    return l;
  }

  private static Sentence wordify(List wList) {
    Sentence s = new Sentence();
    for (int i=0; i<wList.size(); i++) {
      s.add(new Word(wList.get(i).toString()));
    }
    return s;
  }

  private static Sentence cutLast(Sentence s) {
    return new Sentence(s.subList(0, s.size()-1));
  }

  private static Sentence addLast(Sentence s) {
    Sentence s2 = new Sentence(s);
    s2.add(new StringLabel(Lexicon.BOUNDARY));
    return s2;
  }

  /** Not an instantiable class */
  private FactoredParser() {}

}
